/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;

import java.awt.Container;
import java.awt.event.*;
import javax.swing.*;

import sun.misc.*;

import ffe.panels.MainPanel;

/**
 * The Main class is the entry point to Force Field Explorer and does some
 * initializations.
 */
public final class Main
   extends JFrame{

   static MainPanel mainPanel;

   public static boolean MAC_OS_X = (System.getProperty("os.name").toLowerCase().startsWith(
                                    "mac os x"));
   public static boolean LINUX = (System.getProperty("os.name").toLowerCase().startsWith("linux"));
   public static boolean WINDOWS = (System.getProperty("os.name").toLowerCase().startsWith(
                                   "windows"));

   /**
    * Main does some window intializations
    */
   public Main(){
      setTitle("Force Field Explorer");
      URL iconURL = getClass().getClassLoader().getResource("ffe/icons/icon.jpg");
      ImageIcon icon = new ImageIcon(iconURL);
      setIconImage(icon.getImage());
      setDefaultLookAndFeelDecorated(true);
      addWindowListener(new WindowAdapter(){
         public void windowClosing(WindowEvent e){
            if (mainPanel != null){
               mainPanel.exit();
            }
            System.exit(0);
         }
      });

      // Catch 'Control-C' events so FFE can save preferences and shut down cleanly
      SignalHandler handler = new SignalHandler(){
         public void handle(Signal sig){
            System.out.println(" Signal: " + sig.getName() + " - " + sig.getNumber());
            System.out.println(" FFE is shutting down...");
            if (mainPanel != null){
               mainPanel.exit();
            }
            System.exit( -1);
         }
      };

      Signal.handle(new Signal("INT"), handler);
   }

   /**
    * Initialize the main application panel
    */
   private void initialize(){
      // Check for Java3D Extensions
      checkJava3D();

      // Create the MainPanel and Menu, then add them to the JFrame
      mainPanel = new MainPanel(this);
      Container contentPane = getContentPane();
      contentPane.add(mainPanel);
      mainPanel.initialize();
      setJMenuBar(mainPanel.getMenu());
      mainPanel.setPanel(mainPanel.LOGS);
      show();
      mainPanel.setPanel(mainPanel.GRAPHICS);
      //mainPanel.invalidate();
   }

   /**
    * Create an instance of Force Field Explorer
    */
   public static void main(String[] args) throws Exception{
      //long start = System.currentTimeMillis();

      // Some Mac OS X specific features that help FFE look native
      if (MAC_OS_X){
         System.setProperty("apple.mrj.application.apple.menu.about.name", "Force Field Explorer");
         System.setProperty("apple.laf.useScreenMenuBar", "true");
         System.setProperty("apple.awt.showGrowBox", "true");
         System.setProperty("apple.awt.brushMetalLook", "true");
      }

      // If a file was supplied on the command
      File commandLineFile = null;
      if (args != null && args.length > 0){
         commandLineFile = new File(args[0]);
         if (commandLineFile.exists()){
            commandLineFile = new File(commandLineFile.getAbsolutePath());
         }
      }


      Main m = new Main();
      m.initialize();

      if (MAC_OS_X){
         m.macOSXRegistration();
      }

      // Check for a structure file supplied on the command line
      if (commandLineFile != null) {
         if (commandLineFile.exists()) {
            mainPanel.open(commandLineFile);
         } else {
            String message = new String(commandLineFile.toString() + " was not found.");
            JOptionPane.showMessageDialog(null, message, "File Not Found", JOptionPane.ERROR_MESSAGE);
         }
      }

      //long finished = System.currentTimeMillis();
      //double time = ( (double) (finished - start)) / 1000.0;
      //System.out.println("FFE intialization: " + time + " seconds");
      /* for (Enumeration e = System.getProperties().propertyNames();
                           e.hasMoreElements(); ) {
         String s = (String) e.nextElement();
         System.out.println("" + s + ": " + System.getProperty(s));
             } */
   }

   /**
    * Register FFE so that file associations can be handled
    */
   private void macOSXRegistration(){
      try{
         Class osxAdapter = Class.forName("ffe.platforms.macosx.OSXAdapter");
         Class[] defArgs = {MainPanel.class};
         Method registerMethod = osxAdapter.getDeclaredMethod("registerMacOSXApplication", defArgs);
         if (registerMethod != null){
            Object[] args = {mainPanel};
            registerMethod.invoke(osxAdapter, args);
         }
      } catch (NoClassDefFoundError e){
         System.err.println("This version of Mac OS X does not support the Apple EAWT.  Application Menu handling has been disabled (" +
            e + ")");
      } catch (ClassNotFoundException e){
         System.err.println("This version of Mac OS X does not support the Apple EAWT.  Application Menu handling has been disabled (" +
            e + ")");
      } catch (Exception e){
         System.err.println("Exception while loading the OSXAdapter:");
         e.printStackTrace();
      }
   }

   /**
    * Look for Java3D extensions. If not found, directions are printed on where to find them.
    */
   private void checkJava3D(){
      Class java3d = null;
      try{
         java3d = getClass().getClassLoader().loadClass("javax.media.j3d.Canvas3D");
      } catch (Exception e){
      } finally{
         if (java3d == null){
            System.out.println("\nJava3D Extensions have not been installed\n");
            if (MAC_OS_X){
               System.out.println("Please follow Java3D installation instructions given at:");
               System.out.println(
                  "http://www.apple.com/downloads/macosx/apple/java3dandjavaadvancedimagingupdate.html");
            } else if (LINUX){
               System.out.println("Please download Java3D 1.3.1 from a Blackdown FTP Site:");
               System.out.println("http://www.blackdown.org");
            } else if (WINDOWS){
               System.out.println("Please download Java3D 1.3.1 from Sun at:");
               System.out.println("http://java.sun.com/products/java-media/3D/download.html\n");
               System.out.println("Choose the version named:");
               System.out.println("Java 3D for Windows (OpenGL Version) Runtime for the JRE");
            }
            System.out.println("\nForce Field Explorer can not continue\n\n");
            System.exit( -1);
         }
      }
   }
}
