/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.behaviors;

import java.util.Enumeration;
import javax.media.j3d.*;
import javax.vecmath.Vector3d;

import ffe.lang.Atom;

/**
 * The CollisionBehavior class creates a behavior which controls how atoms
 * interact when they collide in the Java3D Universe. It needs work, but
 * might end up being a useful tool for configuring/docking systems.
 */
public final class CollisionBehavior
   extends Behavior{

   private WakeupOnCollisionMovement wMovement;
   private TransformGroup armingFSystemTG, triggeringFSystemTG;
   private TransformGroup armingAtomTG, triggeringAtomTG;
   private Transform3D armingT3D, triggeringT3D;
   private Shape3D armingNode;

   /**
    * CollisionBehavior Constructor
    * @param s is the armingNode atom's Shape3D geometry
    * @param a is the TransformGroup that represents the position of the atom relative to other atoms in the System
    * @param m is the TransformGroup that represents the position of the System in the Java3D VirtualUniverse
    */
   public CollisionBehavior(Shape3D s, TransformGroup a, TransformGroup m){
      armingNode = s;
      armingAtomTG = a;
      armingFSystemTG = m;
   }

   /**
    * This method is called by the Java3D engine to determine what causes this behavior to Wake up
    */
   public void initialize(){
      wMovement = new WakeupOnCollisionMovement(armingNode, WakeupOnCollisionMovement.USE_BOUNDS);
      wakeupOn(wMovement);
   }

   /**
    * This method is called whenever another atom collides with the armingNode atom
    * @param e An emuration of the WakeUp Conditions, currently always a WakeUpCollisionMovement,
    * from which one can obtain information about what caused the collision
    */
   public void processStimulus(Enumeration e){
      //Retrieve the Wakeup Condition info
      if (e.hasMoreElements()){
         WakeupOnCollisionMovement collision = (WakeupOnCollisionMovement) e.nextElement();

         //Get Arming Info
         armingT3D = new Transform3D();
         armingAtomTG.getTransform(armingT3D);
         Transform3D temp = new Transform3D();
         armingFSystemTG.getTransform(temp);
         armingT3D.add(temp);

         Atom a = (Atom) armingNode.getUserData();
         a.print();

         //Get Triggering Info
         Node node;
         boolean first = true;
         SceneGraphPath triggeringpath = collision.getTriggeringPath();
         int count = triggeringpath.nodeCount();
         for (int i = 0; i < count; i++){
            node = triggeringpath.getNode(i);
            if (node instanceof TransformGroup){
               if (first){
                  triggeringAtomTG = (TransformGroup) node;
                  triggeringT3D = new Transform3D();
                  triggeringAtomTG.getTransform(triggeringT3D);
                  first = false;
               } else{
                  triggeringFSystemTG = (TransformGroup) node;
                  triggeringFSystemTG.getTransform(temp);
                  triggeringT3D.add(temp);
               }
            }
         }

         if (triggeringFSystemTG == armingFSystemTG){
            System.out.println("Collision triggered from within System");
            System.out.println("Collision behavior turned off for: " +
               ((Atom) armingNode.getUserData()).toString());
         } else{
            Vector3d move = new Vector3d();
            Vector3d trigger = new Vector3d();
            triggeringT3D.get(trigger);
            Vector3d arming = new Vector3d();
            armingT3D.get(arming);
            move.sub(trigger, arming);
            move.normalize();
            move.scale(0.2);
            triggeringFSystemTG.getTransform(triggeringT3D);
            triggeringT3D.get(trigger);
            move.add(trigger);
            triggeringT3D.setTranslation(move);
            triggeringFSystemTG.setTransform(triggeringT3D);
            wakeupOn(wMovement);
         }
      }
   }
}
