/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides an GUI to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay William Ponder</p>
 * <p>Company: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.behaviors;

import java.util.Enumeration;
import javax.media.j3d.*;
import javax.vecmath.Point3d;

import java.awt.*;
import java.awt.event.MouseEvent;

import com.sun.j3d.utils.picking.*;
import com.sun.j3d.utils.universe.SimpleUniverse;

import ffe.lang.*;
import ffe.panels.*;

/**
 * The EventDispatch class listens for mouse events over the Java3D canvas, dispatching work
 * to more specialized classes.
 */
public class EventDispatch
   extends Behavior{

   // GUI Panels
   private MainPanel ffe;
   private Graphics3D graphics3D;
   private GlobalAxis navigation;

   // Scenegraph Nodes
   private SimpleUniverse universe;
   private TransformGroup viewTG;
   private Bounds bounds;
   private BranchGroup base;
   private TransformGroup baseTransform;
   private Transform3D transform3D;

   // Wake up conditions
   private WakeupOr mouseCriterion;
   private WakeupOr postCriterion;

   // Mouse/Pick state upon wake up event
   //private MouseEvent mouseEvent;
   private boolean buttonPress;
   private int x, y, x_last, y_last;
   private boolean leftButton;
   private boolean rightButton;
   private PickCanvas pickCanvas;
   private PickResult pickResult;
   private Atom atom;
   private boolean axisSelected;

   // Behaviors
   private MouseRotate systemRotate;
   private MouseTranslate systemTranslate;

   private GlobalBehavior viewOrbitBehavior;

   // Behavior Post IDs
   public static int ROTATEPOST = 1;
   public static int TRANSLATEPOST = 2;
   public static int BEHAVIORDONEPOST = 3;

   public EventDispatch(MainPanel f, Graphics3D g, GlobalAxis n, SimpleUniverse u, Bounds b,
      BranchGroup root, TransformGroup tg){

      ffe = f;
      graphics3D = g;
      navigation = n;
      universe = u;
      viewTG = u.getViewingPlatform().getViewPlatformTransform();
      bounds = b;
      setSchedulingBounds(b);
      base = root;
      baseTransform = tg;
      transform3D = new Transform3D();

      // Initialize the System Rotate Behavior
      systemRotate = new MouseRotate(MouseRotate.MANUAL_WAKEUP, viewTG, this, ROTATEPOST);
      base.addChild(systemRotate);
      systemRotate.setFactor(0.025);
      systemRotate.setSchedulingBounds(bounds);

      // Initialize the System Translate Behavior
      systemTranslate = new MouseTranslate(MouseTranslate.MANUAL_WAKEUP, viewTG, this,
                        TRANSLATEPOST);
      base.addChild(systemTranslate);
      systemTranslate.setFactor(0.1);
      systemTranslate.setSchedulingBounds(bounds);

      // Initialize the viewOrbitBehavior
      viewOrbitBehavior = new GlobalBehavior(graphics3D);
      viewOrbitBehavior.setUpCallback(navigation);
      viewOrbitBehavior.setSchedulingBounds(bounds);
      u.getViewingPlatform().setViewPlatformBehavior(viewOrbitBehavior);

      // Initialize the PickCanvas
      pickCanvas = new PickCanvas(graphics3D, universe.getLocale());
      pickCanvas.setMode(PickCanvas.GEOMETRY);
      pickCanvas.setTolerance(10.0f);

   }

   public void processMouseEvent(MouseEvent evt){

      x_last = x;
      y_last = y;
      x = evt.getX();
      y = evt.getY();

      buttonPress = false;
      if (evt.getID() == MouseEvent.MOUSE_PRESSED){
         buttonPress = true;
      }

      atom = null;
      axisSelected = false;
      if (buttonPress){
         // Picking Results
         pickCanvas.setShapeLocation(x, y);
         pickResult = pickCanvas.pickClosest();

         if (pickResult != null){
            SceneGraphPath sgp = pickResult.getSceneGraphPath();
            Node node = sgp.getObject();
            if (node instanceof Shape3D){
               Shape3D s = (Shape3D) node;
               Object o = s.getUserData();
               if (o instanceof FSystem){
                  FSystem sys = (FSystem) o;
                  if (pickResult.numIntersections() > 0){
                     PickIntersection pi = pickResult.getIntersection(0);
                     int coords[] = pi.getPrimitiveCoordinateIndices();
                     atom = sys.getAtomFromWireVertex(coords[0]);
                  }
               } else if (o instanceof Atom || o instanceof Bond){
                  if (o instanceof Bond){
                     Bond b = (Bond) o;
                     atom = b.getAtom(0);
                  } else{
                     atom = (Atom) o;
                  }
               } else if (o instanceof GlobalAxis){
                  axisSelected = true;
               }
            }
         }
      }

      // Left Button
      if ((evt.getModifiersEx() & MouseEvent.BUTTON1_DOWN_MASK) == MouseEvent.BUTTON1_DOWN_MASK){
         leftButton = true;
      } else{
         leftButton = false;
      }

      // Right Button
      if ((evt.getModifiersEx() & MouseEvent.BUTTON3_DOWN_MASK) == MouseEvent.BUTTON3_DOWN_MASK){
         rightButton = true;
      } else{
         rightButton = false;
      }

   }

   // Most of the logic for user interaction with the Scenegraph is here.
   public void processStimulus(Enumeration criteria){

      viewOrbitBehavior.setEnable(false);

      AWTEvent awtEvents[] = null;
      while (criteria.hasMoreElements()){
         WakeupCriterion wakeup = (WakeupCriterion) criteria.nextElement();
         if (wakeup instanceof WakeupOnAWTEvent){
            awtEvents = ((WakeupOnAWTEvent) wakeup).getAWTEvent();
            if (awtEvents == null){
               continue;
            }
            for (int i = 0; i < awtEvents.length; i++){
               MouseEvent mouseEvent = null;
               if (awtEvents[i] instanceof MouseEvent){
                  mouseEvent = (MouseEvent) awtEvents[i];
                  processMouseEvent(mouseEvent);
               } else{
                  continue;
               }

               if (!axisSelected){

                  // Wake Up System Translate Behavior
                  if (rightButton && buttonPress){
                     if (systemTranslate()){
                        wakeupOn(postCriterion);
                        return;
                     }

                  }

                  // Wake Up System Rotate Behavior
                  if (leftButton && buttonPress){
                     if (systemRotate()){
                        wakeupOn(postCriterion);
                        return;
                     }
                  }

                  // Zoom
                  if ((mouseEvent.getID() == MouseEvent.MOUSE_DRAGGED) &&
                     ((mouseEvent.getModifiersEx() &
                     MouseEvent.BUTTON2_DOWN_MASK) == MouseEvent.BUTTON2_DOWN_MASK)){
                     zoom();
                     return;
                  }

               } else{
                  //if (leftButton && buttonPress){
                  viewOrbitBehavior.setEnable(true);
                  wakeupOn(mouseCriterion);
                  return;
                  //}
               }
            }
         }
      }
      wakeupOn(mouseCriterion);
   }

   public void initialize(){
      WakeupCriterion[] behaviorPost = new WakeupCriterion[2];
      behaviorPost[0] = new WakeupOnBehaviorPost(systemRotate, BEHAVIORDONEPOST);
      behaviorPost[1] = new WakeupOnBehaviorPost(systemTranslate, BEHAVIORDONEPOST);
      postCriterion = new WakeupOr(behaviorPost);

      WakeupCriterion awtCriterion[] = new WakeupCriterion[2];
      awtCriterion[0] = new WakeupOnAWTEvent(Event.MOUSE_DOWN);
      awtCriterion[1] = new WakeupOnAWTEvent(Event.MOUSE_UP);
      mouseCriterion = new WakeupOr(awtCriterion);
      wakeupOn(mouseCriterion);
   }

   private boolean systemRotate(){
      TransformGroup tg = null;
      if ((graphics3D.getPickMode() == Graphics3D.SCREEN) && atom != null){
         tg = (TransformGroup) pickResult.getNode(PickResult.TRANSFORM_GROUP);
      } else if (graphics3D.getPickMode() == Graphics3D.ACTIVE){
         if (ffe.getTree().getActive() == null){
            tg = null;
         } else{
            tg = ffe.getTree().getActive().getTransformGroup();
         }
      }
      if ((tg != null) && (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_READ)) &&
         (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_WRITE))){
         systemRotate.setTransformGroup(tg);
         postId(ROTATEPOST);
         return true;
      }
      return false;
   }

   private boolean systemTranslate(){
      TransformGroup tg = null;
      if ((graphics3D.getPickMode() == Graphics3D.SCREEN) && atom != null){
         tg = (TransformGroup) pickResult.getNode(PickResult.TRANSFORM_GROUP);
      } else if (graphics3D.getPickMode() == Graphics3D.ACTIVE){
         if (ffe.getTree().getActive() == null){
            tg = null;
         } else{
            tg = ffe.getTree().getActive().getTransformGroup();
         }
      }
      if ((tg != null) && (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_READ)) &&
         (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_WRITE))){
         systemTranslate.setTransformGroup(tg);
         postId(TRANSLATEPOST);
         return true;
      }
      return false;
   }

   public void centerView(boolean resetRotation, boolean resetTranslation, boolean resetZoom){
      viewOrbitBehavior.centerView(resetRotation, resetTranslation, resetZoom);
   }

   public void setGlobalCenter(double d[]){
      Point3d point = new Point3d(d);
      viewOrbitBehavior.setRotationCenter(point);
   }

   private void zoom(){
      int dy = y - y_last;

      baseTransform.getTransform(transform3D);
      double z = ( -1.0) * dy * 0.0002;
      double scale = transform3D.getScale() + z;
      if (scale > 0){
         transform3D.setScale(scale);
         baseTransform.setTransform(transform3D);
      }
   }

}
