/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.behaviors;

import java.util.*;
import javax.media.j3d.*;
import javax.media.j3d.Node;
import javax.vecmath.Vector3d;

import javax.swing.JFrame;
import javax.swing.tree.TreePath;

import bsh.Interpreter;
import com.sun.j3d.utils.picking.*;

import ffe.lang.*;
import ffe.panels.*;

/**
 * The PickNodeBehavior class is used to make graphical selections and measurements.
 */
public class PickNodeBehavior
   extends PickMouseBehavior{

   // GUI Components
   private MainPanel fpanel;

   // Currently picked atom/node
   private Atom previousAtom = null;
   private FNode previousPick = null;
   private int pickNumber = 0;

   // Picking Level
   private boolean mode = false;

   private String pickLevel = "Atom".intern();
   private String newPickLevel = "Atom".intern();

   // Measuring Level and selected atoms
   private Vector atomcache = new Vector(4);
   private int count = 0;

   /**
    * Constructor
    * @param root Base of the Scenegraph
    * @param canvas Scene Canvas3D
    * @param bounds Behavior bounds
    */
   public PickNodeBehavior(BranchGroup base, Bounds bounds, Graphics3D g, MainPanel f){
      super(g, base, bounds);
      fpanel = f;
      pickCanvas.setMode(PickCanvas.GEOMETRY);
      pickCanvas.setTolerance(3.0f);
   }

   /**
    * Clear currently selected nodes
    */
   public void clear(){
      if (previousPick != null){
         fpanel.getTree().collapsePath(new TreePath(previousPick.getPath()));
         previousPick.setSelected(false);
         previousPick.switchColor(GraphicsContext.bySELECT, null, null);
         previousPick = null;
         pickNumber = 0;
      }
      for (int i = 0; i < atomcache.size(); i++){
         Atom a = (Atom) atomcache.get(i);
         a.setSelected(false);
         a.switchColor(GraphicsContext.bySELECT, null, null);
      }
      atomcache.clear();
   }

   public FNode getPick(){
      return previousPick;
   }

   public void setMode(boolean m){
      mode = m;
      if (mode == false){
         clear();
      }
   }

   public void setPickLevel(String newPick){
      newPickLevel = newPick;
   }

   public String getPickLevel(){
      return pickLevel;
   }

   public void resetCount(){
      count = 0;
   }

   /**
    * Called by Java3D when an atom is picked
    * @param xpos Horizontal mouse position
    * @param ypos Vertical mouse position
    */
   public void updateScene(int xpos, int ypos){

      if (mode == false){
         return;
      }

      // Determine what FNode was picked
      pickCanvas.setShapeLocation(xpos, ypos);
      PickResult result = pickCanvas.pickClosest();

      if (result != null){
         SceneGraphPath sgp = result.getSceneGraphPath();
         Node node = sgp.getObject();
         if (!(node instanceof Shape3D)){
            return;
         }
         Shape3D s = (Shape3D) node;
         Object o = s.getUserData();
         if (o instanceof FSystem){
            FSystem sys = (FSystem) o;
            if (result.numIntersections() > 0){
               PickIntersection pi = result.getIntersection(0);
               int coords[] = pi.getPrimitiveCoordinateIndices();
               o = sys.getAtomFromWireVertex(coords[0]);
            } else{
               return;
            }
         }

         if (o instanceof Atom || o instanceof Bond){
            Atom a;
            if (o instanceof Bond){
               Bond b = (Bond) o;
               a = b.getAtom(0);
            } else{
               a = (Atom) o;
            }

            // Check to see if the pickLevel has changed
            if (!(pickLevel == newPickLevel)){
               pickLevel = newPickLevel;
               pickNumber = 0;
            }

            // Clear selections between measurements
            if (!pickLevel.startsWith("Measure") || count == 0){
               Iterator li = atomcache.iterator();
               while (li.hasNext()){
                  Atom matom = (Atom) li.next();
                  matom.setSelected(false);
                  matom.switchColor(GraphicsContext.bySELECT, null, null);
               }
               atomcache.clear();
               count = 0;
            }

            // If measuring, select the current atom and add it to the cache
            if (pickLevel.startsWith("Measure") && !atomcache.contains(a)){
               atomcache.add(0, a);
               a.setSelected(true);
               a.switchColor(GraphicsContext.byPICK, null, null);
               count++;
               measure();
            }

            if (!pickLevel.startsWith("Measure")){
               // Check to see if the same Atom has been selected twice in a row
               // This allows iteration through the atom's terms
               if (a == previousAtom){
                  pickNumber++;
               } else{
                  previousAtom = a;
                  pickNumber = 0;
               }

               FNode currentPick = null;
               if (pickLevel == "Atom"){
                  currentPick = a;
               } else if (pickLevel == "Bond" || pickLevel == "Angle" || pickLevel == "Dihedral"){

                  ArrayList terms = null;
                  if (pickLevel == "Bond"){
                     terms = a.getBonds();
                  } else if (pickLevel == "Angle"){
                     terms = a.getAngles();
                  } else if (pickLevel == "Dihedral"){
                     terms = a.getDihedrals();

                  }
                  if (terms == null){
                     return;
                  }
                  int num = terms.size();
                  if (pickNumber >= num){
                     pickNumber = 0;
                  }
                  currentPick = (Term) terms.get(pickNumber);
               } else if (pickLevel == "Residue" || pickLevel == "Polymer" || pickLevel == "System" ||
                  pickLevel == "Molecule"){
                  FNode fnode = null;
                  if (pickLevel == "Residue"){
                     fnode = a.getResidueNode();
                  } else if (pickLevel == "Polymer"){
                     fnode = a.getPolymerNode();
                  } else if (pickLevel == "System"){
                     fnode = a.getFSystemNode();
                  } else if (pickLevel == "Molecule"){
                     fnode = a.getMoleculeNode();
                     if (fnode == null){
                        fnode = a.getPolymerNode();
                     }
                  }
                  currentPick = fnode;
               }

               // Add the selected node to the Tree View
               if (currentPick != null){

                  if (controlButton){
                     fpanel.getTree().toggleSelection(currentPick);
                  } else if (currentPick != previousPick){
                     fpanel.getTree().onlySelection(currentPick);
                  }

                  // Color the Current Pick by Picking Color
                  fpanel.getGraphics3D().updateScene(currentPick, false, false, 0, true,
                     GraphicsContext.byPICK);

                  // The Shift + Pick behavior is currently not very intuitive...

                  /*else if (shiftButton){
                     if (previousPick != null && previousPick != currentPick){
                        fpanel.getTree().groupSelection(currentPick, previousPick);
                        previousPick = currentPick;
                     } else{
                        fpanel.getTree().onlySelection(currentPick);
                        previousPick = currentPick;
                     }
                                  } */
               }

               // Remove picking color from the previousPick
               if (previousPick != null && previousPick != currentPick){
                  previousPick.switchColor(GraphicsContext.byREVERT, null, null);
               }
               previousPick = currentPick;
            }
         }
      }
   }

   /* A few static variables for reuse */
   private double a[] = new double[3];
   private double b[] = new double[3];
   private double c[] = new double[3];
   private double d[] = new double[3];

   private Transform3D sys3d = new Transform3D();
   private Vector3d syspos = new Vector3d();
   private Vector3d atpos = new Vector3d();

   /**
    *
    * @param measureLevel
    * @param atoms
    */
   private void measure(){
      Interpreter beanshell = fpanel.getBeanShell();

      if (pickLevel == "Measure Distance"){
         if (atomcache.size() < 2){
            return;
         }
         Atom a1 = (Atom) atomcache.get(0);
         Atom a2 = (Atom) atomcache.get(1);
         distance(a1, a);
         distance(a2, b);
         double dist = ffe.lang.AtomUtils.dist(a, b);
         beanshell.print("\nDistance\t" + a1.xyzindex + ", " + a2.xyzindex + ":   \t" + dist);
      } else if (pickLevel == "Measure Angle"){
         if (atomcache.size() < 3){
            return;
         }
         Atom a1 = (Atom) atomcache.get(0);
         Atom a2 = (Atom) atomcache.get(1);
         Atom a3 = (Atom) atomcache.get(2);
         distance(a1, a);
         distance(a2, b);
         distance(a3, c);
         double angle = ffe.lang.AtomUtils.bondangle(a, b, c);
         angle = Math.toDegrees(angle);
         beanshell.print("\nAngle\t" + a1.xyzindex + ", " + a2.xyzindex + ", " + a3.xyzindex +
            ":   \t" + angle);
      } else if (pickLevel == "Measure Dihedral"){
         if (atomcache.size() < 4){
            return;
         }
         Atom a1 = (Atom) atomcache.get(0);
         Atom a2 = (Atom) atomcache.get(1);
         Atom a3 = (Atom) atomcache.get(2);
         Atom a4 = (Atom) atomcache.get(3);
         distance(a1, a);
         distance(a2, b);
         distance(a3, c);
         distance(a4, d);
         double dihedral = ffe.lang.AtomUtils.dihedralangle(a, b, c, d);
         dihedral = Math.toDegrees(dihedral);
         beanshell.print("\nDihedral\t" + a1.xyzindex + ", " + a2.xyzindex + ", " + a3.xyzindex +
            ", " + a4.xyzindex + ":\t" + dihedral);
      }
      beanshell.refresh();
      JFrame frame = beanshell.getFrame();
      if (!frame.isShowing()){
         frame.show();
      }
      count = 0;
   }

   private void distance(Atom atom, double pos[]){
      atom.getFSystemNode().getTransformGroup().getTransform(sys3d);
      sys3d.get(syspos);
      sys3d.setScale(1.0d);
      sys3d.setTranslation(new Vector3d(0, 0, 0));
      atom.getV3D(atpos);
      sys3d.transform(atpos);
      atpos.add(syspos);
      atpos.get(pos);
   }
}
