/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.behaviors;

import java.util.*;
import java.util.List;
import javax.media.j3d.*;

import java.awt.Cursor;

import ffe.lang.*;
import ffe.panels.*;

/**
 * The UpdateBehavior class allows updating the Java3D Universe between redraws.
 */
public class UpdateBehavior
   extends Behavior{

   private Graphics3D graphics;
   private ArrayList nodesToUpdate = null;
   private ArrayList nodesCache = null;

   private boolean doTransform, doView, doColor;
   private boolean doTransformCache, doViewCache, doColorCache;
   private int view, color;
   private int viewCache, colorCache;
   private WakeupOnBehaviorPost postid;

   /**
    * Constructor
    * @param b Bounds of this behavior
    * @param g 3D Canvas
    */
   public UpdateBehavior(Bounds b, Graphics3D g){
      setSchedulingBounds(b);
      graphics = g;
   }

   private void checkJava3D(){
      // Initialize Atom Java3D Objects
      Hashtable initHash = new Hashtable();
      if (view == GraphicsContext.BALLSTICK || view == GraphicsContext.SPACEFILL ||
         view == GraphicsContext.TUBE || view == GraphicsContext.RMIN){
         for (ListIterator li = nodesToUpdate.listIterator(); li.hasNext(); ){
            FNode nodeToUpdate = (FNode) li.next();
            if (nodeToUpdate == null){
               continue;
            }
            if (nodeToUpdate instanceof Root){
               Enumeration e = nodeToUpdate.children();
               while (e.hasMoreElements()){
                  FSystem t = (FSystem) e.nextElement();
                  t.initAtoms(t);
               }
            } else{
               List nodeAtoms = nodeToUpdate.getAtomList();
               if (nodeAtoms == null || nodeAtoms.size() == 0){
                  continue;
               }
               FSystem sys = nodeToUpdate.getFSystemNode();
               if (initHash.containsKey(sys)){
                  ArrayList atomsToInit = (ArrayList) initHash.get(sys);
                  atomsToInit.addAll(nodeAtoms);
               } else{
                  ArrayList atomsToInit = new ArrayList(nodeAtoms);
                  initHash.put(sys, atomsToInit);
               }
            }
         }
         for (Enumeration e = initHash.keys(); e.hasMoreElements(); ){
            FSystem sys = (FSystem) e.nextElement();
            sys.initAtoms((List) initHash.get(sys));
         }
      }

      // Initialize Bond Java3D Objects
      initHash = new Hashtable();
      if (view == GraphicsContext.BALLSTICK || view == GraphicsContext.TUBE){
         for (ListIterator li = nodesToUpdate.listIterator(); li.hasNext(); ){
            FNode nodeToUpdate = (FNode) li.next();
            if (nodeToUpdate == null){
               continue;
            }
            if (nodeToUpdate instanceof Root){
               Enumeration e = nodeToUpdate.children();
               while (e.hasMoreElements()){
                  FSystem t = (FSystem) e.nextElement();
                  t.initBonds(t);
               }
            } else{
               List nodeBonds = nodeToUpdate.getBondList();
               if (nodeBonds == null || nodeBonds.size() == 0){
                  continue;
               }
               FSystem sys = nodeToUpdate.getFSystemNode();
               if (initHash.containsKey(sys)){
                  ArrayList bondsToInit = (ArrayList) initHash.get(sys);
                  bondsToInit.addAll(nodeBonds);
               } else{
                  ArrayList bondsToInit = new ArrayList(nodeBonds);
                  initHash.put(sys, bondsToInit);
               }
            }
         }
         for (Enumeration e = initHash.keys(); e.hasMoreElements(); ){
            FSystem sys = (FSystem) e.nextElement();
            sys.initBonds((List) initHash.get(sys));
         }
      }
   }

   /**
    * This method is called by Java3D after a graphics operation is loaded in the arm
    * and postID is called
    * @param parm1
    */
   public void processStimulus(Enumeration parm1){
      synchronized (nodesToUpdate){
         if (!doTransform){
            graphics.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         }

         // Make sure we have created Java3D objects for some View Modes
         if (doView){
            checkJava3D();
         }

         // Perform the requested rendering operation
         String viewString = GraphicsContext.toViewString(view);
         String colorString = GraphicsContext.toColorString(color);
         for (ListIterator li = nodesToUpdate.listIterator(); li.hasNext(); ){
            FNode nodeToUpdate = (FNode) li.next();

            if (nodeToUpdate == null){
               continue;
            }

            if (doTransform){
               if (nodeToUpdate instanceof FGroup){
                  FGroup t = (FGroup) nodeToUpdate;
                  t.updateAtoms();
                  t.updateBonds();
               }
            }

            if (doView){
               if (view == GraphicsContext.VRML){
                  if (!(nodeToUpdate instanceof Root || nodeToUpdate instanceof FSystem)){
                     return;
                  }
               }
               nodeToUpdate.switchView(view);
               graphics.getStatusBar().setText("  Style \"" + viewString + "\" Applied to " +
                  nodeToUpdate.toString());
            }

            if (doColor){
               nodeToUpdate.switchColor(color, null, null);
               graphics.getStatusBar().setText("  Color by \"" + colorString + "\" Applied to " +
                  nodeToUpdate.toString());
            }

         }
      }

      nodesToUpdate = null;
      graphics.setCursor(Cursor.getDefaultCursor());
      wakeupOn(postid);

      if (nodesCache != null){
         nodesToUpdate = nodesCache;
         doTransform = doTransformCache;
         doView = doViewCache;
         view = viewCache;
         doColor = doColorCache;
         color = colorCache;
         nodesCache = null;
         postId(1);
      }
   }

   /**
    * This node arms UpdateBehavior with a graphics operation to carry out
    * @param node Node where the operation will be performed
    * @param t True for a change in atomic position
    * @param v True for a change in rendering method
    * @param vtype The rendering method to use
    * @param c True for a change in rendering color
    * @param ctype The coloring method to use
    */
   public void arm(ArrayList nodes, boolean t, boolean v, int vtype, boolean c, int ctype){
      // If the node isn't null, the last rendering
      // operation hasn't finished so one operation will be cached
      if (nodesToUpdate != null){
         nodesCache = nodes;
         doTransformCache = t;
         doViewCache = v;
         viewCache = vtype;
         doColorCache = c;
         colorCache = ctype;
      } else{
         nodesToUpdate = nodes;
         doTransform = t;
         doView = v;
         view = vtype;
         doColor = c;
         color = ctype;
         postId(1);
      }
   }

   public void arm(FNode node, boolean t, boolean v, int vtype, boolean c, int ctype){
      ArrayList temp = new ArrayList();
      temp.add(node);
      arm(temp, t, v, vtype, c, ctype);
   }

   /**
    * Check to see if a graphics operation is pending/executing
    * @return Whether a node has been cued
    */
   public boolean isArmed(){
      if (nodesToUpdate != null){
         return true;
      }
      return false;
   }

   public boolean cacheFull(){
      if (nodesCache != null){
         return true;
      }
      return false;
   }

   /**
    * Initialize this behavior to respond to postID messages
    */
   public void initialize(){
      postid = new WakeupOnBehaviorPost(this, 1);
      wakeupOn(postid);
   }
}
