/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.ArrayList;

/**
 * The Angle class represents an angle formed between three linearly bonded atoms.
 */
public final class Angle
   extends Term{

   private static Atom a1, a2, a3;
   public static ArrayList anglepool = new ArrayList();

   /**
    * Angle contructor
    * @param b1 Bond that forms one leg of the angle
    * @param b2 Bond that forms the other leg of the angle
    */
   public Angle(Bond b1, Bond b2){
      super();
      bonds = new Bond[2];
      bonds[0] = b1;
      bonds[1] = b2;
      b1.setAngleWith(b2);
      b2.setAngleWith(b1);
      a2 = b1.getCommonAtom(b2);
      a1 = b1.getOtherAtom(a2);
      a3 = b2.getOtherAtom(a2);
      atoms = new Atom[3];
      atoms[0] = a1;
      atoms[1] = a2;
      atoms[2] = a3;
      a1.setAngle(this);
      a2.setAngle(this);
      a3.setAngle(this);
      setID_Key(false);
      update();
   }

   /**
    * Angle re-contructor
    * @param b1 Bond that forms one leg of the angle
    * @param b2 Bond that forms the other leg of the angle
    */
   public void reInit(Bond b1, Bond b2){
      bonds = new Bond[2];
      bonds[0] = b1;
      bonds[1] = b2;
      b1.setAngleWith(b2);
      b2.setAngleWith(b1);
      a2 = b1.getCommonAtom(b2);
      a1 = b1.getOtherAtom(a2);
      a3 = b2.getOtherAtom(a2);
      atoms = new Atom[3];
      atoms[0] = a1;
      atoms[1] = a2;
      atoms[2] = a3;
      a1.setAngle(this);
      a2.setAngle(this);
      a3.setAngle(this);
      setID_Key(false);
      update();
   }

   /**
    * Store this Angle in a pool
    */
   public void destroy(){
      super.destroy();
      removeFromParent();
      anglepool.add(this);
   }

   private static double vi[] = new double[3];
   private static double vj[] = new double[3];
   private static double vk[] = new double[3];

   /**
    * Update recomputes the angle's value
    */
   public void update(){
      atoms[0].getXYZ(vi);
      atoms[1].getXYZ(vj);
      atoms[2].getXYZ(vk);
      setValue(Math.toDegrees(AtomUtils.bondangle(vi, vj, vk)));
   }

   /**
    * Finds the common bond between <b>this</b> angle and another
    * @param a An Angle that may have a common bond with <b>this</b> angle
    * @return The common Bond between this Angle and Angle a, or null if this == a or no common bond exists
    */
   public Bond getCommonBond(Angle a){
      if (a == this){
         return null; //Comparing an angle to itself returns null
      }
      if (a.bonds[0] == bonds[0]){
         return bonds[0];
      }
      if (a.bonds[0] == bonds[1]){
         return bonds[1];
      }
      if (a.bonds[1] == bonds[0]){
         return bonds[0];
      }
      if (a.bonds[1] == bonds[1]){
         return bonds[1];
      }
      return null; //No common bond found
   }

   /**
    * Finds the other bond that makes up <b>this</b> angle
    * @param b The bond to find the opposite of
    * @return The other Bond that makes up this Angle, or null if Bond b is not part of this Angle
    */
   public Bond getOtherBond(Bond b){
      if (b == getBond(0)){
         return getBond(1);
      }
      if (b == getBond(1)){
         return getBond(0);
      }
      return null; //b not found in angle
   }
}
