/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.text.*;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import ffe.panels.*;

/**
 * The Atom class represents a single atom.
 */
public final class Atom
   extends FNode{

   // Identification Info
   public int xyzindex; // Index in TINKER Files
   private String id = null;

   // Identifiers currently used for PDB Files
   private String residue = null;
   private int residueNumber = -1;
   private String chain = null;

   // Coordinates
   private Vector3d pos3d;
   private Vector trajectory;

   // Molecular Mechanics Info
   private MMAtom mm;

   private double[] induced;
   private double inducedMag;
   private double[] velocity;
   private double velMag;
   private double[] acceleration;
   private double accelMag;
   private double[] force;
   private double forceMag;
   private static double maxInduced, maxVel, maxAccel, maxForce;

   // Connectivity Informaton, for example an atom "knows" what bonded molecular
   // mechanics terms it contributes to.
   private ArrayList bonds;
   private ArrayList angles;
   private ArrayList dihedrals;

   // Java 3D
   //
   // There are ways to reduce Java3D memory usage - the geometry
   // for all atoms in a system could be stored in one large GeometryArray.
   // However, this will make the code more complicated and may not be
   // necessary in the long run.
   //
   /*
      moltg (from FSystem)
            |
      transformGroup(transform3D)
       _____|_________________________
      |           |                  |
    sphere   cylTG(cylT3D)     coneTG(coneT3D)
                  |                  |
               cylinder             cone

    The Cylinder and Cone (to form a Vector)
    need an extra translation and rotation
    */
   private Shape3D sphere, cylinder, cone;
   private TransformGroup transformGroup, cylTG, coneTG;
   private Transform3D transform3D, cylT3D, coneT3D;

   // All of an Atom's shapes are going to have the same appearance for now
   private Appearance appearance;
   private int polygonType = GraphicsContext.FILL;
   private int colorType = GraphicsContext.byATOM;
   private Color3f currentCol, previousCol;
   private Color3f userColor;
   private double scale;

   // True if this Atom's J3D elements are attached to the FSystem BranchGroup
   private boolean attached;

   // True if this Atom is visible
   private boolean visible;

   // True if this Atom's J3D transforms need to updated before making it visible
   private boolean stale;

   // True if this Atom is not a Hydrogen or Hydrogens are being rendered
   private boolean showHydrogen;

   // True if this Atom's vector object is being rendered
   private boolean showVector;

   // The type of vector to be displayed
   private int vectorType = GraphicsContext.FORCE;
   private int vectorLength = GraphicsContext.UNIT;

   // The current viewtype
   private int view = 0;
   private int detail = 3;

   // A few static objects that are used in computing transforms
   private static AxisAngle4d rot = new AxisAngle4d(); // Rotation from y-axis to induced dipole
   private static double angle;

   public static ArrayList atompool = new ArrayList();

   /**
    * Constructor
    * @param id The Atom's PDB/Type identifier
    */
   public Atom(String s){
      super(s);
      id = s;
   }

   /**
    * Constructor
    */
   public Atom(String id, MMAtom mmdata, double[] d){
      this(id);
      mm = mmdata;
      pos3d = new Vector3d(d);
      setAllowsChildren(false);
      if (mm != null){
         currentCol = previousCol = Graphics3D.colors[mm.atomic_num];
      } else{
         currentCol = previousCol = GraphicsContext.toAtomColor(id);
      }
      userColor = GraphicsContext.whiteColor;
      attached = false;
      visible = false;
      showHydrogen = true;
      stale = false;
      showVector = false;
   }

   /**
    * Constructor used when parsing XYZ files
    */
   public Atom(int xyznum, String id, MMAtom mmdata, double[] d){
      this(id, mmdata, d);
      xyzindex = xyznum;
   }

   /**
    * Constructor used when parsing XYZ files
    */
   public Atom(int xyznum, String id, MMAtom mmdata, double[] d, String r, int n, String c){
      this(xyznum, id, mmdata, d);
      residue = r;
      residueNumber = n;
      chain = c;
   }

   public final void addTrajectoryCoords(Vector3d coords, int position){
      if (trajectory == null){
         trajectory = new Vector();
         trajectory.add(0, pos3d);
      }
      trajectory.add(position, coords);
   }

   /**
    * Store this Atom in a pool
    */
   public final void destroy(){
      super.destroy();
      removeFromParent();
      if (bonds != null){
         bonds.clear();
      }
      if (angles != null){
         angles.clear();
      }
      if (dihedrals != null){
         dihedrals.clear();
      }
      if (trajectory != null){
         trajectory.clear();
      }
      currentCol = null;
      trajectory = null;
      atompool.add(this);
   }

   /**
    * Overidden equals method
    * @param object The Object to compare with <b>this</b>
    * @return True if <b>this</b> atom and object do not reference the same object, are of the same class, and have the same id
    */
   public final boolean equals(Object object){
      if (this == object){
         return true;
      } else{
         return false;
      }
   }

   public void getVelocity(double[] t){
      if (velocity == null || t == null){
         return;
      }
      t[0] = velocity[0];
      t[1] = velocity[1];
      t[2] = velocity[2];
   }

   public void getAcceleration(double[] t){
      if (acceleration == null || t == null){
         return;
      }
      t[0] = acceleration[0];
      t[1] = acceleration[1];
      t[2] = acceleration[2];
   }

   public void getForce(double[] t){
      if (force == null || t == null){
         return;
      }
      t[0] = force[0];
      t[1] = force[1];
      t[2] = force[2];
   }

   public static void setMaxForce(double f){
      maxForce = f;
   }

   public static void setMaxVelocity(double f){
      maxVel = f;
   }

   public static void setMaxAcceleration(double f){
      maxAccel = f;
   }

   public static void setMaxInduced(double f){
      maxInduced = f;
   }

   public static double getMaxForce(){
      return maxForce;
   }

   public static double getMaxVelocity(){
      return maxVel;
   }

   public static double getMaxAcceleration(){
      return maxAccel;
   }

   public static double getMaxInduced(){
      return maxInduced;
   }

   public void getInducedDipole(double[] t){
      if (induced == null || t == null){
         return;
      }
      t[0] = induced[0];
      t[1] = induced[1];
      t[2] = induced[2];
   }

   public void setVeclocity(double x, double y, double z){
      if (velocity == null){
         velocity = new double[3];
      }
      velocity[0] = x;
      velocity[1] = y;
      velocity[2] = z;

      velMag = AtomUtils.r(velocity);
      if (velMag > maxVel){
         maxVel = velMag;
      }
   }

   public void setAcceleration(double x, double y, double z){
      if (acceleration == null){
         acceleration = new double[3];
      }
      acceleration[0] = x;
      acceleration[1] = y;
      acceleration[2] = z;

      accelMag = AtomUtils.r(acceleration);
      if (accelMag > maxAccel){
         maxAccel = accelMag;
      }

   }

   public void setForce(double x, double y, double z){
      if (force == null){
         force = new double[3];
      }
      force[0] = x;
      force[1] = y;
      force[2] = z;

      forceMag = AtomUtils.r(force);
      if (forceMag > maxForce){
         maxForce = forceMag;
      }
   }

   public void setInducedDipole(double x, double y, double z){
      if (induced == null){
         induced = new double[3];
      }
      induced[0] = x;
      induced[1] = y;
      induced[2] = z;

      inducedMag = AtomUtils.r(induced);
      if (inducedMag > maxInduced){
         maxInduced = inducedMag;
      }

   }

   public final ArrayList getAngles(){
      return angles;
   }

   public final Appearance getAppearance(){
      return appearance;
   }

   /**
    * Gets the Atomic Number
    * @return Atomic Number
    */
   public final int getAtomicNum(){
      return mm.atomic_num;
   }

   public final List getAtomList(){
      ArrayList atoms = new ArrayList();
      atoms.add(this);
      return atoms;
   }

   public final String getAtomType(){
      if (mm.stused){
         return mm.specifictype;
      }
      return mm.type;
   }

   public final Bond getBond(Atom a){
      Bond b;
      if (bonds == null){
         return null;
      }
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         if (b.getOtherAtom(a) == this){
            return b;
         }
      }
      return null;
   }

   public final int getBondCount(){
      if (bonds == null){
         return 0;
      } else{
         return bonds.size();
      }

   }

   /**
    * Gets the list of the Bonds <b>this</b> Atom helps to form
    * @return A list of the bonds this atom helps to form
    */
   public final ArrayList getBonds(){
      return bonds;
   }

   /**
    * Gets the partial atomic charge
    * @return partial atomic charge
    */
   public final float getCharge(){
      return mm.charge;
   }

   public final Color3f getColor(){
      return currentCol;
   }

   public final ArrayList getDihedrals(){
      return dihedrals;
   }

   /**
    * Gets the Epsilon value
    * @return Epsilon value
    */
   public final float getEpsilon(){
      return mm.epsilon;
   }

   /**
    * Gets the energy gradient
    * @return energy gradient
    */
   //public Vector3d getGradient(){ return new Vector3d(gradient); }
   /**
    * Gets the Atomic Hybridization
    * @return Atomic Hybridization
    */
   public final int getHybridization(){
      return mm.hybridization;
   }

   /**
    * Gets the ID
    * @return ID
    */
   public final String getID(){
      return id;
   }

   /**
    * Get the residue name
    * @return String
    */
   public final String getResidueName(){
      if (residue != null){
         return residue;
      }
      return getResidueNode().getName();
   }

   public final int getResidueNumber(){
      return residueNumber;
   }

   /**
    * Get the chain name
    * @return String
    */
   public final String getChain(){
      if (chain != null){
         return chain;
      }
      return getPolymerNode().getName();
   }

   /**
    * Gets the atom ID
    * @return atom ID
    */
   public final String getIdent(){
      String s = new String(mm.description);
      return s;
   }

   public final Group getJ3D(){
      return transformGroup;
   }

   /**
    * Gets the atom Key
    * @return atom Key
    */
   public final String getKey(){
      String s = new String(mm.type);
      return s;
   }

   /**
    * Gets the Atomic Mass
    * @return Atomic Mass
    */
   public final float getMass(){
      return mm.mass;
   }

   public final int getNumAngles(){
      return angles.size();
   }

   /**
    * Gets the number of atoms bonded to <b<this</b> Atom
    * @return Number of atoms bonded to this atom
    */
   public final int getNumBonds(){
      return bonds.size();
   }

   public final int getNumDihedrals(){
      return dihedrals.size();
   }

   /**
    * Gets the Sigma value
    * @return Sigma value
    */
   public float getSigma(){
      return mm.sigma;
   }

   public Vector3f getTrajectoryCoords(int position){
      return (Vector3f) trajectory.get(position);
   }

   public int getTrajectoryLength(){
      return trajectory.size();
   }

   /**
    * Gets the Atom's Cartesian Coordinates
    * return The Cartesian Coordinates
    */
   public void getV3D(Vector3d temp){
      temp.set(pos3d);
   }

   //public final  Vector3f getnewV3D(){ return new Vector3f(v3f); }
   /**
    * Gets the van der Waals radius
    * return van der Waals radius
    */
   public float getVDWR(){
      return mm.sigma;
   }

   /**
    * Gets the x coordinate
    * @return x coordinate
    */
   public double getX(){
      return pos3d.x;
   }

   /**
    * Gets the Atom's Cartesian Coordinates
    * return The Cartesian Coordinates
    */

   /*public double[] getXYZ() {
       double[] temp = new double[3];
       pos3d.get(temp);
       return temp;
        }*/

   public void getXYZ(double[] xyz){
      pos3d.get(xyz);
   }

   /**
    * Gets the XYZ Index
    * @return XYZ Index
    */
   public final int getXYZIndex(){
      return xyzindex;
   }

   /**
    * Gets the y coordinate
    * @return y coordinate
    */
   public final double getY(){
      return pos3d.y;
   }

   /**
    * Gets the z coordinate
    * @return z coordinate
    */
   public final double getZ(){
      return pos3d.z;
   }

   /**
    * Checks to see if an Atom is bonded to <b>this</b> Atom
    * @param a Atom to check
    * @return True is Atom a is bonded to <b>this</b>this atom
    */
   public final boolean isBonded(Atom a){
      Bond b;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         if (b.getOtherAtom(a) == this){
            return true;
         }
      }
      return false;
   }

   public boolean isVisible(){
      return visible;
   }

   public boolean isBondVisible(){
      if (getBonds() == null){
         return false;
      }
      for (ListIterator li = getBonds().listIterator(); li.hasNext(); ){
         Bond b = (Bond) li.next();
         if (b.isVisible()){
            return true;
         }
      }
      return false;
   }

   public boolean showHydrogen(){
      return showHydrogen;
   }

   /**
    * Gets whether or not the Atom is under-constrained
    * @return True if the atom is under-constrained (ie has can accept bonds)
    */
   public boolean isDangeling(){
      Integer hybrid = (Integer) hybridTable.get("" + mm.atomic_num);
      if (hybrid == null){
         return false;
      }
      int result = hybrid.compareTo(new Integer(bonds.size()));
      return (result > 0);
   }

   public boolean initJ3D(){
      if (attached){
         return false;
      } else if (sphere != null){
         attached = true;
         update();
         return true;
      }
      appearance = Graphics3D.getAppearance(currentCol, GraphicsContext.FILL);
      detail = GraphicsContext.detail;
      sphere = Graphics3D.createSphere(appearance, detail);
      sphere.setUserData(this);

      transform3D = new Transform3D();
      transform3D.setTranslation(pos3d);
      scale = 1.0d;
      transform3D.setScale(scale);
      transformGroup = new TransformGroup(transform3D);
      transformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      transformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      transformGroup.addChild(sphere);

      setVisible(false);
      attached = true;
      initVec();
      return true;
   }

   public void initVec(){
      appearance = Graphics3D.getAppearance(currentCol, GraphicsContext.FILL);
      detail = GraphicsContext.detail;

      cylinder = Graphics3D.createCylinder(appearance, detail);
      cylinder.setUserData(this);
      cylT3D = new Transform3D();
      cylTG = new TransformGroup(cylT3D);
      cylTG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      cylTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      cylTG.addChild(cylinder);
      transformGroup.addChild(cylTG);

      cone = Graphics3D.createCone(appearance, detail);
      cone.setUserData(this);
      coneT3D = new Transform3D();
      coneT3D.setScale(coneScale);
      coneTG = new TransformGroup(coneT3D);
      coneTG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      coneTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      coneTG.addChild(cone);
      transformGroup.addChild(coneTG);
      updateTransforms();
   }

   /**
    * Add a vector to the Atom's current position vector
    * @param d Vector to add to the current position
    */
   public final void move(double[] d){
      pos3d.x += d[0];
      pos3d.y += d[1];
      pos3d.z += d[2];
      if (trajectory != null){
         Vector3d vt;
         ListIterator li = trajectory.listIterator();
         // Skip the first frame, done above
         vt = (Vector3d) li.next();
         while (li.hasNext()){
            vt = (Vector3d) li.next();
            vt.x += d[0];
            vt.y += d[1];
            vt.z += d[2];
         }
      }
   }

   /**
    * Moves the atom to the specified location
    * @param d Location to move <b>this</b> Atom to
    */
   public final void moveto(double[] d){
      pos3d.set(d);
   }

   public final void moveto(double x, double y, double z){
      pos3d.set(x, y, z);
   }

   /**
    * Prints the atom identity and Cartesian coordinates to stout
    */
   public final void print(){
      System.out.println(toString());
   }

   public final void reInit(int index, String i, MMAtom mmdata, double[] d){
      xyzindex = index;
      id = i;
      setName(id);
      mm = mmdata;
      pos3d.set(d);
      if (mm != null){
         currentCol = previousCol = Graphics3D.colors[mm.atomic_num];
      } else{
         currentCol = previousCol = GraphicsContext.toAtomColor(id);
      }
      userColor = GraphicsContext.whiteColor;
      if (sphere != null){
         appearance = Graphics3D.getAppearance(currentCol, GraphicsContext.FILL);
         sphere.setAppearance(appearance);
         sphere.setUserData(this);
         if (cylinder != null){
            cylinder.setAppearance(appearance);
            cylinder.setUserData(this);
            cone.setAppearance(appearance);
            cone.setUserData(this);
         }
         if (detail != GraphicsContext.detail){
            switchView(GraphicsContext.detail);
         }
      }
      attached = false;
      visible = false;
      showHydrogen = true;
      showVector = false;
      stale = false;
   }

   public final void setAngle(Angle a){
      if (angles == null){
         angles = new ArrayList();
      }
      angles.add(a);
   }

   /**
    * Specify that <b>this</b> Atom is part of a Bond
    * @param b Bond that <b>this</b> Atom is part of
    */
   public final void setBond(Bond b){
      if (bonds == null){
         bonds = new ArrayList();
      }
      bonds.add(b);
      /*if ( b.getAtom(0).getKey().equals(" 0") ){
       hydrogens.add(b.getAtom(0));
         }
         if ( b.getAtom(1).getKey().equals(" 0") ){
       hydrogens.add(b.getAtom(1));
         }*/
   }

   public final void setCurrentCycle(int cycle){
      if (trajectory == null){
         return;
      }

      if (cycle <= 0){
         cycle = trajectory.size();
      } else if (cycle > trajectory.size()){
         cycle = 1;
      }
      if (cycle <= trajectory.size()){
         pos3d = (Vector3d) trajectory.get(cycle - 1);
      }
   }

   public final void setDihedral(Dihedral d){
      if (dihedrals == null){
         dihedrals = new ArrayList();
      }
      dihedrals.add(d);
   }

   public final void setXYZIndex(int index){
      xyzindex = index;
   }

   public void setSelected(boolean b){
      selected = b;
   }

   public final void switchColor(int ctype, Color3f newCol, Material newMat){
      if (ctype != GraphicsContext.byPICK && ctype != GraphicsContext.byREVERT){
         colorType = ctype;
      }
      if (ctype == GraphicsContext.byUSER){
         currentCol = previousCol = userColor;
      } else if (ctype == GraphicsContext.setUSER){
         userColor = GraphicsContext.userColor;
         currentCol = previousCol = userColor;
      } else if (ctype == GraphicsContext.byMONOCHROME){
         currentCol = previousCol = GraphicsContext.whiteColor;
      } else if (ctype == GraphicsContext.byATOM){
         newCol = Graphics3D.getColor(this, GraphicsContext.byATOM);
         if (newCol == currentCol){
            return;
         }
         currentCol = previousCol = newCol;
      } else if (ctype == GraphicsContext.bySELECT){
         if (isSelected()){
            newCol = GraphicsContext.selectionColor;
            if (newCol != currentCol){
               currentCol = newCol;
            } else{
               return;
            }
         } else{
            currentCol = previousCol;
         }
      } else if (ctype == GraphicsContext.byPICK){
         newCol = GraphicsContext.pickingColor;
         if (newCol != currentCol){
            currentCol = newCol;
         } else{
            return;
         }
      } else if (ctype == GraphicsContext.byREVERT){
         if (GraphicsContext.highlightSelections && isSelected()){
            currentCol = GraphicsContext.selectionColor;
         } else{
            currentCol = previousCol;
         }
      } else if (ctype == GraphicsContext.byCHARGE){
         newCol = Graphics3D.getColor(this, GraphicsContext.byCHARGE);
         if (newCol == currentCol){
            return;
         }
         currentCol = previousCol = newCol;
      } else if (ctype == GraphicsContext.byVECTOR){
         newCol = Graphics3D.getColor(this, ctype);
         if (newCol == currentCol){
            return;
         }
         currentCol = previousCol = newCol;
      }

      // Check for a Color Choice sent from a higher level structure (residue,polymer,etc)
      else if (newCol != null){
         if (newCol == currentCol){
            return;
         }
         currentCol = previousCol = newCol;
      }

      appearance = Graphics3D.getAppearance(currentCol, polygonType);

      if (sphere != null && visible && showHydrogen){
         sphere.setAppearance(appearance);
         if (cylinder != null && showVector){
            cylinder.setAppearance(appearance);
            cone.setAppearance(appearance);
         }
      }

      if (bonds != null){
         Bond bond;
         for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
            bond = (Bond) li.next();
            bond.switchColor(this);
         }
      }
   }

   static private Appearance nullAp;
   static{
      nullAp = new Appearance();
      RenderingAttributes ra = new RenderingAttributes();
      ra.setVisible(false);
      nullAp.setRenderingAttributes(ra);
   }

   public final void switchView(int viewtype){
      if (viewtype == GraphicsContext.ACCELERATION || viewtype == GraphicsContext.VELOCITY ||
         viewtype == GraphicsContext.INDUCED || viewtype == GraphicsContext.PERMANENT ||
         viewtype == GraphicsContext.FORCE){
         vectorType = viewtype;
         if (getMagnitude(false) >= 0.0){
            showVector = true;
            stale = true;
         } else{
            showVector = false;
         }
         setVisible(visible);
         return;
      } else if (viewtype == GraphicsContext.ABSOLUTE || viewtype == GraphicsContext.RELATIVE ||
         viewtype == GraphicsContext.UNIT){
         vectorLength = viewtype;
         if (getMagnitude(false) >= 0.0){
            showVector = true;
            stale = true;
         } else{
            showVector = false;
         }
         setVisible(visible);
         return;
      } else if (viewtype != GraphicsContext.DETAIL && viewtype != GraphicsContext.HYDROGENS &&
         viewtype != GraphicsContext.VECTORS){
         view = viewtype;
      }

      if (viewtype == GraphicsContext.VECTORS){
         if (showVector){
            showVector = false;
         } else{
            showVector = true;
         }
         setVisible(visible);
         return;
      } else if (viewtype == GraphicsContext.HYDROGENS){
         if (mm.atomic_num == 1){
            if (!showHydrogen){
               showHydrogen = true;
               switchView(view);
               return;
            } else{
               showHydrogen = false;
               setVisible(false);
            }
         }
      } else if (view == GraphicsContext.RESTRICT){
         if (!isSelected()){
            setVisible(false);
         }
      } else if (view != GraphicsContext.WIREFRAME && view != GraphicsContext.INVISIBLE){

         if (viewtype == GraphicsContext.DETAIL){
            int newdetail = GraphicsContext.detail;
            if (newdetail != detail){
               detail = newdetail;
               Geometry geom = Graphics3D.getSphereGeom(detail);
               sphere.removeAllGeometries();
               sphere.addGeometry(geom);
            }
         }

         if (sphere != null){
            if (view == GraphicsContext.POINTS || view == GraphicsContext.LINES ||
               view == GraphicsContext.FILL){
               polygonType = view;
               appearance = Graphics3D.getAppearance(currentCol, polygonType);
            }
         }

         // See what size the sphere should be
         if (view == GraphicsContext.BALLSTICK){
            scale = Graphics3D.vdw[mm.atomic_num] / 5.0d * GraphicsContext.radius;
         } else if (view == GraphicsContext.SPACEFILL){
            scale = Graphics3D.vdw[mm.atomic_num] * GraphicsContext.radius;
         } else if (view == GraphicsContext.TUBE){
            scale = 0.2d * GraphicsContext.radius;
         } else if (view == GraphicsContext.RMIN){
            scale = (double) mm.sigma * GraphicsContext.radius;
         }

         transform3D.setScale(scale);
         transformGroup.setTransform(transform3D);
         setVisible(true);
      } else{
         if (sphere != null){
            setVisible(false);
         }
      }
   }

   private static Point3d point3d = new Point3d();
   public void drawLabel(Graphics3D graphics, J3DGraphics2D g2d, Node node, boolean labelAtoms,
      boolean labelResidues){
      if (labelAtoms){
         point3d.x = getX();
         point3d.y = getY();
         point3d.z = getZ();
         Point2d point = graphics.getScreenCoordinate(node, point3d);
         g2d.drawString(toShortString(), (float) point.x, (float) point.y);
      }
   }

   public double getMagnitude(boolean relative){
      if (vectorType == GraphicsContext.ACCELERATION){
         if (acceleration == null){
            return -1.0;
         } else{
            if (relative){
               return accelMag / maxAccel;
            } else{
               return accelMag;
            }
         }
      } else if (vectorType == GraphicsContext.VELOCITY){
         if (velocity == null){
            return -1.0;
         } else{
            if (relative){
               return velMag / maxVel;
            } else{
               return velMag;
            }
         }
      } else if (vectorType == GraphicsContext.INDUCED){
         if (induced == null){
            return -1.0;
         } else{
            if (relative){
               return inducedMag / maxInduced;
            } else{
               return inducedMag;
            }
         }
      } else if (vectorType == GraphicsContext.FORCE){
         if (force == null){
            return -1.0;
         } else{
            if (relative){
               return forceMag / maxForce;
            } else{
               return forceMag;
            }

         }
      }
      return -1.0;
   }

   public final void setVisible(boolean v){
      visible = v;
      // The Java3D objects haven't been created
      if (sphere == null){
         return;
      }
      // Hide the atom & vector
      if (!visible){
         sphere.setAppearance(nullAp);
         sphere.setPickable(false);
         if (cylinder != null){
            cylinder.setAppearance(nullAp);
            cylinder.setPickable(false);
            cone.setAppearance(nullAp);
            cone.setPickable(false);
         }
      } else{
         if (!showHydrogen){
            return;
         }
         // If the transform hasn't been updated during dynamics or trajectory
         // playback, it needs to be refreshed
         if (stale){
            update();
         }
         sphere.setAppearance(appearance);
         sphere.setPickable(true);
         // Make the vector visible
         if (cylinder != null){
            if (showVector){
               cylinder.setAppearance(appearance);
               cylinder.setPickable(true);
               cone.setAppearance(appearance);
               cone.setPickable(true);
            } else{
               cylinder.setAppearance(nullAp);
               cylinder.setPickable(false);
               cone.setAppearance(nullAp);
               cone.setPickable(false);
            }
         }
      }
   }

   /* public final  String toPDBString(int num){
    String blanks = new String("                   ");
    String xS, yS, zS;
    int xi,yi,zi;
    StringBuffer temp = new StringBuffer((new Integer(num)).toString());
    temp.insert(0, blanks.substring(0,5 - temp.length()));
    StringBuffer record = new StringBuffer("ATOM  " + temp.toString());
    temp = new StringBuffer(id);
    temp.insert(0, blanks.substring(0,5 - temp.length()));
    record.append(temp.toString() + " " + myResidue.toString() + " " + myPolymer.toString());
    temp = new StringBuffer((new Integer(myResidueNumber)).toString());
    temp.insert(0, blanks.substring(0,5 - temp.length()));
    record.append(temp.toString() + "   ");
    xS = new String("" + v3d.x); xi = xS.indexOf("."); xS = xS.substring(0,xi + 4);
    temp = new StringBuffer(xS); temp.insert(0, blanks.substring(0, 8 - xS.length()));
    xS = temp.toString();
    yS = new String("" + v3d.y); yi = yS.indexOf("."); yS = yS.substring(0,yi + 4);
    temp = new StringBuffer(yS); temp.insert(0, blanks.substring(0, 8 - yS.length()));
    yS = temp.toString();
    zS = new String("" + v3d.z); zi = zS.indexOf("."); zS = zS.substring(0,zi + 4);
    temp = new StringBuffer(zS); temp.insert(0, blanks.substring(0, 8 - zS.length()));
    zS = temp.toString();
    record.append(" " + xS + yS + zS);
    return record.toString();
     }*/

   /**
    * Overidden toString method
    * @return The string: "Atom PDB ID, OPLS ID"
    */
   private static NumberFormat form = NumberFormat.getInstance();
   public String toString(){
      DecimalFormat dform = null;
      StringBuffer s = new StringBuffer(xyzindex + "-" + id + " ");
      if (form instanceof DecimalFormat){
         dform = (DecimalFormat) form;
         dform.applyPattern("###0.00");
         s.append(" (" + dform.format(pos3d.x) + ", " + dform.format(pos3d.y) + ", " +
            dform.format(pos3d.z) + ")");
      }
      return s.toString();
   }

   public String toShortString(){
      StringBuffer s = new StringBuffer("" + xyzindex);
      s.append("-");
      s.append(id);
      return s.toString();
   }

   /**
    * Atomic Positions are not updated unless the atom is visible -> this
    * allows better performance for rendering partial structures during an
    * interactive dynamics run or during trajectory playback.
    */

   public void update(){
      if (visible){
         updateTransforms();
         stale = false;
      } else{
         stale = true;
      }
   }

   public void updateTransforms(){
      if (transform3D == null){
         return;
      }
      transform3D.setTranslation(pos3d);
      transformGroup.setTransform(transform3D);
      if (colorType == GraphicsContext.byVECTOR){
         switchColor(colorType, null, null);
      }
      double mag = 2.0d;
      if (vectorLength != GraphicsContext.UNIT){
         if (vectorLength == GraphicsContext.RELATIVE){
            mag = getMagnitude(true) * 2.0d;
         } else{
            mag = getMagnitude(false);
         }
      }
      if (cylTG != null && showVector){
         if (vectorType == GraphicsContext.FORCE){
            setVectorTransform3d(force, mag);
         } else if (vectorType == GraphicsContext.INDUCED){
            setVectorTransform3d(induced, mag);
         } else if (vectorType == GraphicsContext.VELOCITY){
            setVectorTransform3d(velocity, mag);
         } else if (vectorType == GraphicsContext.ACCELERATION){
            setVectorTransform3d(acceleration, mag);
         }
      }
   }

   private static double coneScale = 0.3d;
   private static double[] bcross = new double[4];
   private static double[] yd = {0.0d, -1.0d, 0.0d};
   private static double[] d3 = new double[3];
   private static Vector3d scaleV3D = new Vector3d();
   private static Vector3d tV3D = new Vector3d();

   public void setVectorTransform3d(double[] orient, double mag){
      // Cylinder and Cone Rotation from the Y-axis
      angle = AtomUtils.angle(orient, yd);
      AtomUtils.cross(yd, orient, bcross);
      bcross[3] = angle + Math.PI;
      rot.set(bcross);

      // Create Local References
      double vec[] = d3;
      Vector3d transV3D = tV3D;
      double vecScale = scale;

      // Cylinder Orientation
      cylT3D.setRotation(rot);
      AtomUtils.norm(orient, vec);
      AtomUtils.scalar(vec, mag / 2.0d / vecScale, vec);
      transV3D.set(vec);
      cylT3D.setTranslation(transV3D);
      scaleV3D.set(0.2d / scale, mag / vecScale, 0.2d / scale);
      cylT3D.setScale(scaleV3D);
      cylTG.setTransform(cylT3D);

      // Cone Orientation
      coneT3D.setRotation(rot);
      AtomUtils.norm(vec, vec);
      AtomUtils.scalar(vec, (mag + coneScale / 2.0d) / vecScale, vec);
      transV3D.set(vec);
      coneT3D.setTranslation(transV3D);
      coneT3D.setScale(coneScale / vecScale);
      coneTG.setTransform(coneT3D);
   }

   public final static Hashtable hybridTable = new Hashtable();
   static{
      hybridTable.put("1", new Integer(1));
      hybridTable.put("6", new Integer(4));
      hybridTable.put("7", new Integer(3));
      hybridTable.put("8", new Integer(2));
      hybridTable.put("15", new Integer(4));
      hybridTable.put("16", new Integer(2));
      hybridTable.put("19", new Integer(0));
      hybridTable.put("26", new Integer(8));
   }
}
