/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.Hashtable;
import javax.vecmath.Color3f;

import java.awt.Color;

/**
 * The AtomUtils class is a simple math library that operates on 3-coordinate vectors.
 * The main design objectives were speed and no memory consumption.
 */
public final class AtomUtils{
   /**
    * Length in Angstroms that is added to Atomic Radii when determining if two
    * Atoms are within bonding distance
    */
   public static final float BUFF = 0.3f;

   /**
    * Atomic Bonding Hybridization
    */
   public static final int SP = 2, SP2 = 3, SP3 = 4;

   /**
    * Bonding Character
    */
   public static final int SINGLEBOND = 1, DOUBLEBOND = 2, TRIPLEBOND = 3;

   /**
    * Colors for display routines
    */
   public static final Hashtable colorlookup;
   public static final Color3f colorlist[] = new Color3f[30];

   private static final Color3f ORANGE = new Color3f(Color.orange);
   private static final Color3f RED = new Color3f(Color.red);
   private static final Color3f BLUE = new Color3f(Color.blue);
   private static final Color3f GRAY = new Color3f(Color.lightGray);
   private static final Color3f YELLOW = new Color3f(Color.yellow);
   private static final Color3f CYAN = new Color3f(Color.CYAN);
   private static final Color3f GREEN = new Color3f(Color.green);
   private static final Color3f WHITE = new Color3f(Color.white);
   private static final Color3f PINK = new Color3f(Color.pink);
   private static final Color3f MAGENTA = new Color3f(Color.magenta);

   static{

      colorlist[0] = WHITE;
      colorlist[1] = RED; // In honor of West Branch & WUSTL - Go Bears
      colorlist[2] = YELLOW; // In honor of U. of Iowa - Go Hawks
      colorlist[3] = BLUE;
      colorlist[4] = GREEN;
      colorlist[5] = CYAN;
      colorlist[6] = MAGENTA;
      colorlist[7] = GRAY;
      colorlist[8] = PINK;
      colorlist[9] = ORANGE;

      colorlist[10] = WHITE;
      colorlist[11] = RED;
      colorlist[12] = YELLOW;
      colorlist[13] = BLUE;
      colorlist[14] = GREEN;
      colorlist[15] = CYAN;
      colorlist[16] = MAGENTA;
      colorlist[17] = GRAY;
      colorlist[18] = PINK;
      colorlist[19] = ORANGE;

      colorlist[20] = WHITE;
      colorlist[21] = RED;
      colorlist[22] = YELLOW;
      colorlist[23] = BLUE;
      colorlist[24] = GREEN;
      colorlist[25] = CYAN;
      colorlist[26] = MAGENTA;
      colorlist[27] = GRAY;
      colorlist[28] = PINK;
      colorlist[29] = ORANGE;

      colorlookup = new Hashtable();

      // Water, blue of course
      colorlookup.put("Water", BLUE);

      // RNA Colors
      colorlookup.put("A", RED);
      colorlookup.put("C", MAGENTA);
      colorlookup.put("G", BLUE);
      colorlookup.put("U", YELLOW);

      // DNA Colors
      colorlookup.put("dA", RED);
      colorlookup.put("dC", MAGENTA);
      colorlookup.put("dG", BLUE);
      colorlookup.put("dT", ORANGE);

      // Amino Acid Colors
      colorlookup.put("GLY", GRAY);
      colorlookup.put("ALA", GRAY);
      colorlookup.put("VAL", GRAY);
      colorlookup.put("LEU", GRAY);
      colorlookup.put("ILE", GRAY);
      colorlookup.put("SER", BLUE);
      colorlookup.put("THR", BLUE);
      colorlookup.put("CYS", YELLOW);
      colorlookup.put("PRO", ORANGE);
      colorlookup.put("PHE", GREEN);
      colorlookup.put("TYR", GREEN);
      colorlookup.put("TRP", GREEN);
      colorlookup.put("HIS", BLUE);
      colorlookup.put("ASP", RED);
      colorlookup.put("ASN", BLUE);
      colorlookup.put("GLU", RED);
      colorlookup.put("GLN", BLUE);
      colorlookup.put("MET", YELLOW);
      colorlookup.put("LYS", BLUE);
      colorlookup.put("ARG", BLUE);
   }

   private static float fa[] = new float[3];
   private static float fb[] = new float[3];
   private static float fc[] = new float[3];
   private static float fd[] = new float[3];

   /**
    * Finds the length of a vector
    * @param d A vector to find the length of.
    * @return Length of vector d.
    */
   public static float r(float[] d){
      return (float) Math.sqrt((double) (d[0] * d[0] + d[1] * d[1] + d[2] * d[2]));
   }

   /**
    * Normalizes a vector
    * @param n A vector to be normalized.
    * @return Normalized vector.
    */
   public static void norm(float[] n, float[] ret){
      float length;
      length = r(n);
      ret[0] = n[0] / length;
      ret[1] = n[1] / length;
      ret[2] = n[2] / length;
   }

   /**
    * Scales a vector
    * @param n A vector to be scaled
    * @param a A scaler value
    * @return Vector n scaled by a
    */
   public static void scalar(float[] n, float a, float[] ret){
      ret[0] = n[0] * a;
      ret[1] = n[1] * a;
      ret[2] = n[2] * a;
   }

   /**
    * Finds the distance between two vectors
    * @param a First vector
    * @param b Second vector
    * @returns The distance between vectors a and b
    */
   public static float dist(float[] a, float[] b){
      float x = 0;
      x += (a[0] - b[0]) * (a[0] - b[0]);
      x += (a[1] - b[1]) * (a[1] - b[1]);
      x += (a[2] - b[2]) * (a[2] - b[2]);
      return (float) Math.sqrt((double) x);
   }

   /**
    * Finds the difference between two vectors
    * @param a First vector
    * @param b Second vector
    * @param ret Return Values
    */
   public static void diff(float[] a, float[] b, float[] ret){
      ret[0] = a[0] - b[0];
      ret[1] = a[1] - b[1];
      ret[2] = a[2] - b[2];
   }

   public static void sum(float[] a, float[] b, float[] ret){
      ret[0] = a[0] + b[0];
      ret[1] = a[1] + b[1];
      ret[2] = a[2] + b[2];

   }

   /**
    * Finds the dot product between two vectors
    * @param a First vector
    * @param b Second vector
    * @return The dot product of a and b
    */
   public static float dot(float[] a, float[] b){
      return ((a[0] * b[0]) + (a[1] * b[1]) + (a[2] * b[2]));
   }

   /**
    * Finds the cross-product between two vectors
    * @param a First vector
    * @param b Second vector
    * @param ret The cross-product a x b
    */
   public static void cross(float[] a, float[] b, float ret[]){
      ret[0] = a[1] * b[2] - a[2] * b[1];
      ret[1] = a[2] * b[0] - a[0] * b[2];
      ret[2] = a[0] * b[1] - a[1] * b[0];
   }

   /**
    * Finds the angle formed by three atoms
    * @param i Atom position vector
    * @param j Atom position vector (central atom)
    * @param k Atom position vector
    * @return The angle in the range [ -pi, pi ]
    */
   public static float bondangle(float[] i, float[] j, float[] k){
      float x;
      diff(i, j, fa);
      diff(k, j, fb);
      norm(fa, fc);
      norm(fb, fd);
      x = dot(fc, fd);
      if (Math.abs(x) > 1){
         System.out.println("abs(dot) > 1 " + x);
         if (x > 0){
            x = 1;
         } else{
            x = -1;
         }
      }
      return (float) Math.acos((double) x);
   }

   public static float angle(float[] i, float[] j){
      float x;
      norm(i, fa);
      norm(j, fb);
      x = dot(fa, fb);
      if (Math.abs(x) > 1){
         System.out.println("abs(dot) > 1 " + x);
         if (x > 0){
            x = 1;
         } else{
            x = -1;
         }
      }
      return (float) Math.acos(x);
   }

   private static float[] fba = new float[3];
   private static float[] fcb = new float[3];
   private static float[] fdc = new float[3];
   private static float[] ft = new float[3];
   private static float[] fu = new float[3];
   private static float[] ftu = new float[3];

   /**
    * Finds the dihedral angle formed between 4 atoms
    * @param a Atom position vector
    * @param b Atom position vector
    * @param c Atom position vector
    * @param d Atom position vector
    * @return The dihedral angle in the range [ -pi, pi ]
    */
   public static float dihedralangle(float[] a, float[] b, float[] c, float[] d){

      diff(b, a, fba);
      diff(c, b, fcb);
      diff(d, c, fdc);

      cross(fba, fcb, ft);
      cross(fcb, fdc, fu);
      cross(ft, fu, ftu);

      float rt = dot(ft, ft);
      float ru = dot(fu, fu);
      float rtu = (float) Math.sqrt(rt * ru);

      if (rtu != 0.0){
         float rcb = r(fcb);
         float cosine = dot(ft, fu) / rtu;
         float sine = dot(fcb, ftu) / (rcb * rtu);
         cosine = Math.min(1.0f, Math.max( -1.0f, cosine));
         float angle = (float) Math.acos((double) cosine);
         if (sine < 0.0){
            angle = -angle;
         }
         return angle;
      }
      return 0;
   }

   private static double da[] = new double[3];
   private static double db[] = new double[3];
   private static double dc[] = new double[3];
   private static double dd[] = new double[3];

   /**
    * Finds the length of a vector
    * @param d A vector to find the length of.
    * @return Length of vector d.
    */
   public static double r(double[] d){
      return Math.sqrt((d[0] * d[0] + d[1] * d[1] + d[2] * d[2]));
   }

   /**
    * Normalizes a vector
    * @param n A vector to be normalized.
    * @return Normalized vector.
    */
   public static void norm(double[] n, double[] ret){
      double length;
      length = r(n);
      ret[0] = n[0] / length;
      ret[1] = n[1] / length;
      ret[2] = n[2] / length;
   }

   /**
    * Scales a vector
    * @param n A vector to be scaled
    * @param a A scaler value
    * @return Vector n scaled by a
    */
   public static void scalar(double[] n, double a, double[] ret){
      ret[0] = n[0] * a;
      ret[1] = n[1] * a;
      ret[2] = n[2] * a;
   }

   /**
    * Finds the distance between two vectors
    * @param a First vector
    * @param b Second vector
    * @returns The distance between vectors a and b
    */
   public static double dist(double[] a, double[] b){
      double x = 0;
      x += (a[0] - b[0]) * (a[0] - b[0]);
      x += (a[1] - b[1]) * (a[1] - b[1]);
      x += (a[2] - b[2]) * (a[2] - b[2]);
      return Math.sqrt(x);
   }

   /**
    * Finds the difference between two vectors
    * @param a First vector
    * @param b Second vector
    * @param ret Return Values
    */
   public static void diff(double[] a, double[] b, double[] ret){
      ret[0] = a[0] - b[0];
      ret[1] = a[1] - b[1];
      ret[2] = a[2] - b[2];
   }

   public static void sum(double[] a, double[] b, double[] ret){
      ret[0] = a[0] + b[0];
      ret[1] = a[1] + b[1];
      ret[2] = a[2] + b[2];

   }

   /**
    * Finds the dot product between two vectors
    * @param a First vector
    * @param b Second vector
    * @return The dot product of a and b
    */
   public static double dot(double[] a, double[] b){
      return ((a[0] * b[0]) + (a[1] * b[1]) + (a[2] * b[2]));
   }

   /**
    * Finds the cross-product between two vectors
    * @param a First vector
    * @param b Second vector
    * @param ret The cross-product a x b
    */
   public static void cross(double[] a, double[] b, double ret[]){
      ret[0] = a[1] * b[2] - a[2] * b[1];
      ret[1] = a[2] * b[0] - a[0] * b[2];
      ret[2] = a[0] * b[1] - a[1] * b[0];
   }

   /**
    * Finds the angle formed by three atoms
    * @param i Atom position vector
    * @param j Atom position vector (central atom)
    * @param k Atom position vector
    * @return The angle in the range [ -pi, pi ]
    */
   public static double bondangle(double[] i, double[] j, double[] k){
      double x;
      diff(i, j, da);
      diff(k, j, db);
      norm(da, dc);
      norm(db, dd);
      x = dot(dc, dd);
      if (Math.abs(x) > 1){
         if (x > 0){
            x = 1;
         } else{
            x = -1;
         }
      }
      return Math.acos(x);
   }

   public static double angle(double[] i, double[] j){
      double x;

      norm(i, da);
      norm(j, db);
      x = dot(da, db);
      if (Math.abs(x) > 1){
         if (x > 0){
            x = 1;
         } else{
            x = -1;
         }
      }
      return Math.acos(x);
   }

   public static void printVector(double v[]){
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < v.length; i++){
         sb.append(v[i] + " ");
      }
      System.out.println("Vector ( " + sb.toString() + ")");
   }

   private static double[] dba = new double[3];
   private static double[] dcb = new double[3];
   private static double[] ddc = new double[3];
   private static double[] dt = new double[3];
   private static double[] du = new double[3];
   private static double[] dtu = new double[3];

   /**
    * Finds the dihedral angle formed between 4 atoms
    * @param a Atom position vector
    * @param b Atom position vector
    * @param c Atom position vector
    * @param d Atom position vector
    * @return The dihedral angle in the range [ -pi, pi ]
    */
   public static double dihedralangle(double[] a, double[] b, double[] c, double[] d){

      diff(b, a, dba);
      diff(c, b, dcb);
      diff(d, c, ddc);

      cross(dba, dcb, dt);
      cross(dcb, ddc, du);
      cross(dt, du, dtu);

      double rt = dot(dt, dt);
      double ru = dot(du, du);
      double rtu = Math.sqrt(rt * ru);

      if (rtu != 0.0){
         double rcb = r(dcb);
         double cosine = dot(dt, du) / rtu;
         double sine = dot(dcb, dtu) / (rcb * rtu);
         cosine = Math.min(1.0, Math.max( -1.0, cosine));
         double angle = Math.acos(cosine);
         if (sine < 0.0){
            angle = -angle;
         }
         return angle;
      }
      return 0;
   }

}
