/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import ffe.panels.*;

/**
 * The Bond class represents a covalent bond formed between two atoms.
 */

public final class Bond
   extends Term{

   private ArrayList bonds = new ArrayList(); // List of Angles that this bond helps form
   private int order; // Order of the bond, 1 for single, 2 for double, 3 for triple

   // Java 3D
   /*
              base  (found in FSystem)
      _________|_________
     |                   |
      cy1tg(cy1t3d)       cy2tg(cy2t3d)
     |                   |
    cy1                 cy2

    */

   private TransformGroup cy1tg;
   private TransformGroup cy2tg;
   private Transform3D cy1t3d;
   private Transform3D cy2t3d;
   private boolean attached;
   private boolean stale = false;
   private Shape3D cy1, cy2;
   private LineArray la;
   private int lineIndex;
   private boolean wireVisible = false;
   private boolean j3dVisible = false;
   private Vector3d scale;
   private int detail = 3;
   private int view = 0;
   private double angle = 0;

   // Bond Object "recycling"
   public static ArrayList bondpool = new ArrayList();

   public Bond(Atom a1, Atom a2, int ord){
      atoms = new Atom[2];
      atoms[0] = a1;
      atoms[1] = a2;
      order = ord;
      a1.setBond(this);
      a2.setBond(this);
      setID_Key(false);
   }

   /**
    * Simple Bond constructor that is intended to be used with the equals method
    * @param n Bond id
    */
   public Bond(String n){
      super(n);
   }

   public void destroy(){
      super.destroy();
      removeFromParent();
      bonds.clear();
      bondpool.add(this);
   }

   /**
    * Check to see if <b>this</b> Bond and another combine to form an angle
    * @return True if Bond b helps form an angle with <b>this</b> Bond
    */
   public boolean formsAngleWith(Bond b){
      ListIterator li;
      for (li = bonds.listIterator(); li.hasNext(); ){
         if (b == (Bond) li.next()){
            return true;
         }
      }
      return false;
   }

   public List getBondList(){
      ArrayList temp = new ArrayList();
      temp.add(this);
      return temp;
   }

   /**
    * Finds the common Atom between <b>this</b> Bond and Bond b
    * @param b Bond to compare with
    * @return The Atom the Bonds have in common or Null if they are the same Bond or have no atom in common
    */
   public Atom getCommonAtom(Bond b){
      if (b == this){
         return null; //undefined when comparing a bond to itself
      }
      if (b.atoms[0] == atoms[0]){
         return atoms[0];
      }
      if (b.atoms[0] == atoms[1]){
         return atoms[1];
      }
      if (b.atoms[1] == atoms[0]){
         return atoms[0];
      }
      if (b.atoms[1] == atoms[1]){
         return atoms[1];
      }
      return null; //Common atom not found
   }

   public Group getJ3D1(){
      return cy1tg;
   }

   public Group getJ3D2(){
      return cy2tg;
   }

   /**
    * Find the other Atom in <b>this</b> Bond
    * @param a The known Atom
    * @return The other Atom that makes up this bond, or Null if Atom a is not part of this bond
    */
   public Atom getOtherAtom(Atom a){
      if (a == atoms[0]){
         return atoms[1];
      }
      if (a == atoms[1]){
         return atoms[0];
      }
      return null; //Atom not found in bond
   }

   /**
    * Find the Atom the <b>this</b> Bond and Bond b don't have in common
    * @param b Bond to compare with
    * @return The Atom that Bond b and <b>this</b> Bond do not have in common, or Null if they have no Atom in common
    */
   public Atom getOtherAtom(Bond b){
      if (b == this){
         return null;
      }
      if (b.atoms[0] == atoms[0]){
         return atoms[1];
      }
      if (b.atoms[0] == atoms[1]){
         return atoms[0];
      }
      if (b.atoms[1] == atoms[0]){
         return atoms[1];
      }
      if (b.atoms[1] == atoms[1]){
         return atoms[0];
      }
      return null;
   }

   public boolean initJ3D(){
      if (attached){
         return false;
      }
      if (cy1 != null){
         update();
         attached = true;
         return true;
      }

      scale = new Vector3d(0.2d, 1.0d, 0.2d);
      cy1t3d = new Transform3D();
      cy1tg = new TransformGroup();
      cy1tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      detail = GraphicsContext.detail;
      cy1 = Graphics3D.createCylinder(atoms[0].getAppearance(), detail);
      cy1.setUserData(atoms[0]);
      cy1tg.addChild(cy1);

      cy2t3d = new Transform3D();
      cy2tg = new TransformGroup();
      cy2tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      cy2 = Graphics3D.createCylinder(atoms[1].getAppearance(), detail);
      cy2.setUserData(atoms[1]);
      cy2tg.addChild(cy2);
      setCylinderVisible(false);
      update();

      attached = true;
      return true;
   }

   public void reInit(Atom a1, Atom a2, int ord){
      atoms = new Atom[2];
      atoms[0] = a1;
      atoms[1] = a2;
      order = ord;
      a1.setBond(this);
      a2.setBond(this);
      setID_Key(false);
      attached = false;
      if (cy1 != null){
         cy1.setAppearance(a1.getAppearance());
         cy1.setUserData(a1);
         cy2.setAppearance(a2.getAppearance());
         cy2.setUserData(a2);
         if (detail != GraphicsContext.detail){
            switchView(GraphicsContext.detail);
         }
      }
   }

   // For hiding parts of the Wireframe
   static private Appearance nullAp;
   static{
      nullAp = new Appearance();
      RenderingAttributes ra = new RenderingAttributes();
      ra.setVisible(false);
      nullAp.setRenderingAttributes(ra);
   }

   public void switchView(int viewtype){
      if (viewtype == GraphicsContext.WIREFRAME){
         view = viewtype;
         setWireVisible(true);
         setCylinderVisible(false);
      } else if (viewtype == GraphicsContext.INVISIBLE ){
         view = viewtype;
         setWireVisible(false);
         setCylinderVisible(false);
      } else if (viewtype == GraphicsContext.RESTRICT){
         if (!atoms[0].isSelected() || !atoms[1].isSelected()){
            setWireVisible(false);
            setCylinderVisible(false);
         }
      } else if (viewtype == GraphicsContext.SPACEFILL) {
         setWireVisible(false);
      } else if (viewtype == GraphicsContext.BALLSTICK || viewtype == GraphicsContext.TUBE){
         view = viewtype;
         if (cy1 == null){
            initJ3D();
         }
         // Get the radius to use
         double rad;
         double len = getValue() / 2.0d;
         if (view == GraphicsContext.BALLSTICK){
            rad = 0.1d * GraphicsContext.radius;
         } else{
            rad = 0.2d * GraphicsContext.radius;
         }
         // Scale the cylinders
         scale.set(rad, len, rad);
         cy1t3d.setScale(scale);
         cy1tg.setTransform(cy1t3d);
         cy2t3d.setScale(scale);
         cy2tg.setTransform(cy2t3d);
         setCylinderVisible(true);
         setWireVisible(false);
      } else if (viewtype == GraphicsContext.DETAIL){
         int res;
         res = GraphicsContext.detail;
         if (res > 3){
            res = 3;
         }
         if (res != detail){
            detail = res;
            if (cy1 == null){
               return;
            }
            Geometry g1 = Graphics3D.getCylinderGeom(0, detail);
            Geometry g2 = Graphics3D.getCylinderGeom(1, detail);
            Geometry g3 = Graphics3D.getCylinderGeom(2, detail);
            cy1.removeAllGeometries();
            cy2.removeAllGeometries();
            cy1.addGeometry(g1);
            cy1.addGeometry(g2);
            cy1.addGeometry(g3);
            cy2.addGeometry(g1);
            cy2.addGeometry(g2);
            cy2.addGeometry(g3);
         }
         switchView(view);
      } else if (viewtype == GraphicsContext.HYDROGENS){
         if (!atoms[0].showHydrogen() || !atoms[1].showHydrogen()){
            setCylinderVisible(false);
            setWireVisible(false);
         } else{
            switchView(view);
         }
      } else if (viewtype == GraphicsContext.FILL || viewtype == GraphicsContext.POINTS ||
         viewtype == GraphicsContext.LINES){
         if (j3dVisible){
            cy1.setAppearance(atoms[0].getAppearance());
            cy2.setAppearance(atoms[1].getAppearance());
         }
      }
   }

   private static float col[] = {0.0f, 0.0f, 0.0f, 0.9f};

   public void switchColor(Atom a){
      if (j3dVisible){
         if (cy1 != null && a == atoms[0]){
            cy1.setAppearance(a.getAppearance());
         } else if (cy2 != null && a == atoms[1]){
            cy2.setAppearance(a.getAppearance());
         }
      }
      if (wireVisible){
         if (a == atoms[0]){
            Color3f col3f = a.getColor();
            if (col3f != null){
               col3f.get(col);
               la.setColor(lineIndex, col);
               la.setColor(lineIndex + 1, col);
            }
         } else if (a == atoms[1]){
            Color3f col3f = a.getColor();
            if (col3f != null){
               col3f.get(col);
               la.setColor(lineIndex + 2, col);
               la.setColor(lineIndex + 3, col);
            }
         }
      }
   }

   // Some static variables used for computing cylinder orientations
   private static double d;
   private static double d3a[] = new double[3];
   private static double d3b[] = new double[3];
   private static double d3c[] = new double[3];
   private static double d3d[] = new double[3];
   private static double d3e[] = new double[3];
   private static double d12[] = new double[12];

   /**
    * Update recomputes the bonds length
    */

   public void update(){

      // Local References
      double a13d[] = d3a;
      double a23d[] = d3b;
      double mid[] = d3c;
      double diff3d[] = d3d;
      double sum3d[] = d3e;
      double coord[] = d12;

      atoms[0].getXYZ(a13d);
      atoms[1].getXYZ(a23d);
      AtomUtils.diff(a13d, a23d, diff3d);
      d = AtomUtils.r(diff3d);
      setValue(d);

      AtomUtils.sum(a13d, a23d, sum3d);
      AtomUtils.scalar(sum3d, 0.5d, mid);

      if (la != null){
         for (int i = 0; i < 3; i++){
            coord[i] = a13d[i];
            coord[3 + i] = mid[i];
            coord[6 + i] = mid[i];
            coord[9 + i] = a23d[i];
         }
         la.setCoordinates(lineIndex, coord);
      }
      if (cy1 != null){
         if (!j3dVisible){
            stale = true;
         } else{
            AtomUtils.norm(diff3d, diff3d);
            scale.y = d / 2.0d;
            setBondTransform3d(cy1t3d, mid, diff3d, d, true);
            AtomUtils.scalar(diff3d, -1.0d, diff3d);
            setBondTransform3d(cy2t3d, mid, diff3d, d, false);
            cy1tg.setTransform(cy1t3d);
            cy2tg.setTransform(cy2t3d);
            stale = false;
         }
      }
   }

   public boolean sameGroup(){
      if (atoms[0].getParent() == atoms[1].getParent()){
         return true;
      } else{
         return false;
      }
   }

   /**
    * Specifies <b>this</b> Bond helps form an angle with the given Bond
    * @param b Bond that forms an angle with <b>this</b> Bond
    */
   public void setAngleWith(Bond b){
      bonds.add(b);
   }

   public void setCylinderVisible(boolean v){
      j3dVisible = v;
      if (cy1 == null){
         return;
      }
      if (!v){
         cy1.setAppearance(nullAp);
         cy1.setPickable(false);
         cy2.setAppearance(nullAp);
         cy2.setPickable(false);
      } else{
         if (!atoms[0].showHydrogen() || !atoms[1].showHydrogen()){
            return;
         }
         if (stale){
            update();
         }
         cy1.setAppearance(atoms[0].getAppearance());
         cy1.setPickable(true);
         cy2.setAppearance(atoms[1].getAppearance());
         cy2.setPickable(true);
      }
   }

   private static double yd[] = {0.0d, 1.0d, 0.0d};
   private static AxisAngle4d rot = new AxisAngle4d();
   private static double[] bcross = new double[4];
   private static double[] cstart = new double[3];
   private static Vector3d pos3d = new Vector3d();

   public void setBondTransform3d(Transform3D t3d, double[] pos, double[] orient, double len,
      boolean newRot){
      //Bond Orientation
      if (newRot){

         angle = AtomUtils.angle(orient, yd);
         AtomUtils.cross(yd, orient, bcross);
         bcross[3] = angle - Math.PI;
         rot.set(bcross);
      }
      //Scale the Orientation to be a fourth the bond length
      //and add it to the position vector of the of the first atom
      AtomUtils.scalar(orient, len / 4.0d, cstart);
      AtomUtils.sum(cstart, pos, cstart);
      pos3d.set(cstart);
      t3d.setTranslation(pos3d);
      t3d.setRotation(rot);
      t3d.setScale(scale);
   }

   public void setWire(LineArray l, int i){
      la = l;
      lineIndex = i;
   }

   private static float a0col[] = {0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,
                                  0.0f, 0.0f, 0.0f, 0.0f, 0.0f};
   private static float f16[] = {0.0f, 0.0f, 0.0f, 0.9f, 0.0f, 0.0f, 0.0f, 0.9f, 0.0f, 0.0f, 0.0f,
                                0.9f, 0.0f, 0.0f, 0.0f, 0.9f};
   private static float f4a[] = {0.0f, 0.0f, 0.0f, 0.9f};
   private static float f4b[] = {0.0f, 0.0f, 0.0f, 0.9f};

   public boolean isVisible(){
      if (wireVisible || j3dVisible){
         return true;
      }
      return false;
   }

   public void setWireVisible(boolean v){
      wireVisible = v;
      if (!v){
         la.setColors(lineIndex, a0col);
      } else{
         float cols[] = f16;
         float col1[] = f4a;
         float col2[] = f4b;
         if (atoms[0].showHydrogen() && atoms[1].showHydrogen()){
            atoms[0].getColor().get(col1);
            atoms[1].getColor().get(col2);
            for (int i = 0; i < 3; i++){
               cols[i] = col1[i];
               cols[4 + i] = col1[i];
               cols[8 + i] = col2[i];
               cols[12 + i] = col2[i];
            }
            la.setColors(lineIndex, cols);
         }
      }
   }
}
