/**
 * <p>Title: Rick RossForce Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

/**
 * The Dihedral class represents a dihedral angle formed between four bonded atoms.
 */
public final class Dihedral
   extends Term{

   /**
    * Dihedral Constructor
    * @param n Dihedral id
    */
   public Dihedral(String n){
      super(n);
   }

   /**
    * Dihedral contructor
    * @param an1 Angle that combines to form the Diherdral Angle
    * @param an2 Angle that combines to form the Diherdral Angle
    * @param mmdi Reference to a hashtable of Molecular Mechanics dihedral angle definitions.
    */
   public Dihedral(Angle an1, Angle an2){
      super();
      bonds = new Bond[3];
      bonds[1] = an1.getCommonBond(an2);
      bonds[0] = an1.getOtherBond(bonds[1]);
      bonds[2] = an2.getOtherBond(bonds[1]);
      Initialize();
   }

   /**
    * Dihedral contructor
    * @param a Angle that has one Atom in common with Bond b
    * @param b Bond that has one Atom in common with Angle A
    * @param mmdi Reference to a hashtable of Molecular Mechanics dihedral angle definitions.
    */
   public Dihedral(Angle a, Bond b){
      super();
      Atom atom;
      bonds = new Bond[3];
      bonds[0] = b;
      bonds[1] = a.getBond(0);
      bonds[2] = a.getBond(1);
      //See if bond 2 or bond 3 is the middle bond
      atom = bonds[1].getCommonAtom(b);
      if (atom == null){
         Bond temp = bonds[1];
         bonds[1] = bonds[2];
         bonds[2] = temp;
      }
      Initialize();
   }

   //Initialization
   private void Initialize(){
      atoms = new Atom[4];
      atoms[1] = bonds[0].getCommonAtom(bonds[1]);
      atoms[0] = bonds[0].getOtherAtom(atoms[1]);
      atoms[2] = bonds[1].getOtherAtom(atoms[1]);
      atoms[3] = bonds[2].getOtherAtom(atoms[2]);
      atoms[0].setDihedral(this);
      atoms[1].setDihedral(this);
      atoms[2].setDihedral(this);
      atoms[3].setDihedral(this);
      setID_Key(false);
      update();
   }

   /**
    * Update recomputes the dihedral angle value
    */
   private static double vi[] = new double[3];
   private static double vj[] = new double[3];
   private static double vk[] = new double[3];
   private static double vl[] = new double[3];

   public void update(){
      atoms[0].getXYZ(vi);
      atoms[1].getXYZ(vj);
      atoms[2].getXYZ(vk);
      atoms[3].getXYZ(vl);
      setValue(Math.toDegrees(AtomUtils.dihedralangle(vi, vj, vk, vl)));
   }
}
