/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.*;
import javax.media.j3d.Material;
import javax.vecmath.Color3f;

import ffe.panels.GraphicsContext;

/**
 * The FGroup class has one subnode containing atoms, and one that contains molecular mechanics/geometry terms.
 */
public abstract class FGroup
   extends FNode{

   // Geometry Nodes
   private FNode tgroups = new FNode("Atoms");
   private FNode terms = new FNode("Connectivity");
   private FNode bonds = new FNode("Bonds");
   private FNode angles = new FNode("Angles");
   private FNode dihedrals = new FNode("Dihedrals");
   private boolean bondsKnown = true;

   // Whether the terms are current
   private boolean finalized;
   private double[] center;

   // List of underconstrained Atoms
   private ArrayList dangelingatoms;

   private static ArrayList fgrouppool = new ArrayList();

   /**
    * Default Constructor initializes a FGroup and a few of its subnodes.
    */
   public FGroup(){
      super("");
      finalized = false;
      terms.add(bonds);
      terms.add(angles);
      terms.add(dihedrals);
      add(tgroups);
      add(terms);
   }

   public void reInit(String n){
      setName(n);
      finalized = false;
   }

   public void destroy(){
      removeFromParent();
      remove(tgroups);
      fgrouppool.add(this);
   }

   /**
    * Constructs a FGroup object with name n.
    */
   public FGroup(String n, boolean bk){
      this();
      setName(n);
      bondsKnown = bk;
   }

   /**
    * Constructs a FGroup object with name n and sets its tgroup child node equals to tnode.
    */
   public FGroup(String n, FNode tnode, boolean bk){
      this(n, bk);
      tgroups = tnode;
      add(tgroups);
   }

   /**
    * Abstract method that should specify how to add various FNode
    * subclasses (such as Atoms, Residues and Polymers) to the FGroup
    */
   public abstract void addFNode(FNode m);

   /**
    * Abstact method that should specify how to finalize a FGroup
    */
   public abstract void finalize(boolean finalizeGroups);

   protected void removeLeaves(){
      if (terms.getParent() == null){
         return;
      }

      if (bonds.getChildCount() == 0){
         if (!(bonds.getParent() == null)){
            terms.remove(bonds);
         }
      }
      if (angles.getChildCount() == 0){
         if (!(angles.getParent() == null)){
            terms.remove(angles);
         }
      }
      if (dihedrals.getChildCount() == 0){
         if (!(dihedrals.getParent() == null)){
            terms.remove(dihedrals);
         }
      }
      if (terms.getChildCount() == 0){
         remove(terms);
      }
   }

   public void updateBonds(){
      Bond b;
      for (ListIterator li = getBondList().listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         b.update();
      }
   }

   public void updateAtoms(){
      Atom a;
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         a.update();
      }
   }

   /**
    * Returns the Moieties name.
    */
   public String toString(){
      return getName();
   }

   /*
     public int writePDB(BufferedWriter bw, int i){
    ArrayList moieties = getFGroupList();
    FNode n;
    Atom a;
    FGroup m;
    for(ListIterator li = moieties.listIterator(); li.hasNext();){
     n = (FNode) li.next();
     if (n  instanceof Atom) {
      a = (Atom) n;
      try {
    bw.write(a.toPDBString(i));
    bw.newLine();
      } catch (IOException e){
    System.out.println("" + e);
      }
      i++;
     } else if (n  instanceof FGroup){
      m = (FGroup) n;
      i = m.writePDB(bw, i);
     }
    }
    return i;
     }*/
   /**
    * Prints the FGroup Term Node's children and Moieties Node children.
    */
   public void print(){
      super.print();
      ListIterator li;
      for (li = tgroups.getListIterator(); li.hasNext(); ){
         ((FNode) li.next()).print();
      }
      for (li = bonds.getListIterator(); li.hasNext(); ){
         ((FNode) li.next()).print();
      }
   }

   /**
    * Constructs the Geometry of this FGroup and stores the terms as
    * children in the Term node. Bonds are made based purely by distance.
    */
   private static double[] da = new double[3];
   private static double[] db = new double[3];
   public void ConstructGeometry(){
      Atom a1, a2, atom13;
      Bond newbond, adjunctbond, bond14;
      ArrayList adjunctbonds, bonds14;
      Angle newangle;
      ListIterator li, lj, lk, ll;
      double d1, d2;

      FNode b = new FNode("Bonds");
      FNode a = new FNode("Angles");
      FNode d = new FNode("Dihedrals");

      List atoms = getAtomList();
      int index = 0;
      for (li = atoms.listIterator(); li.hasNext(); ){
         a1 = (Atom) li.next();
         index++;
         for (lj = atoms.listIterator(index); lj.hasNext(); ){
            a2 = (Atom) lj.next();
            a1.getXYZ(da);
            a2.getXYZ(db);
            d1 = AtomUtils.dist(da, db);
            d2 = AtomUtils.BUFF + a1.getVDWR() / 2 + a2.getVDWR() / 2;
            if (d1 < d2){
               adjunctbonds = new ArrayList();
               if (a1.getBondCount() > 0){
                  adjunctbonds.addAll(a1.getBonds());
               }
               if (a2.getBondCount() > 0){
                  adjunctbonds.addAll(a2.getBonds());
               }
               newbond = new Bond(a1, a2, 1);
               b.add(newbond);
               for (lk = adjunctbonds.listIterator(); lk.hasNext(); ){
                  adjunctbond = (Bond) lk.next();
                  if (newbond == adjunctbond){
                     System.out.println("New Bond = Adjunct Bond");
                  } else{
                     newangle = new Angle(newbond, adjunctbond);
                     a.add(newangle);
                     atom13 = adjunctbond.getOtherAtom(newbond);
                     bonds14 = atom13.getBonds();
                     for (ll = bonds14.listIterator(); ll.hasNext(); ){
                        bond14 = (Bond) ll.next();
                        if (bond14 != adjunctbond){
                           d.add(new Dihedral(newangle, bond14));
                        }
                     }
                  }
               }
            }
         }
      }
      setBonds(b);
      setAngles(a);
      setDihedrals(d);
   }

   public void collectBonds(){
      ListIterator li, lj;

      FNode b = new FNode("Bonds");
      FNode a = new FNode("Angles");
      FNode d = new FNode("Dihedrals");

      // Collect all bonds for which both atoms are in this FGroup
      Bond bond = null;
      Atom a1;
      int index = 0;
      List atoms = getAtomList();
      for (li = atoms.listIterator(); li.hasNext(); ){
         a1 = (Atom) li.next();
         if (!(a1.getBondCount() == 0)){
            for (lj = a1.getBonds().listIterator(); lj.hasNext(); ){
               bond = (Bond) lj.next();
               if (bond.sameGroup()){
                  if (bond.getParent() == null){
                     b.add(bond);
                  }
               }
            }
         }
      }
      b.setName("Bonds (" + b.getChildCount() + ")");
      setBonds(b);

      // Collect Angles
      index = 0;
      Bond b1, b2;
      Angle an1, an2;
      ArrayList angleAL = new ArrayList();
      List newbonds = b.getBondList();
      for (li = newbonds.listIterator(); li.hasNext(); ){
         b1 = (Bond) li.next();
         index++;
         for (lj = newbonds.listIterator(index); lj.hasNext(); ){
            b2 = (Bond) lj.next();
            if (b1.getOtherAtom(b2) != null){
               if (Angle.anglepool.isEmpty()){
                  an1 = new Angle(b1, b2);
               } else{
                  an1 = (Angle) Angle.anglepool.remove(0);
                  an1.reInit(b1, b2);
               }
               a.insert(an1, 0);
               angleAL.add(an1);
            }
         }
      }
      a.setName("Angles (" + a.getChildCount() + ")");
      setAngles(a);

      // Collect Dihedrals
      index = 0;
      Bond cbond, obond1, obond2;
      for (li = angleAL.listIterator(); li.hasNext(); ){
         an1 = (Angle) li.next();
         index++;
         for (lj = angleAL.listIterator(index); lj.hasNext(); ){
            an2 = (Angle) lj.next();
            cbond = an1.getCommonBond(an2);
            if (cbond != null){
               obond1 = an1.getOtherBond(cbond);
               obond2 = an2.getOtherBond(cbond);
               if (obond2.getCommonAtom(obond1) == null){
                  d.add(new Dihedral(an1, an2));
               }
            }
         }
      }
      d.setName("Dihedrals (" + d.getChildCount() + ")");
      setDihedrals(d);

   }

   /**
    * Joiner joins Moieties m1 and m2 and returns the
    * Geometry objects formed in a Joint.
    */
   public Joint joiner(FGroup m1, FGroup m2){
      List m1Atoms = m1.getAtomList();
      List m2Atoms = m2.getAtomList();

      Joint joint = null;
      Atom a1, a2;
      for (ListIterator li = m1Atoms.listIterator(); li.hasNext(); ){
         a1 = (Atom) li.next();
         a1.getXYZ(da);
         for (ListIterator lj = m2Atoms.listIterator(); lj.hasNext(); ){
            a2 = (Atom) lj.next();
            a2.getXYZ(db);
            double d1 = AtomUtils.dist(da, db);
            double d2 = AtomUtils.BUFF + a1.getVDWR() / 2 + a2.getVDWR() / 2;
            if (d1 < d2){
               Bond b = new Bond(a1, a2, 1);
               Joint newJoint = joiner(b);
               if (joint != null){
                  joint.merge(newJoint);
               } else{
                  joint = newJoint;
               }
            }
         }
      }
      return joint;
   }

   /**
    * Joiner joins
    * @param bond Bond
    * @return Joint
    */
   public Joint joiner(Bond bond){
      FNode b = new FNode("Bonds");
      FNode a = new FNode("Angles");
      FNode d = new FNode("Dihedrals");

      ListIterator li, lj;

      // Collect all bonds for which one atom is in each FGroup

      b.add(bond);
      b.setName("Bonds (" + b.getChildCount() + ")");

      // Collect Angles that include the joining bond(s)
      Bond b1, b2;
      Atom a1, a2;
      Angle an1;
      ArrayList angleAL = new ArrayList();
      List newbonds = b.getBondList();

      for (li = newbonds.listIterator(); li.hasNext(); ){
         b1 = (Bond) li.next();
         // FGroup #1
         a1 = b1.getAtom(0);
         for (lj = a1.getBonds().listIterator(); lj.hasNext(); ){
            b2 = (Bond) lj.next();
            if (b1 == b2){
               continue;
            }
            if (b1.getOtherAtom(b2) != null){
               an1 = new Angle(b1, b2);
               a.add(an1);
               angleAL.add(an1);
            }
         }
         // FGroup #2
         a2 = b1.getAtom(1);
         for (lj = a2.getBonds().listIterator(); lj.hasNext(); ){
            b2 = (Bond) lj.next();
            if (b1 == b2){
               continue;
            }
            if (b1.getOtherAtom(b2) != null){
               an1 = new Angle(b1, b2);
               a.add(an1);
               angleAL.add(an1);
            }
         }
      }
      a.setName("Angles (" + a.getChildCount() + ")");

      // todo: Collect Dihedrals

      FGroup m1 = bond.getAtom(0).getResidueNode();
      FGroup m2 = bond.getAtom(1).getResidueNode();
      Joint j = new Joint(m1, m2, b, a, d);
      return j;
   }

   /**
    * This method constructs an ArrayList of atoms which are under-constrained.
    * (ie They can except more bonds)
    */
   public void findDangelingAtoms(){
      Atom a;
      ArrayList d = new ArrayList();
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (a.isDangeling()){
            d.add(a);
         }
      }
      setDangelingAtoms(d);
   }

   /**
    * This method finds the Geometrical center of this FGroup, or the mass-weighted center
    * if w is set to true, and returns it as a double[3].
    */
   public double[] findCenter(boolean w){
      //Find the center of mass if w == true, the center of geometry if w == false
      double[] Rc = {0, 0, 0};
      double sum = 0, mass = 1;
      Atom a;

      List atoms = getAtomList();
      if (atoms == null){
         return Rc;
      }
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (w){
            mass = a.getMass();
            sum += mass;
         }
         Rc[0] += mass * a.getX();
         Rc[1] += mass * a.getY();
         Rc[2] += mass * a.getZ();
      }
      if (!w){
         sum = atoms.size();
      }
      for (int i = 0; i < 3; i++){
         Rc[i] /= sum;
      }
      return Rc;
   }

   //Public Mutators
   /**
    * Specifies whether the FGroup has been finalized.
    */
   public void setFinalized(boolean t){
      finalized = t;
   }

   /**
    * Sets the FGroup's terms node to t.
    */
   public void setTerms(FNode t){
      remove(terms);
      terms = t;
      add(terms);
   }

   /**
    * Sets the FGroup's bonds node to t.
    */
   public void setBonds(FNode t){
      terms.remove(bonds);
      bonds.removeAllChildren();
      bonds = t;
      terms.add(bonds);
   }

   /**
    * Sets the BondsKnown Variable
    */
   public void setBondsKnown(boolean b){
      bondsKnown = b;
   }

   /**
    * Sets the FGroup's angles node to t.
    */
   public void setAngles(FNode t){
      terms.remove(angles);
      angles = t;
      terms.add(angles);
   }

   /**
    * Sets the FGroup's dihedrals node to t.
    */
   public void setDihedrals(FNode t){
      terms.remove(dihedrals);
      dihedrals = t;
      terms.add(dihedrals);
   }

   /**
    * Sets the FGroup's impropers node to t.
    */
   /*public void setImpropers(FNode t){
    terms.remove(impropers);
    impropers = t;
    terms.add(impropers);
     }*/
   /**
    * Sets the FGroup's dangelingatoms member to a.
    */
   public void setDangelingAtoms(ArrayList a){
      dangelingatoms = a;
   }

   /**
    * Set the value of Center to d.
    */
   public void setCenter(double[] d){
      center = d;
   }

   /**
    * Sets the FGroup's moieties node to t.
    */
   public void setFGroups(FNode t){
      remove(tgroups);
      tgroups = t;
      add(tgroups);
   }

   public void switchColor(int colortype, Color3f color, Material mat){
      ListIterator li;
      Atom atom;

      if (colortype == GraphicsContext.byMOLECULE && (color == null || mat == null)){
         return;
      }

      // Otherwise, just pass on information
      for (li = getAtomList().listIterator(); li.hasNext(); ){
         atom = (Atom) li.next();
         atom.switchColor(colortype, color, mat);
      }
   }

   /**
    * Returns true if the FGroup is finalized.
    */
   public boolean isFinalized(){
      return finalized;
   }

   public boolean getBondsKnown(){
      return bondsKnown;
   }

   /**
    * Returns the FGroup's moieties FNode.
    */
   public FNode getFGroups(){
      return tgroups;
   }

   /**
    * Returns the FGroup's terms FNode.
    */
   public FNode getTerms(){
      return terms;
   }

   /**
    * Returns the FGroup's bonds FNode.
    */
   public FNode getBonds(){
      return bonds;
   }

   /**
    * Returns the FGroup's angles FNode.
    */
   public FNode getAngles(){
      return angles;
   }

   /**
    * Returns the FGroup's dihedrals FNode.
    */
   public FNode getDihedrals(){
      return dihedrals;
   }

   /**
    * Returns the FGroup's impropers FNode.
    */
   //public FNode getImpropers(){	return impropers; }
   /**
    * Returns the FGroup's dangelingatoms list.
    */
   public ArrayList getDangelingAtoms(){
      return dangelingatoms;
   }

   /**
    * Returns the FGroup's center as a double[3].
    */
   public double[] getCenter(){
      return center;
   }

   /**
    * Returns an ArrayList of the FGroup's moieties FNode's children.
    */
   public ArrayList getFGroupList(){
      return tgroups.getList();
   }

   /**
    * Returns the Bond at the supplied index.
    */
   public Bond getBond(int index){
      return (Bond) bonds.getChildAt(index);
   }

   /**
    * Returns the Bond with the given id.
    */
   public Bond getBond(String id){
      int i = bonds.getIndex(new Bond(id));
      if (i == -1){
         return null;
      }
      return (Bond) bonds.getChildAt(i);
   }

   /**
    * Returns the FNode at the given index.
    */
   public FNode getFNode(int index){
      return (FNode) getFGroupList().get(index);
   }

   /**
    * Returns the FNode specified by the String n.
    */
   public FNode getFNode(String n){
      int i = getFGroupList().indexOf(new FNode(n));
      if (i == -1){
         return null;
      }
      return (FNode) getFGroupList().get(i);
   }
}
