/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.io.Serializable;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.Color3f;

import javax.swing.tree.*;

import ffe.panels.Graphics3D;

/**
 * The FNode class forms the basic unit that all data classes extend.
 */
public class FNode
   extends DefaultMutableTreeNode
   implements Serializable{

   private String name;
   protected boolean selected = false;

   /**
    * Default FNode Constructor
    */
   public FNode(){
      name = null;
   }

   /**
    * Constructs a FNode with the name n.
    */
   public FNode(String n){
      name = n;
   }

   public void destroy(){
      name = null;
      selected = false;
   }

   /**
    * Overidden equals method that returns true if object is not equal to this
    * object, is of the same class as this, and has the same name.
    */
   public boolean equals(Object object){
      if (this == object){
         return true;
      } else if (object == null || getClass() != object.getClass()){
         return false;
      }

      FNode other = (FNode) object;
      return (name == other.getName());
   }

   /**
    * Prints the FNode's name to stdout
    */
   public void print(){
      System.out.println(name);
   }

   /**
    * Overidden toString method returns the FNode's name
    */
   public String toString(){
      return name;
   }

   /**
    * Returns an ArrayList of the FNode's Children (instead of using an Enumeration).
    */
   public ArrayList getList(){
      ArrayList l = new ArrayList();
      Enumeration e = children();
      while (e.hasMoreElements()){
         l.add(e.nextElement());
      }
      return l;
   }

   /**
    * Returns a ListIterator containing this FNode's children.
    */
   public ListIterator getListIterator(){
      ArrayList l = getList();
      return l.listIterator();
   }

   /**
    * Returns an ArrayList of all Terms below the present FNode.
    */
   public ArrayList getTermList(){
      FNode o;
      ArrayList t = new ArrayList();
      Enumeration e = depthFirstEnumeration();
      while (e.hasMoreElements()){
         o = (FNode) e.nextElement();
         if (o instanceof Term){
            t.add(o);
         }
      }
      return t;
   }

   public void drawLabel(Graphics3D graphics, J3DGraphics2D g2d, Node node, boolean labelAtoms,
      boolean labelResidues){
      if (!isSelected()){
         return;
      } else{
         FNode fnode;
         for (Enumeration e = children(); e.hasMoreElements(); ){
            fnode = (FNode) e.nextElement();
            fnode.drawLabel(graphics, g2d, node, labelAtoms, labelResidues);
         }
      }
   }

   public void switchView(int viewtype){
      Atom a;

      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         a.switchView(viewtype);
      }
      Bond b;
      for (ListIterator li = getBondList().listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         b.switchView(viewtype);
      }
   }

   public void switchColor(int colortype, Color3f color, Material mat){
      FNode node;
      for (Enumeration e = children(); e.hasMoreElements(); ){
         node = (FNode) e.nextElement();
         node.switchColor(colortype, color, mat);
      }
   }

   /*public void switchColor() {
       Atom a;
       GraphicsContext gc = getFSystemNode().getGraphicsContext();
       Material mat = null;
       mat = Graphics.createMaterial(gc.pick);
       for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ) {
           a = (Atom) li.next();
           a.switchColor(gc.colortype, gc.pick, mat);
       }
        }*/

   public double getMW(){
      double weight = 0.0;
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         weight += ((Atom) li.next()).getMass();
      }
      return weight;
   }

   public FSystem getFSystemNode(){
      TreeNode[] nodes = getPath();
      if (nodes == null){
         return null;
      } else if (nodes.length >= 2){
         return (FSystem) nodes[1];
      }
      return null;
   }

   public Polymer getPolymerNode(){
      TreeNode[] nodes = getPath();
      Polymer c = null;
      if (nodes.length >= 4 && nodes[3] instanceof Polymer){
         c = (Polymer) nodes[3];
      }
      return c;
   }

   public Residue getResidueNode(){
      TreeNode[] nodes = getPath();
      Residue r = null;
      if (nodes.length >= 6 && nodes[5] instanceof Residue){
         r = (Residue) nodes[5];
      }
      return r;
   }

   public Molecule getMoleculeNode(){
      TreeNode[] nodes = getPath();
      Molecule m = null;
      if (nodes.length >= 4 && nodes[3] instanceof Molecule){
         m = (Molecule) nodes[3];
      }
      return m;
   }

   /**
    * Returns an ArrayList of all Atoms below the present FNode.
    */
   public List getAtomList(){
      FNode o;
      ArrayList t = new ArrayList();
      Enumeration e = depthFirstEnumeration();
      while (e.hasMoreElements()){
         o = (FNode) e.nextElement();
         if (o instanceof Atom){
            t.add(o);
         }
      }
      return t;
   }

   /**
    * Returns an ArrayList of all Bonds below the present FNode.
    */
   public List getBondList(){
      FNode o;
      ArrayList t = new ArrayList();
      Enumeration e = depthFirstEnumeration();
      while (e.hasMoreElements()){
         o = (FNode) e.nextElement();
         if (o instanceof Bond){
            t.add(o);
         }
      }
      return t;
   }

   /**
    * Sets the name of this NodeObect to n.
    */
   public void setName(String n){
      name = n;
   }

   /**
    * Returns the name of this FNode.
    */
   public String getName(){
      return name;
   }

   public void setSelected(boolean b){
      selected = b;
      FNode node;
      for (Enumeration e = children(); e.hasMoreElements(); ){
         node = (FNode) e.nextElement();
         node.setSelected(b);
      }
   }

   public boolean isSelected(){
      return selected;
   }

   public double[] findCenter(boolean w){
      double[] Rc = {0, 0, 0};
      double sum = 0, mass = 1;
      Atom a;
      ListIterator li;

      for (li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (w){
            mass = a.getMass();
            sum += mass;
         }
         Rc[0] += mass * a.getX();
         Rc[1] += mass * a.getY();
         Rc[2] += mass * a.getZ();
      }
      if (!w){
         sum = getAtomList().size();
      }
      for (int i = 0; i < 3; i++){
         Rc[i] /= sum;
      }
      return Rc;
   }
}
