/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.io.File;
import java.text.*;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import ffe.net.TinkerUpdate;
import ffe.panels.*;
import ffe.parsers.MMFilter;

/**
 * The FSystem class contains the information needed to represent one system in Force Field Explorer.
 */
public final class FSystem
   extends FGroup{

   // FSystem member variables
   private File file;
   private Vector3d offset;

   private byte filetype;
   private File keyFile;
   private Hashtable keywords = new Hashtable();
   private Vector altLoc = null;

   private File logFile;
   private MMFilter mmfilter;
   private String forcefield;
   private boolean closing = false;
   private int cycles = 1;
   private int currentCycle = 1;
   private Trajectory trajectory = null;

   // Data Nodes
   private FNode ions = new FNode("Ions");
   private FNode water = new FNode("Water");
   private FNode molecules = new FNode("Hetero Molecules");

   // TINKER Simulation variables
   private Vector3d box = new Vector3d();
   private Vector3d angle = new Vector3d();
   private int simulation = 0;
   private double extent, energy;
   private double time, temperature;
   private int step;

   // 3D Graphics Nodes - There is a diagram explaining the FSystem Scenegraph below
   private BranchGroup branchGroup;
   private TransformGroup originToRot;
   private Transform3D originToRotT3D;
   private Vector3d originToRotV3D;
   private TransformGroup rotToCOM;
   private Transform3D rotToCOMT3D;
   private Vector3d rotToCOMV3D;
   private BranchGroup base;
   private Switch switchGroup;
   private Shape3D wire;
   private BranchGroup vrml;
   private BranchGroup atomgroup, bondgroup;
   private Atom[] atomLookUp;

   private LineAttributes lineAttributes;
   private boolean vrmlShowing = false;
   private boolean visible = false;
   private boolean j3datoms = false;
   private boolean j3dbonds = false;

   // These lists are used frequently, so they are stored rather than regenerated
   private ArrayList atomlist, bondlist, termlist, moleculeList;

   // Constructors
   public FSystem(String name, boolean bk){
      super(name, bk);
      getFGroups().setName("MacroMolecules");
      add(molecules);
      add(ions);
      add(water);
   }

   public FSystem(String name, FNode Polymers, boolean bk){
      super(name, Polymers, bk);
   }

   public void addKeyword(Keyword k){
      if (keywords.containsKey(k.getKeyword())){
         return;
      }
      keywords.put(k.getKeyword(), k);
   }

   public Keyword getKeyword(String k){
      return (Keyword) keywords.get(k);
   }

   public Hashtable getKeywords(){
      return keywords;
   }

   public void setKeywords(Hashtable k){
      keywords = k;
   }

   protected void removeLeaves(){
      super.removeLeaves();

      FNode macroMol = getFGroups();
      if (macroMol != null && macroMol.getChildCount() > 0){
         getFGroups().setName("Macromolecules " + "(" + macroMol.getChildCount() + ")");
      } else{
         remove(macroMol);
      }

      if (molecules.getChildCount() == 0){
         remove(molecules);
      } else{
         molecules.setName("Hetero Molecules " + "(" + molecules.getChildCount() + ")");
      }
      if (ions.getChildCount() == 0){
         remove(ions);
      } else{
         ions.setName("Ions " + "(" + ions.getChildCount() + ")");
      }
      if (water.getChildCount() == 0){
         remove(water);
      } else{
         water.setName("Water " + "(" + water.getChildCount() + ")");
      }
   }

   public void removeKeyword(Keyword kd){
      if (keywords.containsKey(kd.getKeyword())){
         keywords.remove(kd.getKeyword());
      }
   }

   public void addFNode(FNode o){
      ArrayList Polymers = getFGroupList();
      if (o instanceof Atom){
         Atom a = (Atom) o;
         Residue res = getResidue(a.getChain(), a.getResidueNumber(), a.getResidueName(), true);
         if (res == null){
            System.out.println("Atom with no Home in FSystem.AddFNode()");
            return;
         }
         res.addFNode(a);
      } else if (o instanceof Residue){
         Residue g = (Residue) o;
         String key = g.getPolymer();
         int index = Polymers.indexOf(new Polymer(key, getBondsKnown()));
         if (index != -1){
            Polymer c = (Polymer) Polymers.get(index); //Find the requested Polymer
            c.addFNode(g); //Add the residue to the Polymer
            setFinalized(false);
         } else{
            Polymer newc = new Polymer(key, getBondsKnown()); //Create a new Polymer
            newc.addFNode(g); //Add residue
            getFGroups().add(newc);
            setFinalized(false);
         }
      } else if (o instanceof Polymer){
         Polymer c = (Polymer) o;
         int index = Polymers.indexOf(c);
         if (index != -1){
         } else{
            getFGroups().add(c);
            setFinalized(false);
         }
      } else if (o instanceof Molecule){
         Molecule m = (Molecule) o;
         if (m.getAtomList().size() == 1){
            ions.add(m);
         } else if (XYZUtils.isWater((Atom) m.getAtomList().get(0))){
            water.add(m);
         } else{
            molecules.add(m);
         }
      }
   }

   public void destroy(){
      detach();
      Atom a;
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         a.destroy();
      }
      Term t;
      for (ListIterator li = getTermList().listIterator(); li.hasNext(); ){
         t = (Term) li.next();
         t.destroy();
      }
      if (atomgroup != null){
         atomgroup.removeAllChildren();
      }
      if (bondgroup != null){
         bondgroup.removeAllChildren();
      }
   }

   public BranchGroup getBranchGroup(){
      return branchGroup;
   }

   public TransformGroup getTransformGroup(){
      return originToRot;
   }

   public void finalize(boolean finalizeGroups){
      setFinalized(false);
      ArrayList Polymers = getFGroupList();
      if (finalizeGroups){
         for (ListIterator li = Polymers.listIterator(); li.hasNext(); ){
            ((FGroup) li.next()).finalize(true);
         }
         for (ListIterator li = getMoleculeList().listIterator(); li.hasNext(); ){
            Object o = li.next();
            if (!(o instanceof Molecule)){
               System.out.println("" + o.getClass().toString() + " " + o.toString());
            } else{
               ((Molecule) o).finalize(true);
            }
         }
      }
      createLists();
      updateBonds();
      createScene(!finalizeGroups);
      centerFSystem();
      extent = getExtent();
      removeLeaves();
      if (mmfilter != null){
         forcefield = mmfilter.getForceField();
         if (forcefield != null){
            int index = forcefield.lastIndexOf("\\");
            int index2 = forcefield.lastIndexOf("/");
            if (index2 > index){
               index = index2;
            }
            if (index > 0){
               forcefield = forcefield.substring(index + 1, forcefield.length());
            }
         }
         mmfilter.destroy();
         mmfilter = null;
      }
      setFinalized(true);
   }

   public void createLists(){
      atomlist = new ArrayList();
      bondlist = new ArrayList();
      termlist = new ArrayList();

      FNode o;
      Enumeration e = breadthFirstEnumeration();
      while (e.hasMoreElements()){
         o = (FNode) e.nextElement();
         if (o instanceof Atom){
            atomlist.add(o);
         } else if (o instanceof Bond){
            bondlist.add(o);
            termlist.add(o);
         } else if (o instanceof Term){
            termlist.add(o);
         }
      }
   }

   public void switchView(int viewtype){
      // Just Detach the whole system branch group
      if (viewtype == GraphicsContext.DESTROY){
         switchGroup.setWhichChild(Switch.CHILD_NONE);
         visible = false;
      } else if (viewtype == GraphicsContext.VRML){
         if (vrmlShowing){
            switchGroup.setWhichChild(0);
            vrmlShowing = false;
         } else{
            switchGroup.setWhichChild(Switch.CHILD_ALL);
            vrmlShowing = true;
         }
         return;
      } else{
         setWireWidth(GraphicsContext.bondwidth);
         super.switchView(viewtype);
         for (ListIterator li = moleculeList.listIterator(); li.hasNext(); ){
            Molecule m = (Molecule) li.next();
            m.switchView(viewtype);
         }
         if (viewtype == GraphicsContext.INVISIBLE){
            switchGroup.setWhichChild(0);
         }
      }
   }

   public void setWireWidth(float f){
      if (wire == null){
         return;
      }
      lineAttributes.setLineWidth(f);
   }

   public boolean isVisible(){
      return visible;
   }

   public void setEnergy(double e){
      energy = e;
   }

   public void setTime(double t){
      time = t;
   }

   public void setTemperature(double t){
      temperature = t;
   }

   public void switchColor(int colortype, Color3f color, Material mat){
      FGroup group;
      for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ){
         group = (FGroup) li.next();
         group.switchColor(colortype, color, mat);
      }
      for (ListIterator li = moleculeList.listIterator(); li.hasNext(); ){
         Molecule m = (Molecule) li.next();
         m.switchColor(colortype, color, mat);
      }
   }

   public String toFFString(){
      StringBuffer sb = new StringBuffer();
      if (file != null){
         sb.append(file.getName());
      }
      if (forcefield != null){
         sb.append(" (");
         sb.append(forcefield);
         sb.append(")");
      }
      return sb.toString();
   }

   public String toString(){
      if (getFile() == null){
         return getName();
      } else{
         return toFileString();
      }
   }

   public String toFileString(){
      if (getFile() == null){
         return null;
      }
      StringBuffer sb = new StringBuffer(getFile().getAbsolutePath());
      if (forcefield != null){
         sb.append(" (");
         sb.append(forcefield);
         sb.append(")");
      }
      return sb.toString();
   }

   /**
    * The FSystem Branchgroup has two TransformGroups between it and the "base" node where gemoetry
    * is attached. If the point between the two transformations is where user rotation occurs.
    * For example, if rotating about the center of mass of the system, the RotToCOM transformation
    * will be an identity transformation (ie. none). If rotation is about some atom or group of atoms
    * within the system, then the RotToCOM transformation will be a translation from that point to
    * the COM.
    *
    * @param zero boolean
    * @return BranchGroup
    */

   public BranchGroup createScene(boolean zero){

      /* Scenegraph
           branchGroup
                |
           originToRot              z                 COM
                |                   |    / y           ^
             rotToCOM               |  /               |  RotToCOM
                |                   |/                 |
           switchGroup      VW      \ ---------------> |
           _____|____       Origin   \   originToRot
          |          |                \
         base       vrml               \x
          ________|__________
         |        |          |
        wire  atomgroup bondgroup

       */

      originToRotT3D = new Transform3D();
      originToRotV3D = new Vector3d();
      // Check to see if this FSystem has already been rendered
      // If it has, detach it from the scene
      if (branchGroup != null){
         branchGroup.detach();
         originToRot.getTransform(originToRotT3D);
         //rotToCOM.removeAllChildren();
      }
      originToRot = new TransformGroup(originToRotT3D);

      branchGroup = new BranchGroup();
      rotToCOM = new TransformGroup();
      rotToCOMT3D = new Transform3D();
      rotToCOMV3D = new Vector3d();

      // Set capabilities needed for picking and moving the FSystem
      branchGroup.setCapability(BranchGroup.ALLOW_DETACH);
      originToRot.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      originToRot.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      originToRot.setCapability(TransformGroup.ENABLE_PICK_REPORTING);
      rotToCOM.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      rotToCOM.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

      // Put the FSystem in the middle of the scene
      if (zero){
         originToRotV3D.set(0.0, 0.0, 0.0);
         originToRotT3D.set(originToRotV3D);
         originToRot.setTransform(originToRotT3D);
      }

      wire = renderWire();
      switchGroup = new Switch(Switch.CHILD_NONE);
      base = new BranchGroup();
      base.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
      base.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
      switchGroup.addChild(base);
      if (wire != null){
         base.addChild(wire);
      }
      if (vrml != null){
         switchGroup.addChild(vrml);
         vrmlShowing = true;
      }
      switchGroup.setWhichChild(Switch.CHILD_ALL);
      switchGroup.setCapability(Switch.ALLOW_SWITCH_WRITE);

      rotToCOM.addChild(switchGroup);
      originToRot.addChild(rotToCOM);
      branchGroup.addChild(originToRot);
      branchGroup.compile();
      return branchGroup;
   }

   public Node getWireFrame(){
      return wire;
   }

   public void createBox(){
      int vertices = 8;
      LineArray la = new LineArray(4 * vertices,
                     GeometryArray.COORDINATES | GeometryArray.COLOR_4 | GeometryArray.NORMALS);
      la.setCapability(LineArray.ALLOW_COORDINATE_WRITE);
      la.setCapability(LineArray.ALLOW_COORDINATE_READ);
      la.setCapability(LineArray.ALLOW_COLOR_WRITE);
      la.setCapability(LineArray.ALLOW_COUNT_READ);
      la.setCapability(LineArray.ALLOW_INTERSECT);
      la.setCapability(LineArray.ALLOW_FORMAT_READ);
      // Create a normal
      //for (ListIterator li = bondlist.listIterator(); li.hasNext(); ){
      //   la.setCoordinate(i, a1);
      // la.setColor(i, col);
      //la.setNormal(i++, a1);
      //}
      ColoringAttributes cola = new ColoringAttributes(new Color3f(),
                                ColoringAttributes.SHADE_GOURAUD);
      Appearance app = new Appearance();
      lineAttributes = new LineAttributes();
      lineAttributes.setLineWidth(GraphicsContext.bondwidth);
      lineAttributes.setCapability(LineAttributes.ALLOW_WIDTH_WRITE);
      lineAttributes.setLineAntialiasingEnable(true);
      app.setLineAttributes(lineAttributes);
      app.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_READ);
      app.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
      RenderingAttributes ra = new RenderingAttributes();
      ra.setAlphaTestValue(0.1f);
      ra.setAlphaTestFunction(RenderingAttributes.GREATER);
      ra.setDepthBufferEnable(true);
      ra.setDepthBufferWriteEnable(true);
      app.setRenderingAttributes(ra);
      app.setColoringAttributes(cola);
      Shape3D wire = new Shape3D(la, app);
      //PickTool.setCapabilities(wire, PickTool.INTERSECT_COORD);
      wire.setUserData(this);
      wire.setBounds(new BoundingSphere(new Point3d(0, 0, 0), 10.0));
      try{
         wire.setBoundsAutoCompute(false);
      } catch (Exception e){
         //System.out.println("" + e);
      }
      wire.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      wire.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
      //return wire;
   }

   public void removeJ3D(){
      if (branchGroup != null && branchGroup.isLive()){
         return;
      }
      /*if (rotToCOM != null) {
          originToRot.removeAllChildren();
               }*/
      if (atomgroup != null){
         atomgroup.removeAllChildren();
      }
      if (bondgroup != null){
         bondgroup.removeAllChildren();
      }
      j3datoms = false;
      j3dbonds = false;
   }

   /**
    * Rotate about a point in given in the System's Local Coordinates
    * @param v Vector3d
    */
   public void rotateAbout(Vector3d v){
      Vector3d newRotPoint = new Vector3d(v);

      originToRot.getTransform(originToRotT3D);
      originToRotT3D.get(originToRotV3D);
      originToRotT3D.setTranslation(new Vector3d(0, 0, 0));

      rotToCOM.getTransform(rotToCOMT3D);
      rotToCOMT3D.get(rotToCOMV3D);
      newRotPoint.add(rotToCOMV3D);

      originToRotT3D.transform(newRotPoint);
      newRotPoint.add(originToRotV3D);
      originToRotT3D.setTranslation(newRotPoint);

      rotToCOMV3D.set(v);
      rotToCOMV3D.negate();
      rotToCOMT3D.setTranslation(rotToCOMV3D);

      originToRot.setTransform(originToRotT3D);
      rotToCOM.setTransform(rotToCOMT3D);
   }

   public void centerView(boolean rot, boolean trans){
      originToRot.getTransform(originToRotT3D);
      if (rot){
         Matrix3d m3d = new Matrix3d();
         m3d.setIdentity();
         originToRotT3D.setRotation(m3d);
         //rotToCOMT3D.setRotation(m3d);
      }
      if (trans){
         originToRotV3D.set(offset);
         originToRotT3D.set(originToRotV3D);
      }
      originToRot.setTransform(originToRotT3D);
      //rotToCOM.setTransform(rotToCOMT3D);
   }

   public void initAtoms(FNode f){
      initAtoms(f.getAtomList());
      if (f == this){
         j3datoms = true;
      }
   }

   public void initAtoms(List f){
      if (j3datoms == true){
         return;
      }

      if (atomgroup == null){
         atomgroup = new BranchGroup();
         atomgroup.setCapability(BranchGroup.ALLOW_DETACH);
      } else{
         atomgroup.detach();
      }

      Atom a;
      for (ListIterator li = f.listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (a.initJ3D()){
            if (a.getJ3D() != null){
               atomgroup.addChild(a.getJ3D());
            }
         }
      }
      if (atomgroup.numChildren() > 0){
         base.addChild(atomgroup);
      }
      return;
   }

   public void initBonds(FNode f){
      initBonds(f.getBondList());
      if (f == this){
         j3dbonds = true;
      }
   }

   public void initBonds(List f){
      if (j3dbonds == true){
         return;
      }
      if (bondgroup == null){
         bondgroup = new BranchGroup();
         bondgroup.setCapability(BranchGroup.ALLOW_DETACH);
      } else{
         bondgroup.detach();
      }

      Bond b;
      for (ListIterator li = f.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         if (b.initJ3D()){
            bondgroup.addChild(b.getJ3D1());
            bondgroup.addChild(b.getJ3D2());
         }
      }
      if (bondgroup.numChildren() > 0){
         base.addChild(bondgroup);
      }
   }

   public void detach(){
      if (branchGroup.isLive()){
         branchGroup.detach();
      }
   }

   public double getEnergy(){
      return energy;
   }

   public double getExtent(){

      double[] Rc = {0, 0, 0};
      int num = getAtomList().size();
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).getXYZ(a); Rc[0] += a[0];
         Rc[1] += a[1];
         Rc[2] += a[2];
      }
      for (int i = 0; i < 3; i++){
         Rc[i] /= num;
      }

      double r, d = 0;
      double[] xyz = new double[3];
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).getXYZ(xyz); AtomUtils.diff(xyz, Rc, xyz);
         r = AtomUtils.r(xyz);
         if (d < r){
            d = r;
         }
      }
      return d;
   }

   public static final double KCAL_TO_KJ = 4.184;

   public void centerFSystem(){
      double center[] = findCenter(false);
      offset = new Vector3d(center);

      offset.negate();
      originToRotV3D.set(offset);
      originToRotT3D.setTranslation(originToRotV3D);
      originToRot.setTransform(originToRotT3D);

      rotToCOMT3D.setIdentity();
      rotToCOM.setTransform(rotToCOMT3D);

      offset.negate();
      rotateAbout(offset);
      originToRotT3D.get(offset);

      return;
   }

   private static double[] a = new double[3];
   public void CenterAt(double[] d){
      double[] Rc = {0, 0, 0};
      double[] c = new double[3];
      ListIterator li;
      int i, num = getAtomList().size();
      for (li = getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).getXYZ(a); Rc[0] += a[0];
         Rc[1] += a[1];
         Rc[2] += a[2];
      }
      for (i = 0; i < 3; i++){
         Rc[i] /= num;
      }
      AtomUtils.diff(d, Rc, c);
      for (li = getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).move(c);
      }
   }

   /*
     public void formLinks(){
    ListIterator li;
    String[] connection;
    Residue g1, g2;
    int num;
    Joint j;
    FNode joints = getTerms();
    joints.removeAllChildren();
    for(li = links.listIterator(); li.hasNext();){
     connection = (String[]) li.next();
     num = (new Integer(connection[2])).intValue();
     g1 = getResidue(connection[0], num);
     num = (new Integer(connection[3])).intValue();
     g2 = getResidue(connection[1], num);
     if (g1 != null && g2 != null){
      j = Joiner(g1, g2);
      joints.add(j);
     } else
      System.out.println("Link not formed" + g1 + " " +  g2);
    }
     }*/

   public void MoveCenter(double[] d){
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).move(d);
      }
   }

   /*
        public double[][] getPhiPsi() {
       double[][] phipsi;
       ArrayList a[], b[] = new ArrayList[2];
       b[0] = new ArrayList();
       b[1] = new ArrayList();
    for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ) {
           a = ( (Polymer) li.next()).getPhiPsiList();
           b[0].addAll(a[0]);
           b[1].addAll(a[1]);
       }
       phipsi = new double[2][b[0].size()];
       for (int i = 0; i < b[0].size(); i++) {
           phipsi[0][i] = ( (Dihedral) b[0].get(i)).getValue();
           phipsi[1][i] = ( (Dihedral) b[1].get(i)).getValue();
       }
       return phipsi;
        }*/

   public ArrayList getBackBoneAtoms(){
      ArrayList backbone = new ArrayList();
      Atom a;
      for (ListIterator li = atomlist.listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (a.equals(new Atom("CA"))){
            backbone.add(a);
//			else if (a.equals(new Atom("C"))) backbone.add(a);
//			else if (a.equals(new Atom("N"))) backbone.add(a);
         }
      }
      return backbone;
   }

   public ArrayList getResidueList(){
      ArrayList residues = new ArrayList();
      ListIterator li, lj;
      FNode o;
      Polymer c;
      for (li = getFGroupList().listIterator(); li.hasNext(); ){
         o = (FNode) li.next();
         if (o instanceof Polymer){
            c = (Polymer) o;
            for (lj = c.getFGroupList().listIterator(); lj.hasNext(); ){
               o = (FNode) lj.next();
               if (o instanceof Residue){
                  residues.add(o);
               }
            }
         }
      }
      return residues;
   }

   public void sidePolymerCOM(){
      ArrayList residues = getResidueList();
      Residue r;
      ListIterator li;
      for (li = residues.listIterator(); li.hasNext(); ){
         r = (Residue) li.next();
         r.printSidePolymerCOM();
      }
   }

   public Polymer getPolymer(String chainName, boolean create){
      for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ){
         FNode node = (FNode) li.next();
         if (node instanceof Polymer){
            if (node.getName().equals(chainName)){
               return (Polymer) node;
            }
         }
      }
      if (create){
         Polymer polymer = new Polymer(chainName, false, true);
         addFNode(polymer);
         return polymer;
      }
      return null;
   }

   public Residue getResidue(String chainName, int resNum, String resName, boolean create){
      // Find/Create the chain
      Polymer polymer = getPolymer(chainName, create);
      if (polymer == null){
         return null;
      }
      // If the chain was found/created, find/create the residue
      return polymer.getResidue(resName, resNum, create);
   }

   public void setLogFile(File f){
      logFile = f;
   }

   public File getLogFile(){
      if (logFile == null){
         if (file == null){
            return null;
         } else{
            String fileName = file.getName();
            int dot = fileName.lastIndexOf(".");
            fileName = fileName.subSequence(0, dot) + ".log";
            logFile = new File(fileName);
         }
      }
      return logFile;
   }

   public void setFile(File f){
      if (f == null || !f.exists()){
         return;
      }
      file = f;
   }

   public File getFile(){
      return file;
   }

   public void setKeyFile(File f){
      keyFile = f;
   }

   public File getKeyFile(){
      return keyFile;
   }

   public void setMMFilter(MMFilter m){
      mmfilter = m;
   }

   public void setOffset(Vector3d o){
      offset = o;
   }

   public Vector3d getOffset(){
      return offset;
   }

   public MMFilter getMMFilter(){
      return mmfilter;
   }

   public List getAtomList(){
      if (atomlist == null){
         createLists();
      }
      return atomlist;
   }

   public List getBondList(){
      if (bondlist == null){
         createLists();
      }
      return bondlist;
   }

   public List getMoleculeList(){
      if (moleculeList == null){
         moleculeList = new ArrayList();
         if (ions.getChildCount() > 0){
            moleculeList.addAll(ions.getList());
         }
         if (water.getChildCount() > 0){
            moleculeList.addAll(water.getList());
         }
         if (molecules.getChildCount() > 0){
            moleculeList.addAll(molecules.getList());
         }
      }
      return moleculeList;
   }

   public double getTemperature(){
      return temperature;
   }

   public ArrayList getTermList(){
      if (termlist == null){
         createLists();
      }
      return termlist;
   }

   public double getTime(){
      return time;
   }

   public Trajectory getTrajectory(){
      return trajectory;
   }

   public void setTrajectory(Trajectory t){
      trajectory = t;
   }

   public void setStep(int s){
      step = s;
   }

   public void setAngle(double a[]){
      if (a == null){
         return;
      }
      if (angle == null){
         angle = new Vector3d();
      }
      angle.set(a);
   }

   public void setBox(double b[]){
      if (b == null){
         return;
      }
      if (box == null){
         box = new Vector3d();
      }
      box.set(b);
   }

   public void setClosing(boolean b){
      closing = b;
   }

   public boolean isSimulation(){
      if (simulation == TinkerUpdate.SIMULATION){
         return true;
      }
      return false;
   }

   public boolean isOptimization(){
      if (simulation == TinkerUpdate.OPTIMIZATION){
         return true;
      }
      return false;
   }

   public void setSimulation(int type){
      simulation = type;
   }

   public String getTimeString(){
      DecimalFormat dform = (DecimalFormat) NumberFormat.getInstance();
      dform.applyPattern("0.000E0");
      StringBuffer sb = new StringBuffer("Time: ");
      sb.append(dform.format(time)).append(" picoseconds");
      return sb.toString();
   }

   public String getStepString(){
      StringBuffer sb = new StringBuffer("Step: ");
      sb.append(Integer.toString(step));
      return sb.toString();
   }

   public String getEnergyString(){
      StringBuffer sb = new StringBuffer("Energy: ");
      DecimalFormat dform = (DecimalFormat) NumberFormat.getInstance();
      dform.applyPattern("0.00000E0");
      sb.append(dform.format(energy)).append(" kcal/mole");
      return sb.toString();
   }

   public boolean isClosing(){
      return closing;
   }

   public void setFileType(byte ftype){
      filetype = ftype;
   }

   public void setForceField(String f){
      if (f != null){
         forcefield = f;
      } else{
         Keyword kd = (Keyword) keywords.get("PARAMETERS");
         if (kd != null){
            forcefield = kd.getEntry(0);
         } else{
            forcefield = null;
         }
      }
   }

   public byte getFileType(){
      return filetype;
   }

   public void addAltLocation(String s){
      if (altLoc == null){
         altLoc = new Vector();
      }
      if (!altLoc.contains(s)){
         altLoc.add(s);
      }
   }

   public String[] getAltLocations(){
      if (altLoc == null || altLoc.size() == 0){
         return null;
      } else{
         String[] names = new String[altLoc.size()];
         for (int i = 0; i < altLoc.size(); i++){
            names[i] = (String) altLoc.get(i);
         }
         return names;
      }
   }

   public String[] getChainNames(){
      Vector temp = new Vector();
      for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ){
         FNode node = (FNode) li.next();
         if (node instanceof Polymer){
            temp.add(((Polymer) node).getName());
         }
      }
      if (temp.size() == 0){
         return null;
      } else{
         String[] names = new String[temp.size()];
         for (int i = 0; i < temp.size(); i++){
            names[i] = (String) temp.get(i);
         }
         return names;
      }
   }

   public String getParameterFile(){
      return forcefield;
   }

   public Atom getAtomFromWireVertex(int i){
      if (atomLookUp != null && atomLookUp.length > i){
         return atomLookUp[i];
      }
      return null;
   }

   public void setCurrentCycle(int c){
      if (c <= cycles && c > 0){
         currentCycle = c;
         ListIterator li;
         Atom a;
         for (li = getAtomList().listIterator(); li.hasNext(); ){
            a = (Atom) li.next();
            a.setCurrentCycle(currentCycle);
         }
      }
   }

   public void setVRML(BranchGroup v){
      vrml = v;
   }

   public int getCurrentCycle(){
      return currentCycle;
   }

   public void setCycles(int c){
      cycles = c;
   }

   public int getCycles(){
      return cycles;
   }

   private Shape3D renderWire(){
      int numbonds = bondlist.size();

      if (numbonds < 1){
         return null;
      }

      Vector3d bondmidpoint = new Vector3d();
      double[] mid = {0, 0, 0};
      Vector3d v1 = new Vector3d();
      Vector3d v2 = new Vector3d();
      float[] a1 = {0, 0, 0};
      float[] a2 = {0, 0, 0};
      float[] col = new float[4];
      Bond bond;
      Atom atom1, atom2;

      LineArray la = new LineArray(4 * numbonds,
                     GeometryArray.COORDINATES | GeometryArray.COLOR_4 | GeometryArray.NORMALS);
      la.setCapability(LineArray.ALLOW_COORDINATE_WRITE);
      la.setCapability(LineArray.ALLOW_COORDINATE_READ);
      la.setCapability(LineArray.ALLOW_COLOR_WRITE);
      la.setCapability(LineArray.ALLOW_COUNT_READ);
      la.setCapability(LineArray.ALLOW_INTERSECT);
      la.setCapability(LineArray.ALLOW_FORMAT_READ);
      atomLookUp = new Atom[4 * numbonds];

      int i = 0;
      col[3] = 0.9f;
      for (ListIterator li = bondlist.listIterator(); li.hasNext(); ){
         bond = (Bond) li.next();
         bond.setWire(la, i);
         atom1 = bond.getAtom(0);
         atom2 = bond.getAtom(1);
         atom1.getV3D(v1);
         atom2.getV3D(v2);
         a1[0] = (float) v1.x;
         a1[1] = (float) v1.y;
         a1[2] = (float) v1.z;
         a2[0] = (float) v2.x;
         a2[1] = (float) v2.y;
         a2[2] = (float) v2.z;

         // Find the bond center
         bondmidpoint.add(v1, v2);
         bondmidpoint.scale(0.5d);
         bondmidpoint.get(mid);
         // Atom #1
         Graphics3D.colors[atom1.getAtomicNum()].get(col);
         atomLookUp[i] = atom1;
         la.setCoordinate(i, a1);
         la.setColor(i, col);
         la.setNormal(i, a2);
         i++;
         atomLookUp[i] = atom1;
         la.setCoordinate(i, mid);
         la.setColor(i, col);
         la.setNormal(i, a2);
         i++;
         // Atom #2
         Graphics3D.colors[atom2.getAtomicNum()].get(col);
         atomLookUp[i] = atom2;
         la.setCoordinate(i, a2);
         la.setColor(i, col);
         la.setNormal(i, a1);
         i++;
         atomLookUp[i] = atom2;
         la.setCoordinate(i, mid);
         la.setColor(i, col);
         la.setNormal(i, a1);
         i++;
      }
      ColoringAttributes cola = new ColoringAttributes(new Color3f(),
                                ColoringAttributes.SHADE_GOURAUD);
      Appearance app = new Appearance();
      lineAttributes = new LineAttributes();
      lineAttributes.setLineWidth(GraphicsContext.bondwidth);
      lineAttributes.setCapability(LineAttributes.ALLOW_WIDTH_WRITE);
      lineAttributes.setLineAntialiasingEnable(true);
      app.setLineAttributes(lineAttributes);
      app.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_READ);
      app.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
      RenderingAttributes ra = new RenderingAttributes();
      ra.setAlphaTestValue(0.1f);
      ra.setAlphaTestFunction(RenderingAttributes.GREATER);
      ra.setDepthBufferEnable(true);
      ra.setDepthBufferWriteEnable(true);
      app.setRenderingAttributes(ra);
      app.setColoringAttributes(cola);
      Shape3D wire = new Shape3D(la, app);
      //PickTool.setCapabilities(wire, PickTool.INTERSECT_COORD);
      wire.setUserData(this);
      wire.setBounds(new BoundingSphere(new Point3d(0, 0, 0), 1000.0));
      try{
         wire.setBoundsAutoCompute(false);
      } catch (Exception e){
         //System.out.println("" + e);
      }
      wire.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      wire.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
      wire.setCapability(Shape3D.ALLOW_LOCAL_TO_VWORLD_READ);
      return wire;
   }

}
