/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.ListIterator;

/**
 * The FSystemUtils class contains methods for manipulating a system of atoms.
 */
public final class FSystemUtils{

   public static final double tol = 0.000001;

   /**
    * Finds the Root Mean Square Coordinate Deviation between FSystems m1 and m2
    * provided they have the same number of atoms.
    */

   private static double[] da = new double[3];
   private static double[] db = new double[3];
   public static double RMSCoordDev(FSystem m1, FSystem m2){
      Atom a1, a2;
      double[] d = new double[3];
      if (m1 == null){
         System.out.println("FSystem 1 not loaded");
         return 0;
      } else if (m2 == null){
         System.out.println("FSystem 1 not loaded");
         return 0;
      }
      int n1 = m1.getAtomList().size();
      int n2 = m2.getAtomList().size();
      double rms = 0;
      if (n1 == n2){
         ListIterator li, lj;
         for (li = m1.getAtomList().listIterator(), lj = m2.getAtomList().listIterator();
            li.hasNext(); ){
            a1 = (Atom) li.next();
            a2 = (Atom) lj.next();
            a1.getXYZ(da);
            a2.getXYZ(db);
            AtomUtils.diff(da, db, d);
            rms += d[0] * d[0] + d[1] * d[1] + d[2] * d[2];
         }
         return Math.sqrt(rms / n1);
      }
      System.out.println("Can't do RMS, number of Atoms not equal");
      return 0;
   }

   /**
    * Centers the FSystem m at (0,0,0), using either geometry or mass-weighted
    * moments if w is true.
    */
   public static void Center(FSystem m, boolean w){
      //First find the center of mass if w == true, the center of geometry if w == false
      double[] Rc = {0, 0, 0};
      double sum = 0, mass = 1;
      Atom a;
      ListIterator li;

      for (li = m.getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (w){
            mass = a.getMass();
            sum += mass;
         }
         Rc[0] += mass * a.getX();
         Rc[1] += mass * a.getY();
         Rc[2] += mass * a.getZ();
      }
      if (!w){
         sum = m.getAtomList().size();
      }
      for (int i = 0; i < 3; i++){
         Rc[i] /= -sum;
      }
      for (li = m.getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).move(Rc);
      }
   }

   /**
    * Finds the center of mass (if w is true) or geometry and returns it.
    */
   public static double[] FindCenter(FSystem m, boolean w){
      //First find the center of mass if w == true, the center of geometry if w == false
      double[] Rc = {0, 0, 0};
      double sum = 0, mass = 1;
      Atom a;
      ListIterator li;

      for (li = m.getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (w){
            mass = a.getMass();
            sum += mass;
         }
         Rc[0] += mass * a.getX();
         Rc[1] += mass * a.getY();
         Rc[2] += mass * a.getZ();
      }
      if (!w){
         sum = m.getAtomList().size();
      }
      for (int i = 0; i < 3; i++){
         Rc[i] /= sum;
      }
      return Rc;
   }

   /**
    * Performs a principle axis transform on FSystem m.
     public static void PrincipalAxisTransform(FSystem m){
    if (m == null){
     System.out.println("Principle Axis Transform: No FSystem Loaded");
     return;
    }
    Matrix r, i, eig, u;
    double rsquared;
    Matrix inertia = new Matrix(3,3);
    int n = m.getAtomList().size();
    double[][] coords = new double[3][n];
    double[][] d = new double[3][1];
    double[] c = new double[3];
    int j = 0;
    for(ListIterator li = m.getAtomList().listIterator(); li.hasNext(); ){
     c = ((Atom) li.next()).getXYZ();
     coords[0][j] = d[0][0] = c[0];
     coords[1][j] = d[1][0] = c[1];
     coords[2][j++] = d[2][0] = c[2];
     r = new Matrix(d);
     i = Matrix.identity(3,3);
     rsquared = AtomUtils.r(c);
     rsquared *= rsquared;
     i.timesEquals(rsquared);
     r = r.times(r.transpose());
     inertia.plusEquals(i.minus(r));
    }
    eig = (new EigenvalueDecomposition(inertia)).getV();
    //Change the sign (if necessary) of the first eigenvector so that the determinant is positive.
    //This maintains the "handedness" of the FSystem
    if (eig.det() < 0)
     for(j=0; j<3; j++)
      eig.set(j,0,-eig.get(j,0));
    u = eig.transpose();
    Rotate(m, u);
     }*/

   /**
    * Uses Kabsch et al. routine for optimally superimposing two FSystems m1 and m2,
    * provided they have an equal number of atoms
     public static void Kabsch(FSystem m1, FSystem m2){
    if (m1 == null || m2 == null){
     System.out.println("Kabsch: m1 or m2 is null");
     return;
    }
    double[] c = FindCenter(m1, true);
    int numatoms = m1.getAtomList().size();
    if (numatoms != m2.getAtomList().size()){
    System.out.println("Kabsch: m1 and m2 don't have the same number of atoms");
     return;
    }
    double[][] coords = new double[numatoms][];
    int j = 0;
    for(ListIterator li = m1.getAtomList().listIterator(); li.hasNext(); )
     coords[j++] = AtomUtils.diff(((Atom) li.next()).getXYZ(), c);
    Center(m2, true);
    //Compute Kabsch matrix for m1 and m2
    Atom a;
    double sum, m;
    double[] n;
    Matrix kabsch = new Matrix(3,3);
    for(int i=0; i<3; i++)
     for(j=0; j<3; j++){
      sum = 0;
      int k = 0;
      for(ListIterator li = m2.getAtomList().listIterator(); li.hasNext(); ){
    a = (Atom) li.next();
    n = a.getXYZ();
    m = a.getMass();
    sum += m * coords[k++][j] * n[i];
      }
      kabsch.set(j,i,sum);
     }
    //Compute Kabsch rotation matrix
    //Construct a positive definite matrix by multiplying kabsch by its transform
    Matrix kabschkT = kabsch.times(kabsch.transpose());
    EigenvalueDecomposition eig = kabschkT.eig();
    Matrix e = eig.getV();
    double[] ev = eig.getRealEigenvalues();
    //Reset the eigenvector with the lowest eigenvalue
    e.set(0, 0, e.get(1,2)*e.get(2,1) - e.get(2,2)*e.get(1,1));
    e.set(1, 0, e.get(2,2)*e.get(0,1) - e.get(0,2)*e.get(2,1));
    e.set(2, 0, e.get(0,2)*e.get(1,1) - e.get(1,2)*e.get(0,1));
    Matrix RA = kabsch.times(e);
    //Normalize the two highest RA vectors by the inverse square root of the eigenvalues
    double[][] b = new double[3][3];
    for(int i=1; i<3; i++)
     for(j=0; j<3; j++)
      b[j][i] = RA.get(j,i) / Math.sqrt(ev[i]);
    //Calculate the b vector with the lowest eigenvalue
    b[0][0] = b[1][2]*b[2][1] - b[2][2]*b[1][1];
    b[1][0] = b[2][2]*b[0][1] - b[0][2]*b[2][1];
    b[2][0] = b[0][2]*b[1][1] - b[1][2]*b[0][1];
    //If necessary, change the sign of the lowest eigenvector.
    double[] d1 = new double[3];
    double[] d2 = new double[3];
    for(int i=0; i<3; i++){
     d1[i] = b[i][0];
     d2[i] = RA.get(i,0);
    }
    if(AtomUtils.dot(d1,d2) < 0)
     for(int i=0; i<3; i++)
      b[i][0] = -b[i][0];
    Matrix bm = new Matrix(b);
    Matrix u = bm.times(e.transpose());
    //System.out.println("Kabsch Rotation Matrix: ");
    //u.print(3,3);
    //System.out.println("Kabsch Rotation Matrix Determinant (Should be close to 1.000): " + u.det());
    //Apply the rotation matrix to FSystem 2, then center it over FSystem 1
    Rotate(m2, u);
    Translate(m2, c);
     } */

   /**
    * Rotates the given FSystem m by the rotation Matrix u, where u is a
    * 3x3 rotation matrix in JAMA format
     public static void Rotate(FSystem m, Matrix u){
    if (m == null){
     System.out.println("Rotate: FSystem is null");
     return;
    }
    int n = m.getAtomList().size();
    double[][] coords = new double[3][n];
    double[] c = new double[3];
    //Get the current coordinates and insert them into a matrix
    int i = 0;
    for(ListIterator li = m.getAtomList().listIterator(); li.hasNext(); ){
     c = ((Atom) li.next()).getXYZ();
     coords[0][i] = c[0];
     coords[1][i] = c[1];
     coords[2][i++] = c[2];
    }
    Matrix r = new Matrix(coords);
    //Apply the rotation matrix
    r = u.times(r);
    //Use the results to reposition the atoms
    i = 0;
    for(ListIterator li = m.getAtomList().listIterator(); li.hasNext(); ){
     c[0] = r.get(0,i);
     c[1] = r.get(1,i);
     c[2] = r.get(2,i++);
     ((Atom) li.next()).moveto(c);
    }
     }*/

   /**
    * Translates FSystem m by the vector specified in d
    */
   public static void translate(FSystem m, double[] d){
      if (m == null){
         return;
      }
      for (ListIterator li = m.getAtomList().listIterator(); li.hasNext(); ){
         ((Atom) li.next()).move(d);
      }
   }
}
