/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.*;
import javax.media.j3d.Material;
import javax.vecmath.Color3f;

import ffe.panels.*;

/**
 * The Polymer class encapsulates a polypeptide or polynucleotide chain.
 */
public final class Polymer
   extends FGroup{

   private boolean link = false;
   private int maxResidue = 0;

   private static int count = 0;
   private int polymerNumber;

   /**
    * Default Polymer Construtor
    */
   public Polymer(){
      super();
      polymerNumber = ++count;
   }

   public Polymer(String n, boolean bk){
      super(n, bk);
      polymerNumber = ++count;
   }

   /**
    * Polymer Constructor
    * @param n Polymer indentifier, generally a letter in PDB files (ie A,B,C,etc)
    */
   public Polymer(String n, boolean bk, boolean l){
      this(n, bk);
      link = l;
   }

   /**
    * Polymer Constructor
    * @param n Polymer indentifier, generally a letter in PDB files (ie A,B,C,etc)
    * @param residues Represents a FNode where the Polymer's residues have been attached
    */
   public Polymer(String n, FNode residues, boolean bk){
      super(n, residues, bk);
      polymerNumber = ++count;
   }

   /**
    * Overidden equals method.
    * @param object Object to compare
    * @return True if object is not <b>this</b> Polymer, is of Class Polymer, and both object and this Polymer have identical names
    */
   public boolean equals(Object object){
      if (this == object){
         return true;
      } else if (object == null || getClass() != object.getClass()){
         return false;
      }

      Polymer other = (Polymer) object;
      return (getName() == other.getName());
   }

   /**
    * A generic method for adding a FNode to the Polymer.
    * @param o If the FNode is an Atom, an attempt is made to pass it to the residue
    * to which it belongs. Otherwise, if the FNode is a Residue, it will be added to the Polymer,
    * so long as its sequence number is not already in use
    */
   public void addFNode(FNode o){
      ArrayList residues = getFGroupList();
      //Add an Atom
      if (o instanceof Atom){
         Atom a = (Atom) o;
         System.out.println("Attempt to add Atom Polymer");
      } else if (o instanceof Residue){
         Residue g = (Residue) o;
         int num = g.getResidueNumber();
         Residue key = new Residue(num);
         if (residues.indexOf(key) != -1){
         } else{

            if (num > maxResidue){
               maxResidue = num;
            }
            getFGroups().add(g);
         }
      }
   }

   /**
    * Finalize should be called after all the Residues have been added to the Polymer.
    * This method in turn calls the Finalize method of each Residue, then forms Joints
    * between adjacent Residues in the Polymer
    */
   public void finalize(boolean finalizeGroups){
      ListIterator li, lj;
      ArrayList res = getFGroupList();
      setFinalized(false);
      //Finalize the residues in the Polymer
      if (finalizeGroups){
         for (li = res.listIterator(); li.hasNext(); ){
            ((Residue) li.next()).finalize(true);
         }
      }

      // Join the residues in the Polymer
      if (link){
         getFGroups().setName("Residues " + "(" + res.size() + ")");
         Joint j;
         FNode joints = getTerms();
         joints.removeAllChildren();
         List atoms = getAtomList();
         if (getBondsKnown()){
            Atom a;
            Bond b;
            for (li = atoms.listIterator(); li.hasNext(); ){
               a = (Atom) li.next();
               if (a.getBondCount() > 0){
                  for (lj = a.getBonds().listIterator(); lj.hasNext(); ){
                     b = (Bond) lj.next();
                     if (!b.sameGroup()){
                        if (b.getParent() == null){
                           j = joiner(b);
                           joints.add(j);
                        }
                     }
                  }
               }
            }
         } else{
            Residue next = getResidue(0);
            Residue current = null;
            for (int i = 1; i <= maxResidue; i++){
               current = next;
               next = getResidue(i);
               if (current != null && next != null){
                  j = joiner(current, next);
                  if (j != null){
                     joints.add(j);
                  }
               }
            }
         }
         getTerms().setName("Linkages " + "(" + joints.getChildCount() + ")");
      } else{
         getFGroups().setName("Sub-Groups " + "(" + res.size() + ")");
         if (getTerms().getParent() != null){
            remove(getTerms());
         }
      }
      removeLeaves();
      setFinalized(true);
   }

   public Residue getResidue(int resNum){
      for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ){
         FNode node = (FNode) li.next();
         if (node instanceof Residue){
            Residue residue = (Residue) node;
            if (residue.getResidueNumber() == resNum){
               return residue;
            }
         }
      }
      return null;
   }

   public Residue getResidue(String resName, int resNum, boolean create){
      for (ListIterator li = getFGroupList().listIterator(); li.hasNext(); ){
         FNode node = (FNode) li.next();
         if (node instanceof Residue){
            Residue residue = (Residue) node;
            if (residue.getName().equalsIgnoreCase(resName) && residue.getResidueNumber() == resNum){
               return residue;
            }
         }
      }
      if (!create){
         return null;
      }
      Residue residue = new Residue(resName, resNum, false);
      addFNode(residue);
      return residue;
   }

   public void setLink(boolean t){
      link = t;
   }

   public boolean getLink(){
      return link;
   }

   public void switchColor(int colortype, Color3f color, Material mat){
      ListIterator li;
      FGroup tgroup;

      // If coloring by Polymer, pass this Polymer's color
      if (colortype == GraphicsContext.byPOLYMER){
         int index = polymerNumber % 10;
         color = (Color3f) AtomUtils.colorlist[index];
         mat = (Material) Graphics3D.createMaterial(color);
      }

      for (li = getFGroupList().listIterator(); li.hasNext(); ){
         tgroup = (FGroup) li.next();
         tgroup.switchColor(colortype, color, mat);
      }
   }

   /**
    * Get the Phi Psi List for the Polymer
    * @return An ArrayList of Dihedral objects representing the Phi/Psi angles of the Polymer,
    * useful for creating Ramachandran plots
        public ArrayList[] getPhiPsiList() {
       FNode dihedrals;
       ListIterator li, lj;
       ArrayList[] phipsi = new ArrayList[2];
       phipsi[0] = new ArrayList();
       phipsi[1] = new ArrayList();
       Dihedral d;

       FNode joints = getTerms();
       for (li = joints.getListIterator(); li.hasNext(); ) {
           for (lj = ( (Joint) li.next()).dihedrals.getListIterator();
                lj.hasNext(); ) {
               d = (Dihedral) lj.next();
               if (d.equals(new Dihedral("CNCAC")) ||
                   d.equals(new Dihedral("CCANC"))) { //Phi
                   phipsi[0].add(d);
               }
               else if (d.equals(new Dihedral("NCCAN")) ||
                        d.equals(new Dihedral("NCACN"))) { //Psi
                   phipsi[1].add(d);
               }
           }
       }
       return phipsi;
        }*/
}
