/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import ffe.panels.*;

/**
 * The Residue class represents individual amino acids or nucleic acid bases.
 */
public final class Residue
   extends FGroup{

   private int resNumber; // Residue number in the Polymer
   private int ssType = NONE; // Secondary Structure Type
   private String chainName = null;

   /**
    * Default Constructor where num is this Residue's position in the Polymer.
    */
   public Residue(int num){
      resNumber = num;
   }

   public Residue(String name, boolean bk){
      super(name, bk);
   }

   /**
    * Name is the residue's 3 letter abbreviation and num is its position in the Polymer.
    */
   public Residue(String name, int num, boolean bk){
      this(name, bk);
      resNumber = num;
   }

   /**
    * As above, with atoms being a FNode with this Residue's atoms as child nodes
    */
   public Residue(String name, int num, FNode atoms, boolean bk){
      super(name, atoms, bk);
      resNumber = num;
      finalize(true);
   }

   /**
    * Overidden equals method that return true if object is not equals to this, is of
    * the same class, has the same parent Polymer and the same sequence number.
    */
   public boolean equals(Object object){
      if (this == object){
         return true;
      } else if (object == null || getClass() != object.getClass()){
         return false;
      }

      Residue other = (Residue) object;
      if (getParent() == null || other.getParent() == null){
         return (getResidueNumber() == other.getResidueNumber());
      } else if (getParent() == other.getParent()){
         return (getResidueNumber() == other.getResidueNumber());
      } else{
         return false;
      }
   }

   /**
    * Allows adding Atom FNodes to the Residue.
    */
   public void addFNode(FNode o){
      if (o instanceof Atom){
         getFGroups().add(o);
         setFinalized(false);
      } else{
         System.out.println("Can't add FNode to Residue, not of type Atom");
      }
   }

   static StringBuffer s = new StringBuffer();
   public String toString(){

      if (s.length() > 0){
         s.replace(0, s.length(), getName());
      } else{
         s.append(getName());
      }
      s.append(": " + resNumber + " ");

      /*
             double[] pos = findCenter(false);
       NumberFormat form = NumberFormat.getInstance();
             if (form  instanceof DecimalFormat) {
         DecimalFormat dform = (DecimalFormat) form;
         dform.applyPattern("##0.00");
         s.append(" (" + dform.format(pos[0]) + ", " + dform.format(pos[1]) +
                  ", " + dform.format(pos[2]) + ")");
             } */return s.toString();
   }

   /**
    * Prints "Residue Number: x" to stdout.
    */
   public void print(){
      System.out.println(toString());
      super.print();
   }

   private static Point3d point3d = new Point3d();
   public void drawLabel(Graphics3D graphics, J3DGraphics2D g2d, Node node, boolean labelAtoms,
      boolean labelResidues){
      if (labelResidues){
         Atom a = (Atom)this.getFGroups().getFirstChild();
         if (a != null && (a.isVisible() || a.isBondVisible())){
            double d[] = getCenter();
            point3d.x = d[0];
            point3d.y = d[1];
            point3d.z = d[2];
            Point2d point = graphics.getScreenCoordinate(node, point3d);
            g2d.drawString(getName(), (float) point.x, (float) point.y);
         }
      }
      if (labelAtoms){
         super.drawLabel(graphics, g2d, node, labelAtoms, labelResidues);
      }
   }

   public void switchColor(int colortype, Color3f color, Material mat){

      // If Color by Residue, pass this Residue's Color
      if (colortype == GraphicsContext.byRESIDUE){
         color = (Color3f) AtomUtils.colorlookup.get(getName());
         if (color == null){
            return;
         }
         mat = (Material) Graphics3D.createMaterial(color);
      } else if (colortype == GraphicsContext.bySTRUCTURE){
         color = ssColors[ssType];
         mat = Graphics3D.createMaterial(color);
      }

      super.switchColor(colortype, color, mat);
   }

   public void printSidePolymerCOM(){
      Atom a;
      Vector3d v = new Vector3d();
      Vector3d v2 = new Vector3d();
      int count = 0;
      for (ListIterator li = getAtomList().listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         if (!a.getID().equals("CA")){
            if (!a.getID().equals("N")){
               if (!a.getID().equals("C")){
                  if (!a.getID().equals("O")){
                     a.getV3D(v2);
                     v.add(v2);
                     count++;
                  }
               }
            }
         } else if (a.getID().equals("CA")){
            a.print();
         }
      }
      v.scale(1.0f / count);
      System.out.println(getName() + " " + v.toString());
   }

   /**
    * The Finalize method should be called once all atoms have been added to the Residue.
    * Geometry objects (Bonds, Angles, etc) are then formed, followed by a determination of
    * under-constrained (Dangeling) atoms.
    */
   public void finalize(boolean finalizeGeometry){
      setFinalized(false);
      getFGroups().setName("Atoms: " + getAtomList().size());
      if (finalizeGeometry){
         if (!getBondsKnown()){
            ConstructGeometry();
         } else{
            collectBonds();
         }
         removeLeaves();
      }
      //findDangelingAtoms();
      setCenter(findCenter(false));
      setFinalized(true);
   }

   //Public data access methods
   /**
    * Returns this Residue's sequence number.
    */
   public int getResidueNumber(){
      return resNumber;
   }

   /**
    * Returns this Residues Parent Polymer name.
    */
   public String getPolymer(){
      return new String(chainName);
   }

   /**
    * Returns the position of this Residue's Alpha Carbon (if it is an amino acid).
    */
   public Vector3d getAlpha3d(){
      Atom a = (Atom) getFNode("CA");
      if (a == null){
         return null;
      }
      Vector3d temp = new Vector3d();
      a.getV3D(temp);
      return temp;
   }

   /**
    * Returns this Residue's Hydrophobicity value.
    */
   public double getHydrophobicity(){
      Prop p = (Prop) aaprop.get(getName());
      if (p == null){
         return 0.7; //neutral
      }
      return p.hydrophobicity;
   }

   /**
    * Returns the Hydrophobicity of the Residue provided as an arguement.
    * (Must be the 3-letter abreviation)
    */
   public static double getHydrophobicity(String aa){
      Prop p = (Prop) aaprop.get(aa);
      if (p == null){
         return 0.7; //neutral
      }
      return p.hydrophobicity;
   }

   public void setNumber(int n){
      resNumber = n;
   }

   public void setPolymer(String n){
      chainName = new String(n);
   }

   public void setSSType(int i){
      ssType = i;
   }

   public static int NONE = 0;
   public static int HELIX = 1;
   public static int TURN = 2;
   public static int SHEET = 3;

   public static Color3f[] ssColors = new Color3f[4];
   static{
      ssColors[0] = GraphicsContext.whiteColor;
      ssColors[1] = GraphicsContext.helixColor;
      ssColors[2] = GraphicsContext.turnColor;
      ssColors[3] = GraphicsContext.sheetColor;
   }

   public static int parseSSType(String ss){
      String type = ss.trim().toUpperCase();
      if (type.equals("HELIX")){
         return HELIX;
      }
      if (type.equals("SHEET")){
         return SHEET;
      }
      if (type.equals("TURN")){
         return TURN;
      }
      return NONE;
   }

   //***Amino Acid Descriptors - Hydrophobicity/Tendency to form Secondary Structure***
    private static Hashtable aaprop = new Hashtable();
    static{
      aaprop.put("GLY", new Prop(0.83125, 0.56, 0.92, 1.64));
      aaprop.put("ALA", new Prop(0.86875, 1.29, 0.90, 0.78));
      aaprop.put("VAL", new Prop(0.93125, 0.91, 1.49, 0.47));
      aaprop.put("LEU", new Prop(0.94375, 1.30, 1.02, 0.59));
      aaprop.put("ILE", new Prop(0.96250, 0.97, 1.45, 0.51));
      aaprop.put("SER", new Prop(0.80625, 0.82, 0.95, 1.33));
      aaprop.put("THR", new Prop(0.84375, 0.82, 1.21, 1.03));
      aaprop.put("CYS", new Prop(0.89375, 1.11, 0.74, 0.80));
      aaprop.put("PRO", new Prop(0.75625, 0.52, 0.64, 1.91));
      aaprop.put("PHE", new Prop(1.00000, 1.07, 1.32, 0.58));
      aaprop.put("TYR", new Prop(0.72500, 0.72, 1.25, 1.05));
      aaprop.put("TRP", new Prop(0.88750, 0.99, 1.14, 0.75));
      aaprop.put("HIS", new Prop(0.75000, 1.22, 1.08, 0.69));
      aaprop.put("ASP", new Prop(0.19375, 1.04, 0.72, 1.41));
      aaprop.put("ASN", new Prop(0.46875, 0.90, 0.76, 1.23));
      aaprop.put("GLU", new Prop(0.25625, 1.44, 0.75, 1.00));
      aaprop.put("GLN", new Prop(0.51250, 1.27, 0.80, 0.97));
      aaprop.put("MET", new Prop(0.98125, 1.47, 0.97, 0.39));
      aaprop.put("LYS", new Prop(0.21875, 1.23, 0.77, 0.96));
      aaprop.put("ARG", new Prop(0.00000, 0.96, 0.99, 0.88));
   }

   private static class Prop{
      public double hydrophobicity, helix, sheet, turn;

      public Prop(double hy, double h, double s, double t){
         hydrophobicity = hy;
         helix = h;
         sheet = s;
         turn = t;
      }
   }
}
