/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.text.*;
import java.util.*;
import javax.media.j3d.Material;
import javax.vecmath.Color3f;

/**
 * The Term class is extended by all bonded geometry classes (bond, angle, dihedral, improper).
 */
public abstract class Term
   extends FNode{

   private String id;
   private String key; // Concatenated molecular mechanics keys
   protected Atom atoms[]; // Atoms that are used to form the term
   protected Bond bonds[]; // Bonds that are used to form the term
   private double value; // Value of the term
   private MMTerm mm; // Molecular Mechanics parameters

   /**
    * Default Constructor
    */
   public Term(){
      setAllowsChildren(false);
   }

   /**
    * Constructor which sets the Term's id.
    */
   public Term(String i){
      this();
      this.id = i;
   }

   public void destroy(){
      super.destroy();
      id = null;
      key = null;
      atoms = null;
      bonds = null;
      value = 0;
      mm = null;
   }

   /**
    * Overidden method that returns true if object is not equals to this,
    * is of the same Class and has the same id.
    */
   public final boolean equals(Object object){
      if (this == object){
         return true;
      } else if (object == null || getClass() != object.getClass()){
         return false;
      }

      Term other = (Term) object;
      return getID() == other.getID();
   }

   /**
    * Overidden toString Method returns the Term's id.
    */
   static StringBuffer s = new StringBuffer();
   public String toString(){
      if (s.length() > 0){
         s.replace(0, s.length(), id);
      } else{
         s.append(id);

      }
      NumberFormat form = NumberFormat.getInstance();
      DecimalFormat dform = null;
      update();
      if (form instanceof DecimalFormat){
         dform = (DecimalFormat) form;
         dform.applyPattern("###0.00");
         s.append(": " + dform.format(value));
      }
      return s.toString();
   }

   /**
    * This method sets the Term's id and key by concatenating the respective id and keys
    * of the Atoms that are used in forming the term. Order can be reversed for help in
    * finding corresponding Molecular Mechanics file entries for the Term.
    */
   private static StringBuffer idtemp = new StringBuffer();
   private static StringBuffer keytemp = new StringBuffer();
   public final void setID_Key(boolean reverse){
      Atom a;
      // Reuse the string buffers
      if (idtemp.length() > 0){
         idtemp.delete(0, idtemp.length() - 1);
      }
      if (keytemp.length() > 0){
         keytemp.delete(0, keytemp.length() - 1);
      }
      if (atoms != null){
         if (!reverse){
            for (int i = 0; i < atoms.length; i++){
               a = atoms[i];
               if (i != 0){
                  idtemp.append("-").append(a.getID());
               } else{
                  idtemp.append(a.getID());
               }
               keytemp.append(a.getKey());
            }
         } else{ //Reverse Order
            for (int i = 0; i < atoms.length; i++){
               a = atoms[i];
               if (i != 0){
                  idtemp.append("-").append(a.getID());
               } else{
                  idtemp.append(a.getID());
               }
               keytemp.append(a.getKey());
            }
         }
         id = idtemp.substring(1).toString().intern();
         key = keytemp.toString().intern();
      }
   }

   /**
    * Sets a Reference to Molecular Mechanics data (MMTerm) for the Term
    */
   public void setMM(Hashtable data){
      mm = (MMTerm) data.get(key);
      if (mm == null){
         setID_Key(true);
         mm = (MMTerm) data.get(key);
      }
   }

   public abstract void update();

   /**
    * Prints the toString method to stdout
    */
   public void print(){
      System.out.println(this);
   }

   /**
    * Sets the Term's value.
    */
   public void setValue(double v){
      value = v;
   }

   /**
    * Sets the Term's energy.
    */
   //public void setEnergy(double e)     { energy = e; }
   /**
    * Sets the Term's key.
    */
   public void setKey(String k){
      key = new String(k);
   }

   /**
    * Sets the Term's id.
    */
   public void setID(String i){
      id = new String(i);
   }

   public void switchColor(int colortype, Color3f color, Material mat){
      Atom atom;

      if (atoms == null){
         return;
      }

      for (int i = 0; i < atoms.length; i++){
         atom = atoms[i];
         atom.switchColor(colortype, color, mat);
      }
   }

   public void switchView(int viewtype){
      int i;
      for (i = 0; i < atoms.length; i++){
         atoms[i].switchView(viewtype);
      }
      for (i = 0; i < bonds.length; i++){
         bonds[i].switchView(viewtype);
      }
   }

   /**
    * Sets a reference to Molecular Mechcanics Data for this term.
    */
   public void setMM(MMTerm mmterm){
      mm = mmterm;
   }

   /**
    * Add a constituent Atom to the Term.
    */
   public void setAtoms(Atom a[]){
      atoms = a;
   }

   /**
    * Add a constituent Bond to the Term.
    */
   public void setBonds(Bond b[]){
      bonds = b;
   }

   /**
    * Get the Term's key.
    */
   public String getKey(){
      return new String(key);
   }

   /**
    * Get the Term's id.
    */
   public String getID(){
      return new String(id);
   }

   /**
    * Get the constituent Atom specified by index.
    */
   public Atom getAtom(int index){
      if (index >= 0 && index < atoms.length){
         return atoms[index];
      } else{
         return null;
      }
   }

   /**
    * Get the constituent Bond specified by index.
    */
   public Bond getBond(int index){
      if (index >= 0 && index < atoms.length){
         return bonds[index];
      } else{
         return null;
      }
   }

   /**
    * Returns an ArrayList of the Atoms that constitute this Term.
    */
   public List getAtomList(){
      if (atoms == null){
         return null;
      }
      return java.util.Arrays.asList(atoms);
   }

   /**
    * Returns an ArrayList of the Bonds that constitute this Term.
    */

   public List getBondList(){
      if (bonds == null){
         return null;
      }
      return (ArrayList) java.util.Arrays.asList(bonds);
   }

   /**
    * Get the Term's value.
    */
   public double getValue(){
      return value;
   }

   /**
    * Get the Term's energy.
    */
   //public double getEnergy()       { return energy; }
   /**
    * Get the Term's Molecular Mechanics reference.
    */
   public MMTerm getMM(){
      return mm;
   }

   public void setSelected(boolean b){
      super.setSelected(b);
      Atom a;
      for (int i = 0; i < atoms.length; i++){
         a = atoms[i];
         a.setSelected(b);
      }
      if (!(this instanceof Bond)){
         Bond bond;
         for (int i = 0; i < bonds.length; i++){
            bond = bonds[i];
            bond.setSelected(b);
         }
      }
   }
}
