/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.lang;

import java.text.*;
import java.util.*;

import ffe.panels.MainPanel;

/**
 * The XYZUtils class provides methods to locate functional units of a system.
 */
public final class XYZUtils{

   public static ArrayList arraylistpool = new ArrayList();

   public static void addArraytoPool(ArrayList a){
      a.clear();
      arraylistpool.add(a);
   }

   static int count = 0;
   public static ArrayList getArrayfromPool(){
      if (arraylistpool.isEmpty()){
         return new ArrayList();
      } else{
         return (ArrayList) arraylistpool.remove(0);
      }
   }

   // This routine sub-divides a TINKER system into groups of ions, water,
   // hetero molecules, and polynucleotides/polypeptides.
   public static void groupXYZ(FSystem m, ArrayList atoms, MainPanel main){
      Atom atom, seed = null;
      int num = 0;
      int waterNum = 0;
      int ionNum = 0;
      int heteroNum = 0;
      while (atoms.size() > 0){

         // Find a "seed" nitrogen. Carbons can be separated from the
         // chain backbone by sulfurs (ie. MET). Since sulfur is a termination condition
         // (don't cross disulfides) nitrogen is the easiest seed atom choice.
         for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
            seed = (Atom) li.next();
            if (seed.getAtomicNum() == 7){
               break;
            }
         }

         // No more nitrogens, so no DNA/RNA/Protein molecules can be left
         if (seed.getAtomicNum() != 7){
            ArrayList resatoms;
            while (atoms.size() > 0){
               atom = (Atom) atoms.get(0);
               // Check for a lone ion/noble gas
               if (atom.getBondCount() == 0){
                  ionNum++;
                  Molecule ion = new Molecule(atom.getName() + ": " + ionNum, true);
                  ion.addFNode(atom);
                  atoms.remove(0);
                  m.addFNode(ion);
                  continue;
                  // Check for water
               } else if (atom.getAtomicNum() == 8){
                  if (isWater(atom)){
                     waterNum++;
                     Molecule water = new Molecule("H2O: " + waterNum, true);
                     water.addFNode(atom);
                     atoms.remove(0);
                     ArrayList bonds = atom.getBonds();
                     for (ListIterator lj = bonds.listIterator(); lj.hasNext(); ){
                        Bond b = (Bond) lj.next();
                        Atom o = b.getOtherAtom(atom);
                        water.addFNode(o);
                        atoms.remove(o);
                     }
                     m.addFNode(water);
                     continue;
                  }
               }
               // Otherwise classify the molecule as a hetero
               heteroNum++;
               Molecule hetero = new Molecule("Hetero: " + heteroNum, true);
               resatoms = getArrayfromPool();
               collectAtoms((Atom) atoms.get(0), resatoms);
               while (resatoms.size() > 0){
                  atom = (Atom) resatoms.get(0);
                  resatoms.remove(0);
                  hetero.addFNode(atom);
                  atoms.remove(atom);
               }
               NumberFormat form = NumberFormat.getInstance();
               if (form instanceof DecimalFormat){
                  DecimalFormat dform = (DecimalFormat) form;
                  dform.applyPattern("0.##E0");
                  hetero.setName("Hetero " + dform.format(hetero.getMW()) + ": " + heteroNum);
               }
               m.addFNode(hetero);
            }
            seed = null;
            break;
         }

         if (seed != null){
            ArrayList backbone = findPolymer(atoms, seed, null);
            if (backbone.size() > 0){
               for (ListIterator li = backbone.listIterator(backbone.size()); li.hasPrevious(); ){
                  seed = (Atom) li.previous();
                  if (seed.getAtomicNum() == 7){
                     break;
                  }
               }
               backbone = findPolymer(atoms, seed, null);
            }

            Polymer c = new Polymer(PolymerLookup(num), true, true);
            if (backbone.size() > 2 && divideBackbone(backbone, c)){
               for (ListIterator li = c.getAtomList().listIterator(); li.hasNext(); ){
                  atom = (Atom) li.next();
                  atoms.remove(atom);
               }
               if (main != null){
                  main.getStatusBar().setText("  Sequenced Chain " + c.getName());
               }
               m.addFNode(c);
               num++;
            } else{
               heteroNum++;
               Molecule hetero = new Molecule("Hetero: " + heteroNum, true);
               atom = (Atom) backbone.get(0);
               ArrayList nonpolymer = getArrayfromPool();
               collectAtoms(atom, nonpolymer);
               for (ListIterator li = nonpolymer.listIterator(); li.hasNext(); ){
                  atom = (Atom) li.next();
                  hetero.addFNode(atom);
               }
               for (ListIterator li = hetero.getAtomList().listIterator(); li.hasNext(); ){
                  atom = (Atom) li.next();
                  atoms.remove(atom);
               }
               NumberFormat form = NumberFormat.getInstance();
               if (form instanceof DecimalFormat){
                  DecimalFormat dform = (DecimalFormat) form;
                  dform.applyPattern("0.##E0");
                  hetero.setName("Hetero " + dform.format(hetero.getMW()) + ": " + heteroNum);
               }
               m.addFNode(hetero);
            }
         }
      }
   }

   public static ArrayList findPolymer(ArrayList atoms, Atom currentAtom, ArrayList path){

      // End of Recursion conditions
      if (currentAtom.getParent() != null){
         return null;
      }

      int anum = currentAtom.getAtomicNum();
      // Only C,N,O,P in a DNA/RNA/protein backbone
      if (anum != 6 && anum != 7 && anum != 8 && anum != 15){
         return null;
      }

      // Allow the search to make it out of side Polymers, but not enter them...
      if (path != null && path.size() > 7){
         // Oxygen is only in the backbone for Nucleic Acids in a phosphate group
         if (anum == 8){
            if (!bondsX(currentAtom, 15)){
               return null;
            }
            // Nitrogen is only in the backbone in peptide bonds
         } else if (anum == 7){
            Atom carbonyl = findCarbonyl(currentAtom);
            if (carbonyl == null){
               return null;
            }
            // Avoid more than 3 carbons in a row (phenyl groups, etc.)
         } else if (anum == 6){
            Atom a;
            int anum2, anum3, anum4;
            int size = path.size();
            a = (Atom) path.get(size - 1);
            anum2 = a.getAtomicNum();
            if (anum2 == 6){
               a = (Atom) path.get(size - 2);
               anum3 = a.getAtomicNum();
               if (anum3 == 6){
                  a = (Atom) path.get(size - 3);
                  anum4 = a.getAtomicNum();
                  if (anum4 == 6){
                     return null;
                  }
               }
            }
         }
      }

      // Atoms with only one bond are at the end of a Polymer
      Atom previousAtom = null;
      if (path != null){
         previousAtom = (Atom) path.get(path.size() - 1);
      }
      ArrayList bonds = currentAtom.getBonds();
      if (bonds.size() == 1 && previousAtom != null){
         return null;
      }
      // Initialization
      if (path == null){
         path = getArrayfromPool();
         previousAtom = null;
         // Or Continuation
      } else{
         ArrayList pathclone = getArrayfromPool();
         for (ListIterator li = path.listIterator(); li.hasNext(); ){
            pathclone.add(li.next());
         }
         path = pathclone;
      }
      // Add the currentAtom to the growing path
      path.add(currentAtom);
      // Continue search in each bond direction, but no backtracking over previousAtom
      Bond b;
      Atom nextAtom;
      ArrayList newPolymer, maxPolymer = getArrayfromPool();
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         nextAtom = b.getOtherAtom(currentAtom);
         // Check to avoid returning in the same direction and loops
         if (nextAtom != previousAtom && !path.contains(nextAtom)){
            newPolymer = findPolymer(atoms, nextAtom, path);
            if (newPolymer != null){
               //				if (maxPolymer == null) {
               //				maxPolymer = newPolymer;
               // Check to see if the Polymers contain any of the same atoms
               // and if so, use the shorter Polymer (avoids loops)
               if (havesameAtom(newPolymer, maxPolymer)){
                  if (newPolymer.size() < maxPolymer.size()){
                     addArraytoPool(maxPolymer);
                     maxPolymer = newPolymer;
                  }
               } else if (newPolymer.size() > maxPolymer.size()){
                  addArraytoPool(maxPolymer);
                  maxPolymer = newPolymer;
               }
            }
         }
      }
      // Add the currentAtom to the longest discovered Polymer and return
      maxPolymer.add(0, currentAtom);
      return maxPolymer;
   }

   private static boolean havesameAtom(ArrayList Polymer1, ArrayList Polymer2){
      for (ListIterator li = Polymer1.listIterator(); li.hasNext(); ){
         if (Polymer2.contains(li.next())){
            return true;
         }
      }
      return false;
   }

   public static boolean divideBackbone(ArrayList backbone, Polymer c){

      Atom match;
      Residue res = null;
      int i, an, n, p;
      int length = backbone.size();
      int start;

      //Try to find a Phosphorus or Nitrogen in the backbone
      n = p = 0;
      for (i = 0; i < length; i++){
         match = (Atom) backbone.get(i);
         an = match.getAtomicNum();
         if (an == 15){
            p++;
         } else if (an == 7){
            n++;
         }
      }
      int type;
      if (p >= n && p > 1){
         type = NUCLEICACID;
      } else if (n > p && n > 2){
         type = AMINOACID;
      } else{
         return false;
      }
      start = -1;
      for (i = 0; i < length; i++){
         res = patternMatch(i, backbone, type);
         if (res != null){
            for (ListIterator li = res.getAtomList().listIterator(); li.hasNext(); ){
               Atom a = (Atom) li.next();
               a.setParent(null);
            }
            if (!(res.getName() == "Unknown")){
               start = i;
               // Want 5' to 3'
               if (type == NUCLEICACID){
                  Atom carbon5 = (Atom) backbone.get(start + 1);
                  if (bondsXnum(carbon5, 6) != 1){
                     start = -1;
                  }
               }
               break;
            }
         }
      }
      if (start == -1){
         backbone = reverseBackbone(backbone);
         for (i = 0; i < length; i++){
            res = patternMatch(i, backbone, type);
            if (res != null){
               for (ListIterator li = res.getAtomList().listIterator(); li.hasNext(); ){
                  Atom a = (Atom) li.next();
                  a.setParent(null);
               }
               if (!(res.getName() == "Unknown")){
                  start = i;
                  break;
               }
            }
         }
      }
      if (start == -1){
         return false;
      }

      //Potential Polypeptide
      if (type == AMINOACID){
         Atom nitrogen, alpha, carbonyl = null;
         Atom nitrogen2, carbonyl2;
         Residue aa = null;
         ArrayList aaArray = getArrayfromPool();
         int lastRes = 0;
         int firstRes = -1;
         while (start < length){
            aa = patternMatch(start, backbone, AMINOACID);
            if (aa != null){
               if (firstRes == -1){
                  firstRes = start;
                  carbonyl = findCarbonyl((Atom) backbone.get(start));
               }
               aaArray.add(aa);
               lastRes = start;
            }
            start += 3;
         }
         //Make sure the fisrt residue is found
         aa = null;
         if (carbonyl != null){
            alpha = findAlphaCarbon(carbonyl);
            if (alpha != null){
               nitrogen = findX(alpha, 7);
               if (nitrogen != null){
                  nitrogen2 = findX(carbonyl, 7);
                  ArrayList firstAtoms = getArrayfromPool();
                  firstAtoms.add(nitrogen);
                  firstAtoms.add(alpha);
                  firstAtoms.add(carbonyl);
                  firstAtoms.add(nitrogen2);
                  aa = patternMatch(0, firstAtoms, AMINOACID);
                  addArraytoPool(firstAtoms);
                  if (aa != null){
                     addCap(alpha, nitrogen, aa);
                     aaArray.add(0, aa);
                  }
               }
            }
         }
         // Add the remaining atoms to the end of the Polymer
         if (aa == null){
            nitrogen = (Atom) backbone.get(firstRes);
            alpha = (Atom) backbone.get(firstRes + 1);
            addCap(alpha, nitrogen, (Residue) aaArray.get(0));
         }
         // Make sure the last residue is found
         aa = null;
         carbonyl = findCarbonyl((Atom) backbone.get(lastRes + 1));
         if (carbonyl != null){
            nitrogen = findX(carbonyl, 7);
            if (nitrogen != null){
               alpha = findAlphaCarbon(nitrogen);
               if (alpha != null){
                  carbonyl2 = findCarbonyl(alpha);
                  if (carbonyl2 != null){
                     ArrayList lastAtoms = getArrayfromPool();
                     lastAtoms.add(carbonyl);
                     lastAtoms.add(nitrogen);
                     lastAtoms.add(alpha);
                     lastAtoms.add(carbonyl2);
                     aa = patternMatch(1, lastAtoms, AMINOACID);
                     addArraytoPool(lastAtoms);
                     if (aa != null){
                        addCap(alpha, carbonyl2, aa);
                        aaArray.add(aa);
                     }
                  }
               }
            }
         }
         if (aa == null){
            carbonyl = (Atom) backbone.get(lastRes + 2);
            alpha = (Atom) backbone.get(lastRes + 1);
            addCap(alpha, carbonyl, (Residue) aaArray.get(aaArray.size() - 1));
         }
         int index = 1;
         for (ListIterator li = aaArray.listIterator(); li.hasNext(); ){
            aa = (Residue) li.next();
            aa.setNumber(index++);
            c.addFNode(aa);
         }
         addArraytoPool(aaArray);
         //Potential DNA/RNA
      } else if (type == NUCLEICACID){
         Residue base;
         int lastRes = 0;
         boolean firstBase = true;
         Atom phos = null;
         Atom oxygen1 = null;
         Atom phosphate1 = null;
         ArrayList na = getArrayfromPool();
         while (start < length){
            base = patternMatch(start, backbone, NUCLEICACID);
            if (base != null){
               phos = (Atom) backbone.get(start - 1);
               if (phos != null && phos.getAtomicNum() == 15){
                  addPhosphate(phos, base);
               }
               na.add(base);
               if (firstBase){
                  firstBase = false;
                  phosphate1 = (Atom) backbone.get(start - 1);
                  oxygen1 = (Atom) backbone.get(start);
               }
               lastRes = start;
            }
            start += 6;
         }
         //Make sure the fisrt base is found
         Atom o2, o3;
         Atom c1, c2, c3;
         if (phosphate1 != null && oxygen1 != null){
            o2 = findOtherOxygen(phosphate1, oxygen1);
            if (o2 != null){
               c1 = findX(o2, 6);
               if (c1 != null){
                  c2 = findCO(c1);
                  if (c2 != null){
                     c3 = findC5(c2);
                     if (c3 != null){
                        o3 = findX(c3, 8);
                        if (o3 != null){
                           ArrayList firstAtoms = getArrayfromPool();
                           firstAtoms.add(o3);
                           firstAtoms.add(c3);
                           firstAtoms.add(c2);
                           firstAtoms.add(c1);
                           firstAtoms.add(o2);
                           firstAtoms.add(phosphate1);
                           base = patternMatch(0, firstAtoms, type);
                           if (base != null){
                              addCap(c3, o3, base);
                              na.add(0, base);
                           }
                        }
                     }
                  }

               }
            }
         }
         // Make sure the last base is found
         oxygen1 = (Atom) backbone.get(lastRes + 4);
         phosphate1 = (Atom) backbone.get(lastRes + 5);
         if (phosphate1 != null && oxygen1 != null){
            o2 = findOtherOxygen(phosphate1, oxygen1);
            if (o2 != null){
               c1 = findX(o2, 6);
               if (c1 != null){
                  c2 = findX(c1, 6);
                  if (c2 != null){
                     c3 = findCCO(c2);
                     if (c3 != null){
                        o3 = findX(c3, 8);
                        if (o3 != null){
                           ArrayList lastAtoms = getArrayfromPool();
                           lastAtoms.add(phosphate1);
                           lastAtoms.add(o2);
                           lastAtoms.add(c1);
                           lastAtoms.add(c2);
                           lastAtoms.add(c3);
                           lastAtoms.add(o3);
                           base = patternMatch(1, lastAtoms, type);
                           if (base != null){
                              addPhosphate(phosphate1, base);
                              addCap(c3, o3, base);
                              na.add(base);
                           }
                        }
                     }
                  }
               }
            }

         }

         int index = 1;
         for (ListIterator li = na.listIterator(); li.hasNext(); ){
            base = (Residue) li.next();
            base.setNumber(index++);
            c.addFNode(base);
         }
      } else{
         return false;
      }
      return true;
   }

   static private ArrayList reverseBackbone(ArrayList backbone){
      ArrayList reversed = getArrayfromPool();
      for (ListIterator li = backbone.listIterator(); li.hasNext(); ){
         reversed.add(0, li.next());
      }
      return reversed;
   }

   //Check to see if a portion of the backbone matches that of a
   //biological polymer, and if so determine the respective residue
   private static Residue patternMatch(int start, ArrayList backbone, int type){
      int pattern[];
      Atom a;

      // Initialization
      if (type == AMINOACID){
         pattern = AAPATTERN;
         if (backbone.size() < start + pattern.length){
            return null;
         }

         // Check for correct Carbonyl placement
         a = (Atom) backbone.get(start + 1);
         if (bondsX(a, 8)){
            return null;
         }
         a = (Atom) backbone.get(start + 2);
         if (!bondsX(a, 8)){
            return null;
         }

      } else if (type == NUCLEICACID){
         pattern = NAPATTERN;
         if (backbone.size() < start + pattern.length){
            return null;
         }
      } else{
         return null;
      }

      int length = pattern.length;
      ArrayList atoms = getArrayfromPool();
      ArrayList sidePolymer = getArrayfromPool();

      for (int i = 0; i < length; i++){
         a = (Atom) backbone.get(start + i);
         // add backbone atoms to terminate sidePolymer search
         sidePolymer.add(a);
         if (a.getAtomicNum() != pattern[i]){
            return null;
         }
      }

      // Collect all the atoms in the Residue
      // Add the atom before and after the backbone pattern to
      // terminate the search, then remove them
      if (start > 0){
         atoms.add((Atom) backbone.get(start - 1));
      }
      if (start + length < backbone.size()){
         atoms.add((Atom) backbone.get(start + length));

      }

      collectAtoms((Atom) backbone.get(start), atoms);

      if (start > 0){
         atoms.remove(0);
      }
      if (start + length < backbone.size()){
         atoms.remove(0);

         // Collect Just Side chain atoms, then remove backbone termination
      }
      if (type == AMINOACID){
         collectAtoms((Atom) sidePolymer.get(1), sidePolymer);
      } else if (type == NUCLEICACID){
         Atom seed = null;
         for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
            seed = (Atom) li.next();
            if (seed.getAtomicNum() == 7){
               break;
            }
         }
         if (seed != null && seed.getAtomicNum() == 7){
            sidePolymer.add(seed);
            collectAtoms(seed, sidePolymer);
         } else{
            return null;
         }
      }
      for (int i = 0; i <= length; i++){
         sidePolymer.remove(0);
      }
      Residue res = assignResidue(backbone, start, atoms, sidePolymer);
      return res;
   }

   private static Residue assignResidue(ArrayList backbone, int start, ArrayList atoms,
      ArrayList sidePolymer){

      Atom a;
      int atomicnum;
      int bins[] = new int[5]; //0 = S, 1 = P, 2 = O, 3 = N, 4 = C
      char chars[] = {'S', 'P', 'O', 'N', 'C'};
      for (ListIterator li = sidePolymer.listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         atomicnum = a.getAtomicNum();
         switch (atomicnum){
            case 1:

               //ignore hydrogens
               break;
            case 6:

               //Carbon
               bins[4]++;
               break;
            case 7:

               //Nitrogen
               bins[3]++;
               break;
            case 8:

               //Oxygen
               bins[2]++;
               break;
            case 15:

               //Phosphorus
               bins[1]++;
               break;
            case 16:

               //Sulfur
               bins[0]++;
               break;
            default:
               return null;
         }
      }

      StringBuffer key = new StringBuffer();
      int count = 0;
      for (int i = 0; i < 5; i++){
         if (bins[i] != 0){
            count += bins[i];
            key.append(chars[i]);
            key.append(Integer.toString(bins[i]));
         }
      }
      if (count == 0){
         key.append("H"); //Glycine
      }
      String resname = (String) structLookup.get(key.toString());

      if (resname == null){
         resname = "Unknown";
      } else{
         resname = resname.intern();
      }

      if (resname == "1" || resname == "2"){
         //Special case where atom string keys aren't unique
         Atom alpha = (Atom) backbone.get(start + 1);
         Atom carbonyl = (Atom) backbone.get(start + 2);
         Atom beta = null;
         ArrayList alphabonds = alpha.getBonds();
         Bond abond;
         for (ListIterator li = alphabonds.listIterator(); li.hasNext(); ){
            abond = (Bond) li.next();
            beta = abond.getOtherAtom(alpha);
            // Don't want the peptide nitrogen or alpha hydrogen
            if (beta.getAtomicNum() != 7 && beta.getAtomicNum() != 1){
               // Don't want carbonyl carbon
               if (beta != carbonyl){
                  break;
               }
            }
            beta = null;
         }

         if (beta == null){
            return null;
         }

         ArrayList betabonds = beta.getBonds();
         Atom gamma;
         int carboncount = 0;
         for (ListIterator li = betabonds.listIterator(); li.hasNext(); ){
            abond = (Bond) li.next();
            gamma = (Atom) abond.getOtherAtom(beta);
            if (gamma.getAtomicNum() == 6){
               carboncount++;
            }
         }
         if (resname == "1"){
            if (carboncount == 2){
               resname = "PRO";
            } else{
               resname = "VAL";
            }
         } else{
            if (carboncount == 2){
               resname = "LEU";
            } else{
               resname = "ILE";
            }
         }
      } else if (resname == "3"){
         Atom c3 = (Atom) backbone.get(start + 3);
         int num = countCO(c3);
         if (num == 2){
            resname = "A";
         } else{
            resname = "dG";
         }
      }

      Residue res = new Residue(resname, true);
      // Create the Residue group
      for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         res.addFNode(a);
      }
      return res;
   }

   // Given an array of bonded atoms, this function recursively collects all other
   // connected atoms, without backtracking over atoms already in the list.
   // (ie, atoms can be preloaded into the ArrayList to provide search termination)
   private static void collectAtoms(Atom seed, ArrayList atoms){
      atoms.add(seed);
      ArrayList bonds = seed.getBonds();
      Bond b;
      Atom nextAtom;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         nextAtom = b.getOtherAtom(seed);
         if (nextAtom.getParent() != null){
            continue;
         }
         // avoid crossing disulfides
         if (nextAtom.getAtomicNum() != 16 || seed.getAtomicNum() != 16){
            if (!atoms.contains(nextAtom)){
               collectAtoms(nextAtom, atoms);
            }
         }
      }
   }

   // Returns the oxygen that is bonded to p and a carbon, but not o
   public static Atom findOtherOxygen(Atom p, Atom o){
      ArrayList bonds = p.getBonds();
      Bond b;
      Atom oxygen;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         oxygen = b.getOtherAtom(p);
         if (oxygen.getAtomicNum() == 8 && oxygen != o){
            if (bondsX(oxygen, 6)){
               return oxygen;
            }
         }
      }
      return null;
   }

   public static int countCO(Atom adjacent){
      Atom carbonyl, oxygen;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      int count = 0;
      ArrayList carbonylbonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         carbonyl = b.getOtherAtom(adjacent);
         if (carbonyl.getAtomicNum() == 6){
            carbonylbonds = carbonyl.getBonds();
            for (ListIterator lj = carbonylbonds.listIterator(); lj.hasNext(); ){
               b = (Bond) lj.next();
               oxygen = (Atom) b.getOtherAtom(carbonyl);
               if (oxygen.getAtomicNum() == 8){
                  count++;
               }
            }
         }
      }
      return count;
   }

   //Returns a carbonyl carbon that is bonded to the adjacent atom
   public static Atom findCO(Atom adjacent){
      Atom carbonyl, oxygen;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      ArrayList carbonylbonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         carbonyl = b.getOtherAtom(adjacent);
         if (carbonyl.getAtomicNum() == 6){
            carbonylbonds = carbonyl.getBonds();
            for (ListIterator lj = carbonylbonds.listIterator(); lj.hasNext(); ){
               b = (Bond) lj.next();
               oxygen = (Atom) b.getOtherAtom(carbonyl);
               if (oxygen.getAtomicNum() == 8){
                  return carbonyl;
               }
            }
         }
      }
      return null;
   }

   //Returns a carbon that is bonded to the adjacent atom; plus 1 carbons and 1 oxygen
   public static Atom findCCO(Atom adjacent){
      Atom carbon, oxygen;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      ArrayList carbonbonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         carbon = b.getOtherAtom(adjacent);
         if (carbon.getAtomicNum() == 6){
            if (bondsXnum(carbon, 6) == 2){
               carbonbonds = carbon.getBonds();
               for (ListIterator lj = carbonbonds.listIterator(); lj.hasNext(); ){
                  b = (Bond) lj.next();
                  oxygen = (Atom) b.getOtherAtom(carbon);
                  if (oxygen.getAtomicNum() == 8){
                     return carbon;
                  }
               }
            }
         }
      }
      return null;
   }

   // Returns a carbon that is bonded to the adjacent atom,
   // including 1 carbon and 1 oxygen
   public static Atom findC5(Atom adjacent){
      Atom carbon, oxygen;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      ArrayList carbonbonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         carbon = b.getOtherAtom(adjacent);
         if (carbon.getAtomicNum() == 6){
            if (bondsXnum(carbon, 6) == 1 && bondsXnum(carbon, 8) == 1){
               return carbon;
            }
         }
      }
      return null;
   }

   //Returns a carbon that is bonded to the adjacent atom and an oxygen
   public static Atom findCarbonyl(Atom adjacent){
      Atom carbonyl, oxygen;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      ArrayList carbonylbonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         carbonyl = b.getOtherAtom(adjacent);
         if (carbonyl.getAtomicNum() == 6){
            carbonylbonds = carbonyl.getBonds();
            for (ListIterator lj = carbonylbonds.listIterator(); lj.hasNext(); ){
               b = (Bond) lj.next();
               oxygen = (Atom) b.getOtherAtom(carbonyl);
               if (oxygen.getAtomicNum() == 8){
                  if (oxygen.getBonds().size() == 1){
                     return carbonyl;
                  }
               }
            }
         }
      }
      return null;
   }

   //Returns the Alpha carbon that is bonded to the adjacent atom
   //It must be bonded to a carbonyl group and a nitrogen
   public static Atom findAlphaCarbon(Atom adjacent){
      Atom alpha, next;
      ArrayList bonds = adjacent.getBonds();
      Bond b;
      ArrayList alphabonds;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         alpha = b.getOtherAtom(adjacent);
         if (alpha.getAtomicNum() == 6){
            if (findCO(alpha) != null && bondsX(alpha, 7)){
               return alpha;
            }
         }
      }
      return null;
   }

   // Returns the first atom of atomic number x bonded to Atom a
   public static Atom findX(Atom a, int x){
      ArrayList bonds = a.getBonds();
      Bond b;
      Atom atomx;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         atomx = b.getOtherAtom(a);
         if (atomx.getAtomicNum() == x){
            return atomx;
         }
      }
      return null;
   }

   // Returns true if a bond exists to am atom with atomic number x
   public static boolean bondsX(Atom a, int x){
      ArrayList bonds = a.getBonds();
      Bond b;
      Atom oxygen;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         oxygen = b.getOtherAtom(a);
         if (oxygen.getAtomicNum() == x){
            return true;
         }
      }
      return false;
   }

   // This adds a phosphate and any oxygens that are not bonded to a carbon
   public static void addPhosphate(Atom p, Residue r){
      if (p == null){
         return;
      }
      r.addFNode(p);
      ArrayList bonds = p.getBonds();
      Bond b;
      Atom a;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         a = b.getOtherAtom(p);
         // Add oxygens not bonded to a Carbon
         if (bondsXnum(a, 6) == 0){
            r.addFNode(a);
            // Add hydrogens atoms for protonated oxygen groups
            a = findX(a, 1);
            if (a != null){
               r.addFNode(a);
            }
         }
      }
   }

   public static Atom findSeed(Atom end, Atom other){
      ArrayList bonds = end.getBonds();
      Bond b;
      Atom seed;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         seed = b.getOtherAtom(end);
         if (seed != other){
            return seed;
         }
      }
      return null;
   }

   // Collect all the atoms in a Polymer beyound the end atom, including the
   // seed, and put them into the residue. This assumes that the cap is only
   // linked to the rest of the Polymer through the end atom.
   public static void addCap(Atom end, Atom seed, Residue r){
      ArrayList cap = getArrayfromPool();
      cap.add(end);
      collectAtoms(seed, cap);
      // Assume the end & seed atoms are already part of the residue
      cap.remove(0);
      cap.remove(0);
      Atom a;
      for (ListIterator li = cap.listIterator(); li.hasNext(); ){
         a = (Atom) li.next();
         r.addFNode(a);
      }
   }

   // This function returns the number of times an atom is bonded to
   // an atom of atomic number X
   public static int bondsXnum(Atom a, int x){
      ArrayList bonds = a.getBonds();
      Bond b;
      Atom atomx;
      int count = 0;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         atomx = b.getOtherAtom(a);
         if (atomx.getAtomicNum() == x){
            count++;
         }
      }
      return count;
   }

   public static boolean isWater(Atom a){
      if (a.getAtomicNum() != 8){
         return false;
      }

      ArrayList bonds = a.getBonds();
      Bond b;
      Atom h;
      for (ListIterator li = bonds.listIterator(); li.hasNext(); ){
         b = (Bond) li.next();
         h = b.getOtherAtom(a);
         if (h.getAtomicNum() != 1){
            return false;
         }
      }
      return true;
   }

   public static String PolymerLookup(int i){
      String add = new String();
      if (i > 25){
         i = i % 25;
         add = new String("" + i / 25);
      }
      return String.valueOf((char) (i + 65)) + add;
   }

   /*
     //Eventually going to used in Ribbon Construction...
     public static double polint(double xa[], double ya[], int n, double x){
    int ns = 0;
    double y, dy;
    double den, dif, dift, ho, hp, w;
    double c[] = new double[n];
    double d[] = new double[n];
    dif = Math.abs(x-xa[0]);
    for (int i=0; i<n; i++){
     if ((dift = Math.abs(x=xa[i]))<dif){
      ns = i;
      dif = dift;
     }
     c[i] = ya[i];
     d[i] = ya[i];
    }
    y = ya[ns--];
    for (int m=0; m<n-1; m++){
     for (int i=0; i<n-m; i++){
      ho = xa[i] - x;
      hp = xa[i+m] - x;
      w = c[i+1] - d[i];
      if ((den = ho-hp) == 0.0)
      den = w/den;
      d[i] = hp*den;
      c[i] = ho*den;
     }
     y += (dy = (2*ns < (n-m) ? c[ns+1] : d[ns--]));
    }
    return y;
     }
    */

   public static final int AMINOACID = 0;
   public static final int NUCLEICACID = 1;
   public static final int AAPATTERN[] = {7, 6, 6};
   public static final int NAPATTERN[] = {8, 6, 6, 6, 8};

   static private Hashtable structLookup = new Hashtable();
   static{
      // In no particular order
      // Amino Acid Side Chain Polymers
      structLookup.put("S1C3", "MET");
      structLookup.put("S1C1", "CYS");

      structLookup.put("O1C1", "SER");
      structLookup.put("O1C2", "THR");
      structLookup.put("O1C7", "TYR");
      structLookup.put("O2C2", "ASP");
      structLookup.put("O2C3", "GLU");
      structLookup.put("O1N1C2", "ASN");
      structLookup.put("O1N1C3", "GLN");

      structLookup.put("N3C4", "ARG");
      structLookup.put("N2C4", "HIS");
      structLookup.put("N1C9", "TRP");
      structLookup.put("N1C4", "LYS");

      structLookup.put("C7", "PHE");
      structLookup.put("C3", "1"); //Proline & Valine
      structLookup.put("C4", "2"); //Leucine Isoleucine
      structLookup.put("H", "GLY");
      structLookup.put("C1", "ALA");

      // DNA
      structLookup.put("O2N5C7", "3"); // Both DNA Gaunine and RNA Adenine
      structLookup.put("O2N3C6", "dC");
      structLookup.put("O1N5C7", "dA");
      structLookup.put("O3N2C7", "dT");

      // RNA
      structLookup.put("O3N5C7", "G");
      structLookup.put("O3N3C6", "C");
      // structLookup.put("O2N5C7", "A");
      structLookup.put("O4N2C6", "U");

   }
}
