/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.net;

import java.io.*;
import java.net.*;
import javax.swing.*;

/**
 * The FClient class encapsulates a socket connection to an FServer started by an
 * executing TINKER program. TinkerSystem and TinkerUpdate objects are sent
 * by the FServer to the FClient on request.
 */

public class FClient{

   private Socket client; // Socket connection to the server
   private InetSocketAddress address; // Server address

   private InputStream in; // Input from the server
   private ObjectInputStream oin; // Input form the server
   private OutputStream out; // Output to the server
   private ObjectOutputStream oout; // Output to the server

   private TinkerSystem system; // Tinker system definition
   private TinkerUpdate update; // Tinker update information
   private FMessage message; // Message Passed Between Client & Server

   //  If true, the Client will request a System Definition from the Server
   private boolean needSystem = false;

   //  If true, the Client will request an Update from the Server
   private boolean needUpdate = true;

   //  Various connection status variables
   private int retryCount = 0; // Count of Client attempts to Connect
   private int retryLimit = 10000; // Maximum number of retries
   private boolean connectionMade = false; // True when a connection has been made

   // closed is True if the server closes an open connection
   // or if the retryLimit is reached
   private boolean closed = false;

   public FClient(){
      address = new InetSocketAddress(2000);
   }

   public FClient(int port){
      address = new InetSocketAddress(port);
   }

   public FClient(InetSocketAddress a){
      address = a;
   }

   /**
    * Specifies that a request should (not) be made for a System Definition
    * @param b
    */
   public void setNeedSystem(boolean b){
      needSystem = b;
   }

   /**
    * Specifies that a request should (not) be made for a Tinker Update
    * @param b
    */
   public void setNeedUpdate(boolean b){
      needUpdate = b;
   }

   /**
    * Attempts to connect to a Tinker FServer. If this FClient is already
    * connected, the connection will be closed.
    */
   public void connect(){
      if (client != null && client.isConnected()){
         release(null);
      }
      closed = false;
      client = new Socket();
      try{
         // System.out.println("Attempting to connect: " +
         //                   System.currentTimeMillis());
         //System.out.println("" + address);
         client.connect(address, 100);
         client.setTcpNoDelay(true);
         out = client.getOutputStream();
         oout = new ObjectOutputStream(out);
         in = client.getInputStream();
         oin = new ObjectInputStream(in);
         connectionMade = true;
         //System.out.println("Client Connected: " + client);
         //System.out.println("Current Time: " + System.currentTimeMillis());
      } catch (Exception e){
         release(e);
      }
      if (client == null || !client.isConnected()){
         release(null);
      }
   }

   public TinkerSystem getSystem(){
      return system;
   }

   public TinkerUpdate getUpdate(){
      return update;
   }

   public boolean isClosed(){
      return closed;
   }

   public boolean isConnected(){
      if (client != null && client.isConnected()){
         return true;
      }
      return false;
   }

   public void readSocket(){
      try{
         while (oin != null && in.available() > 0){
            Object o = oin.readObject();
            if (o instanceof FMessage){
               message = (FMessage) o;
               if (message.getMessage() == FMessage.SYSTEM){
                  system = (TinkerSystem) oin.readObject();
                  system.read = false;
                  needSystem = false;
               } else if (message.getMessage() == FMessage.UPDATE){
                  update = (TinkerUpdate) oin.readObject();
                  update.read = false;
                  needUpdate = false;
               } else if (message.getMessage() == FMessage.CLOSING){
                  closed = true;
                  release(null);
               }
            }
         }

         if (needSystem){
            message = new FMessage(FMessage.SYSTEM);
            oout.reset();
            oout.writeObject(message);
            oout.flush();
         }

         if (needUpdate){
            message = new FMessage(FMessage.UPDATE);
            if (update != null){
               if (update.type == TinkerUpdate.SIMULATION){
                  message.setTime(update.time);
               } else{
                  message.setStep(update.step);
               }
            }
            oout.reset();
            oout.writeObject(message);
            oout.flush();
         }
      } catch (Exception e){
         release(e);
      }
   }

   public void release(){
      release(null);
   }

   private void release(Exception ex){
      if (client == null){
         return;
      }
//        if (ex != null) {
//            System.out.println("Cycle: " + retryCount + ", Socket Release Exception: " + ex);
//        }
      retryCount++;
      if (retryCount > retryLimit || connectionMade){
//            System.out.println("Client closed: " + address.toString());
         closed = true;
      }
      if (client != null && client.isConnected() && oout != null){
         try{
            FMessage close = new FMessage(FMessage.CLOSING);
            oout.reset();
            oout.writeObject(close);
            oout.flush();
         } catch (Exception e){
         }
      }
      try{
         if (oin != null){
            oin.close();
         }
         if (in != null){
            in.close();
         }
         if (oout != null){
            oout.close();
         }
         if (out != null){
            out.close();
         }
         if (client != null){
            client.close();
         }
      } catch (Exception e){
      }
      in = null;
      oin = null;
      out = null;
      oout = null;
      client = null;
   }
}
