/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.net;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;

/**
 * The FServer is launched by TINKER executables to allow Force Field Explorer
 * clients to connect.
 */

public class FServer
   implements Runnable{

   private ServerSocket server;
   private int serverPort = 2000;
   private int serverTimeout = 100;
   private Thread thread;
   private int sleepTime = 1;
   private long checkTime;
   private boolean init = true;
   private int cycle = 0;

   private boolean shutdown = false;

//    private boolean pause = false;
   private boolean request = false;

   private Vector clients = new Vector();
   private Vector outputs = new Vector();
   private Vector inputs = new Vector();

   private TinkerSystem system = null;
   private TinkerUpdate update = null;

   public FServer(TinkerSystem s){
      system = s;
      checkTime = System.currentTimeMillis();
   }

   public boolean isAlive(){
      if (thread == null){
         return false;
      } else if (thread.isAlive()){
         return true;
      } else{
         return false;
      }
   }

   public void loadUpdate(TinkerUpdate u){
      update = u;
   }

   public boolean needUpdate(){
      if (clients.size() == 0){
         return false;
      }
      if (request != true){
         return false;
      }
      long currentTime = System.currentTimeMillis();
      sleepTime = (int) (currentTime - checkTime);
      checkTime = currentTime;
      return true;
   }

   public void run(){
      startServer();
      while (!shutdown){
         accept();
         send();
         try{
            thread.sleep(sleepTime);
         } catch (Exception e){
         }
         if (init){
            cycle++;
            if (cycle >= 10){
               init = false;
               serverTimeout = 1;
            }
         }
      }
      accept();
      send();
      closeServer();
   }

   public void setUpdated(){
      request = false;
   }

   public void start(){
      if (thread == null || !thread.isAlive()){
         thread = new Thread(this);
         thread.setPriority(Thread.MAX_PRIORITY);
         thread.start();
      }
   }

   public void stop(){
//        pause = false;
      shutdown = true;
   }

   private void accept(){
      Socket client = null;
      ObjectOutputStream oout = null;
      ObjectInputStream oin = null;
      if (server != null){
         try{
            client = server.accept();
            if (client != null && client.isConnected()){
               client.setTcpNoDelay(true);
               clients.add(client);
               oout = new ObjectOutputStream(client.getOutputStream());
               outputs.add(oout);
               oin = new ObjectInputStream(client.getInputStream());
               inputs.add(oin);
               //System.out.println("Client Accepted: " + client);
               //System.out.println("Current Time: " + System.currentTimeMillis());
            }
         } catch (Exception e){
            if (client != null){
               clients.remove(client);
            }
            if (oout != null){
               outputs.remove(oout);
            }
            if (oin != null){
               inputs.remove(oin);
            }
         }
      }
   }

   /* Catching an exception is a bad way to check for a closed socket,
       but I can't figure out a better way...
    */
   private static FMessage closing = new FMessage(FMessage.CLOSING);
   private void closeClient(int index){
      Socket client = (Socket) clients.get(index);
      ObjectOutputStream oout = (ObjectOutputStream) outputs.get(index);
      ObjectInputStream oin = (ObjectInputStream) inputs.get(index);

      try{
         oout.reset();
         oout.writeObject(closing);
         oout.flush();
      } catch (Exception e){
         try{
            outputs.remove(index);
            inputs.remove(index);
            clients.remove(index);
            if (oout != null){
               oout.close();
            }
            if (oin != null){
               oin.close();
            }
            if (client != null){
               client.close();
            }
         } catch (Exception ex){

         }
      }

   }

   private void closeServer(){

      for (int i = 0; i < clients.size(); i++){
         lastUpdate(i);
      }
      while (!clients.isEmpty()){
         closeClient(0);
         try{
            synchronized (this){
               wait(10);
            }
         } catch (Exception e){
            System.out.println("" + e);
         }
      }
      try{
         if (server != null){
            server.close();
            server = null;
         }
      } catch (Exception e){

      }
   }

   private void lastUpdate(int index){
      try{
         ObjectOutputStream oout = (ObjectOutputStream) outputs.get(index);
         Socket client = (Socket) clients.get(index);
         if (client != null && client.isConnected() && oout != null){
            FMessage last = new FMessage(FMessage.SYSTEM);
            if (system != null){
               oout.reset();
               oout.writeObject(last);
               oout.writeObject(system);
               oout.flush();
            }
            if (update != null){
               oout.reset();
               last.setMessage(FMessage.UPDATE);
               oout.writeObject(last);
               oout.writeObject(update);
               oout.flush();
            }
            last.setMessage(FMessage.CLOSING);
            oout.reset();
            oout.writeObject(last);
            oout.flush();
         }
      } catch (Exception e){
//            System.out.println("" + e);
      }
   }

   private void send(){
      if (system == null){
         return;
      }
      if (clients.size() == 0){
         return;
      }
      ObjectOutputStream oout;
      ObjectInputStream oin;
      Socket client;
      ListIterator lout, lin, lclient;
      Vector closed = new Vector();

      for (lout = outputs.listIterator(), lin = inputs.listIterator(),
         lclient = clients.listIterator(); lout.hasNext(); ){
         oout = (ObjectOutputStream) lout.next();
         oin = (ObjectInputStream) lin.next();
         client = (Socket) lclient.next();
         if (!client.isConnected() || client.isClosed()){
            closed.add(client);
         } else{
            try{
               FMessage message = null;
               while (oin != null && client.getInputStream().available() > 0){
                  Object o = oin.readObject();
                  if (o instanceof FMessage){
                     message = (FMessage) o;
                     if (message.getMessage() == FMessage.CLOSING){
                        closed.add(client);
                        message = null;
                        break;
                     }
                  }
               }
               if (message != null){
                  if (message.getMessage() == FMessage.SYSTEM){
                     synchronized (system){
                        oout.reset();
                        oout.writeObject(message);
                        oout.writeObject(system);
                        oout.flush();
                     }
                  } else if (message.getMessage() == FMessage.UPDATE){
                     request = true;
                     if (update == null){

                     } else{
                        if (update.isNewer(message)){
                           synchronized (update){
                              //System.out.println("Sending Update");
                              oout.reset();
                              oout.writeObject(message);
                              oout.writeObject(update);
                              oout.flush();
                           }
                        }
                     }
                  }
               }
            } catch (Exception e){
               closed.add(client);
            }
         }
      }

      ListIterator li;
      for (li = closed.listIterator(); li.hasNext(); ){
         client = (Socket) li.next();
         int index = clients.indexOf(client);
         closeClient(index);
      }
   }

   private void startServer(){
      try{
         server = new ServerSocket();
         server.setSoTimeout(serverTimeout);
         server.setReuseAddress(true);
         server.bind(new InetSocketAddress(InetAddress.getLocalHost(), serverPort));
      } catch (Exception e){
         try{
            server.bind(new InetSocketAddress(InetAddress.getByName(null), serverPort));
         } catch (Exception ex){
            System.out.println(" SERVER -- Could Not Start: " + ex);
            return;
         }
      }
      //System.out.println("" + server.getLocalSocketAddress());
      //JOptionPane.showMessageDialog(null, server.getLocalSocketAddress(), "Server Address", JOptionPane.ERROR_MESSAGE);
//        System.out.println("Server Address: " + server.getLocalSocketAddress());
   }
}
