/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.net;

import java.net.InetSocketAddress;
import java.util.*;

import java.awt.event.*;
import javax.swing.Timer;

import ffe.lang.*;
import ffe.panels.MainPanel;
import ffe.parsers.SocketFilter;

/**
 * This Simulation class oversees loading information from an executing TINKER program into Force Field Explorer.
 */
public class Simulation
   implements ActionListener{

   // The client monitors a socket based connection to an executing TINKER program.
   private FClient client;
   private SocketFilter sktfilter;
   private InetSocketAddress address;

   // If the TINKER program was launched from the GUI, the job will be alive until the program exits.
   private Thread job = null;

   // Once the simulation is finished, this flag will be true.
   private boolean finished = false;

   // The reader thread contains a SocketFilter instance that will read a simulation into the FFE.
   private Thread reader;
   private MainPanel fpanel;
   private FSystem fsystem;
   private boolean firstUpdate = true;

   private Timer timer;
   private int delay = 10;
   private double time = 0.0;
   private int step = 0;

   // Constructor
   public Simulation(FSystem m, Thread j, MainPanel f, InetSocketAddress a){
      fsystem = m;
      job = j;
      fpanel = f;
      address = a;
      if (address == null){
         finished = true;
         return;
      }
   }

   public void actionPerformed(ActionEvent e){
      if (isFinished()){
         return;
      }

      if (!isConnected()){
         connect();
         return;
      }

      // Check if we need to initialize the Simulation System
      if (fsystem == null){
         client.readSocket();
         TinkerSystem sys = client.getSystem();
         if (sys != null){
            if (sktfilter == null){
               sktfilter = new SocketFilter(fpanel, sys);
               reader = new Thread(sktfilter);
               reader.start();
            } else if (sktfilter.FileRead()){
               fsystem = sktfilter.getFSystem();
               sktfilter = null;
            }
         }
      }
      // We have a Simulation System, attempt to get an Update
      else{
         client.readSocket();
         TinkerUpdate update = client.getUpdate();
         if (update != null && !update.read){
            if (!fpanel.getGraphics3D().sceneRendering()){
               update();
            }
            client.setNeedUpdate(true);
         }
      }
   }

   public void connect(){
      if (isFinished() || isConnected()){
         return;
      }
      if (client == null){
         client = new FClient(address);
      }
      if (fsystem == null){
         client.setNeedSystem(true);
      }
      client.connect();
      if (timer == null){
         timer = new Timer(delay, this);
         timer.setCoalesce(true);
         timer.setDelay(1);
         timer.start();
      }
      if (client.isConnected()){
         timer.setDelay(delay);
      } else{
         if (timer.getDelay() < 200){
            timer.setDelay(timer.getDelay() * 2);
         }
      }
   }

   public FSystem getFSystem(){
      return fsystem;
   }

   public boolean isFinished(){
      checkDone();
      return finished;
   }

   public boolean isConnected(){
      if (client == null){
         return false;
      } else if (!client.isConnected()){
         return false;
      }
      return true;
   }

   // Release the FSystemUpdate Thread
   public void release(){
      finished = true;
      if (timer != null){
         timer.stop();
      }
      if (client != null){
         client.release();
      }
      if (fsystem != null){
         fsystem.setSimulation(TinkerUpdate.NONE);
      }
      fpanel.getMenu().setConnect(true);
   }

   private void checkDone(){
      if (client != null && client.isClosed()){
         finished = true;
      }
      if (job != null && !job.isAlive()){
         finished = true;
      }
      if (finished){
         if (timer != null){
            timer.stop();
         }
         update();
         release();
      }
   }

   private void update(){
      if (client == null){
         return;
      }
      TinkerUpdate update = client.getUpdate();
      if (update == null || update.read == true){
         return;
      }
      fsystem.setSimulation(update.type);

      // Check that FFE and TINKER agree on the number of atoms - should be true...
      List atoms = fsystem.getAtomList();
      int n = atoms.size();
      if (update.numatoms != n){
         finished = true;
         return;
      }

      fsystem.setEnergy(update.energy);
      if (update.type == update.SIMULATION){
         if (update.time == time){
            return;
         } else{
            time = update.time;
            fsystem.setTime(time);
            fsystem.setTemperature(update.temperature);
         }
      } else if (update.type == update.OPTIMIZATION){
         if (update.step == step){
            return;
         } else{
            step = update.step;
            fsystem.setStep(step);
         }
      }

      Atom.setMaxForce(0.0);
      Atom.setMaxAcceleration(0.0);
      Atom.setMaxVelocity(0.0);
      Atom.setMaxInduced(0.0);
      double d[] = new double[3];
      for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
         Atom a = (Atom) li.next();
         int index = a.getXYZIndex() - 1;
         d[0] = update.coordinates[0][index];
         d[1] = update.coordinates[1][index];
         d[2] = update.coordinates[2][index];
         a.moveto(d);
         if (update.amoeba){
            a.setInducedDipole( -1.0 * update.induced[0][index], -1.0 * update.induced[1][index],
               -1.0 * update.induced[2][index]);
         }
         if (update.type == TinkerUpdate.SIMULATION){
            a.setVeclocity(update.velocity[0][index], update.velocity[1][index],
               update.velocity[2][index]);
            a.setAcceleration(update.acceleration[0][index], update.acceleration[1][index],
               update.acceleration[2][index]);
         } else if (update.type == TinkerUpdate.OPTIMIZATION){
            a.setForce( -1.0 * update.gradients[0][index], -1.0 * update.gradients[1][index],
               -1.0 * update.gradients[2][index]);
         }
      }
      if (firstUpdate){
         fsystem.centerFSystem();
         firstUpdate = false;
      }
      fpanel.getGraphics3D().updateScene(fsystem, true, false, 0, false, 0);
      fpanel.getTree().updateStatus();
      fpanel.getTree().repaint();
      update.read = true;
   }
}
