/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.File;

import ffe.lang.FSystem;

/**
 * FExec is a native replacement for the JDK System.Exec() method. This class executes
 * TINKER programs in their own thread through a call to the native function "system()".
 * The reason we aren't using the System.Exec() methods is that
 * some TINKER routines execute indefinitely. Users may want to exit the
 * GUI & JVM after launching a dynamics run, for example. In this case
 * the thread should not be dependent on a JVM instance.
 */
public final class FExec
   implements Runnable{

   private FSystem system;
   private String name;
   private String args;
   private String dir;
   private String path;
   private String classpath;
   private String ld_library_path;
   private MainPanel mainPanel;
   private File newFile;
   private boolean openOnto;

   static private boolean nativeLibFound = false;

   private void loadNativeFExec(){

      try{
         String absPath = System.getProperty("java.library.path", ".");
         File absNativePath = null;

         if (ffe.Main.MAC_OS_X){
            absNativePath = new File(absPath + File.separator + "libffe.jnilib");
         } else if (ffe.Main.LINUX){
            absNativePath = new File(absPath + File.separator + "libffe.so");

         } else if (ffe.Main.WINDOWS){
            absNativePath = new File(absPath + File.separator + "ffe.dll");
         }

         if (absNativePath != null && absNativePath.exists()){
            System.load(absNativePath.getAbsolutePath());

            nativeLibFound = true;
         }

         if (nativeLibFound == false){
            System.loadLibrary("ffe");
            nativeLibFound = true;
         }

      } catch (UnsatisfiedLinkError e){
         nativeLibFound = false;
         System.out.println(
            "The Native Library used to launch TINKER executables could not be found");
         System.out.println("in the 'java.library.path': " + System.getProperty("java.library.path"));
         System.out.println("Please reset the path to TINKER Binaries using the 'Set Path' button.");
         System.out.println("Alternatively, the following statement can be included in your \nForce Field Explorer startup command: ");

         if (ffe.Main.MAC_OS_X){
            System.out.println("-Djava.library.path=tinker/jar/macosx\n");
            System.out.println(
               "Replace 'tinker' in the above statement with the full path to the \nTINKER installation directory.");
            System.out.println("For example, replace 'tinker' with '/Applications/TINKER'\n");
         } else if (ffe.Main.LINUX){
            System.out.println("-Djava.library.path=tinker/jar/linux\n");
            System.out.println(
               "Replace 'tinker' in the above statement with the full path to the \nTINKER installation directory.");
            System.out.println("For example, replace 'tinker' with '/opt/TINKER'\n");
         } else if (ffe.Main.WINDOWS){
            System.out.println("-Djava.library.path=tinker/jar/windows\n");
            System.out.println(
               "Replace 'tinker' in the above statement with the full path to the \nTINKER installation directory.");
            System.out.println("For example, replace 'tinker' with 'C:\\TINKER'\n");
         }
      }
   }

   private native int fexec(String argv, String dir, String path, String classpath, String jre);

   public boolean isNativeLibFound(){
      return nativeLibFound;
   }

   /**
    * Constructor
    * @param f File the TINKER command will execute on
    * @param n Name of the log file
    * @param a Command to execute
    * @param d Directory to launch the command in
    * @param p PATH
    * @param c CLASSPATH
    * @param l LD_LIBRARY_PATH
    * @param m MainPanel
    * @param file File to open
    * @params o Open file onto an already open system
    */
   public FExec(FSystem s, String n, String a, String d, String p, String c, String l, MainPanel m,
      File file, boolean o){
      system = s;
      name = n;
      args = a;
      dir = d;
      path = p;
      classpath = c;
      ld_library_path = l;
      mainPanel = m;
      newFile = file;
      openOnto = o;

      if (!nativeLibFound){
         loadNativeFExec();
      }
   }

   /**
    * Executes the native call to "System()" and notifies the ResultPanel
    * upon completion. This should only be called indirectly by Thread.Start()
    */
   public void run(){
      //System.out.println("FExec - PATH: " + path + "\nCLASSPATH: " + classpath +
      //                "\nLD_LIBRARY_PATH: " + ld_library_path);

      if (!nativeLibFound){
         return;
      }

      fexec(args, dir, path, classpath, ld_library_path);
      if (newFile != null){
         if (openOnto){
            mainPanel.openOn(newFile, system);
         } else{
            mainPanel.open(newFile);
         }
      }
      mainPanel.getLogPanel().setDone(name);
   }
}
