/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.prefs.Preferences;
import javax.imageio.ImageIO;
import javax.media.j3d.*;
import javax.media.j3d.Raster;
import javax.vecmath.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import javax.swing.*;

import org.web3d.j3d.loaders.VRML97Loader;
import com.sun.j3d.loaders.Scene;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.picking.PickTool;
import com.sun.j3d.utils.universe.SimpleUniverse;

import ffe.behaviors.*;
import ffe.lang.*;

/**
 * The Graphics class provides 3D Graphics capibilities, making heavy use of Java3D and
 * its utility packages. The following display types are currently supported:
 * Wireframe, Ball & Stick, Spacefill/CPK and Tube models.
 */
public final class Graphics3D
   extends Canvas3D
   implements ActionListener{

   // 3D Canvas Setup
   private MainPanel fpanel;

   private SimpleUniverse universe;
   private EventDispatch eventDispatch;
   private GlobalAxis sceneAxis;

   private Background background;
   private BranchGroup baseBranchGroup;
   private TransformGroup baseTransformGroup;
   private Transform3D baseTransform3D = new Transform3D();
   private PickNodeBehavior pickNodeBehavior;
   private UpdateBehavior updateBehavior;
   private Bounds bounds;

   private boolean imageCapture = false;
   private String imageFormat = "png";
   private File imageName;
   private int mouseMode = ACTIVE;

   private static final Hashtable materials = new Hashtable();
   public static final ColoringAttributes coloringAttributes = new ColoringAttributes();
   public static final RenderingAttributes renderingAttributes = new RenderingAttributes();
   public static final LineAttributes lineAttributes = new LineAttributes();
   public static final TransparencyAttributes transparencyAttributes = new TransparencyAttributes();
   public static final PolygonAttributes fillPolygonAttributes = new PolygonAttributes();
   public static final PolygonAttributes pointPolygonAttributes = new PolygonAttributes();
   public static final PolygonAttributes linePolygonAttributes = new PolygonAttributes();

   public Graphics3D(GraphicsConfiguration config, MainPanel f){
      super(config);
      fpanel = f;
      initialize();
   }

   /**
    * The Graphics Constructor initializes the Java3D SimpleUniverse, scene graph, lighting and
    * Mouse Behaviors.
    * @param r Root arguement provides a reference to loaded FSystems.
    */
   public Graphics3D(MainPanel f){
      this(GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().
         getBestConfiguration(new GraphicsConfigTemplate3D()), f);
   }

   /**
    * Handles actions performed on various menu items. This
    * includes changes to the viewing and coloring types,
    * as well as capturing the 3D view to a file.
    */
   public void actionPerformed(ActionEvent evt){
      String arg = evt.getActionCommand();

      if (arg.equals("Reset Rotation")){
         FSystem sys = fpanel.getTree().getActive();
         if (sys != null){
            sys.centerView(true, false);
         }
      } else if (arg.equals("Reset Translation")){
         FSystem sys = fpanel.getTree().getActive();
         if (sys != null){
            sys.centerView(false, true);
         }
      } else if (arg.equals("Reset Rotation and Translation")){
         FSystem sys = fpanel.getTree().getActive();
         if (sys != null){
            sys.centerView(true, true);
         }
      } else if (arg.equals("Label Selected Residues")){
         if (GraphicsContext.labelResidues == true){
            GraphicsContext.labelResidues = false;
            getStatusBar().setText("  Residue Labeling Turned Off");
         } else{
            GraphicsContext.labelResidues = true;
            getStatusBar().setText("  Residue Labeling Turned On");
         }
         repaint();
      } else if (arg.equals("Label Selected Atoms")){
         if (GraphicsContext.labelAtoms == true){
            GraphicsContext.labelAtoms = false;
            getStatusBar().setText("  Atom Labeling Turned Off");
         } else{
            GraphicsContext.labelAtoms = true;
            getStatusBar().setText("  Atom Labeling Turned On");
         }
         repaint();
      } else if (arg.equals("Set Label Font Color...")){
         Color color = getGraphics2D().getColor();
         JColorChooser c = new JColorChooser();
         Color newColor = c.showDialog(this, "Choose Font Color", color);
         if (newColor != null && newColor != color){
            getGraphics2D().setPaint(newColor);
            if (GraphicsContext.labelAtoms || GraphicsContext.labelResidues){
               repaint();
            }
            getStatusBar().setText("  Label Font Color Changed to (" + newColor.getRed() + "," +
               newColor.getGreen() + "," + newColor.getBlue() + ")");
         }
      } else if (arg.equals("Set Label Font Size...")){
         Font currentFont = getGraphics2D().getFont();
         int currentSize = currentFont.getSize();
         String size = new String("" + currentSize);
         size = JOptionPane.showInputDialog("Set the Font Size (8 to 64)", size);
         try{
            int f = Integer.parseInt(size);
            if (f < 8 || f > 64 || f == currentSize){
               return;
            }
            Font newFont = new Font(currentFont.getName(), currentFont.getStyle(), f);
            getGraphics2D().setFont(newFont);
            if (GraphicsContext.labelAtoms || GraphicsContext.labelResidues){
               repaint();
            }
            getStatusBar().setText("  Label Font Size Changed to " + newFont.getSize());
         } catch (NumberFormatException e){
            //System.out.println("Enter an integer between 1 and 100");
         }
      } else if (arg.equals("Set Background Color...")){
         Color3f col = new Color3f();
         background.getColor(col);
         JColorChooser c = new JColorChooser();
         Color newcolor = c.showDialog(this, "Choose Background Color", col.get());
         if (newcolor != null && newcolor != col.get()){
            background.setColor(new Color3f(newcolor));
         }
      } else if (arg.equals("JPEG")){
         imageFormat = new String("jpg").intern();
      } else if (arg.equals("PNG")){
         imageFormat = new String("png").intern();
      } else if (arg.equals("Capture Graphics")){
         FSystem active = fpanel.getTree().getActive();
         if (active == null){
            return;
         }
         imageName = null;
         String name = active.getName();
         JFileChooser fileChooser = MainPanel.getFileChooser();
         fileChooser.setAcceptAllFileFilterUsed(true);
         if (fpanel.getTree().getActive() != null){
            imageName = fpanel.getTree().getActive().getFile();
         } else{
            imageName = null;
         }
         if (imageName != null){
            if (name.indexOf(".") > 0){
               name = name.substring(0, name.indexOf("."));
            }
            imageName = new File(imageName.getParentFile() + File.separator + name + "." + imageFormat);
            fileChooser.setSelectedFile(imageName);
         }
         fileChooser.setDialogTitle("Select Name for Screen Capture " + "(" + imageFormat + ")");
         fileChooser.setCurrentDirectory(fpanel.getCWD());
         int result = fileChooser.showSaveDialog(this);
         if (result == JFileChooser.APPROVE_OPTION){
            imageName = fileChooser.getSelectedFile();
            fpanel.setCWD(fileChooser.getCurrentDirectory());
            imageCapture = true;
         }
         //repaint();
      } else if (arg.equalsIgnoreCase("Zoom in")){
         baseTransformGroup.getTransform(baseTransform3D);
         double scale = baseTransform3D.getScale() + 0.01;
         baseTransform3D.setScale(scale);
         baseTransformGroup.setTransform(baseTransform3D);
      } else if (arg.equalsIgnoreCase("Zoom out")){
         baseTransformGroup.getTransform(baseTransform3D);
         double scale = baseTransform3D.getScale() - 0.01;
         if (scale > 0.0){
            baseTransform3D.setScale(scale);
            baseTransformGroup.setTransform(baseTransform3D);
         }
      } else if (arg.equalsIgnoreCase("Reset Global View")){
         centerView();
      } else if (arg.equalsIgnoreCase("Reset Global Translation")){
         eventDispatch.centerView(false, true, false);
      } else if (arg.equalsIgnoreCase("Reset Global Rotation")){
         eventDispatch.centerView(true, false, false);
      } else if (arg.equalsIgnoreCase("Reset Global Zoom")){
         double radius = fpanel.getRoot().getExtent();
         baseTransformGroup.getTransform(baseTransform3D);
         baseTransform3D.setScale(1.0d / (1.2d * radius));
         baseTransformGroup.setTransform(baseTransform3D);
      } else if (arg.equalsIgnoreCase("Rotate About Pick")){
         FNode node = pickNodeBehavior.getPick();
         if (node != null){
            double center[] = node.findCenter(false);
            node.getFSystemNode().rotateAbout(new Vector3d(center));
         }
      } else if (arg.equalsIgnoreCase("Rotate About Center")){
         FSystem sys = fpanel.getTree().getActive();
         double center[] = sys.findCenter(false);
         sys.rotateAbout(new Vector3d(center));
      } else if (arg.equalsIgnoreCase("Graphics Picking")){
         if (evt.getSource() instanceof JButton){
            FMenuBar m = fpanel.getMenu();
            boolean picking = m.getPicking();
            if (picking){
               pickNodeBehavior.clear();
               pickNodeBehavior.setMode(false);
               m.setPickBehavior(false);
            } else{
               pickNodeBehavior.setMode(true);
               m.setPickBehavior(true);
            }
         } else if (evt.getSource() instanceof JCheckBoxMenuItem){
            JCheckBoxMenuItem jcbmi = (JCheckBoxMenuItem) evt.getSource();
            if (jcbmi.isSelected()){
               pickNodeBehavior.setMode(true);
            } else{
               pickNodeBehavior.setMode(false);
            }
         }
      } else if (arg.equals("Atom") || arg.equals("Bond") || arg.equals("Angle") ||
         arg.equals("Dihedral") || arg.equals("Residue") || arg.equals("Polymer") ||
         arg.equals("Molecule") || arg.equals("System")){
         pickNodeBehavior.setPickLevel(arg.intern());
      } else if (GraphicsContext.toView(arg) >= 0){
         int viewtype = GraphicsContext.toView(arg);
         if (viewtype == GraphicsContext.RESTRICT){
            updateBehavior.arm(fpanel.getRoot(), false, true, viewtype, false, 0);
            return;
         }
         ArrayList active = fpanel.getTree().getActiveNodes();
         if (active == null){
            return;
         }
         updateBehavior.arm(active, false, true, viewtype, false, 0);
      } else if (GraphicsContext.toColor(arg) >= 0){
         int colortype = GraphicsContext.toColor(arg);
         ArrayList active = fpanel.getTree().getActiveNodes();
         if (active == null){
            return;
         }
         updateBehavior.arm(active, false, false, 0, true, colortype);
      } else if (arg.equals("System Below Mouse")){
         mouseMode = SCREEN;
      } else if (arg.equals("Active System")){
         mouseMode = ACTIVE;
      } else if (arg.equals("Set Graphics Picking Color...")){
         Color3f col = new Color3f();
         GraphicsContext.pickingColor.get(col);
         JColorChooser c = MainPanel.getColorChooser();
         Color newcolor = c.showDialog(this, "Choose Picking Color", col.get());
         if (newcolor != null){
            GraphicsContext.pickingColor = new Color3f(newcolor);
         }
      } else if (arg.equals("Set Selection Color...")){
         Color3f col = new Color3f();
         GraphicsContext.selectionColor.get(col);
         JColorChooser c = MainPanel.getColorChooser();
         Color newcolor = c.showDialog(this, "Choose Selection Color", col.get());
         if (newcolor != null){
            GraphicsContext.selectionColor = new Color3f(newcolor);
         }
         if (GraphicsContext.highlightSelections){
            this.updateScene(fpanel.getRoot(), false, false, 0, true, GraphicsContext.bySELECT);
         }
      } else if (arg.equals("Set User Color...")){
         Color3f col = new Color3f();
         GraphicsContext.selectionColor.get(col);
         JColorChooser c = MainPanel.getColorChooser();
         Color newcolor = c.showDialog(this, "Choose User Color", col.get());
         if (newcolor != null){
            GraphicsContext.userColor = new Color3f(newcolor);
         }
      } else if (arg.lastIndexOf("Measure") >= 0){
         pickNodeBehavior.setPickLevel(arg.intern());
         pickNodeBehavior.setMode(true);
         fpanel.getMenu().setPickBehavior(true);
         clearPicks();
         pickNodeBehavior.resetCount();
      } else{
         System.out.println("Menu command not found: " + arg.toString());
      }
   }

   public BranchGroup loadVRML(File file){
      if (file == null){
         return null;
      }
      try{
         BufferedReader br = new BufferedReader(new FileReader(file));
         return loadVRML(br);
      } catch (Exception e){
         System.out.println("" + e);
         return null;
      }

   }

   public BranchGroup loadVRML(URL pdburl){
      try{
         InputStream is = pdburl.openStream();
         BufferedReader br = new BufferedReader(new InputStreamReader(is));
         return loadVRML(br);
      } catch (Exception e){
         return null;
      }
   }

   public BranchGroup loadVRML(BufferedReader br){
      try{
         int retry = 0;
         while (!br.ready() && retry < 20){
            synchronized (this){
               wait(50);
               retry++;
            }
         }
         if (!br.ready()){
            return null;
         }
         int flag = VRML97Loader.LOAD_VIEW_GROUPS;
         VRML97Loader loader = new VRML97Loader(flag);
         Scene scene = loader.load(br);
         BranchGroup bg = scene.getSceneGroup();
         recurseVRML(bg);

         bg.setCapability(BranchGroup.ALLOW_DETACH);
         bg.setCapability(BranchGroup.ALLOW_BOUNDS_READ);
         bg.compile();

         return bg;
      } catch (Exception e){
         return null;
      }
   }

   public void recurseVRML(Node node){
      if (node instanceof Shape3D){
         Shape3D s3d = (Shape3D) node;
         PickTool.setCapabilities(s3d, PickTool.INTERSECT_COORD);
         return;
      } else if (node instanceof SharedGroup){
         SharedGroup sg = (SharedGroup) node;
         for (Enumeration e = sg.getAllChildren(); e.hasMoreElements(); ){
            recurseVRML((Node) e.nextElement());
         }
         return;
      } else if (node instanceof BranchGroup){
         BranchGroup bg = (BranchGroup) node;
         for (Enumeration e = bg.getAllChildren(); e.hasMoreElements(); ){
            recurseVRML((Node) e.nextElement());
         }
         return;
      } else if (node instanceof TransformGroup){
         TransformGroup tg = (TransformGroup) node;
         for (Enumeration e = tg.getAllChildren(); e.hasMoreElements(); ){
            recurseVRML((Node) e.nextElement());
         }
         return;
      } else if (node instanceof Link){
         Link link = (Link) node;
         recurseVRML(link.getSharedGroup());
         return;
      } else{
         return;
      }
   }

   public void attach(BranchGroup bg){
      baseBranchGroup.addChild(bg);
   }

   /**
    * This functions trys to center the molecular scene on the 3D canvas
    */
   public void centerView(){
      double radius = fpanel.getRoot().getExtent();
      Transform3D t3d = new Transform3D();
      t3d.setScale(1.0d / (1.2d * radius));
      baseTransformGroup.setTransform(t3d);
      eventDispatch.centerView(true, true, true);
   }

   public void clearPicks(){
      pickNodeBehavior.clear();
   }

   /**
    * Create an appearance object
    */
   static private PointAttributes pointAttributes = new PointAttributes();
   static{
      coloringAttributes.setShadeModel(ColoringAttributes.NICEST);
      coloringAttributes.setColor(new Color3f(0, 0, 0));

      lineAttributes.setLineAntialiasingEnable(true);
      lineAttributes.setLinePattern(LineAttributes.PATTERN_SOLID);
      lineAttributes.setLineWidth(1.0f);

      pointAttributes.setPointAntialiasingEnable(true);
      pointAttributes.setPointSize(1.0f);

      fillPolygonAttributes.setPolygonMode(PolygonAttributes.POLYGON_FILL);
      fillPolygonAttributes.setCullFace(PolygonAttributes.CULL_BACK);
      linePolygonAttributes.setPolygonMode(PolygonAttributes.POLYGON_LINE);
      pointPolygonAttributes.setPolygonMode(PolygonAttributes.POLYGON_POINT);

      renderingAttributes.setVisible(true);
      renderingAttributes.setDepthBufferEnable(true);
      renderingAttributes.setDepthBufferWriteEnable(true);
      renderingAttributes.setIgnoreVertexColors(true);

      transparencyAttributes.setTransparencyMode(TransparencyAttributes.NONE);
   }

   static public Material createMaterial(Color3f col){
      if (col == null){
         return null;
      }
      Material mat = (Material) materials.get(col);
      if (mat == null){
         mat = new Material(col, GraphicsContext.blackColor, col, GraphicsContext.whiteColor, 75.0f);
         mat.setLightingEnable(true);
         materials.put(col, mat);
      }
      return mat;
   }

   static public Appearance getAppearance(Color3f col, int type){
      if (col == null){
         return null;
      }
      Appearance ap;
      if (type == GraphicsContext.FILL){
         ap = (Appearance) fillAppearances.get(col);
      } else if (type == GraphicsContext.POINTS){
         ap = (Appearance) pointAppearances.get(col);
      } else{
         ap = (Appearance) lineAppearances.get(col);
      }
      if (ap == null){
         ap = createNewAppearance(col, type);
      }
      return ap;
   }

   static private Hashtable pointAppearances = new Hashtable();
   static private Hashtable lineAppearances = new Hashtable();
   static private Hashtable fillAppearances = new Hashtable();

   static private Appearance createNewAppearance(Color3f col, int type){
      Appearance ap = new Appearance();
      Material mat = createMaterial(col);
      ap.setMaterial(mat);
      ap.setRenderingAttributes(renderingAttributes);
      ap.setColoringAttributes(coloringAttributes);
      ap.setLineAttributes(lineAttributes);
      ap.setPointAttributes(pointAttributes);
      if (type == GraphicsContext.FILL){
         ap.setPolygonAttributes(fillPolygonAttributes);
         fillAppearances.put(col, ap);
      } else if (type == GraphicsContext.POINTS){
         ap.setPolygonAttributes(pointPolygonAttributes);
         pointAppearances.put(col, ap);
      } else{
         ap.setPolygonAttributes(linePolygonAttributes);
         lineAppearances.put(col, ap);
      }
      return ap;
   }

   private static Geometry conegeom[][] = new Geometry[2][4];
   private static void initConeGeom(int res){
      Appearance ap = new Appearance();
      Cone cone = new Cone(1.0f, 1.0f, Cylinder.GENERATE_NORMALS, (res + 1) * 4, 1, ap);
      for (int i = 0; i < 2; i++){
         conegeom[i][res] = cone.getShape(i).getGeometry();
         conegeom[i][res].setCapability(Geometry.ALLOW_INTERSECT);
         conegeom[i][res].setCapability(GeometryArray.ALLOW_FORMAT_READ);
         conegeom[i][res].setCapability(GeometryArray.ALLOW_COUNT_READ);
         conegeom[i][res].setCapability(GeometryArray.ALLOW_COORDINATE_READ);
      }
   }

   public static Geometry getConeGeom(int num, int res){
      if (res > 3){
         res = 3;
      }
      if (conegeom[num][res] == null){
         initConeGeom(res);
      }
      return conegeom[num][res];
   }

   /**
    * This method creates a Cylinder
    */
   static public Shape3D createCone(Appearance ap, int res){
      if (res > 3){
         res = 3;
      }
      Shape3D cone = new Shape3D();
      cone.setAppearance(ap);
      cone.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      cone.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
      cone.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
      cone.setCapability(Shape3D.ALLOW_PICKABLE_WRITE);
      cone.addGeometry(getConeGeom(0, res));
      cone.addGeometry(getConeGeom(1, res));
      return cone;
   }

   private static Geometry cylgeom[][] = new Geometry[3][4];
   private static void initCylinderGeom(int res){
      Appearance ap = new Appearance();
      Cylinder cyl = new Cylinder(1.0f, 1.0f, Cylinder.GENERATE_NORMALS, (res + 1) * 4, 1, ap);
      for (int i = 0; i < 3; i++){
         cylgeom[i][res] = cyl.getShape(i).getGeometry();
         cylgeom[i][res].setCapability(Geometry.ALLOW_INTERSECT);
         cylgeom[i][res].setCapability(GeometryArray.ALLOW_FORMAT_READ);
         cylgeom[i][res].setCapability(GeometryArray.ALLOW_COUNT_READ);
         cylgeom[i][res].setCapability(GeometryArray.ALLOW_COORDINATE_READ);
      }
   }

   public static Geometry getCylinderGeom(int num, int res){
      if (res > 3){
         res = 3;
      }
      if (cylgeom[num][res] == null){
         initCylinderGeom(res);
      }
      return cylgeom[num][res];
   }

   /**
    * This method creates a Cylinder
    */
   static public Shape3D createCylinder(Appearance ap, int res){
      if (res > 3){
         res = 3;
      }
      Shape3D cyl = new Shape3D();
      cyl.setAppearance(ap);
      cyl.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      cyl.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
      cyl.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
      cyl.setCapability(Shape3D.ALLOW_PICKABLE_WRITE);
      cyl.addGeometry(getCylinderGeom(0, res));
      cyl.addGeometry(getCylinderGeom(1, res));
      cyl.addGeometry(getCylinderGeom(2, res));
      return cyl;
   }

   static public Shape3D createPolar(Appearance ap, int div){
      return createSphere(ap, div);
   }

   /**
    * This method creates a single Sphere from the given appearance
    */
   static public Shape3D createSphere(Appearance ap, int div){
      Shape3D sphere = new Shape3D();
      sphere.setAppearance(ap);
      sphere.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      sphere.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
      sphere.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
      sphere.setCapabilityIsFrequent(Shape3D.ALLOW_APPEARANCE_WRITE);
      sphere.setCapability(Shape3D.ALLOW_PICKABLE_WRITE);
      sphere.addGeometry(getSphereGeom(div));
      return sphere;
   }

   private static Geometry sphereGeom[] = new Geometry[6];
   private static void initSphereGeom(int res){
      Appearance ap = new Appearance();
      Sphere sphere;
      sphere = new Sphere(1.0f, Sphere.GENERATE_NORMALS, 5 * (res + 1), ap);
      sphere.getShape().setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
      sphereGeom[res] = sphere.getShape().getGeometry();
      sphereGeom[res].setCapability(Geometry.ALLOW_INTERSECT);
      GeometryArray g = (GeometryArray) sphereGeom[res];
      g.setCapability(g.ALLOW_FORMAT_READ);
      g.setCapability(g.ALLOW_COUNT_READ);
      g.setCapability(g.ALLOW_COORDINATE_READ);
   }

   /**
    * Finds the desired color for an Atom
    */
   static public Color3f getColor(Atom a, int colortype){
      Color3f col = null;
      if (colortype == GraphicsContext.byATOM){
         col = colors[a.getAtomicNum()];
      } else if (colortype == GraphicsContext.byGROUP){
         col = (Color3f) AtomUtils.colorlookup.get(a.getResidueNode().getName());
      } else if (colortype == GraphicsContext.byMOLECULE){
         col = (Color3f) AtomUtils.colorlookup.get(a.getPolymerNode().getName());
      } else if (colortype == GraphicsContext.byPICK){
         col = GraphicsContext.pickingColor;
      } else if (colortype == GraphicsContext.bySELECT){
         col = GraphicsContext.selectionColor;
      } else if (colortype == GraphicsContext.byCHARGE){
         int index = 0;
         float charge = a.getCharge();
         if (charge < 0){
            index = Math.round(charge * -10f);
            if (index > 9){
               index = 9;
            }
            col = negCharge[index];
         } else if (charge == 0){
            col = GraphicsContext.whiteColor;
         } else{
            index = Math.round(charge * 10f);
            if (index > 9){
               index = 9;
            }
            col = posCharge[index];
         }
      } else if (colortype == GraphicsContext.byVECTOR){
         int index = 0;
         double mag = a.getMagnitude(true);
         if (mag > 0){
            index = (int) Math.round(mag * 10.0);
            if (index > 9){
               index = 9;
            }
            col = negCharge[index];
         } else if (mag <= 0){
            col = GraphicsContext.whiteColor;
         }
      }
      if (col == null){
         col = GraphicsContext.nullColor;
      }
      return col;
   }

   private static Color3f negCharge[] = new Color3f[10];
   private static Color3f posCharge[] = new Color3f[10];
   static{
      float value;
      for (int i = 1; i <= 10; i++){
         value = i * 0.1f;
         negCharge[i - 1] = new Color3f(1.0f, 1.0f - value, 1.0f - value);
         posCharge[i - 1] = new Color3f(1.0f - value, 1.0f - value, 1.0f);
      }

   }

   public GlobalAxis getNavigation(){
      return sceneAxis;
   }

   public int getPickMode(){
      return mouseMode;
   }

   public static Geometry getPolarGeom(int res){
      return getSphereGeom(res);
   }

   public JLabel getStatusBar(){
      return fpanel.getStatusBar();
   }

   public static Geometry getSphereGeom(int res){
      if (sphereGeom[res] == null){
         initSphereGeom(res);
      }
      return sphereGeom[res];
   }

   public void paint(java.awt.Graphics g){
      super.paint(g);
      Toolkit.getDefaultToolkit().sync();
   }

   /**
    * This function facilitates image capture from the 3D Canvas.
    */
   public void postSwap(){
      if (!imageCapture || fpanel.getTree().getActive() == null){
         return;
      }

      GraphicsContext3D ctx = getGraphicsContext3D();
      Rectangle rect = this.getBounds();
      BufferedImage img = new BufferedImage(rect.width, rect.height, BufferedImage.TYPE_INT_RGB);
      ImageComponent2D comp = new ImageComponent2D(ImageComponent.FORMAT_RGB, img, true, false);
      Raster ras = new Raster(new Point3f( -1.0f, -1.0f, -1.0f), Raster.RASTER_COLOR, 0, 0,
                   rect.width, rect.height, comp, null);

      ctx.readRaster(ras);

      try{
         ImageIO.write(img, imageFormat, imageName);
      } catch (IOException e){
         // System.out.println("" + e);
      }
      imageCapture = false;
   }

   public Point2d getScreenCoordinate(Node node, Point3d point3d){
      // Get the transform to put the node in the world coordinate system
      Transform3D localToVworld = new Transform3D();
      node.getLocalToVworld(localToVworld);

      // Get the image plate transform
      Transform3D vWorldToImagePlate = new Transform3D();
      View view = getView();
      view.getCanvas3D(0).getVworldToImagePlate(vWorldToImagePlate);

      // Transform into world coordinates
      localToVworld.transform(point3d);
      // Transform into imageplate coordinates
      vWorldToImagePlate.transform(point3d);
      // Final step to the 2D Screen.
      Point2d point = new Point2d();
      view.getCanvas3D(0).getPixelLocationFromImagePlate(point3d, point);

      /**
       Now we have the location where the point will be rendered on the screen depending
       on resize, placement, size, and eye point policies.
       This should only be called on points that reside within the clipping planes.
       */
      return point;
   }

   public void postRender(){
      super.postRender();

      if (GraphicsContext.labelAtoms || GraphicsContext.labelResidues){
         J3DGraphics2D g2D = getGraphics2D();

         synchronized (fpanel.getTree()){
            ArrayList nodes = fpanel.getTree().getActiveNodes();
            if (nodes != null && nodes.size() > 0){
               for (ListIterator li = nodes.listIterator(); li.hasNext(); ){
                  FNode node = (FNode) li.next();
                  FSystem sys = node.getFSystemNode();
                  if (sys != null){
                     node.drawLabel(this, g2D, sys.getWireFrame(), GraphicsContext.labelAtoms,
                        GraphicsContext.labelResidues);
                  }
               }
            } else{
               return;
            }
         }
         g2D.flush(true);
      }
      //System.out.println("Frame Time: " + this.getView().getLastFrameDuration());
   }

   public void setLabelsUpdated(){
      repaint();
   }

   /**
    * This method removes a FSystem from the Scene Graph
    * @param m FSystem to remove
    */
   public void remove(FSystem m){
      BranchGroup bg = m.getBranchGroup();
      bg.detach();
   }

   /**
    * Save preferences to the user node
    */
   public void savePrefs(){
      Preferences prefs = Preferences.userNodeForPackage(MainPanel.class);
      prefs.putInt("Graphics_bondwidth", GraphicsContext.bondwidth);
      prefs.putInt("Graphics_detail", GraphicsContext.detail);
      prefs.putDouble("Graphics_radius", GraphicsContext.radius);
      prefs.putInt("Graphics_mouse", fpanel.getMenu().getMouseMode());
      prefs.putBoolean("Graphics_highlight", GraphicsContext.highlightSelections);
      Color col = GraphicsContext.selectionColor.get();
      prefs.put("Graphics_highlightColor",
         "" + col.getRed() + " " + col.getGreen() + " " + col.getBlue());
      prefs.putBoolean("Graphics_labelAtoms", GraphicsContext.labelAtoms);
      prefs.putBoolean("Graphics_labelResidues", GraphicsContext.labelResidues);
      prefs.putInt("Graphics_labelSize", getGraphics2D().getFont().getSize());
      Color fontColor = getGraphics2D().getColor();
      prefs.put("Graphics_labelColor",
         "" + fontColor.getRed() + " " + fontColor.getGreen() + " " + fontColor.getBlue());
      col = GraphicsContext.pickingColor.get();
      prefs.put("Graphics_pickColor", "" + col.getRed() + " " + col.getGreen() + " " + col.getBlue());
      prefs.putBoolean("Graphics_picking", fpanel.getMenu().isPickingActive());
      prefs.put("Graphics_pickLevel", pickNodeBehavior.getPickLevel());
      col = GraphicsContext.userColor.get();
      prefs.put("Graphics_userColor", "" + col.getRed() + " " + col.getGreen() + " " + col.getBlue());
      Color3f temp = new Color3f();
      background.getColor(temp);
      col = temp.get();
      prefs.put("Graphics_backgroundColor",
         "" + col.getRed() + " " + col.getGreen() + " " + col.getBlue());
   }

   /**
    * Load preferences from the user node
    */
   public void loadPrefs(){
      Preferences prefs = Preferences.userNodeForPackage(MainPanel.class);
      GraphicsContext.bondwidth = prefs.getInt("Graphics_bondwidth", 3);
      GraphicsContext.detail = prefs.getInt("Graphics_detail", 3);
      GraphicsContext.radius = prefs.getDouble("Graphics_radius", 1.0d);
      int mouseMode = prefs.getInt("Graphics_mouse", Graphics3D.ACTIVE);
      fpanel.getMenu().setMouseMode(mouseMode);
      GraphicsContext.highlightSelections = prefs.getBoolean("Graphics_highlight", false);
      fpanel.getMenu().setHighlighting(GraphicsContext.highlightSelections);
      String[] hlColor = prefs.get("Graphics_highlightColor", "153 153 255").trim().split(" +");
      Color newColor = new Color(Integer.parseInt(hlColor[0]), Integer.parseInt(hlColor[1]),
                       Integer.parseInt(hlColor[2]));
      GraphicsContext.selectionColor = new Color3f(newColor);
      GraphicsContext.labelAtoms = prefs.getBoolean("Graphics_labelAtoms", false);
      fpanel.getMenu().setAtomLabels(GraphicsContext.labelAtoms);
      GraphicsContext.labelResidues = prefs.getBoolean("Graphics_labelResidues", false);
      fpanel.getMenu().setResidueLabels(GraphicsContext.labelResidues);
      int fontSize = prefs.getInt("Graphics_labelSize", 12);
      J3DGraphics2D j2D = getGraphics2D();
      Font currentFont = j2D.getFont();
      Font newFont = new Font(currentFont.getName(), currentFont.getStyle(), fontSize);
      j2D.setFont(newFont);
      String[] fontColor = prefs.get("Graphics_labelColor", "255 255 255").trim().split(" +");
      newColor = new Color(Integer.parseInt(fontColor[0]), Integer.parseInt(fontColor[1]),
                 Integer.parseInt(fontColor[2]));
      j2D.setPaint(newColor);
      String[] pickColor = prefs.get("Graphics_pickColor", "102 255 102").trim().split(" +");
      newColor = new Color(Integer.parseInt(pickColor[0]), Integer.parseInt(pickColor[1]),
                 Integer.parseInt(pickColor[2]));
      GraphicsContext.pickingColor = new Color3f(newColor);
      String pickLevel = prefs.get("Graphics_pickLevel", "Atom");
      fpanel.getMenu().setPickLevel(pickLevel);
      boolean pickMode = prefs.getBoolean("Graphics_picking", false);
      if (pickMode){
         pickNodeBehavior.setMode(true);
      }
      fpanel.getMenu().setPickBehavior(pickMode);
      String[] userColor = prefs.get("Graphics_userColor", "255 255 255").trim().split(" +");
      newColor = new Color(Integer.parseInt(userColor[0]), Integer.parseInt(userColor[1]),
                 Integer.parseInt(userColor[2]));
      GraphicsContext.userColor = new Color3f(newColor);
      String[] bgColor = prefs.get("Graphics_backgroundColor", "0 0 0").trim().split(" +");
      newColor = new Color(Integer.parseInt(bgColor[0]), Integer.parseInt(bgColor[1]),
                 Integer.parseInt(bgColor[2]));
      background.setColor(new Color3f(newColor));
   }

   /**
    * This method renders a FSystem to the 3DCanvas
    * @param m FSystem to render
    */
   public void Render(FSystem s, boolean update){
      if (s == null){
         return;
      }
      BranchGroup bg = s.getBranchGroup();
      centerView();
      if (!bg.isLive()){
         baseBranchGroup.addChild(bg);
         if (update){
            if (s.getBondList().size() != 0){
               updateScene(s, false, true, GraphicsContext.WIREFRAME, true, GraphicsContext.byATOM);
            } else{
               updateScene(s, false, true, GraphicsContext.SPACEFILL, true, GraphicsContext.byATOM);
            }
         }
      }
   }

   public void setAxisShowing(boolean b){
      if (b && sceneAxis == null){
         sceneAxis = new GlobalAxis(universe.getViewingPlatform(), this, bounds);
      } else if (sceneAxis != null){
         sceneAxis.showAxis(b);
      }
   }

   public void setCapture(boolean c){
      imageCapture = c;
   }

   public void selected(){

   }

   public void updateScene(FNode n, boolean t, boolean v, int viewtype, boolean c, int colortype){
      if (n != null){
         updateBehavior.arm(n, t, v, viewtype, c, colortype);
      }
   }

   public void updateScene(ArrayList n, boolean t, boolean v, int viewtype, boolean c,
      int colortype){
      if (n != null){
         updateBehavior.arm(n, t, v, viewtype, c, colortype);
      }
   }

   public void setView(String style){
      setView(fpanel.getTree().getActive(), style);
   }

   public void setView(FNode node, String style){
      int v = GraphicsContext.toView(style);
      if (v != -1){
         updateScene(node, false, true, v, false, 0);
      }
   }

   public void setColor(String colortype){
      setColor(fpanel.getTree().getActive(), colortype);
   }

   public void setColor(FNode node, String colortype){
      int c = GraphicsContext.toColor(colortype);
      if (c != -1){
         updateScene(node, false, false, 0, true, c);
      }
   }

   public void setPosition(){
      setPosition(fpanel.getTree().getActive());
   }

   public void setPosition(FNode node){
      updateScene(node, true, false, 0, true, 0);
   }

   public boolean sceneRendering(){
      return updateBehavior.isArmed();
   }

   public boolean cacheFull(){
      return updateBehavior.cacheFull();
   }

   //  Initialization of the Branch Group Hierarchy
   private void initialize(){

      setBackground(Color.black);
      universe = new SimpleUniverse(this);
      universe.setJ3DThreadPriority(Thread.MAX_PRIORITY);
      universe.getViewingPlatform().setNominalViewingTransform();
      baseTransformGroup = new TransformGroup();

      //  Create the Scene Root BranchGroup
      BranchGroup objRoot = new BranchGroup();
      Transform3D t3d = new Transform3D();
      t3d.setScale(0.1d);
      baseTransformGroup.setTransform(t3d);
      baseTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      baseTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);

      //  Set the Background
      Color3f bgColor = new Color3f(GraphicsContext.backgroundColor);
      bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 2000.0);
      background = new Background(bgColor);
      background.setCapability(Background.ALLOW_COLOR_READ);
      background.setCapability(Background.ALLOW_COLOR_WRITE);
      background.setApplicationBounds(bounds);

      // Create lights
      // Ambient
      Color3f alColor = new Color3f(0.2f, 0.2f, 0.2f);
      AmbientLight aLgt = new AmbientLight(alColor);
      aLgt.setInfluencingBounds(bounds);

      // Directional
      Vector3f dir = new Vector3f(0.0f, -0.5f, -.5f);
      DirectionalLight dLgt = new DirectionalLight(GraphicsContext.lightGray, dir);
      dLgt.setInfluencingBounds(bounds);

      // Create the Base of the Molecular Scene
      baseBranchGroup = new BranchGroup();
      baseBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
      baseBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
      baseBranchGroup.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
      baseBranchGroup.setCapability(BranchGroup.ALLOW_BOUNDS_READ);
      // Add children created above to the base TransformGroup
      baseTransformGroup.addChild(background);
      baseTransformGroup.addChild(baseBranchGroup);
      objRoot.addChild(baseTransformGroup);

      // Position the view platmform and add lights
      View v = universe.getViewer().getView();
      v.setProjectionPolicy(View.PARALLEL_PROJECTION);
      v.setFrontClipPolicy(View.VIRTUAL_EYE);
      v.setFrontClipDistance(0.1);
      v.setBackClipPolicy(View.VIRTUAL_EYE);
      v.setBackClipDistance(1000.0);
      v.setTransparencySortingPolicy(View.TRANSPARENCY_SORT_NONE);

      //  v.setFieldOfView(Math.PI / 4.0);
      Transform3D trans = new Transform3D();
      trans.set(new Vector3d(0.0d, 0.0d, 2.0d));
      TransformGroup vptg = universe.getViewingPlatform().getViewPlatformTransform();
      vptg.setTransform(trans);
      BranchGroup viewBranch = new BranchGroup();
      viewBranch.addChild(aLgt);
      viewBranch.addChild(dLgt);
      vptg.addChild(viewBranch);

      // Initialize Behaviors
      sceneAxis = new GlobalAxis(universe.getViewingPlatform(), this, bounds);
      eventDispatch = new EventDispatch(fpanel, this, sceneAxis, universe, bounds, baseBranchGroup,
                      baseTransformGroup);
      baseBranchGroup.addChild(eventDispatch);
      pickNodeBehavior = new PickNodeBehavior(baseBranchGroup, bounds, this, fpanel);
      baseBranchGroup.addChild(pickNodeBehavior);
      updateBehavior = new UpdateBehavior(bounds, this);
      baseBranchGroup.addChild(updateBehavior);

      //  Compile the Root BranchGroup and add it to the Universe
      objRoot.compile();
      universe.addBranchGraph(objRoot);
   }

   // More Image Writing Capabilities are on the way from Sun
   public static final int JPEG = 1;
   public static final int PNG = 2;
   public static final int SCREEN = 1;
   public static final int ACTIVE = 2;

   public static Color3f colors[] = new Color3f[110];
   public static float vdw[] = new float[110];
   static{
      colors[0] = new Color3f(Color.lightGray);

      colors[1] = new Color3f(Color.white);
      colors[2] = new Color3f(Color.green);

      colors[3] = new Color3f(Color.magenta);
      colors[4] = colors[3];
      colors[5] = colors[3];
      colors[6] = colors[0];
      colors[7] = new Color3f(Color.blue);
      colors[8] = new Color3f(Color.red);
      colors[9] = colors[2];
      colors[10] = colors[2];

      colors[11] = colors[3];
      colors[12] = colors[3];
      colors[13] = colors[3];
      colors[14] = colors[0];
      colors[15] = new Color3f(Color.orange);
      colors[16] = new Color3f(Color.yellow);
      colors[17] = colors[2];
      colors[18] = colors[2];

      colors[19] = colors[3];
      colors[20] = colors[3];
      colors[21] = colors[3];
      colors[22] = colors[3];
      colors[22] = colors[3];
      colors[23] = colors[3];
      colors[24] = colors[3];
      colors[25] = colors[3];
      colors[26] = colors[3];
      colors[27] = colors[3];
      colors[28] = colors[3];
      colors[29] = colors[3];
      colors[30] = colors[3];
      colors[31] = colors[3];
      colors[32] = colors[3];
      colors[33] = colors[3];
      colors[34] = colors[3];
      colors[35] = colors[3];
      colors[36] = colors[2];

      colors[37] = colors[3];
      colors[38] = colors[3];
      colors[39] = colors[3];
      colors[40] = colors[3];
      colors[41] = colors[3];
      colors[42] = colors[3];
      colors[43] = colors[3];
      colors[44] = colors[3];
      colors[45] = colors[3];
      colors[46] = colors[3];
      colors[47] = colors[3];
      colors[48] = colors[3];
      colors[49] = colors[3];
      colors[50] = colors[3];
      colors[51] = colors[3];
      colors[52] = colors[3];

      colors[53] = colors[3];
      colors[54] = colors[3];
      colors[55] = colors[3];
      colors[56] = colors[3];
      colors[57] = colors[3];
      colors[58] = colors[3];
      colors[59] = colors[3];
      colors[60] = colors[3];
      colors[61] = colors[3];
      colors[62] = colors[3];
      colors[63] = colors[3];
      colors[64] = colors[3];
      colors[65] = colors[3];
      colors[66] = colors[3];
      colors[67] = colors[3];
      colors[68] = colors[3];
      colors[69] = colors[3];
      colors[70] = colors[3];
      colors[71] = colors[3];
      colors[72] = colors[3];
      colors[73] = colors[3];
      colors[74] = colors[3];
      colors[75] = colors[3];
      colors[76] = colors[3];
      colors[77] = colors[3];
      colors[78] = colors[3];
      colors[79] = colors[3];
      colors[80] = colors[3];
      colors[81] = colors[3];
      colors[82] = colors[3];
      colors[83] = colors[3];
      colors[84] = colors[3];
      colors[85] = colors[3];
      colors[86] = colors[2];

      colors[87] = colors[3];
      colors[88] = colors[3];
      colors[89] = colors[3];
      colors[90] = colors[3];
      colors[91] = colors[3];
      colors[92] = colors[3];
      colors[93] = colors[3];
      colors[94] = colors[3];
      colors[95] = colors[3];
      colors[96] = colors[3];
      colors[97] = colors[3];
      colors[98] = colors[3];
      colors[99] = colors[3];
      colors[100] = colors[3];
      colors[101] = colors[3];
      colors[102] = colors[3];
      colors[103] = colors[3];
      colors[104] = colors[3];
      colors[105] = colors[3];
      colors[106] = colors[3];
      colors[107] = colors[3];
      colors[108] = colors[3];
      colors[109] = colors[3];

      vdw[0] = 1.0f;

      vdw[1] = 1.20f;
      vdw[2] = 1.22f;

      vdw[3] = 0.78f;
      vdw[4] = 0.34f;
      vdw[5] = 2.08f;
      vdw[6] = 1.85f;
      vdw[7] = 1.54f;
      vdw[8] = 1.40f;
      vdw[9] = 1.35f;
      vdw[10] = 1.60f;

      vdw[11] = 0.98f;
      vdw[12] = 0.78f;
      vdw[13] = 0.57f;
      vdw[14] = 2.00f;
      vdw[15] = 1.90f;
      vdw[16] = 1.85f;
      vdw[17] = 1.81f;
      vdw[18] = 1.91f;

      vdw[19] = 1.33f;
      vdw[20] = 1.06f;
      vdw[21] = 0.91f;
      vdw[22] = 0.83f;
      vdw[22] = 0.82f;
      vdw[23] = 0.78f;
      vdw[24] = 1.0f;
      vdw[25] = 1.0f;
      vdw[26] = 1.0f;
      vdw[27] = 1.0f;
      vdw[28] = 1.0f;
      vdw[29] = 1.0f;
      vdw[30] = 1.0f;
      vdw[31] = 1.0f;
      vdw[32] = 1.0f;
      vdw[33] = 1.0f;
      vdw[34] = 1.0f;
      vdw[35] = 1.95f;
      vdw[36] = 1.89f;

      vdw[37] = 1.0f;
      vdw[38] = 1.0f;
      vdw[39] = 1.0f;
      vdw[40] = 1.0f;
      vdw[41] = 1.0f;
      vdw[42] = 1.0f;
      vdw[43] = 1.0f;
      vdw[44] = 1.0f;
      vdw[45] = 1.0f;
      vdw[46] = 1.0f;
      vdw[47] = 1.0f;
      vdw[48] = 1.0f;
      vdw[49] = 1.0f;
      vdw[50] = 1.0f;
      vdw[51] = 1.0f;
      vdw[52] = 1.0f;
      vdw[53] = 1.0f;

      vdw[54] = 1.0f;
      vdw[55] = 1.0f;
      vdw[56] = 1.0f;
      vdw[57] = 1.0f;
      vdw[58] = 1.0f;
      vdw[59] = 1.0f;
      vdw[60] = 1.0f;
      vdw[61] = 1.0f;
      vdw[62] = 1.0f;
      vdw[63] = 1.0f;
      vdw[64] = 1.0f;
      vdw[65] = 1.0f;
      vdw[66] = 1.0f;
      vdw[67] = 1.0f;
      vdw[68] = 1.0f;
      vdw[69] = 1.0f;
      vdw[70] = 1.0f;
      vdw[71] = 1.0f;
      vdw[72] = 1.0f;
      vdw[73] = 1.0f;
      vdw[74] = 1.0f;
      vdw[75] = 1.0f;
      vdw[76] = 1.0f;
      vdw[77] = 1.0f;
      vdw[78] = 1.0f;
      vdw[79] = 1.0f;
      vdw[80] = 1.0f;
      vdw[81] = 1.0f;
      vdw[82] = 1.0f;
      vdw[83] = 1.0f;
      vdw[84] = 1.0f;
      vdw[85] = 1.0f;
      vdw[86] = 1.0f;

      vdw[87] = 1.0f;
      vdw[88] = 1.0f;
      vdw[89] = 1.0f;
      vdw[90] = 1.0f;
      vdw[91] = 1.0f;
      vdw[92] = 1.0f;
      vdw[93] = 1.0f;
      vdw[94] = 1.0f;
      vdw[95] = 1.0f;
      vdw[96] = 1.0f;
      vdw[97] = 1.0f;
      vdw[98] = 1.0f;
      vdw[99] = 1.0f;
      vdw[100] = 1.0f;
      vdw[101] = 1.0f;
      vdw[102] = 1.0f;
      vdw[103] = 1.0f;
      vdw[104] = 1.0f;
      vdw[105] = 1.0f;
      vdw[106] = 1.0f;
      vdw[107] = 1.0f;
      vdw[108] = 1.0f;
      vdw[109] = 1.0f;

   }

}
