/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.prefs.Preferences;
import javax.xml.parsers.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import org.w3c.dom.*;
import org.xml.sax.SAXException;

import ffe.lang.FSystem;
import ffe.parsers.*;
import ffe.xml.DTDResolver;

/**
 * The JobPanel class encapsulates functionality needed to execute
 * TINKER routines.
 */
public class JobPanel
   extends JPanel
   implements ActionListener, MouseListener, ChangeListener{

   // Active System Data
   private FSystem currentSystem = null;
   private String currentCommand; // Currently Selected Command
   private byte fileTypes; // File Types
   private Vector jobs = new Vector(); // Executing Tinker Jobs

   // Log Settings
   private JComboBox logMode = new JComboBox();
   private String overwrite = new String("Overwrite Existing Log");
   private String append = new String("Append to an Existing Log");
   private String createNew = new String("Create a New Log");
   private String logString = null;

   // TINKER Commands for supported file types
   private NodeList commandList;
   private JComboBox xyzCommands;
   private JComboBox intCommands;
   private JComboBox arcCommands;
   private JComboBox pdbCommands;
   private JComboBox currentCommandBox;

   // Swing Components
   private MainPanel fpanel;
   private JLabel status;
   private EtchedBorder eb = new EtchedBorder(EtchedBorder.RAISED);
   private JTabbedPane jtpJobParams;
   private JSplitPane jspCommand;
   private JTextArea taDescription;
   private JPanel commandPanel;
   private JToolBar toolbar;
   private JCheckBoxMenuItem jcbDescript;
   private Hashtable conditionals = new Hashtable();
   private String action = "NONE";

   // Environment Variables
   private String classpath;
   private String ld_library_path;
   private String path;
   private String pathjre;

   /**
    * Constructor
    */
   public JobPanel(MainPanel f){
      super();
      fpanel = f;
      initialize();
   }

   public void actionPerformed(ActionEvent evt){

      String arg = evt.getActionCommand();
      if (evt.getSource() instanceof JComboBox){
         // A change to the selected TINKER Command
         if (arg.equals("TinkerCommand")){
            JComboBox jcb = (JComboBox) toolbar.getComponentAtIndex(2);
            String com = jcb.getSelectedItem().toString();
            if (!com.equals(currentCommand)){
               currentCommand = com.toLowerCase();
               loadCommandPanel();
            }
         } else{
            logString = getLogString();
            status.setText("  " + readCommand());
         }
         // Launch the selected command on the active XYZ File
      } else if (arg.equalsIgnoreCase("Launch")){
         executePanel();
      } else if (arg.equalsIgnoreCase("End")){
         setEnd();
      } else if (arg.equalsIgnoreCase("Set Path...")){
         setPath();
      } else if (arg.equalsIgnoreCase("Set Log File...")){
         setLog();
      } else if (arg.equalsIgnoreCase("Set Key File...")){
         setKey();
      } else if (arg.equals("Description")){
         JCheckBoxMenuItem box = (JCheckBoxMenuItem) evt.getSource();
         setDivider(box.isSelected());
      }
   }

   /*
       public void execute(String c, String params){
      if (!setCommand(c)){
         return;
      }
      FSystem sys = fpanel.getTree().getActive();
      if (sys == null){
         return;
      }
      if ((sys.getFileType() & fileTypes) == 0){
         return;
      }
      File f = sys.getFile();
      if (f == null){
         return;
      }
      String s = c + " " + f.getAbsolutePath() + " " + params;
      String dir = f.getAbsoluteFile().getParent();
      if (dir == null){
         return;
      }
      launch(s, dir);
       }*/

   /**
    * Get a Vector of executing TINKER jobs
    * @return a Vector containing Thread objects
    */
   public Vector getJobs(){
      return jobs;
   }

   /**
    * Execute a command on the active system in the specified directory
    * @param command The command to be excuted. The command is ignored if it
    * isn't found in the command.xml file
    * @param dir The directory to execute the command in.
    */
   public void launch(String command, String dir){

      // Get the active system
      FSystem m = fpanel.getTree().getActive();
      if (m == null){
         return;
      }

      // Check that the TINKER *.exe exists
      File exe = new File(path + File.separator + currentCommand);
      if (!exe.exists()){
         exe = new File(exe.getAbsolutePath() + ".exe");
         if (!exe.exists()){
            //System.out.println("" + exe.toString());
            String message = new String("The " + currentCommand + " executable was not found in " +
                             path + ". Please use the 'Set Path...' dialog.");
            JOptionPane.showMessageDialog(null, message, "Could not launch " + currentCommand,
               JOptionPane.ERROR_MESSAGE);
            return;
         }
      }

      // Check that a key file exists
      if (m.getKeyFile() == null){
         fpanel.createKeyFile(m);
         if (m.getKeyFile() == null){
            return;
         }
      }

      // Check that the directory to execute the command in is valid
      File dirf = new File(dir);
      if (!dirf.exists()){
         return;
      }

      // Check that this TINKER command executes on this file type
      if ((fileTypes & m.getFileType()) == 0){
         String message = new String(currentCommand.toUpperCase() + " does not execute on " +
                          SystemFilter.toFileString(m.getFileType()) + " files.");
         JOptionPane.showMessageDialog(null, message, "Could not launch " + currentCommand,
            JOptionPane.ERROR_MESSAGE);
         return;
      }

      // Check for an interactive job
      boolean connect = false;
      if (action.toUpperCase().indexOf("CONNECT") >= 0){
         connect = true;
      }

      // Check to see if a structure file will be created
      File newFile = null;
      if (action.toUpperCase().indexOf("LOAD") >= 0){
         File oldFile = currentSystem.getFile();
         if (action.toUpperCase().indexOf("LOADXYZ") >= 0 ){
            String fileName = oldFile.getAbsolutePath();
            int dot = fileName.lastIndexOf(".");
            fileName = fileName.substring(0, dot) + ".xyz";
            oldFile = new File(fileName);
         } else if (action.toUpperCase().indexOf("LOADINT") >= 0 ){
            String fileName = oldFile.getAbsolutePath();
            int dot = fileName.lastIndexOf(".");
            fileName = fileName.substring(0, dot) + ".int";
            oldFile = new File(fileName);
         } else if (action.toUpperCase().indexOf("LOADPDB") >= 0){
            String fileName = oldFile.getAbsolutePath();
            int dot = fileName.lastIndexOf(".");
            fileName = fileName.substring(0, dot) + ".pdb";
            oldFile = new File(fileName);
         }
         newFile = SystemFilter.version(oldFile);
      }

      // Save any changes that have been made to the key file
      fpanel.getKeywordPanel().saveChanges();

      // Decide on a Log file
      if (((String) logMode.getSelectedItem()).startsWith("Create")){
         File newLog = SystemFilter.version(currentSystem.getLogFile());
         currentSystem.setLogFile(newLog);
         command = readCommand();
      }
      String logName = currentSystem.getLogFile().getAbsolutePath();

      // Remove any TINKER *.END files
      removeEnd();

      // Finally, create and execute the command in a new thread.
      FExec fexec = new FExec(m, logName, command, dir, pathjre, classpath, ld_library_path, fpanel,
                    newFile, connect);
      if (!fexec.isNativeLibFound()){
         return;
      }

      Thread t = new Thread(fexec);
      t.setPriority(Thread.NORM_PRIORITY);
      t.setName(logName);

      // If the job progressively modifies coordinates, connect to it.
      if (connect){
         fpanel.connectToTINKER(m, t);
      } else{
         t.start();
      }

      // If some action should be taken when the job finishes, add it to the Jobs Vector
      if (!action.equalsIgnoreCase("NONE")){
         jobs.add(t);
         fpanel.getLogPanel().refreshStatus();
      }
   }

   /**
    * Load the active system into the JobPanel. This should be called whenever
    * the active system changes.
    */
   public void loadActive(FSystem active){
      currentSystem = active;

      // No Open Molecules
      if (active == null || active.isClosing()){
         status.setText("  ");
         return;
      }

      if (toolbar == null){
         initLazy();
      }

      refreshLogLabels();

      int type = currentSystem.getFileType();

      if (currentCommandBox != null){
         toolbar.remove(2);
      }

      if (type == SystemFilter.XYZ){
         toolbar.add(xyzCommands);
         currentCommandBox = xyzCommands;
      } else if (type == SystemFilter.INT){
         toolbar.add(intCommands);
         currentCommandBox = intCommands;
      } else if (type == SystemFilter.ARC){
         toolbar.add(arcCommands);
         currentCommandBox = arcCommands;
      } else if (type == SystemFilter.PDB){
         toolbar.add(pdbCommands);
         currentCommandBox = pdbCommands;
      } else{
         currentCommandBox = null;
      }

      if (currentCommandBox != null){
         toolbar.add(currentCommandBox, 2);
         currentCommandBox.setSelectedIndex(0);
      }

      //status.setText("  " + readCommand());
      jtpJobParams.setPreferredSize(jtpJobParams.getPreferredSize());
   }

   /**
    * Load misc. user preferences. These currently include:
    * 1.) The path to the TINKER executables
    * 2.) The location of the Desription divider
    */
   private boolean lazyDone = false;
   public void loadPrefs(){

      // Set the Classpath
      URL url = getClass().getClassLoader().getResource("ffe/icons/launch.png");
      classpath = url.getPath();
      int end = classpath.indexOf('!');
      classpath = classpath.substring(5, end);
      if (classpath.lastIndexOf("%20") > 0) {
         classpath = classpath.replaceAll("%20", " ");

      }
      File tempClass = new File(classpath);
      classpath = tempClass.getAbsolutePath();

      //JOptionPane.showMessageDialog(null, classpath, "Classpath" + currentCommand,
      //         JOptionPane.ERROR_MESSAGE);

      boolean pathSet = false;

      // First set the path using "-Dtinker.dir=path" from the Java command line
      String temp = System.getProperty("tinker.dir");
      if (temp != null){
         File tinker = new File(temp);
         File params = new File(tinker.getAbsolutePath() + File.separator + "params");
         if (tinker.exists() && params.exists()){
            setPaths(tinker.getAbsolutePath());
            pathSet = true;
         }
      }

      // Next, try working back from the location of tinker/ffe-src/ffe.jar to the bin & params directories
      if (!pathSet) {
         File ffeFile = new File(classpath);
         if (ffeFile.exists()){
            File ffeDir = ffeFile.getParentFile();
            if (ffeDir.exists()){
               File tinkerDir = ffeDir.getParentFile();
               if (tinkerDir.exists()){
                  setPaths(tinkerDir.getAbsolutePath());
                  pathSet = true;
               }
            }
         }
      }

      // Otherwise, let the setPaths method assume we are in tinker/test
      if (!pathSet){
         setPaths(null);
      }

      Preferences prefs = Preferences.userNodeForPackage(ffe.panels.JobPanel.class);

      jcbDescript.setSelected(!prefs.getBoolean("JobPanel_description", true));
      jcbDescript.doClick();

      //System.out.println("java.ext.dirs: " + System.getProperty("java.ext.dirs","no extension directory"));
      //System.out.println("java.library.path: " + System.getProperty("java.library.path","no ld_library_path"));
      //System.out.println("java.classpath: " + System.getProperty("java.classpath","no classpath"));

   }

   public void setPaths(String tinkerDir){

      if (tinkerDir == null || !(new File(tinkerDir)).exists()){
         Preferences prefs = Preferences.userNodeForPackage(ffe.panels.JobPanel.class);
         path = prefs.get("JobPanel_path", fpanel.getCWD().getParentFile().getAbsolutePath());
         File f = new File(path);
         if (f.exists()){
            tinkerDir = f.getParent();
         } else{
            tinkerDir = ".";
         }
      } else{
         KeywordPanel keyPanel = fpanel.getKeywordPanel();
         if (keyPanel != null){
            keyPanel.setParamPath(tinkerDir);
         }
         path = tinkerDir + File.separator + "bin";
         pathjre = path + File.pathSeparator + tinkerDir + "\\jre\\bin\\client";
      }
      if (path.indexOf("PATH=") > 0){
         path = path.substring(5, path.length());
      }

      File tempDir = new File(tinkerDir);
      if (tinkerDir != null && tempDir.exists()){
         if (ffe.Main.MAC_OS_X){
            String libPath = new String(tinkerDir + "/jar/macosx");
            File tempNative = new File(libPath + "/libffe.jnilib");
            if (tempNative.exists()){
               System.setProperty("java.library.path", libPath);
            }
         } else if (ffe.Main.LINUX){
            String libPath = new String(tinkerDir + "/jar/linux");
            File tempNative = new File(libPath + "/libffe.so");
            if (tempNative.exists()){
               System.setProperty("java.library.path", libPath);
            }
         } else if (ffe.Main.WINDOWS){
            String libPath = new String(tinkerDir + "\\jar\\windows");
            File tempNative = new File(libPath + "\\ffe.dll");
            if (tempNative.exists()){
               System.setProperty("java.library.path", libPath);
            }
         }
      }

      if (!ffe.Main.WINDOWS){
         ld_library_path = System.getProperty("java.home", ".");
         ld_library_path = ld_library_path + "/lib/i386/client:" + ld_library_path + "/lib/i386:" +
                           ld_library_path + "/lib/i386/native_threads";
      } else{
         ld_library_path = ld_library_path + "\\bin\\client\\jvm.dll";
      }

      //System.out.println("ffe.jar: " + classpath);
      //System.out.println("TINKER Binaries: " + path);
      //System.out.println("LD_Library_Path: " + ld_library_path);
   }

   /**
    * Mouse events are used to trigger status bar updates
    * @param evt A mouse event
    */
   public void mouseClicked(MouseEvent evt){
      status.setText("  " + readCommand());
   }

   public void mouseEntered(MouseEvent evt){
      status.setText("  " + readCommand());
   }

   public void mouseExited(MouseEvent evt){
      status.setText("  " + readCommand());
   }

   public void mousePressed(MouseEvent evt){
      status.setText("  " + readCommand());
   }

   public void mouseReleased(MouseEvent evt){
      status.setText("  " + readCommand());
   }

   /**
    * Save misc. user preferences
    */
   public void savePrefs(){
      if (!lazyDone){
         return;
      }
      Preferences prefs = Preferences.userNodeForPackage(JobPanel.class);
      prefs.put("JobPanel_path", path);
      prefs.putBoolean("JobPanel_description", jcbDescript.isSelected());
   }

   public void selected(){
      refreshLogLabels();
   }

   /**
    * Set the selected command.
    * @param command The command that will be active
    * @return
    */
   private boolean setCommand(String command){
      JComboBox jcb = (JComboBox) toolbar.getComponentAtIndex(2);
      int count = jcb.getItemCount();

      String s;
      command = command.trim();
      for (int i = 0; i < count; i++){
         s = (String) jcb.getItemAt(i);
         if (s.equalsIgnoreCase(command)){
            currentCommand = command.toLowerCase();
            jcb.setSelectedIndex(i);
            loadCommandPanel();
            return true;
         }
      }
      return false;
   }

   /**
    * Prompt the user to toggle the existence of a TINKER END file.
    */
   private void setEnd(){
      FSystem m = fpanel.getTree().getActive();
      if (m == null){
         return;
      }
      File f = m.getFile();
      if (f == null){
         return;
      }

      File end;
      try{
         if (f.getName().indexOf(".") > 0){
            String name = f.getName().substring(0, f.getName().lastIndexOf("."));
            end = new File(f.getParent(), name.toString() + ".end");
         } else{
            end = new File(f.getParent(), f.getName() + ".end");
         }
         if (end.exists()){
            int i = JOptionPane.showConfirmDialog(this, "Delete " + end.getName() + "?",
                    "Delete TINKER End File", JOptionPane.YES_NO_OPTION);
            if (i == JOptionPane.YES_OPTION){
               end.delete();
            }
         } else{
            int i = JOptionPane.showConfirmDialog(this, "Create " + end.getName() + "?",
                    "Create TINKER End File", JOptionPane.YES_NO_OPTION);
            if (i == JOptionPane.YES_OPTION){
               end.createNewFile();
            }
         }
      } catch (IOException e){
//            System.out.println("State could not be changed: " + end.toString());
//            System.out.println("" + e);
      }
   }

   /**
    * Set the description divider location
    * @param b True to show the command description
    */
   private void setDivider(boolean b){
      jcbDescript.setSelected(b);
      if (b){
         int spDivider = (int) (this.getHeight() * (3.0f / 5.0f));
         jspCommand.setDividerLocation(spDivider);
      } else{
         jspCommand.setDividerLocation(1.0);
      }
   }

   public void stateChanged(ChangeEvent evt){
      Object source = evt.getSource();
      if (source instanceof JRadioButtonMenuItem){
         JRadioButtonMenuItem jrbmi = (JRadioButtonMenuItem) source;
         if (!jrbmi.isSelected()){
            return;
         }
         for (Enumeration e = conditionals.keys(); e.hasMoreElements(); ){
            Object controller = e.nextElement();
            if (controller instanceof ButtonGroup){
               ButtonGroup bg = (ButtonGroup) controller;
               JLabel label = (JLabel) conditionals.get(bg);
               JTextField jtf = (JTextField) label.getLabelFor();
               String cupon = jtf.getName().toLowerCase();
               String selection = jrbmi.getText().toLowerCase();
               if (cupon.indexOf(selection) >= 0){
                  label.setEnabled(true);
                  jtf.setEnabled(true);
               } else{
                  label.setEnabled(false);
                  jtf.setEnabled(false);
               }
               status.setText("  " + readCommand());
            }
         }
      }
   }

   /**
    * Launch the TINKER command specified by the JobPanel
    */
   private void executePanel(){
      FSystem s = fpanel.getTree().getActive();
      if (s != null){
         File f = s.getFile();
         String dir = null;
         if (f != null){
            dir = f.getParent();
         }
         launch(status.getText(), dir);
      }
   }

   /**
    * If a TINKER END file exists for the active system & command, remove it.
    */
   private void removeEnd(){
      FSystem m = fpanel.getTree().getActive();
      if (m == null){
         return;
      }
      File f = m.getFile();
      if (f == null){
         return;
      }
      File end = null;
      try{
         if (f.getName().indexOf(".") > 0){
            String name = f.getName().substring(0, f.getName().lastIndexOf("."));
            end = new File(f.getParent(), name.toString() + ".end");
         } else{
            end = new File(f.getParent(), f.getName() + ".end");
         }
      } catch (Exception e){
      }
      if (end != null && end.exists()){
         end.delete();
      }
   }

   /**
    * Prompt the user to select an alternate key file.
    */
   private void setKey(){
      JFileChooser d = MainPanel.getFileChooser();
      d.setFileFilter(new KeyFileFilter());
      d.setAcceptAllFileFilterUsed(false);

      FSystem sys = fpanel.getTree().getActive();
      if (sys != null){
         File cwd = sys.getFile();
         if (cwd != null && cwd.getParentFile() != null){
            d.setCurrentDirectory(cwd.getParentFile());
         }
      } else{
         return;
      }

      int result = d.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         File f = d.getSelectedFile();
         sys.setKeyFile(f);
         fpanel.getKeywordPanel().keyOpen(sys);
      }
   }

   /**
    * Prompt the user to select the path to TINKER executbles. This is necessary
    * because FFE run using Java Web Start does not inherit environment variables.
    */
   private void setPath(){
      JFileChooser d = MainPanel.getFileChooser();
      d.setDialogTitle("Select the TINKER Executable Directory");
      if (path != null){
         d.setSelectedFile(new File(path));
      } else{
         d.setCurrentDirectory(fpanel.getCWD());
      }
      d.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
      int result = d.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         String s = d.getSelectedFile().getParent();
         if (s != null){
            setPaths(s);
         }
      }
      d.setFileSelectionMode(JFileChooser.FILES_ONLY);
   }

   /**
    * Prompt the user to select an alternate log file
    */
   private void setLog(){
      JFileChooser d = MainPanel.getFileChooser();
      FSystem sys = fpanel.getTree().getActive();
      if (sys != null){
         File cwd = sys.getFile();
         if (cwd != null && cwd.getParentFile() != null){
            d.setCurrentDirectory(cwd.getParentFile());
         }
      } else{
         return;
      }
      d.setDialogTitle("Select a log file");
      d.setAcceptAllFileFilterUsed(true);
      int result = d.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         File f = d.getSelectedFile();
         if (f != null){
            sys.setLogFile(f);
            fpanel.setCWD(d.getCurrentDirectory());
            refreshLogLabels();
         }
      }
   }

   private void loadXMLCommands(){
      try{
         DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
         DocumentBuilder db = dbf.newDocumentBuilder();
         db.setEntityResolver(new DTDResolver());

         URL comURL = getClass().getClassLoader().getResource("ffe/xml/commands.xml");

         Document doc = db.parse(comURL.openStream());
         NodeList nodelist = doc.getChildNodes();
         Node commandroot = null;
         for (int i = 0; i < nodelist.getLength(); i++){
            commandroot = nodelist.item(i);
            if (commandroot.getNodeName().equals("TinkerCommands") &&
               commandroot instanceof Element){
               break;
            }
         }
         if (commandroot == null || !(commandroot instanceof Element)){
            commandList = null;
            return;
         }
         commandList = ((Element) commandroot).getElementsByTagName("Command");
      } catch (ParserConfigurationException e){
         System.out.println("Commands not read in: " + e);
      } catch (SAXException e){
         System.out.println("Commands not read in: " + e);
      } catch (IOException e){
         System.out.println("Commands not read in: " + e);
      }
   }

   private void loadCommandPanel(){
      JPanel modifierPanel;
      JTextArea descriptionTextArea;

      // TINKER Command
      NodeList commandFileTypes, commandActions;
      NodeList commandModifiers;
      Element command, commandDescription, commandModifier;

      // Modifying Arguements
      NodeList modifierValues;
      Element modifierName, modifierDescription;
      Element modifierGUIComponent, modifierDefault;

      // Modifying Arguements may be Conditional (not always supplied)
      Element conditional, conditionalUpon;
      Element conditionalDescription, conditionalGUIComponent;

      // Clear the previous components
      jtpJobParams.removeAll();
      conditionals.clear();

      if (currentCommandBox == null){
         return;
      }

      String currentCommand = (String) currentCommandBox.getSelectedItem();
      if (currentCommand == null){
         return;
      }

      int numcommands = commandList.getLength();
      command = null;
      for (int i = 0; i < numcommands; i++){
         command = (Element) commandList.item(i);
         Element commandname = (Element) command.getElementsByTagName("CommandName").item(0);
         String commandString = commandname.getFirstChild().getNodeValue();
         if (commandString.equalsIgnoreCase(currentCommand)){
            break;
         }
      }

      commandDescription = (Element) command.getElementsByTagName("CommandDescription").item(0);
      taDescription.setText(commandDescription.getFirstChild().getNodeValue());
      taDescription.setSize(taDescription.getPreferredSize());
      commandActions = command.getElementsByTagName("CommandAction");
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < commandActions.getLength(); i++){
         Node actionNode = commandActions.item(i);
         String newAction = actionNode.getFirstChild().getNodeValue().toUpperCase();
         sb.append(newAction + " ");
      }
      action = sb.toString().trim();
      commandFileTypes = command.getElementsByTagName("CommandFileType");
      fileTypes = 0;
      for (int i = 0; i < commandFileTypes.getLength(); i++){
         Node filetype = commandFileTypes.item(i);
         String temp = filetype.getFirstChild().getNodeValue().toUpperCase();
         if (temp.indexOf("XYZ") >= 0){
            fileTypes = 1;
         }
         if (temp.indexOf("INT") >= 0){
            fileTypes = (byte) (fileTypes + 2);
         }
         if (temp.indexOf("ARC") >= 0){
            fileTypes = (byte) (fileTypes + 4);
         }
         if (temp.indexOf("PDB") >= 0){
            fileTypes = (byte) (fileTypes + 8);
         }
      }

      commandModifiers = command.getElementsByTagName("CommandModifier");
      int length = commandModifiers.getLength();
      for (int i = 0; i < length; i++){
         //Modifier
         boolean addModifier = true;
         commandModifier = (Element) commandModifiers.item(i);
         //Modifier Elements
         modifierName = (Element) commandModifier.getElementsByTagName("ModifierName").item(0);
         modifierDescription = (Element) commandModifier.getElementsByTagName("ModifierDescription").
                               item(0);
         modifierGUIComponent = (Element) commandModifier.getElementsByTagName(
                                "ModifierGUIComponent").item(0);
         modifierDefault = (Element) commandModifier.getElementsByTagName("ModifierDefault").item(0);
         conditional = (Element) commandModifier.getElementsByTagName("ModifierConditional").item(0);
         //Description
         String commanddescript = new String("");
         if (modifierDescription != null){
            Node n = modifierDescription.getFirstChild();
            if (n != null){
               commanddescript = n.getNodeValue();
            }
         }

         descriptionTextArea = new JTextArea(" " + commanddescript.trim());
         descriptionTextArea.setEditable(false);
         descriptionTextArea.setBorder(eb);
         descriptionTextArea.setPreferredSize(descriptionTextArea.getPreferredSize());
         // Default Value
         String sdefault = null;
         if (modifierDefault != null){
            sdefault = modifierDefault.getFirstChild().getNodeValue();
         }
         // Modifier Panel
         modifierPanel = new JPanel(new BorderLayout());
         modifierPanel.add(descriptionTextArea, BorderLayout.NORTH);
         // Switch Based on Swing Object to Use
         String sswingobj = modifierGUIComponent.getFirstChild().getNodeValue();
         JPanel temppanel = new JPanel(new FlowLayout());
         temppanel.setBorder(eb);
         ButtonGroup bg = null;

         // Use a check box for each value
         if (sswingobj.equalsIgnoreCase("CHECKBOXES")){
            modifierValues = commandModifier.getElementsByTagName("ModifierValue");
            int vlength = modifierValues.getLength();
            for (int j = 0; j < vlength; j++){
               JCheckBox jcb = new JCheckBox(modifierValues.item(j).getFirstChild().getNodeValue());
               jcb.addMouseListener(this);
               if (sdefault != null && jcb.getActionCommand().equalsIgnoreCase(sdefault)){
                  jcb.setSelected(true);
               }
               temppanel.add(jcb);
            }
         } else if (sswingobj.equalsIgnoreCase("SYSTEMS")){
            JComboBox jcb = new JComboBox(fpanel.getTree().getSystems());
            jcb.setSize(jcb.getMaximumSize());
            jcb.addActionListener(this);
            temppanel.add(jcb);
         } else if (sswingobj.equalsIgnoreCase("TEXTFIELD")){
            JTextField jtf = new JTextField(10);
            jtf.addMouseListener(this);
            if (sdefault != null && sdefault.equals("ATOMS")){
               FSystem sys = fpanel.getTree().getActive();
               if (sys != null){
                  jtf.setText("" + sys.getAtomList().size());
               }
            } else if (sdefault != null){
               jtf.setText(sdefault);
            }
            temppanel.add(jtf);
         } else if (sswingobj.equalsIgnoreCase("RADIOBUTTONS")){
            bg = new ButtonGroup();
            modifierValues = commandModifier.getElementsByTagName("ModifierValue");
            int vlength = modifierValues.getLength();
            for (int j = 0; j < vlength; j++){
               String rString = modifierValues.item(j).getFirstChild().getNodeValue();
               JRadioButtonMenuItem jrbmi = new JRadioButtonMenuItem(rString);
               jrbmi.addMouseListener(this);
               bg.add(jrbmi);
               if (sdefault != null && jrbmi.getActionCommand().equalsIgnoreCase(sdefault)){
                  jrbmi.setSelected(true);
               }
               temppanel.add(jrbmi);
            }
         }

         // Check for Conditional Modifiers
         if (conditional != null){
            addModifier = false;
            conditionalDescription = (Element) commandModifier.getElementsByTagName(
                                     "ConditionalDescription").item(0);
            conditionalGUIComponent = (Element) commandModifier.getElementsByTagName(
                                      "ConditionalGUIComponent").item(0);
            conditionalUpon = (Element) commandModifier.getElementsByTagName("ConditionalUpon").
                              item(0);
            if (conditionalGUIComponent != null && conditionalDescription != null){
               addModifier = true;
               for (int j = 0; j < temppanel.getComponentCount(); j++){
                  JRadioButtonMenuItem jrbmi = (JRadioButtonMenuItem) temppanel.getComponent(j);
                  jrbmi.addChangeListener(this);
               }
               JPanel condpanel = new JPanel();
               condpanel.setBorder(eb);
               JLabel condlabel = new JLabel((String) conditionalDescription.getFirstChild().
                                  getNodeValue());
               condlabel.setEnabled(false);
               JTextField condtext = new JTextField(10);
               condlabel.setLabelFor(condtext);
               condtext.setName(conditionalUpon.getFirstChild().getNodeValue());
               condtext.setEnabled(false);
               condpanel.add(condlabel);
               condpanel.add(condtext);
               conditionals.put(bg, condlabel);
               modifierPanel.add(condpanel, BorderLayout.SOUTH);
            } else if (currentSystem != null && currentSystem.getFileType() == SystemFilter.PDB){
               String cond_req = conditionalUpon.getFirstChild().getNodeValue();
               if (cond_req.toLowerCase().startsWith("system") && cond_req.split(" +").length > 1){
                  cond_req = cond_req.split(" +")[1];
                  String[] condModifiers = null;
                  if (cond_req.equalsIgnoreCase("AltLoc")){
                     condModifiers = currentSystem.getAltLocations();
                     if (condModifiers != null && condModifiers.length > 1){
                        addModifier = true;
                        bg = new ButtonGroup();
                        for (int j = 0; j < condModifiers.length; j++){
                           JRadioButtonMenuItem jrbmi = new JRadioButtonMenuItem(condModifiers[j]);
                           jrbmi.addMouseListener(this);
                           bg.add(jrbmi);
                           temppanel.add(jrbmi);
                           if (j == 0){
                              jrbmi.setSelected(true);
                           }
                        }
                     }
                  } else if (cond_req.equalsIgnoreCase("Chains")){
                     condModifiers = currentSystem.getChainNames();
                     if (condModifiers != null && condModifiers.length > 0){
                        addModifier = true;
                        for (int j = 0; j < condModifiers.length; j++){
                           JRadioButtonMenuItem jrbmi = new JRadioButtonMenuItem(condModifiers[j]);
                           jrbmi.addMouseListener(this);
                           bg.add(jrbmi);
                           temppanel.add(jrbmi, j);
                        }
                     }
                  }
               }
            }
         }
         if (addModifier){
            modifierPanel.add(temppanel, BorderLayout.CENTER);
            modifierPanel.setPreferredSize(modifierPanel.getPreferredSize());
            jtpJobParams.addTab(modifierName.getFirstChild().getNodeValue(), modifierPanel);
         }
      }
      jtpJobParams.setPreferredSize(jtpJobParams.getPreferredSize());
      status.setText("  " + readCommand());
   }

   private void createCommandCB(){
      loadXMLCommands();

      xyzCommands = new JComboBox();
      intCommands = new JComboBox();
      arcCommands = new JComboBox();
      pdbCommands = new JComboBox();
      xyzCommands.setActionCommand("TinkerCommand");
      intCommands.setActionCommand("TinkerCommand");
      arcCommands.setActionCommand("TinkerCommand");
      pdbCommands.setActionCommand("TinkerCommand");

      xyzCommands.setMaximumSize(xyzCommands.getPreferredSize());
      xyzCommands.setEditable(false);
      xyzCommands.setToolTipText("Select a Modeling Command");
      intCommands.setMaximumSize(intCommands.getPreferredSize());
      intCommands.setEditable(false);
      intCommands.setToolTipText("Select a Modeling Command");
      arcCommands.setMaximumSize(arcCommands.getPreferredSize());
      arcCommands.setEditable(false);
      arcCommands.setToolTipText("Select a Modeling Command");
      pdbCommands.setMaximumSize(pdbCommands.getPreferredSize());
      pdbCommands.setEditable(false);
      pdbCommands.setToolTipText("Select a Modeling Command");

      Element command;
      NodeList filetypes;
      Element commandname;
      String commandString;

      int numcommands = commandList.getLength();
      for (int i = 0; i < numcommands; i++){
         command = (Element) commandList.item(i);
         commandname = (Element) command.getElementsByTagName("CommandName").item(0);
         commandString = commandname.getFirstChild().getNodeValue();

         filetypes = command.getElementsByTagName("CommandFileType");
         for (int j = 0; j < filetypes.getLength(); j++){
            Node filetype = filetypes.item(j);
            String temp = filetype.getFirstChild().getNodeValue().toUpperCase();
            if (temp.indexOf("XYZ") >= 0){
               xyzCommands.addItem(commandString);
            }
            if (temp.indexOf("INT") >= 0){
               intCommands.addItem(commandString);
            }
            if (temp.indexOf("ARC") >= 0){
               arcCommands.addItem(commandString);
            }
            if (temp.indexOf("PDB") >= 0){
               pdbCommands.addItem(commandString);
            }
         }

      }
      xyzCommands.setSelectedIndex(0);
      intCommands.setSelectedIndex(0);
      arcCommands.setSelectedIndex(0);
      pdbCommands.setSelectedIndex(0);
      xyzCommands.addActionListener(this);
      intCommands.addActionListener(this);
      arcCommands.addActionListener(this);
      pdbCommands.addActionListener(this);

      currentCommand = null;
      return;
   }

   private void initToolBar(){
      // Toolbar
      toolbar = new JToolBar("Modeling Commands", JToolBar.HORIZONTAL);
      toolbar.setLayout(new FlowLayout(FlowLayout.LEFT));

      JButton jblaunch = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                         "ffe/icons/launch.png")));
      jblaunch.setActionCommand("Launch");
      jblaunch.setToolTipText("Launch the TINKER Command");
      jblaunch.addActionListener(this);
      Insets insets = jblaunch.getInsets();
      insets.top = 2;
      insets.bottom = 2;
      insets.left = 2;
      insets.right = 2;
      jblaunch.setMargin(insets);
      toolbar.add(jblaunch);

      JButton jbend = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                      "ffe/icons/cancel.png")));
      jbend.setActionCommand("End");
      jbend.setToolTipText("Toggle the Existence of a TINKER *.END File");
      jbend.addActionListener(this);
      jbend.setMargin(insets);
      toolbar.add(jbend);

      toolbar.addSeparator();

      JButton jmiPath = new JButton("Set Path...");
      jmiPath.setToolTipText("Set the PATH to the TINKER Executables");
      jmiPath.addActionListener(this);
      jmiPath.setMargin(insets);
      toolbar.add(jmiPath);
      JButton jmiKey = new JButton("Set Key File...");
      jmiKey.setToolTipText("Select an Alternate *.Key File");
      jmiKey.addActionListener(this);
      jmiKey.setMargin(insets);
      toolbar.add(jmiKey);
      JButton jmiLog = new JButton("Set Log File...");
      jmiLog.addActionListener(this);
      jmiLog.setToolTipText("Set the Name for an Output Log");
      jmiLog.setMargin(insets);
      toolbar.add(jmiLog);
      toolbar.add(logMode);
      toolbar.addSeparator();

      ImageIcon icinfo = new ImageIcon(getClass().getClassLoader().getResource("ffe/icons/info.png"));
      jcbDescript = new JCheckBoxMenuItem(icinfo);
      jcbDescript.addActionListener(this);
      jcbDescript.setActionCommand("Description");
      jcbDescript.setToolTipText("Show/Hide Modeling Command Descriptions");
      jcbDescript.setMargin(insets);
      toolbar.add(jcbDescript);
      toolbar.add(new JLabel(""));

      toolbar.setBorderPainted(false);
      toolbar.setFloatable(false);
      toolbar.setRollover(true);
   }

   private void initialize(){
      EtchedBorder eb = new EtchedBorder(EtchedBorder.RAISED);

      // Command Description
      taDescription = new JTextArea();
      taDescription.setEditable(false);
      JScrollPane tap = new JScrollPane(taDescription);
      tap.setBorder(new TitledBorder(eb, " Description "));

      // Command Parameters
      jtpJobParams = new JTabbedPane();
      createCommandCB();
      currentCommandBox = null;
      commandPanel = new JPanel(new BorderLayout());
      status = new JLabel();
      status.setBorder(eb);
      status.setToolTipText("  Command to be executed");
      commandPanel.add(jtpJobParams, BorderLayout.CENTER);
      jspCommand = new JSplitPane(JSplitPane.VERTICAL_SPLIT, commandPanel, tap);
      jspCommand.setContinuousLayout(true);
      jspCommand.setResizeWeight(1.0d);
      jspCommand.setOneTouchExpandable(true);
      setLayout(new BorderLayout());

      add(jspCommand, BorderLayout.CENTER);
      add(status, BorderLayout.SOUTH);

      // Log File Control
      logMode.addItem(append);
      logMode.addItem(overwrite);
      logMode.addItem(createNew);
      logMode.addActionListener(this);
      logMode.setSelectedItem(append);

      return;
   }

   public void initLazy(){
      if (lazyDone == true){
         return;
      }

      initToolBar();
      add(toolbar, BorderLayout.NORTH);
      loadCommandPanel();
      loadPrefs();
      validate();
      lazyDone = true;
   }

   private void refreshLogLabels(){
      String selected = (String) logMode.getSelectedItem();

      logMode.removeActionListener(this);
      logMode.removeAllItems();

      // If there is no current system, list choices anyway so the box isn't empty
      if (currentSystem == null){
         append = new String("Append to the Existing Log File");
         overwrite = new String("Overwrite the Existing Log File");
         createNew = new String("Create a New Log File");
         logMode.addItem(append);
         logMode.addItem(overwrite);
         logMode.addItem(createNew);
         logMode.setSelectedItem(selected);
         logString = getLogString();
         logMode.addActionListener(this);
         return;
      }

      File tempLog = null;
      File newLog = null;

      File currentLog = currentSystem.getLogFile();
      String fileName = currentLog.getName();
      File logDir = currentLog.getParentFile();
      File systemDir = currentSystem.getFile().getParentFile();

      if (logDir == null || !logDir.equals(systemDir)){
         tempLog = new File(systemDir.getAbsolutePath() + File.separator + fileName);
         currentSystem.setLogFile(tempLog);
      } else{
         tempLog = currentLog;
      }

      // Simple Case - default log file doesn't exist yet
      if (!tempLog.exists()){
         createNew = new String("Create " + fileName);
         currentSystem.setLogFile(tempLog);
         logMode.addItem(createNew);
         logMode.setSelectedItem(createNew);
         logString = getLogString();
         logMode.addActionListener(this);
         return;
      }

      // The default log exists, so we can appned to it, overwrite it, or create a new one
      append = new String("Append to " + fileName);
      overwrite = new String("Overwrite " + fileName);
      newLog = SystemFilter.version(tempLog);
      logMode.addItem(append);
      logMode.addItem(overwrite);
      if (!newLog.equals(tempLog)){
         createNew = new String("Create " + newLog.getName());
         logMode.addItem(createNew);
      } else{
         createNew = null;
      }

      if (selected.startsWith("Append")){
         logMode.setSelectedItem(append);
      } else if (selected.startsWith("Overwrite")){
         logMode.setSelectedItem(overwrite);
      } else{
         if (createNew != null){
            logMode.setSelectedItem(createNew);
         } else{
            logMode.setSelectedItem(append);
         }
      }
      logString = getLogString();
      logMode.addActionListener(this);
   }

   private String getLogString(){
      if (currentSystem == null){
         return null;
      }

      File currentLog = currentSystem.getLogFile();
      String currentMode = (String) logMode.getSelectedItem();
      if (currentMode.startsWith("Create")){
         currentLog = SystemFilter.version(currentLog);
         return new String(" >> \"" + currentLog.getAbsolutePath() + "\"");
      } else if (currentMode.startsWith("Append")){
         return new String(" >> \"" + currentLog.getAbsolutePath() + "\"");
      } else{
         return new String(" > \"" + currentLog.getAbsolutePath() + "\"");
      }
   }

   private String readCommand(){

      if (currentSystem == null){
         return new String("");
      }
      if ((currentSystem.getFileType() & fileTypes) == 0){
         return new String("");
      }
      File f = currentSystem.getFile();
      if (f == null){
         return new String("");
      }

      StringBuffer commandlineparams = new StringBuffer(currentCommand.toLowerCase() + " \"" +
                                       f.getAbsolutePath() + "\" ");
      int numcomponents, numparams = jtpJobParams.getTabCount();
      JPanel modifierpanel, temppanel;
      Component param;
      for (int i = 0; i < numparams; i++){
         modifierpanel = (JPanel) jtpJobParams.getComponentAt(i);
         int modcount = modifierpanel.getComponentCount();
         temppanel = (JPanel) modifierpanel.getComponent(modcount - 1);
         numcomponents = temppanel.getComponentCount();
         for (int j = 0; j < numcomponents; j++){
            param = temppanel.getComponent(j);
            if (param instanceof JCheckBox){
               JCheckBox jcbox = (JCheckBox) param;
               if (jcbox.isSelected()){
                  commandlineparams.append(jcbox.getText());
               }
            } else if (param instanceof JTextField){
               JTextField jtfield = (JTextField) param;
               commandlineparams.append(" " + jtfield.getText() + " ");
            } else if (param instanceof JComboBox){
               JComboBox jcb = (JComboBox) param;
               Object object = jcb.getSelectedItem();
               if (object instanceof FSystem){
                  FSystem system = (FSystem) object;
                  if (system != null){
                     File file = system.getFile();
                     if (file != null){
                        commandlineparams.append(" " + file.getAbsolutePath() + " ");
                     }
                  }
               }
            } else if (param instanceof JRadioButtonMenuItem){
               JRadioButtonMenuItem jrbutton = (JRadioButtonMenuItem) param;
               if (jrbutton.isSelected()){
                  if (!jrbutton.getText().equalsIgnoreCase("none")){
                     commandlineparams.append(" " + jrbutton.getText() + " ");
                  }
               }
            }
         }
         if (modifierpanel.getComponentCount() == 3){
            temppanel = (JPanel) modifierpanel.getComponent(1);
            JTextField jtf = (JTextField) temppanel.getComponent(1);
            if (jtf.isEnabled()){
               commandlineparams.append(" " + jtf.getText() + " 0 ");
            }
         }
      }
      if (currentSystem.getKeyFile() != null){
         commandlineparams.append(" -k \"" + currentSystem.getKeyFile().getAbsolutePath() + "\"");
      }
      commandlineparams.append(logString);
      return commandlineparams.toString();
   }
}
