/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.util.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

import ffe.lang.Keyword;

/**
 * The KeywordComponent class is used to represent one TINKER keyword.
 */
public final class KeywordComponent
   implements MouseListener, ActionListener, ChangeListener, DocumentListener{

   /**
    * Stores the TINKER Keyword String.
    */
   private String key;

   /**
    * Specifies which Group of Related TINKER Keywords this Keyword belongs to.
    */
   private String keygroup;

   /**
    * An ArrayList of Components used to represent the Keyword Modifiers
    */
   private ArrayList values;
   private String options[];

   /**
    * The number of Swing Components used in representing this Keyword
    */
   private int count;

   /**
    * The type of Swing Conponent used in representing this Keyword
    */
   private int type;

   private String description;
   private JTextArea output;
   private boolean active;
   private boolean init = false;

   /**
    * This is used to test if ANY keyword has been modified, so it is static
    * across all keyword objects
    */
   private static boolean modified = false;
   private static String spaces = new String("                             ");

   /**
    * The Default Constructor
    * k - Keyword String
    * kg - Keyword Group
    * t - Type of GUI Components to represent Keyword modifiers
    * d - Keyword description
    */
   public KeywordComponent(String k, String kg, int t, String d, JTextArea jta){
      key = k;
      keygroup = kg;
      values = new ArrayList();
      type = t;
      description = d;
      output = jta;
      active = false;
   }

   public KeywordComponent(String k, String kg, int t, String d, JTextArea jta, String o[]){
      this(k, kg, t, d, jta);
      options = o;
   }

   /**
    * Overidden equals method return true if object equals this, or if it of the same
    * class and has the same Tinker Keyword.
    */
   public boolean equals(Object object){
      if (this == object){
         return true;
      } else if (object == null || getClass() != object.getClass()){
         return false;
      }

      KeywordComponent other = (KeywordComponent) object;
      return key.equalsIgnoreCase(other.key);
   }

   public void clear(){
      active = false;
      if (init == false){
         return;
      }
      if (type == CHECKBOX){
         ((JCheckBox) values.get(0)).setSelected(false);
      } else if (type == RADIOBUTTONS){
         int i = values.size();
         ((JRadioButtonMenuItem) values.get(i - 1)).setSelected(true);
      } else if (type == CHECKBOXES){
         for (ListIterator li = values.listIterator(); li.hasNext(); ){
            ((JCheckBox) li.next()).setSelected(false);
         }
      } else if (type == COMBOBOX){
         ((JComboBox) values.get(1)).removeAllItems();
      } else{
         ((JTextField) values.get(1)).setText("");
      }
   }

   public static boolean isModified(){
      return modified;
   }

   public static void setModified(boolean b){
      modified = b;
   }

   public boolean isActive(){
      return active;
   }

   public String getKeyGroup(){
      return keygroup;
   }

   public String getKey(){
      return key;
   }

   public String getDescription(){
      return description;
   }

   public void keyEntry(String s){
      Component c;
      active = true;

      if (!init){
         initSwing();

      }
      for (ListIterator li = values.listIterator(); li.hasNext(); ){
         c = (Component) li.next();
         if (c instanceof JCheckBox){
            JCheckBox cb = (JCheckBox) c;
            if (s != null && s.equalsIgnoreCase("FALSE")){
               cb.setSelected(false);
            } else{
               cb.setSelected(true);
            }
            break;
         } else if (c instanceof JRadioButtonMenuItem){
            JRadioButtonMenuItem rbmi = (JRadioButtonMenuItem) c;
            if (s != null && rbmi.getText().equalsIgnoreCase(s)){
               rbmi.setSelected(true);
               break;
            }
         } else if (c instanceof JTextField){
            JTextField tf = (JTextField) c;
            if (s != null){
               tf.setText(s);
            } else{
               tf.setText("");
            }
            break;
         } else if (c instanceof JComboBox){
            if (s != null){
               JComboBox cb = (JComboBox) c;
               cb.addItem(s);
               break;
            }
         }
      }
      if (toString() == null){
         active = false;
      }
   }

   private void initSwing(){
      if (type == CHECKBOX){
         checkBoxInit();
      } else if (type == RADIOBUTTONS){
         radioButtonInit();
      } else if (type == CHECKBOXES){
         checkBoxesInit();
      } else if (type == COMBOBOX){
         comboInit();
      } else{ // Else create Text fields equal to t
         labelInit();
      }
      init = true;
   }

   private void checkBoxInit(){
      JCheckBox cb = new JCheckBox(key, false);
      cb.setMaximumSize(cb.getPreferredSize());
      cb.addMouseListener(this);
      cb.addChangeListener(this);
      values.add(cb);
      count = 1;
   }

   private void checkBoxInit(String label){
      JCheckBox cb = new JCheckBox(label, false);
      cb.setMaximumSize(cb.getPreferredSize());
      cb.addMouseListener(this);
      cb.addChangeListener(this);
      values.add(cb);
   }

   private void comboInit(){
      JLabel jl = new JLabel(key);
      values.add(jl);
      jl.addMouseListener(this);
      jl.setMaximumSize(jl.getPreferredSize());
      JComboBox cb = new JComboBox();
      cb.setEditable(true);
      cb.setFont(output.getFont());
      cb.addMouseListener(this);
      cb.addActionListener(this);
      cb.setMaximumSize(jl.getPreferredSize());
      cb.getEditor().addActionListener(this);
      values.add(cb);
      count = 2;
   }

   private void radioButtonInit(){
      JLabel jl = new JLabel(key);
      jl.setMaximumSize(jl.getPreferredSize());
      jl.addMouseListener(this);
      values.add(jl);
      ButtonGroup buttongroup = new ButtonGroup();
      for (int i = 0; i < options.length - 1; i++){
         initRadioButton(options[i], buttongroup, jl, false);
      }
      initRadioButton(options[options.length - 1], buttongroup, jl, true);
      count = options.length + 1;
   }

   private void checkBoxesInit(){
      int option = 0;
      if (options[0].equals("PRESENT")){
         checkBoxInit();
         option = 1;
      } else{

      }
      while (option < options.length){
         checkBoxInit(options[option++]);
      }
      count = options.length;
   }

   private void initRadioButton(String label, ButtonGroup bg, JLabel jl, boolean b){
      JRadioButtonMenuItem button = new JRadioButtonMenuItem(label);
      button.addMouseListener(this);
      button.addChangeListener(this);
      button.setMaximumSize(jl.getPreferredSize());
      bg.add(button);
      values.add(button);
      button.setSelected(b);
   }

   private void labelInit(){
      JLabel jl = new JLabel(key);
      jl.addMouseListener(this);
      jl.setMaximumSize(jl.getPreferredSize());
      values.add(jl);
      JTextField tf;
      type = TEXTFIELD;
      for (int i = 0; i < type; i++){
         tf = new JTextField("");
         tf.setColumns(11);
         tf.setActionCommand(key);
         tf.setMaximumSize(jl.getPreferredSize());
         tf.addMouseListener(this);
         tf.getDocument().addDocumentListener(this);
         values.add(tf);
      }
      count = type + 1;
   }

   public void loadData(Keyword keywordData){
      for (ListIterator li = values.listIterator(); li.hasNext(); ){
         Component c = (Component) li.next();
         if (c instanceof JRadioButtonMenuItem){
            JRadioButtonMenuItem rbmi = (JRadioButtonMenuItem) c;
            String text = rbmi.getText();
            if (rbmi.isSelected() && !text.equalsIgnoreCase("ABSENT") &&
               !text.equalsIgnoreCase("PRESENT")){
               keywordData.append(rbmi.getText());
            }
         } else if (c instanceof JTextField){
            JTextField tf = (JTextField) c;
            if (!tf.getText().equals("")){
               keywordData.append(tf.getText());
            }
         } else if (c instanceof JComboBox){
            JComboBox cb = (JComboBox) c;
            int num = cb.getItemCount();
            for (int i = 0; i < num; i++){
               String s = (String) cb.getItemAt(i);
               keywordData.append(s);
            }
         } else if (c instanceof JCheckBox){
            JCheckBox cb = (JCheckBox) c;
            String text = cb.getText();
            if (cb.isSelected() && !text.equalsIgnoreCase(key) && !text.equalsIgnoreCase("PRESENT")){
               keywordData.append(text);
            }
         }
      }
   }

   /**
    * Overridden toString methods facilitates Keyword output to a file.
    */
   public String toString(){
      if (!active || !init){
         return null;
      }

      StringBuffer s;
      if (!key.equalsIgnoreCase("TORSION")){
         s = new StringBuffer(key + spaces.substring(0, 18 - key.length()));
      } else{
         s = new StringBuffer(key);
      }

      Component c;
      for (ListIterator li = values.listIterator(); li.hasNext(); ){
         c = (Component) li.next();
         if (c instanceof JCheckBox){
            JCheckBox cb = (JCheckBox) c;
            if (cb.getText().equals(key)){
               if (!cb.isSelected()){
                  return null;
               } else{
                  return s.toString();
               }
            } else{
               if (cb.isSelected()){
                  if (s.toString().equals(key)){
                     if (cb.getText().equalsIgnoreCase("PRESENT")){
                        continue;
                     } else{
                        s.append(" " + cb.getText());
                     }
                  } else{
                     if (cb.getText().equalsIgnoreCase("PRESENT")){
                        s.append("\n" + key);
                     } else{
                        s.append("\n" + key + " " + cb.getText());
                     }
                  }
               } else{
                  continue;
               }
            }
         } else if (c instanceof JRadioButtonMenuItem){
            JRadioButtonMenuItem rbmi = (JRadioButtonMenuItem) c;
            if (rbmi.isSelected()){
               if (rbmi.getText().equalsIgnoreCase("PRESENT")){
                  return s.toString();
               } else if (rbmi.getText().equalsIgnoreCase("ABSENT")){
                  return null;
               } else{
                  return s.append(" " + rbmi.getText()).toString();
               }
            }
         } else if (c instanceof JTextField){
            JTextField tf = (JTextField) c;
            if (tf.getText().equals("")){
               return null;
            } else{
               String v = tf.getText();
               if (v.length() < 8){
                  s.append(v + spaces.substring(0, 8 - v.length()));
               } else{
                  s.append(v);
               }
            }
         } else if (c instanceof JComboBox){
            JComboBox cb = (JComboBox) c;
            int count = cb.getItemCount();
            if (count == 0){
               return null;
            }
            String entries[] = new String[count];
            for (int i = 0; i < count; i++){
               entries[i] = (String) cb.getItemAt(i);
            }
            java.util.Arrays.sort(entries);

            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < count; i++){
               sb.append(key + spaces.substring(0, 18 - key.length()));
               sb.append(entries[i].toUpperCase());
               if (i < count - 1){
                  sb.append("\n");
               }
            }
            return sb.toString();
         }
      }
      return s.toString();
   }

   /**
    * Returns a JPanel with a GridLayout LayoutManager that contains a Swing
    * representation of the Keyword and Modifiers in a single row.
    */
   public JPanel getPanel(){

      if (!init){
         initSwing();
      }
      GridBagLayout gl = new GridBagLayout();
      GridBagConstraints c = new GridBagConstraints();
      c.gridx = c.gridy = 0;
      c.fill = c.HORIZONTAL;
      c.anchor = c.WEST;
      c.gridwidth = c.gridheight = 1;
      c.weightx = 1.0;
      c.insets = new Insets(2, 5, 0, 5);
      JPanel p = new JPanel(gl);
      Component comp = (Component) values.get(0);
      gl.setConstraints(comp, c);
      p.add(comp);
      if (type != MULTIPOLE){
         if (count == 1){
            return p;
         }
         c.anchor = c.EAST;
         c.weightx = 0.0;
         for (int i = 1; i < count; i++){
            c.gridx = i;
            comp = (Component) values.get(i);
            gl.setConstraints(comp, c);
            p.add(comp);
         }
         return p;
      } else{
         return p;
      }
   }

   /*
    public static void fillRow(JPanel p, GridBagLayout g, GridBagConstraints c) {
       int grid = c.gridwidth;
       JLabel jfill = new JLabel("");
       c.gridwidth = GridBagConstraints.REMAINDER;
       g.setConstraints(jfill, c);
       p.add(jfill);
       c.gridwidth = grid;
        } */

   public static void fillPanel(JPanel p, GridBagLayout g, GridBagConstraints c){
      JLabel jfill = new JLabel(" ");
      c.weightx = 1;
      c.weighty = 1;
      c.fill = GridBagConstraints.BOTH;
      c.gridwidth = GridBagConstraints.REMAINDER;
      c.gridheight = GridBagConstraints.REMAINDER;
      g.setConstraints(jfill, c);
      p.add(jfill);
   }

   public void actionPerformed(ActionEvent evt){
      modified = true;
      if ((evt.getSource() instanceof JTextField) && (values.get(1) instanceof JComboBox)){
         JComboBox cb = (JComboBox) values.get(1);
         JTextField tf = (JTextField) evt.getSource();
         String data = (String) tf.getText();
         if (data.equals("")){ //Remove Selected
            int index = cb.getSelectedIndex();
            int count = cb.getItemCount();
            if (index == -1){
               cb.grabFocus();
               return; //No Items in ComboBox
            }
            if (index >= count){
               cb.grabFocus();
               return; //Trying to remove an Item that isn't there!!
            } else{
               if (index == 0){
                  cb.removeItemAt(0);
               } else{
                  cb.setSelectedIndex(0);
                  cb.removeItemAt(index);
               }
            }
         } else{ //Add Selected
            cb.addItem(data);
            cb.setSelectedIndex(cb.getItemCount() - 1);
         }
      }

   }

   public void stateChanged(ChangeEvent evt){
      modified = true;
   }

   public void changedUpdate(DocumentEvent evt){
      modified = true;
   }

   public void removeUpdate(DocumentEvent evt){
      modified = true;
   }

   public void insertUpdate(DocumentEvent evt){
      modified = true;
   }

   private Rectangle rect = new Rectangle(0, 0, 1, 1);
   public void mouseClicked(MouseEvent evt){
      active = true;
      if (this.toString() == null){
         active = false;
      }
      output.setText(key + ": " + description);
      JViewport jsp = (JViewport) output.getParent();
      jsp.setViewPosition(new Point(20, 20));
   }

   public void mousePressed(MouseEvent evt){}

   public void mouseReleased(MouseEvent evt){}

   public void mouseEntered(MouseEvent evt){
      active = true;
      if (this.toString() == null){
         active = false;
      }
      output.setText(key + ": " + description);
      JViewport jsp = (JViewport) output.getParent();
      jsp.setViewPosition(new Point(20, 20));
   }

   public void mouseExited(MouseEvent evt){
      active = true;
      if (this.toString() == null){
         active = false;
      }
      output.setText(key + ": " + description);
      JViewport jsp = (JViewport) output.getParent();
      jsp.setViewPosition(new Point(20, 20));
   }

   public static final int TEXTFIELD = 1;
   public static final int CHECKBOX = 0;
   public static final int RADIOBUTTONS = -1;
   public static final int CHECKBOXES = -2;
   public static final int COMBOBOX = -3;
   public static final int MULTIPOLE = 4;
}
