/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.prefs.Preferences;
import javax.xml.parsers.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

import org.w3c.dom.*;
import org.xml.sax.SAXException;

import ffe.lang.*;
import ffe.parsers.KeyFileFilter;
import ffe.xml.DTDResolver;

/**
 * The KeywordPanel class is an editor for TINKER Keyword (*.KEY) files.
 */
public final class KeywordPanel
   extends JPanel
   implements ActionListener{

   // Keywords and Keyword Groups
   private LinkedHashMap keywordHash;
   private LinkedHashMap keywordGroups;
   private JComboBox jcbGroups;

   // Currently open TINKER key file
   private File currentKeyFile;
   private FSystem currentSystem;
   private boolean fileOpen;

   // Swing Components
   private MainPanel mainPanel;
   private JPanel keywordPanel;
   private JToolBar toolbar;
   private JSplitPane splitPane;
   private JScrollPane spKeywords;
   private JTextArea taComments;
   private JCheckBoxMenuItem jcbDescript;
   private JScrollPane spDescript;
   private JTextArea taDescription;
   private JLabel status = new JLabel("  ");

   // Location of TINKER parameter files
   private File paramDir = null;
   private File paramFiles[] = null;
   private Hashtable paramHash = null;

   //private JComboBox paramCombo = null;
   private String paramNames[] = null;

   /**
    * Default Construtor where parent is the tinker Window Frame object.
    */
   public KeywordPanel(MainPanel f){
      super();
      mainPanel = f;
      initialize();
   }

   public String[] getParamFiles(){
      return paramNames;
   }

   public String getParamPath(String key){
      return (String) paramHash.get(key);
   }

   public boolean isFileOpen(){
      return fileOpen;
   }

   public boolean saveChanges(){
      if (KeywordComponent.isModified()){
         if (currentKeyFile != null){
            keySave(currentKeyFile);
         }
      }
      return false;
   }

   public String getKeywordValue(String key){
      KeywordComponent keyword = (KeywordComponent) keywordHash.get(key.toUpperCase());
      if (keyword == null){
         return null;
      } else{
         String value = keyword.toString();
         if (value == null){
            return null;
         }
         String values[] = value.split(" +");
         if (values.length > 1){
            return values[1];
         } else{
            return null;
         }
      }
   }

   public void setKeywordValue(String key, String value){
      KeywordComponent keyword = (KeywordComponent) keywordHash.get(key.toUpperCase());
      if (keyword == null){
         return;
      } else{
         keyword.keyEntry(value);
         String keygroup = keyword.getKeyGroup();
         if (!jcbGroups.getSelectedItem().equals(keygroup)){
            jcbGroups.setSelectedItem(keygroup);
            getKeywordGroupPanel(keygroup);
         }
      }
   }

   public void setKeywordGroup(String keygroup){
      keygroup = (String) keywordGroups.get(keygroup.toUpperCase());
      if (keygroup == null){
         return;
      }
      if (!jcbGroups.getSelectedItem().equals(keygroup)){
         jcbGroups.setSelectedItem(keygroup);
         getKeywordGroupPanel(keygroup);
      }
   }

   /**
    * Handles input from TinkerKeywordDialog buttons.
    */
   public void actionPerformed(ActionEvent evt){
      String arg = evt.getActionCommand();
      if (arg.equals("Open...")){
         keyOpen();
      } else if (arg.equals("Close")){
         keyClose();
      } else if (arg.equals("Save")){
         keySave(currentKeyFile);
      } else if (arg.equals("Save As...")){
         keySaveAs();
      } else if (arg.equals("Description")){
         JCheckBoxMenuItem box = (JCheckBoxMenuItem) evt.getSource();
         setDivider(box.isSelected());
      }
      if (evt.getSource() instanceof JComboBox){
         String keygroup = ((JComboBox) evt.getSource()).getSelectedItem().toString();
         getKeywordGroupPanel(keygroup);
      }
   }

   public void keyOpen(){
      if (fileOpen && KeywordComponent.isModified()){
         int option = JOptionPane.showConfirmDialog(this, "Save Changes First", "Opening New File",
                      JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
         if (option == JOptionPane.CANCEL_OPTION){
            return;
         } else if (option == JOptionPane.YES_OPTION){
            keySave(currentKeyFile);
         }
         keyClear();
      }
      JFileChooser d = MainPanel.getFileChooser();
      if (currentSystem != null){
         File cwd = currentSystem.getFile();
         if (cwd != null && cwd.getParentFile() != null){
            d.setCurrentDirectory(cwd.getParentFile());
         }
      }
      d.setAcceptAllFileFilterUsed(false);
      d.setFileFilter(MainPanel.keyfilefilter);
      d.setDialogTitle("Open KEY File");
      int result = d.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         currentKeyFile = d.getSelectedFile();
         currentSystem.setKeyFile(currentKeyFile);
         keyOpen(currentSystem);
      }
   }

   public void store(FSystem sys){
      FSystem back = currentSystem;

      currentSystem = sys;
      storeActive();

      currentSystem = back;
   }

   public void storeActive(){
      if (currentSystem == null){
         return;
      }
      if (!KeywordComponent.isModified()){
         return;
      }

      Hashtable currentKeys = currentSystem.getKeywords();
      Hashtable newKeys = new Hashtable();

      for (Iterator li = keywordHash.values().iterator(); li.hasNext(); ){
         KeywordComponent keyword = (KeywordComponent) li.next();
         if (keyword.isActive() == true){
            Keyword keywordData = null;
            if (currentKeys != null){
               keywordData = (Keyword) currentKeys.get(keyword.getKey());
            }
            if (keywordData == null){
               keywordData = new Keyword(keyword.getKey());
            } else{
               keywordData.clear();
            }
            keyword.loadData(keywordData);
            newKeys.put(keyword.getKey(), keywordData);
         }
      }
      Keyword comments = new Keyword("COMMENTS", taComments.getText());
      newKeys.put("COMMENTS", comments);
      currentSystem.setKeywords(newKeys);
   }

   public boolean loadActive(FSystem sys){
      if (currentSystem != null){
         if (currentSystem == sys){
            return false;
         } else{
            storeActive();
         }
      }
      keyClear();

      if (sys == null){
         return false;
      }

      Hashtable sysKeys = sys.getKeywords();

      if (sysKeys == null || sysKeys.size() == 0){
         return false;
      }

      Keyword comments = sys.getKeyword("COMMENTS");
      Vector keysToRemove = new Vector();
      for (Iterator li = sysKeys.values().iterator(); li.hasNext(); ){
         Keyword keywordData = (Keyword) li.next();
         String label = keywordData.getKeyword();
         Vector data = keywordData.getEntries();
         if (label.equals("COMMENTS")){
            continue;
         } else{
            KeywordComponent tk = (KeywordComponent) keywordHash.get(label.toUpperCase());
            // If the keyword label isn't recognized, preserve the line as a Comment
            if (tk == null){
               keysToRemove.add(keywordData);
               if (data.size() == 0){
                  comments.append(label);
               } else{
                  for (ListIterator li2 = data.listIterator(); li2.hasNext(); ){
                     String entry = (String) li2.next();
                     comments.append(label + " " + entry);
                  }
               }
               continue;
            }
            if (data.size() == 0){
               tk.keyEntry(null);
            }
            for (ListIterator li2 = data.listIterator(); li2.hasNext(); ){
               String entry = (String) li2.next();
               tk.keyEntry(entry);
            }
         }
      }
      Vector entries = comments.getEntries();
      if (entries != null){
         for (ListIterator li2 = entries.listIterator(); li2.hasNext(); ){
            String entry = (String) li2.next();
            if (!keywordGroups.containsKey(entry.substring(1).toUpperCase().trim())){
               taComments.append(entry + "\n");
            }
         }
      }
      if (keysToRemove.size() != 0){
         for (ListIterator li = keysToRemove.listIterator(); li.hasNext(); ){
            sys.removeKeyword((Keyword) li.next());
         }
      }

      fileOpen = true;
      currentKeyFile = sys.getKeyFile();
      currentSystem = sys;
      getKeywordGroupPanel(jcbGroups.getSelectedItem().toString());
      KeywordComponent.setModified(false);
      return true;
   }

   public boolean keyOpen(FSystem sys){
      if (sys == null){
         return false;
      }
      File key = sys.getKeyFile();
      if (key == null || !key.exists() || !key.canRead()){
         return false;
      }
      keyClear();
      try{
         FileReader fr = new FileReader(key);
         BufferedReader br = new BufferedReader(fr);
         String s, keyword;
         int firstspace;
         KeywordComponent tk;
         while (br.ready()){
            s = br.readLine();
            if (s == null){
               continue;
            }
            s = s.trim();
            if (s.equals("")){
               continue; //Skip blank lines
            }
            // Put comments in the Comments TextArea for editing
            if (s.startsWith("#") || s.toLowerCase().startsWith("ECHO")){
               if (!keywordGroups.containsKey(s.substring(1).toUpperCase().trim())){
                  taComments.append(s + "\n");
               }
            } else{
               firstspace = s.indexOf(" ");
               if (firstspace == -1){ //no parameters
                  keyword = s.trim();
               } else{
                  keyword = s.substring(0, firstspace);
               }
               tk = (KeywordComponent) keywordHash.get(keyword.toUpperCase());
               if (tk == null){
                  if (keyword.equalsIgnoreCase("Multipole")){
                     taComments.append(s + "\n");
                     for (int i = 0; i < 4; i++){
                        if (!br.ready()){
                           System.out.println("Check for invalid MULTIPOLE keyword at EOF");
                           break;
                        }
                        s = br.readLine();
                        taComments.append(s + "\n");
                     }
                  } else{
                     taComments.append(s + "\n");
                  }
                  continue; //keyword not found
               }
               if (firstspace != -1){
                  s = s.substring(firstspace).trim();
               } else{
                  s = new String("");
               }
               tk.keyEntry(s);
            }
         }

         fileOpen = true;
         currentKeyFile = key;
         getKeywordGroupPanel(jcbGroups.getSelectedItem().toString());
         KeywordComponent.setModified(false);
         return true;
      } catch (IOException e){
         System.out.println("Problem reading Key File: " + e);
         return false;
      }

   }

   public void keyClear(){
      KeywordComponent kw;
      taComments.setText("  ");
      for (Iterator li = keywordHash.values().iterator(); li.hasNext(); ){
         kw = ((KeywordComponent) li.next());
         kw.clear();
      }
      KeywordComponent.setModified(false);
      currentKeyFile = null;
      fileOpen = false;
      status.setText("  ");
      setKeywordGroup("Active Keywords");
      getKeywordGroupPanel("Active Keywords");
   }

   public void keyClose(){
      /*
       int option = JOptionPane.showConfirmDialog(this, "Save Changes First", "Closing current File",
       JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
         if (option == JOptionPane.CANCEL_OPTION) return;
         else if (option == JOptionPane.YES_OPTION) keysave();
       */
      keyClear();
   }

   public void keySaveAs(){
      if (!fileOpen){
         return;
      }

      JFileChooser d = MainPanel.getFileChooser();
      d.setDialogTitle("Save KEY File");
      d.setAcceptAllFileFilterUsed(false);
      if (currentKeyFile != null){
         d.setCurrentDirectory(currentKeyFile.getParentFile());
         d.setSelectedFile(currentKeyFile);
      }
      d.setFileFilter(new KeyFileFilter());
      int result = d.showSaveDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         currentKeyFile = d.getSelectedFile();
         keySave(currentKeyFile);
      }
   }

   public void keySave(File f){

      if (f != null){
         currentKeyFile = f;
      }
      if (!fileOpen || currentKeyFile == null){
         return;
      }

      storeActive();

      try{
         FileWriter fw = new FileWriter(currentKeyFile);
         BufferedWriter bw = new BufferedWriter(fw);
         boolean writegroup = false;
         String comments, group, pgroup = null;
         String line;
         KeywordComponent keyword;

         // Write out keywords in groups
         for (Iterator li = keywordHash.values().iterator(); li.hasNext(); ){
            keyword = (KeywordComponent) li.next();
            group = keyword.getKeyGroup();
            if (pgroup == null || !group.equalsIgnoreCase(pgroup)){
               writegroup = true;
               pgroup = group;
            }
            line = keyword.toString();
            if (line != null){
               if (writegroup == true){
                  bw.newLine();
                  bw.write("# " + group);
                  bw.newLine();
                  writegroup = false;
               }
               bw.write(line);
               bw.newLine();
            }
         }
         bw.newLine();
         comments = taComments.getText();
         if (comments != null && !comments.trim().equals("")){
            bw.write(comments.trim());
         }
         bw.newLine();
         bw.flush();
         bw.close();

         KeywordComponent.setModified(false);
      } catch (FileNotFoundException e){
         System.out.println("File not found");
      } catch (IOException e){
         System.out.println("Can't write a line");
      }
   }

   public void initToolBar(){
      toolbar = new JToolBar("Keyword Editor");
      toolbar.setLayout(new FlowLayout(FlowLayout.LEFT));

      JButton jbopen = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                       "ffe/icons/fileopen.png")));
      jbopen.setActionCommand("Open...");
      jbopen.setToolTipText("Open a *.KEY File for Editing");
      jbopen.addActionListener(this);
      Insets insets = jbopen.getInsets();
      insets.top = 2;
      insets.bottom = 2;
      insets.left = 2;
      insets.right = 2;
      jbopen.setMargin(insets);
      toolbar.add(jbopen);

      JButton jbsave = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                       "ffe/icons/filesave.png")));
      jbsave.setActionCommand("Save");
      jbsave.setToolTipText("Save the Active *.KEY File");
      jbsave.addActionListener(this);
      jbsave.setMargin(insets);
      toolbar.add(jbsave);

      JButton jbsaveas = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                         "ffe/icons/filesaveas.png")));
      jbsaveas.setActionCommand("Save As...");
      jbsaveas.setToolTipText("Save the Active *.KEY File Under a New Name");
      jbsaveas.addActionListener(this);
      jbsaveas.setMargin(insets);
      toolbar.add(jbsaveas);

      JButton jbclose = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                        "ffe/icons/fileclose.png")));
      jbclose.setActionCommand("Close");
      jbclose.setToolTipText("Close the Active *.KEY File");
      jbclose.addActionListener(this);
      jbclose.setMargin(insets);
      toolbar.add(jbclose);
      toolbar.addSeparator();

      jcbGroups.setMaximumSize(jcbGroups.getPreferredSize());
      jcbGroups.addActionListener(this);
      jcbGroups.setEditable(false);
      toolbar.add(jcbGroups);
      toolbar.addSeparator();

      ImageIcon icinfo = new ImageIcon(getClass().getClassLoader().getResource("ffe/icons/info.png"));
      jcbDescript = new JCheckBoxMenuItem(icinfo);
      jcbDescript.setActionCommand("Description");
      jcbDescript.setToolTipText("Show/Hide Keyword Descriptions");
      jcbDescript.addActionListener(this);
      jcbDescript.setMargin(insets);
      toolbar.add(jcbDescript);
      toolbar.add(new JLabel(""));

      toolbar.setBorderPainted(false);
      toolbar.setFloatable(false);
      toolbar.setRollover(true);
      toolbar.setOrientation(JToolBar.HORIZONTAL);
   }

   public void setParamPath(String tinkerDir){
      String params = tinkerDir;

      if (params == null){
         Preferences prefs = Preferences.userNodeForPackage(KeywordPanel.class);
         params = prefs.get("KeywordPanel_params", mainPanel.getCWD().getAbsolutePath());
         if (params != null){
            paramDir = new File(params);
         }
      } else{
         paramDir = new File(params + File.separator + "params");
      }

      if (paramDir != null && paramDir.exists()){
         //System.out.println("Parameter Directory: " + paramDir.getAbsolutePath());
         paramFiles = paramDir.listFiles();
         paramHash = new Hashtable();
         for (int i = 0; i < paramFiles.length; i++){
            File f = paramFiles[i];
            if (f.exists() && f.canRead() && mainPanel.prmFileFilter.accept(f)){
               paramHash.put(f.getName(), f.getAbsolutePath());
            }
         }
         int num = paramHash.size();
         paramNames = new String[num + 1];
         int i = 1;
         for (Enumeration e = paramHash.keys(); e.hasMoreElements(); ){
            paramNames[i] = (String) e.nextElement();
            i++;
         }
         paramNames[0] = "AAA";
         java.util.Arrays.sort(paramNames);
         paramNames[0] = "Use an existing TINKER Key file".intern();
      }

   }

   public void loadPrefs(){
      Preferences prefs = Preferences.userNodeForPackage(KeywordPanel.class);
      jcbDescript.setSelected(!prefs.getBoolean("KeywordPanel_description", true));
      jcbDescript.doClick();
      setParamPath(System.getProperty("tinker.dir"));
   }

   public void savePrefs(){
      if (!lazyDone){
         return;
      }
      Preferences prefs = Preferences.userNodeForPackage(ffe.panels.KeywordPanel.class);
      prefs.putInt("KeywordPanel_divider", splitPane.getDividerLocation());
      prefs.putBoolean("KeywordPanel_description", jcbDescript.isSelected());
      prefs.put("KeywordPanel_params", paramDir.getAbsolutePath());
   }

   public void setDivider(boolean b){
      if (b){
         jcbDescript.setSelected(b);
         int spDivider = (int) (this.getHeight() * (3.0f / 5.0f));
         splitPane.setDividerLocation(spDivider);
      } else{
         splitPane.setDividerLocation(1.0);
      }
   }

   public void selected(){

   }

   private void initialize(){
      NodeList keywords, values, keygroupList = null;
      Element keygroup, keygroupname, keyword, keywordValue, keywordDefault;
      String skeygroupname, skeyword, skeydescript, skeyswing;
      int length;
      try{
         DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
         DocumentBuilder db = dbf.newDocumentBuilder();
         db.setEntityResolver(new DTDResolver());
         URL keyURL = getClass().getClassLoader().getResource("ffe/xml/keywords.xml");
         Document doc = db.parse(keyURL.openStream());
         NodeList nodelist = doc.getChildNodes();
         Node keywordroot = null;
         length = nodelist.getLength();
         for (int i = 0; i < length; i++){
            keywordroot = nodelist.item(i);
            if (keywordroot.getNodeName().equals("Keywords") && keywordroot instanceof Element){
               break;
            }
         }
         if (keywordroot == null){
            keygroupList = null;
            return;
         }
         keygroupList = ((Element) keywordroot).getElementsByTagName("KeywordGroup");
      } catch (ParserConfigurationException e){
         System.out.println("" + e);
      } catch (SAXException e){
         System.out.println("" + e);
      } catch (IOException e){
         System.out.println("" + e);
      }

      length = keygroupList.getLength();
      taDescription = new JTextArea();

      //Iterate through the Keyword Groups
      jcbGroups = new JComboBox();
      jcbGroups.addItem("Active Keywords");
      jcbGroups.addItem("Comments");

      KeywordComponent key;
      keywordHash = new LinkedHashMap();
      keywordGroups = new LinkedHashMap();
      keywordGroups.put("ACTIVE KEYWORDS", "Active Keywords");
      keywordGroups.put("COMMENTS", "Comments");
      int type = 0;

      for (int i = 0; i < length; i++){
         keygroup = (Element) keygroupList.item(i);
         keygroupname = (Element) keygroup.getElementsByTagName("GroupName").item(0);
         skeygroupname = keygroupname.getFirstChild().getNodeValue();
         jcbGroups.addItem(skeygroupname);
         keywords = keygroup.getElementsByTagName("Keyword");
         int klength = keywords.getLength();
         for (int j = 0; j < klength; j++){
            keyword = (Element) keywords.item(j);
            skeyword = keyword.getElementsByTagName("KeywordName").item(0).getFirstChild().
                       getNodeValue();
            skeydescript = keyword.getElementsByTagName("KeywordDescription").item(0).getFirstChild().
                           getNodeValue();
            skeyswing = keyword.getElementsByTagName("KeywordGUIComponent").item(0).getFirstChild().
                        getNodeValue();
            if (skeyswing.equalsIgnoreCase("TEXTFIELD")){
               type = KeywordComponent.TEXTFIELD;
            } else if (skeyswing.equalsIgnoreCase("COMBOBOX")){
               type = KeywordComponent.COMBOBOX;
            } else if (skeyswing.equalsIgnoreCase("CHECKBOX")){
               type = KeywordComponent.CHECKBOX;
            } else if (skeyswing.equalsIgnoreCase("RADIOBUTTONS")){
               type = KeywordComponent.RADIOBUTTONS;
            } else if (skeyswing.equalsIgnoreCase("CHECKBOXES")){
               type = KeywordComponent.CHECKBOXES;
            }
            if (type == KeywordComponent.CHECKBOXES || type == KeywordComponent.RADIOBUTTONS){
               values = keyword.getElementsByTagName("KeywordValue");
               String labels[] = new String[values.getLength()];
               for (int k = 0; k < values.getLength(); k++){
                  keywordValue = (Element) values.item(k);
                  labels[k] = keywordValue.getFirstChild().getNodeValue();
               }
               key = new KeywordComponent(skeyword, skeygroupname, type, skeydescript,
                     taDescription, labels);
            } else{
               key = new KeywordComponent(skeyword, skeygroupname, type, skeydescript,
                     taDescription);
            }
            keywordHash.put(skeyword.toUpperCase(), key);
            keywordGroups.put(skeygroupname.toUpperCase(), skeygroupname);
         }
      }
      jcbGroups.setSelectedIndex(0);

      // Keyword Group Selection Panel
      setLayout(new BorderLayout());

      // Keywords from a Group
      taComments = new JTextArea();
      taComments.setEditable(true);
      keywordPanel = new JPanel(new BorderLayout());
      spKeywords = new JScrollPane(keywordPanel);
      EtchedBorder eb = new EtchedBorder(EtchedBorder.RAISED);
      //spKeywords.setBorder(eb);

      // Keyword Description Panel
      taDescription.setEditable(false);
      spDescript = new JScrollPane(taDescription);
      TitledBorder tbDescript = new TitledBorder(eb, " Description ");
      spDescript.setBorder(tbDescript);

      // Add the Keyword Panel and Decription Panels to a JSplitPane
      splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, spKeywords, spDescript);
      splitPane.setResizeWeight(1.0);
      splitPane.setOneTouchExpandable(true);
      status.setBorder(eb);
      add(splitPane, BorderLayout.CENTER);
      add(status, BorderLayout.SOUTH);
   }

   private boolean lazyDone = false;
   public void initLazy(){
      if (lazyDone){
         return;
      }

      initToolBar();
      add(toolbar, BorderLayout.NORTH);
      getKeywordGroupPanel("Active Keywords");
      loadPrefs();
      validate();
      lazyDone = true;
   }

   private void getKeywordGroupPanel(String selgroup){
      JPanel grouppanel = null;
      KeywordComponent firstkey = null;
      boolean first = true;

      keywordPanel.removeAll();

      if (selgroup.equalsIgnoreCase("Comments")){
         grouppanel = new JPanel(new GridLayout(1, 1));
         grouppanel.add(taComments);
      } else if (selgroup.equalsIgnoreCase("Active Keywords")){
         GridBagLayout g = new GridBagLayout();
         GridBagConstraints c = new GridBagConstraints();
         c.gridx = c.gridy = 0;
         c.anchor = c.WEST;
         c.gridheight = c.gridwidth = 1;
         c.fill = c.HORIZONTAL;

         grouppanel = new JPanel(g);
         KeywordComponent kwtemp = null;
         JPanel jptemp;
         for (Iterator li = keywordHash.values().iterator(); li.hasNext(); ){
            kwtemp = (KeywordComponent) li.next();
            if (kwtemp.isActive() == true){
               if (first){
                  firstkey = kwtemp;
                  first = false;
               }
               jptemp = kwtemp.getPanel();
               g.setConstraints(jptemp, c);
               grouppanel.add(jptemp);
               c.gridy++;
            }
         }
         KeywordComponent.fillPanel(grouppanel, g, c);
      } else{
         GridBagLayout g = new GridBagLayout();
         GridBagConstraints c = new GridBagConstraints();
         c.gridx = c.gridy = 0;
         c.anchor = c.WEST;
         c.gridheight = c.gridwidth = 1;
         c.fill = c.HORIZONTAL;
         grouppanel = new JPanel(g);
         KeywordComponent kwtemp = null;
         JPanel jptemp;
         for (Iterator li = keywordHash.values().iterator(); li.hasNext(); ){
            kwtemp = (KeywordComponent) li.next();
            if (kwtemp.getKeyGroup().equalsIgnoreCase(selgroup)){
               if (first){
                  firstkey = kwtemp;
                  first = false;
               }
               jptemp = kwtemp.getPanel();
               g.setConstraints(jptemp, c);
               grouppanel.add(jptemp);
               c.gridy++;
            }
         }
         KeywordComponent.fillPanel(grouppanel, g, c);
      }
      if (firstkey != null){
         taDescription.removeAll();
         taDescription.setText(firstkey.getKey() + ": " + firstkey.getDescription());
         taDescription.validate();
      }
      keywordPanel.add(grouppanel, BorderLayout.CENTER);
      if (currentKeyFile != null){
         status.setText("  " + currentKeyFile.toString());
      } else{
         status.setText("  ");
      }
      splitPane.validate();
   }
}
