/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.*;
import java.util.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.EtchedBorder;

/**
 * The LogPanel is a simple editor that displays log files
 * created by TINKER jobs. Any text file can be edited.
 */
public class LogPanel
   extends JPanel
   implements ActionListener{

   private Vector jobs;
   private Hashtable files = new Hashtable();

   private MainPanel fpanel;
   private JobPanel jobPanel;

   private JTabbedPane results = new JTabbedPane(JTabbedPane.BOTTOM);
   private JProgressBar statusBar;
   private Font font;

   private JToolBar toolbar;

   /**
    * Constructor
    * @param f
    */
   public LogPanel(MainPanel f){
      fpanel = f;
      jobPanel = f.getJobPanel();
      jobs = jobPanel.getJobs();

      initToolBar();
      setLayout(new BorderLayout());
      EtchedBorder eb = new EtchedBorder(EtchedBorder.RAISED);

      statusBar = new JProgressBar();
      statusBar.setBorder(eb);
      statusBar.setStringPainted(true);

      add(toolbar, BorderLayout.NORTH);
      add(results, BorderLayout.CENTER);

      if (System.getProperty("os.name").startsWith("Windows")){
         font = new Font("Monospaced", Font.PLAIN, 12);
      } else if (System.getProperty("os.name").startsWith("Linux")){
         font = new Font("Lucida Sans Typewriter", Font.PLAIN, 12);
      }
      if (font == null){
         font = new Font("Monospaced", Font.PLAIN, 12);
      }
      refreshStatus();
   }

   public void actionPerformed(ActionEvent evt){
      if (evt.getSource() instanceof javax.swing.Timer){
         if (results.getTabCount() > 0){
            refresh();
         }
         return;
      }
      String arg = evt.getActionCommand();
      if (arg == null){
         return;
      }
      if (arg.equalsIgnoreCase("Refresh")){
         refresh();
      } else if (arg.equalsIgnoreCase("Close")){
         int index = results.getSelectedIndex();
         if (index >= 0){
            String title = results.getTitleAt(index);
            results.remove(index);
            files.remove(title);
            for (int i = 0; i < jobs.size(); i++){
               Thread t = (Thread) jobs.get(i);
               String name = t.getName();
               if (name.equals(title)){
                  jobs.remove(i);
                  break;
               }
            }
         }
      } else if (arg.equalsIgnoreCase("Close All")){
         results.removeAll();
         files.clear();
         jobs.clear();
      } else if (arg.equalsIgnoreCase("Open...")){
         JFileChooser d = MainPanel.getFileChooser();
         d.setAcceptAllFileFilterUsed(true);
         d.setDialogTitle("Open Log File");
         int result = d.showOpenDialog(this);
         if (result == JFileChooser.APPROVE_OPTION){
            File f = d.getSelectedFile();
            addPane(f);
         }
      } else if (arg.equalsIgnoreCase("Save")){
         saveSelected();
      } else if (arg.equalsIgnoreCase("Save As...")){
         saveSelectedAs();
      }
   }

   public void refresh(){
      for (Enumeration e = files.elements(); e.hasMoreElements(); ){
         ((JTextArea) e.nextElement()).setFont(font);
      }

      synchronized (jobs){
         for (ListIterator li = jobs.listIterator(); li.hasNext(); ){
            Thread t = (Thread) li.next();
            File f = new File(t.getName());
            if (!f.exists()){
               continue;
            }
            // Refresh an already existing JTextArea
            if (files.containsKey(f.getAbsolutePath())){
               JTextArea ta = (JTextArea) files.get(f.getAbsolutePath());
               ta.setFont(font);
               loadText(ta, f);
            }
            // Create a new TextArea for the file
            else{
               addPane(f);
            }
         }
      }
      refreshStatus();
   }

   public void refreshStatus(){
      int count = jobs.size();
      if (count == 0){
         statusBar.setString("");
         statusBar.setIndeterminate(false);
      } else if (count == 1){
         statusBar.setString("1 Job Running");
         statusBar.setIndeterminate(true);
      } else{
         statusBar.setString("" + count + " Jobs Running");
         statusBar.setIndeterminate(true);
      }
   }

   public void saveSelected(){
      int index = results.getSelectedIndex();
      if (index < 0){
         return;
      }
      String title = results.getTitleAt(index);
      JTextArea ta = (JTextArea) files.get(title);
      File f = new File(title);
      if (ta != null && f != null && f.exists() && f.canWrite()){
         try{
            FileOutputStream os = new FileOutputStream(f);
            if (os != null){
               BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(os));
               bw.write(ta.getText());
               bw.close();
            }
            os.close();
         } catch (IOException e){
            System.out.println("" + e);
         }
      }
   }

   public void saveSelectedAs(){
      synchronized (results){
         int index = results.getSelectedIndex();
         if (index < 0){
            return;
         }
         String title = results.getTitleAt(index);
         JTextArea ta = (JTextArea) files.get(title);
         File f = new File(title);

         if (ta != null && f != null && f.exists() && f.canWrite()){
            JFileChooser fileChooser = MainPanel.getFileChooser();
            fileChooser.setSelectedFile(f);
            fileChooser.setAcceptAllFileFilterUsed(true);
            int result = fileChooser.showSaveDialog(this);
            if (result == JFileChooser.APPROVE_OPTION){
               f = fileChooser.getSelectedFile();
            } else{
               return;
            }
            try{
               files.remove(title);
               FileOutputStream os = new FileOutputStream(f);
               if (os != null){
                  BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(os));
                  bw.write(ta.getText());
                  bw.close();
               }
               os.close();

               results.setTitleAt(index, f.getAbsolutePath());
               files.put(f.getAbsolutePath(), ta);
            } catch (IOException e){
               System.out.println("" + e);
            }
         }
      }
   }

   public void selected(){
      return;
   }

   public void setDone(String name){
      if (name == null){
         return;
      }
      File f = new File(name);
      for (int i = 0; i < jobs.size(); i++){
         Thread t = (Thread) jobs.get(i);
         String jobname = t.getName();
         if (jobname.equals(name)){
            jobs.remove(i);
            break;
         }
      }
      if (!f.exists()){
         return;
      }
      if (files.containsKey(f.getAbsolutePath())){
         JTextArea ta = (JTextArea) files.get(f.getAbsolutePath());
         loadText(ta, f);
      }
      // Create a new TextArea for the file
      else{
         addPane(f);
      }
      results.setSelectedIndex(results.indexOfTab(f.getAbsolutePath()));
      refreshStatus();
   }

   EtchedBorder eb = new EtchedBorder(EtchedBorder.RAISED);
   private void addPane(File f){
      if (!f.exists()){
         return;
      }
      JTextArea ta = new JTextArea();
      ta.setEditable(f.canWrite());
      ta.setFont(font);
      files.put(f.getAbsolutePath(), ta);
      JScrollPane js = new JScrollPane(ta);
      js.setBorder(eb);
      results.add(js, f.getAbsolutePath());
      loadText(ta, f);
   }

   private void loadText(JTextArea target, File data){
      if (target == null || (!data.exists()) || (!data.canRead())){
         return;
      }

      target.setText("");
      try{
         FileInputStream is = new FileInputStream(data);
         if (is != null){
            BufferedReader br = new BufferedReader(new InputStreamReader(is));
            while (br.ready()){
               target.append(br.readLine() + "\n");
            }
            br.close();
         }
         is.close();
      } catch (IOException e){
         System.out.println("" + e);
      }
      fpanel.setPanel(MainPanel.LOGS);
   }

   public JProgressBar getProgressBar(){
      return statusBar;
   }

   public void initToolBar(){
      toolbar = new JToolBar("Results");
      toolbar.setLayout(new FlowLayout(FlowLayout.LEFT));

      JButton jbrefresh = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                          "ffe/icons/refresh.png")));
      jbrefresh.setActionCommand("Refresh");
      jbrefresh.setToolTipText("Refresh the Results Panel");
      jbrefresh.addActionListener(this);
      Insets insets = jbrefresh.getInsets();
      insets.top = 2;
      insets.bottom = 2;
      insets.left = 2;
      insets.right = 2;
      jbrefresh.setMargin(insets);
      toolbar.add(jbrefresh);
      toolbar.addSeparator();

      JButton jbopen = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                       "ffe/icons/fileopen.png")));
      jbopen.setActionCommand("Open...");
      jbopen.setToolTipText("Open any Text File");
      jbopen.addActionListener(this);
      jbopen.setMargin(insets);
      toolbar.add(jbopen);

      JButton jbsave = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                       "ffe/icons/filesave.png")));
      jbsave.setActionCommand("Save");
      jbsave.setToolTipText("Save the Active File");
      jbsave.addActionListener(this);
      jbsave.setMargin(insets);
      toolbar.add(jbsave);

      JButton jbsaveas = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                         "ffe/icons/filesaveas.png")));
      jbsaveas.setActionCommand("Save As...");
      jbsaveas.setToolTipText("Save the Active Text File Under a New Name");
      jbsaveas.addActionListener(this);
      jbsaveas.setMargin(insets);
      toolbar.add(jbsaveas);

      JButton jbclose = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                        "ffe/icons/fileclose.png")));
      jbclose.setActionCommand("Close");
      jbclose.setToolTipText("Close the Active Text File");
      jbclose.addActionListener(this);
      jbclose.setMargin(insets);
      toolbar.add(jbclose);

      JButton jbcloseall = new JButton(new ImageIcon(getClass().getClassLoader().getResource(
                           "ffe/icons/closeall.png")));
      jbcloseall.setActionCommand("Close All");
      jbcloseall.setToolTipText("Close All Open Text Files");
      jbcloseall.addActionListener(this);
      jbcloseall.setMargin(insets);
      toolbar.add(jbcloseall);

      toolbar.setFloatable(false);
      toolbar.setRollover(true);
      toolbar.setOrientation(JToolBar.HORIZONTAL);
   }
}
