/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.prefs.Preferences;
import javax.media.j3d.*;
import javax.vecmath.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import bsh.Interpreter;
import bsh.util.JConsole;

import ffe.lang.*;
import ffe.net.Simulation;
import ffe.parsers.*;

/**
 * The MainPanel class is the main container for holding Force Field Explorer components.
 */
public final class MainPanel
   extends JPanel
   implements ActionListener, ChangeListener{

   /**
    * MainPanel Constructor
    * @param f Application Frame
    */
   public MainPanel(JFrame f){
      frame = f;
   }

   /**
    * Handle most File, Edit, View and Help Menu Commands
    * This should probably be partioned between a few different handlers
    */
   public void actionPerformed(ActionEvent evt){
      String arg = evt.getActionCommand();

      if (arg.equals("Exit")){
         exit();
      } else if (arg.equals("Open...")){
         open();
      } else if (arg.equals("Download from PDB...")){
         openFromPDB();
      } else if (arg.equals("Save as...")){
         save(null);
      } else if (arg.equals("Merge")){
         merge();
      } else if (arg.equals("Close")){
         closeActive();
      } else if (arg.equals("Close All")){
         closeAll();
      } else if (arg.equals("Load Restart Data...")){
         openRestart();
      } else if (arg.equals("Load Induced Data...")){
         openInduced();
      } else if (arg.equals("Select All")){
         if (root.getChildCount() == 0){
            return;
         }
         tree.selectAll();
      } else if (arg.equals("Highlight Selections")){
         if (evt.getSource() instanceof JCheckBoxMenuItem){
            JCheckBoxMenuItem jcb = (JCheckBoxMenuItem) evt.getSource();
            if (jcb.isSelected()){
               tree.setHighlighting(true);
            } else{
               tree.setHighlighting(false);
            }
         } else{
            boolean highlighting = GraphicsContext.highlightSelections;
            if (highlighting){
               tree.setHighlighting(false);
               menu.setHighlighting(false);
            } else{
               tree.setHighlighting(true);
               menu.setHighlighting(true);
            }
         }
      } else if (arg.equals("Center System")){
         FSystem m = tree.getActive();
         if (m != null){
            TransformGroup tg = m.getTransformGroup();
            Transform3D t3d = new Transform3D();
            Vector3f v = new Vector3f(0, 0, 0);
            t3d.set(v);
            tg.setTransform(t3d);
         }
      } else if (arg.equals("Center Coordinate System")){
         graphics.centerView();
      } else if (arg.equals("Full Screen")){
         if (fullScreenWindow == null){
            fullScreenWindow = new FullScreenWindow(frame, graphPanel, graphics);
         }
         fullScreenWindow.enterFullScreen();
      } else if (arg.equals("Play")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         traj.start();
      } else if (arg.equals("Stop")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         traj.stop();
      } else if (arg.equals("Step Forward")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         traj.stop();
         traj.forward();
      } else if (arg.equals("Step Back")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         traj.stop();
         traj.back();
      } else if (arg.equals("Reset")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         traj.stop();
         traj.rewind();
      } else if (arg.equals("Oscillate")){
         oscillate = ((JCheckBoxMenuItem) evt.getSource()).isSelected();
         FSystem[] system = tree.getSystems();
         for (int i = 0; i < system.length; i++){
            Trajectory traj = system[i].getTrajectory();
            if (traj != null){
               traj.setOscillate(oscillate);
            }
         }
      } else if (arg.equals("Frame...")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         String frame = new String("" + traj.getFrame());
         frame = JOptionPane.showInputDialog("Enter the Frame ", frame);
         try{
            int f = Integer.parseInt(frame);
            traj.setFrame(f);
         } catch (NumberFormatException e){
         }
      } else if (arg.equals("Speed...")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         String frame = new String("" + (int) traj.getRate());
         frame = JOptionPane.showInputDialog("Enter the Frame Rate (1-100)", frame);
         try{
            int f = Integer.parseInt(frame);
            traj.setRate(f);
         } catch (NumberFormatException e){
            //System.out.println("Enter an integer between 1 and 100");
         }
      } else if (arg.equals("Skip...")){
         Trajectory traj = getTrajectory();
         if (traj == null){
            return;
         }
         String skip = new String("" + (int) traj.getSkip());
         skip = JOptionPane.showInputDialog("Enter the number of frames to skip", skip);
         try{
            int f = Integer.parseInt(skip);
            traj.setSkip(f);
         } catch (NumberFormatException e){
         }
      } else if (arg.equals("Connect to Local Job")){
         connectToTINKER(null, null);
      } else if (arg.equals("Connect to Remote Job")){
         connect();
      } else if (arg.equals("Release Job")){
         release();
      } else if (arg.equals("Preferences...")){
         preferences();
      } else if (arg.equals("Set Port...")){
         String s = new String("" + port);
         s = JOptionPane.showInputDialog("Enter a port number", s);
         if (s == null){
            return;
         }
         int temp;
         try{
            temp = Integer.parseInt(s);
         } catch (NumberFormatException e){
            return;
         }
         port = temp;
         socketAddress = new InetSocketAddress(address, port);
      } else if (arg.equals("Set Remote Job Address...")){
         if (address == null){
            try{
               address = InetAddress.getLocalHost();
            } catch (Exception e){
               try{
                  address = InetAddress.getByName(null);
               } catch (Exception ex){
                  return;
               }
            }
         }
         String s = new String("" + address.getHostAddress());
         s = JOptionPane.showInputDialog("Enter an IP Address (XXX.XXX.XXX.XXX)", s);
         if (s == null){
            return;
         }
         InetAddress newAddress;
         InetSocketAddress newSocketAddress;
         try{
            newAddress = InetAddress.getByName(s);
            newSocketAddress = new InetSocketAddress(newAddress, port);
         } catch (NumberFormatException e){
            return;
         } catch (Exception e){
            return;
         }
         address = newAddress;
         socketAddress = newSocketAddress;
      } else if (arg.equals("Show Tool Bar")){
         JCheckBoxMenuItem jcb = (JCheckBoxMenuItem) evt.getSource();
         if (jcb.isSelected()){
            add(menu.getToolBar(), BorderLayout.NORTH);
            frame.validate();
         } else{
            remove(menu.getToolBar());
            frame.validate();
         }
      } else if (arg.equals("Show Tree")){
         JCheckBoxMenuItem jcb = (JCheckBoxMenuItem) evt.getSource();
         if (jcb.isSelected()){
            if (spDivider < frame.getWidth() * (1.0f / 4.0f)){
               spDivider = (int) (frame.getWidth() * (1.0f / 4.0f));
            }
            splitPane.setDividerLocation(spDivider);
         } else{
            spDivider = splitPane.getDividerLocation();
            splitPane.setDividerLocation(0.0);
         }
      } else if (arg.equals("Show Global Axes")){
         JCheckBoxMenuItem jcb = (JCheckBoxMenuItem) evt.getSource();
         graphics.setAxisShowing(jcb.isSelected());
      } else if (arg.equals("Show Console")){
         getBeanShell();
         shellFrame.show();
      } else if (arg.equals("Reset Panes")){
         resizePanes(0);
      } else if (arg.equals("Java Metal Look & Feel")){
         try{
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
            SwingUtilities.updateComponentTreeUI(SwingUtilities.getRoot(this));
         } catch (Exception e){
            System.out.println("Can't set look and feel: " + e);
         }
      } else if (arg.equals("Native")){
         try{
            String osname = System.getProperty("os.name", "unknown");
            if (osname.toLowerCase().startsWith("l")){
               UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
            } else{
               UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            }
            SwingUtilities.updateComponentTreeUI(SwingUtilities.getRoot(this));
         } catch (Exception e){
            System.out.println("Can't set look and feel: " + e);
         }
      } else if (arg.equals("Split Right")){
         resizePanes(20);
      } else if (arg.equals("Split Left")){
         resizePanes( -20);
      } else if (arg.equals("Garbage Collect")){
         Runtime.getRuntime().gc();
      } else if (arg.equals("About...")){
         about();
      } else{
         System.out.println("" + arg.toString());
      }
   }

   private JDialog jdabout = null;
   public void about(){
      if (jdabout == null){
         jdabout = new JDialog(frame, "About... ", true);
         URL tinkerURL = getClass().getClassLoader().getResource("ffe/icons/tinker.jpg");
         ImageIcon i = new ImageIcon(tinkerURL);
         JLabel jltinker = new JLabel(i);
         jltinker.setBorder(new EtchedBorder(EtchedBorder.RAISED));
         Container contentpane = jdabout.getContentPane();
         contentpane.setLayout(new BorderLayout());
         if (help == null){
            initHelp();
         }
         contentpane.add(help, BorderLayout.SOUTH);
         contentpane.add(jltinker, BorderLayout.CENTER);
         contentpane.setBackground(Color.white);
         jdabout.pack();
         Dimension dim = getToolkit().getScreenSize();
         Dimension ddim = jdabout.getSize();
         jdabout.setLocation((dim.width - ddim.width) / 2, (dim.height - ddim.height) / 2);
         jdabout.setResizable(false);
      }
      jdabout.show();
   }

   private boolean jobPanelLazy = true;
   private boolean keyPanelLazy = true;
   public void stateChanged(ChangeEvent evt){
      JTabbedPane jtp = (JTabbedPane) evt.getSource();
      int index = jtp.getSelectedIndex();
      if (index == 0){
         graphics.selected();
      } else if (index == 1){
         if (keyPanelLazy){
            keyPanel.initLazy();
            keyPanelLazy = false;
         }
         keyPanel.selected();
      } else if (index == 2){
         if (jobPanelLazy){
            jobPanel.initLazy();
            jobPanelLazy = false;
         }
         jobPanel.selected();
      } else if (index == 3){
         logPanel.selected();
      }
   }

   /**
    * Detach the active FSystem's BranchGroup from the Scene and clear that FSystem's data
    */
   public void closeActive(){
      FSystem m = tree.getActive();
      if (m != null && m.getParent() != null){
         m.switchView(GraphicsContext.DESTROY);
         tree.removeTreeNode(m);
         tree.removeFSystem(m);
         keyPanel.keyClose();
         Trajectory traj = m.getTrajectory();
         if (traj != null){
            traj.stop();
         }
         if (simulation != null && simulation.getFSystem() == m){
            release();
         }
         CloseFile cf = new CloseFile(m, this, graphics);
         Thread t = new Thread(cf);
         t.setPriority(Thread.MIN_PRIORITY);
         t.start();
      }
      tree.setActive(0);
   }

   /**
    * Close all open systems
    */
   public void closeAll(){
      while (tree.getActive() != null){
         closeActive();
      }
   }

   /**
    * Attempt to connect to a TINKER Simulation
    */
   private void connect(){
      if (simulation == null || simulation.isFinished()){
         if (simulation != null){
            simulation.release();
         }
         simulation = new Simulation(null, null, this, socketAddress);
         simulation.connect();
         menu.setConnect(false);
         setPanel(GRAPHICS);
      }
   }

   public void openRestart(){
      FSystem active = tree.getActive();
      JFileChooser fileChooser = getFileChooser();
      fileChooser.setCurrentDirectory(cwd);
      fileChooser.setSelectedFile(tree.getActive().getFile());
      fileChooser.setDialogTitle("Choose Restart File");
      fileChooser.addChoosableFileFilter(dynFileFilter);
      fileChooser.setAcceptAllFileFilterUsed(true);
      int result = fileChooser.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         File f = fileChooser.getSelectedFile();
         DYNFilter dynFilter = new DYNFilter(active, f);
         dynFilter.read();
      }
   }

   private GraphicsDialog sgd = null;
   public void preferences(){
      if (sgd == null){
         sgd = new GraphicsDialog(frame, root);
         sgd.setModal(false);
      }
      sgd.show();
   }

   public void connectToTINKER(FSystem m, Thread t){
      if (simulation == null || simulation.isFinished()){
         if (simulation != null){
            simulation.release();
         }
         InetSocketAddress tempAddress = null;
         try{
            tempAddress = new InetSocketAddress(InetAddress.getLocalHost(), port);
         } catch (Exception e){
            try{
               tempAddress = new InetSocketAddress(InetAddress.getByName(null), port);
            } catch (Exception ex){
               System.out.println("Could not determine Local Host: " + ex);
               return;
            }
         }
         if (tempAddress == null){
            return;
         }
         simulation = new Simulation(m, t, this, tempAddress);
         if (t != null){
            t.start();
         }
         simulation.connect();
         menu.setConnect(false);
         setPanel(GRAPHICS);
      }
   }

   /**
    * Get the Menu Bar
    * @return menu
    */
   public FMenuBar getMenu(){
      return menu;
   }

   public Tree getTree(){
      return tree;
   }

   /**
    * Returns the molecular root node
    */
   public Root getRoot(){
      return root;
   }

   public JLabel getStatusBar(){
      return status;
   }

   private Interpreter beanshell;
   public Interpreter getBeanShell(){
      if (beanshell == null){
         jconsole = new JConsole();
         shellFrame = new JFrame("Force Field Explorer Shell");
         beanshell = new Interpreter(jconsole, shellFrame);
         Thread t = new Thread(beanshell);
         URL iconURL = getClass().getClassLoader().getResource("ffe/icons/icon.jpg");
         ImageIcon icon = new ImageIcon(iconURL);
         shellFrame.setIconImage(icon.getImage());
         shellFrame.getContentPane().add(jconsole);
         shellFrame.pack();
         jconsole.loadprefs();
         t.start();
         try{
            beanshell.getNameSpace().setVariable("gui", this, false);
            beanshell.getNameSpace().setVariable("data", root, false);
            beanshell.getNameSpace().setVariable("graphics", graphics, false);
         } catch (Exception e){
            System.out.println("" + e);
         }
      }
      return beanshell;
   }

   public Graphics3D getGraphics3D(){
      return graphics;
   }

   public KeywordPanel getKeywordPanel(){
      return keyPanel;
   }

   public JobPanel getJobPanel(){
      return jobPanel;
   }

   public LogPanel getLogPanel(){
      return logPanel;
   }

   public static File getCWD(){
      if (cwd == null){
         cwd = new File(System.getProperty("user.dir", ""));
      }
      return cwd;
   }

   public void setCWD(File f){
      if ((f == null) || (!f.exists())){
         return;
      }
      cwd = f;
   }

   /**
    * Get the Trajectory wrapper for the active system
    * @return trajectory
    */
   public Trajectory getTrajectory(){
      FSystem sys = tree.getActive();
      if (sys == null){
         return null;
      }
      if (sys.getFileType() != SystemFilter.ARC){
         return null;
      }

      Trajectory traj = sys.getTrajectory();
      if (traj != null){
         return traj;
      }
      traj = new Trajectory(sys, this);
      traj.setOscillate(oscillate);
      sys.setTrajectory(traj);
      return traj;
   }

   public void openInduced(){
      FSystem active = tree.getActive();
      JFileChooser fileChooser = getFileChooser();
      fileChooser.setCurrentDirectory(cwd);
      fileChooser.setSelectedFile(active.getFile());
      fileChooser.setDialogTitle("Choose Induced Dipole File");
      fileChooser.addChoosableFileFilter(indFileFilter);
      fileChooser.setAcceptAllFileFilterUsed(true);
      fileChooser.setFileFilter(indFileFilter);
      int result = fileChooser.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         File f = fileChooser.getSelectedFile();
         InducedFilter indFilter = new InducedFilter(active, f);
         indFilter.read();
      }
   }

   /**
    * Initialize all the sub-Panels and put them together
    */
   private boolean init = false;
   public void initialize(){
      if (init){
         return;
      }
      init = true;

      // Splash Screen
      JFrame.setDefaultLookAndFeelDecorated(true);
      JDialog d = new JDialog(frame, "Force Field Explorer - Starting Up...", false);
      ClassLoader loader = getClass().getClassLoader();
      ImageIcon i = new ImageIcon(loader.getResource("ffe/icons/tinker.jpg"));
      JLabel jltinker = new JLabel(i);
      jltinker.setBorder(new EtchedBorder(EtchedBorder.RAISED));
      Container contentpane = d.getContentPane();
      contentpane.setLayout(new BorderLayout());
      contentpane.add(jltinker, BorderLayout.CENTER);
      d.setUndecorated(true);
      d.pack();
      Dimension dim = getToolkit().getScreenSize();
      Dimension ddim = d.getSize();
      d.setLocation((dim.width - ddim.width) / 2, (dim.height - ddim.height) / 2);
      d.setResizable(false);
      d.setDefaultCloseOperation(d.DISPOSE_ON_CLOSE);
      d.show();

      // Make all pop-up Menus Heavyweight so they play nicely with Java3D
      JPopupMenu.setDefaultLightWeightPopupEnabled(false);

      // Create the Root Node
      root = new Root(this);
      EtchedBorder bb = new EtchedBorder(EtchedBorder.RAISED);

      status = new JLabel("  ");
      JLabel step = new JLabel("  ");
      step.setHorizontalAlignment(JLabel.RIGHT);
      JLabel energy = new JLabel("  ");
      energy.setHorizontalAlignment(JLabel.RIGHT);
      JPanel statusPanel = new JPanel(new GridLayout(1, 3));
      statusPanel.setBorder(bb);
      statusPanel.add(status);
      statusPanel.add(step);
      statusPanel.add(energy);

      GraphicsConfigTemplate3D template3D = new GraphicsConfigTemplate3D();
      template3D.setDoubleBuffer(GraphicsConfigTemplate.PREFERRED);
      GraphicsConfiguration gc = GraphicsEnvironment.getLocalGraphicsEnvironment().
                                 getDefaultScreenDevice().getBestConfiguration(template3D);
      graphics = new Graphics3D(gc, this);

      graphPanel = new GraphicsPanel(graphics, statusPanel);

      // Initialize various Panels
      tree = new Tree(this);
      tree.setStatus(status, step, energy);
      keyPanel = new KeywordPanel(this);
      jobPanel = new JobPanel(this);
      logPanel = new LogPanel(this);
      JPanel treePane = new JPanel(new BorderLayout());
      JScrollPane scrollPane = new JScrollPane(tree);
      treePane.add(scrollPane, BorderLayout.CENTER);
      treePane.add(logPanel.getProgressBar(), BorderLayout.SOUTH);

      ImageIcon icdisplay = new ImageIcon(loader.getResource("ffe/icons/display.png"));
      ImageIcon ickey = new ImageIcon(loader.getResource("ffe/icons/keywords.png"));
      ImageIcon icexec = new ImageIcon(loader.getResource("ffe/icons/commands.png"));
      ImageIcon icdata = new ImageIcon(loader.getResource("ffe/icons/data.png"));

      // Put everything together
      tabbedPane = new JTabbedPane();
      tabbedPane.addTab("Graphics", icdisplay, graphPanel);
      tabbedPane.addTab("Keyword Editor", ickey, keyPanel);
      tabbedPane.addTab("Modeling Commands", icexec, jobPanel);
      tabbedPane.addTab("Logs", icdata, logPanel);
      tabbedPane.addChangeListener(this);

      splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, false, treePane, tabbedPane);
      splitPane.setResizeWeight(0.25);
      splitPane.setOneTouchExpandable(true);
      setLayout(new BorderLayout());
      add(splitPane, BorderLayout.CENTER);
      menu = new FMenuBar(this);
      add(menu.getToolBar(), BorderLayout.NORTH);
      loadPrefs();
      SwingUtilities.updateComponentTreeUI(SwingUtilities.getRoot(this));
      d.dispose();
   }

   /**
    * Attempt to open a TINKER *.key file
    * @param f XYZ File that needs an associated Key File
    * @return Key file that was found, or null if nothing could be found
    */
   public boolean openKey(FSystem newSystem, boolean createKey){
      String keyFileName = newSystem.getName() + ".key";
      String path = null;
      File keyfile = null;

      if (cwd != null){
         path = cwd.getAbsolutePath() + File.separator;
      }

      keyfile = new File(path + keyFileName);
      if (keyfile.exists()){
         newSystem.setKeyFile(keyfile);
         KeyFilter.open(newSystem);
         newSystem.setForceField(null);
         return true;
      }

      keyfile = new File(path + "tinker.key");
      if (keyfile.exists()){
         newSystem.setKeyFile(keyfile);
         KeyFilter.open(newSystem);
         newSystem.setForceField(null);
         return true;
      }

      if (createKey){
         return createKeyFile(newSystem);
      }
      return false;
   }

   public void exit(){
      savePrefs();
      System.exit(0);
   }

   public boolean createKeyFile(FSystem system){
      if (jobPanelLazy) {
         jobPanel.initLazy();
         jobPanelLazy = false;
      }
      if (keyPanelLazy){
         keyPanel.initLazy();
         keyPanelLazy = false;
      }

      String message = new String("Please select a parameter file " +
                       "and a TINKER Key file will be created");
      String params = (String) JOptionPane.showInputDialog(this, message,
                      "A TINKER Key file is needed", JOptionPane.QUESTION_MESSAGE, null,
                      keyPanel.getParamFiles(), null);
      if (params != null){
         if (params.equalsIgnoreCase("Use an existing TINKER Key file")){
            JFileChooser fileChooser = getFileChooser();
            fileChooser.setDialogTitle("Choose a KEY File");
            fileChooser.setCurrentDirectory(cwd);
            fileChooser.setSelectedFile(null);
            fileChooser.setFileFilter(keyfilefilter);
            int result = fileChooser.showOpenDialog(this);
            if (result == JFileChooser.APPROVE_OPTION){
               File keyfile = fileChooser.getSelectedFile();
               if (keyfile.exists()){
                  system.setKeyFile(keyfile);
                  KeyFilter.open(system);
                  system.setForceField(null);
                  return true;
               }
            }
         } else{
            File tempFile = system.getFile();
            if (!tempFile.exists()){
               return false;
            }
            if (tempFile.getParentFile().canWrite()){
               String path = system.getFile().getParent() + File.separatorChar;
               String keyFileName = system.getName() + ".key";
               File keyfile = new File(path + keyFileName);

               try{
                  FileWriter fw = new FileWriter(keyfile);
                  BufferedWriter bw = new BufferedWriter(fw);
                  bw.write("\n");
                  bw.write("# Force Field Selection\n");
                  String tempParm = keyPanel.getParamPath(params);
                  if (tempParm.indexOf(" ") > 0){
                     tempParm = "\"" + keyPanel.getParamPath(params) + "\"";
                  }
                  bw.write("PARAMETERS        " + tempParm + "\n");
                  bw.close();
                  fw.close();
                  system.setKeyFile(keyfile);
                  KeyFilter.open(system);
                  system.setForceField(null);
                  return true;
               } catch (Exception e){
                  System.out.println("" + e);
                  message = new String("There was an error creating " + keyfile.getAbsolutePath());
                  JOptionPane.showMessageDialog(this, message);

               }
            } else{
               message = new String("Could not create a Key file because " + cwd.getAbsolutePath() +
                         " is not writable");
               JOptionPane.showMessageDialog(this, message);
            }
         }
      }
      return false;
   }

   /**
    * Load preferences from the user node
    */
   public void loadPrefs(){
      JFrame frame = (JFrame) SwingUtilities.getRoot(this);
      Toolkit toolkit = getToolkit();
      Dimension winSize = toolkit.getScreenSize();
      Preferences prefs = Preferences.userNodeForPackage(MainPanel.class);
      int x = prefs.getInt("MainPanel_x", winSize.width / 8);
      int y = prefs.getInt("MainPanel_y", winSize.height / 8);
      int width = prefs.getInt("MainPanel_width", winSize.width * 3 / 4);
      int height = prefs.getInt("MainPanel_height", winSize.height * 3 / 4);
      if (width < winSize.width * 4 / 5 && height < winSize.height * 4 / 5){
         frame.setSize(width, height);
      } else{
         frame.setSize(winSize.width * 4 / 5, winSize.height * 4 / 5);
      }

      if (x > 0 && x < winSize.width / 2 && y > 0 && y < winSize.height / 2){
         frame.setLocation(x, y);
      } else{
         frame.setLocation(winSize.width / 8, winSize.height / 8);
      }

      spDivider = prefs.getInt("MainPanel_divider", 200);
      if (spDivider < frame.getWidth() * (1.0f / 4.0f)){
         spDivider = (int) (frame.getWidth() * (1.0f / 4.0f));
      }
      splitPane.setDividerLocation(spDivider);
      if (!prefs.getBoolean("MainPanel_system", true)){
         menu.setSystemShowing(false);
         splitPane.setDividerLocation(0);
      } else{
         menu.setSystemShowing(true);
      }
      if (!prefs.getBoolean("MainPanel_menu", true)){
         remove(menu.getToolBar());
         menu.setMenuShowing(false);
         validate();
      } else{
         menu.setMenuShowing(true);
      }

      /*
             if (!prefs.getBoolean("MainPanel_axis", false)){
         menu.setAxisShowing(false);
         graphPanel.setAxis(false);
             } else{
         menu.setAxisShowing(true);
         graphPanel.setAxis(true);
             }*/

      try{
         port = prefs.getInt("MainPanel_port", 2000);
         ip = prefs.get("MainPanel_ip", InetAddress.getLocalHost().getHostAddress());
         if (ip != null){
            address = InetAddress.getByName(ip);
            socketAddress = new InetSocketAddress(address, port);
         }
         socketAddress = new InetSocketAddress(port);
      } catch (Exception e){
      }
      String dir = prefs.get("MainPanel_cwd", "");
      cwd = new File(dir);
      graphics.loadPrefs();
   }

   /**
    * Merge two or more selected FSystem Nodes into one FSystem node. There are a few gotchas that
    * need to be fixed
    */
   public void merge(){
      ArrayList anodes = tree.getActiveNodes();
      if (anodes.size() < 2){
         return;
      }
      Object[] nodes = anodes.toArray();
      if (nodes == null || nodes.length < 2){
         return;
      }
      FSystem s;
      for (int i = 0; i < nodes.length; i++){
         if (!(nodes[i] instanceof FSystem)){
            return;
         } else{
            s = (FSystem) nodes[i];
            if (s.getFileType() != SystemFilter.XYZ && s.getFileType() != SystemFilter.INT){
               return;
            }
         }
      }
      Atom a;
      double d[] = new double[3];
      Vector3d atompos = new Vector3d();
      Vector3d syspos = new Vector3d();
      Transform3D t3dm = new Transform3D();
      TransformGroup tgm;

      FSystem merge = (FSystem) nodes[0];
      int atomnum = merge.getAtomList().size();

      for (int i = 0; i < nodes.length; i++){
         s = (FSystem) nodes[i];
         graphics.remove(s);
         s.removeJ3D();
         tgm = s.getTransformGroup();
         tgm.getTransform(t3dm);
         t3dm.get(syspos);
         t3dm.setScale(1.0d);
         //t3dm.setTranslation(new Vector3d(0, 0, 0));
         for (ListIterator li = s.getAtomList().listIterator(); li.hasNext(); ){
            a = (Atom) li.next();
            if (i > 0){
               a.setXYZIndex(++atomnum);
            }
            a.getV3D(atompos);
            t3dm.transform(atompos);
            atompos.add(syspos);
            atompos.get(d);
            a.moveto(d);
         }
      }

      merge.getTransformGroup().setTransform(new Transform3D());
      for (int i = 1; i < nodes.length; i++){
         s = (FSystem) nodes[i];
         tree.removeFSystem(s);
         ArrayList polymers = s.getFGroupList();
         int len = merge.getFGroupList().size();
         Polymer c;
         for (ListIterator li = polymers.listIterator(); li.hasNext(); ){
            c = (Polymer) li.next();
            c.setName("" + XYZUtils.PolymerLookup(len++));
            merge.addFNode(c);
         }
      }
      merge.finalize(false);
      update(false);
   }

   /**
    * Trys to open a file picked from a JFileChooser
    * A successful read is followed by attachment of the FSystem to the root node
    * and the update of the various displays
    */
   public void open(){
      if (openThread != null && openThread.isAlive()){
         return;
      }
      JFileChooser fileChooser = getFileChooser();
      fileChooser.setDialogTitle("Choose XYZ/Archive File");
      fileChooser.addChoosableFileFilter(xyzFileFilter);
      fileChooser.addChoosableFileFilter(pdbFileFilter);
      fileChooser.addChoosableFileFilter(intFileFilter);
      fileChooser.addChoosableFileFilter(arcFileFilter);
      fileChooser.setAcceptAllFileFilterUsed(true);
      int result = fileChooser.showOpenDialog(this);
      if (result == JFileChooser.APPROVE_OPTION){
         File f = fileChooser.getSelectedFile();
         open(f);
      }
   }

   public void openOn(File f, FSystem oldSystem){
      if (oldSystem.getFileType() == SystemFilter.XYZ){
         XYZFilter.readOnto(f, oldSystem);
      } else if (oldSystem.getFileType() == SystemFilter.INT){
         return;
      }
      graphics.updateScene(oldSystem, true, false, 0, false, 0);
      getTree().updateStatus();
      getTree().repaint();
   }

   /**
    * Opens the supplied File
    * @param f File to open
    */
   public void open(File f){
      if (f != null && f.exists()){
         cwd = f.getParentFile();
         // Initialize a FSystem for the new System
         int dot = f.getName().lastIndexOf(".");
         String fileName = f.getName().substring(0, dot);

         SystemFilter molfile;
         try{
            FileReader fr = new FileReader(f);
            BufferedReader br = new BufferedReader(fr);
            if (!br.ready()){
               graphics.getStatusBar().setText("  Could not create BufferedReader for: " + f.getName());
               return;
            }

            // If the first thing read isn't an integer, the PDB parser will be used
            boolean isTINKER = true;
            String rawdata = br.readLine();
            String header[] = rawdata.trim().split(" +");
            if (header != null && header.length > 0){
               try {
                  int numatoms = Integer.parseInt(header[0]);
               } catch (Exception e){
                  isTINKER = false;
               }
            }

            // PDB File
            FSystem newSystem = null;
            if (!isTINKER){
               newSystem = new FSystem(fileName, false);
               newSystem.setFile(f);
               newSystem.setFileType(SystemFilter.PDB);
               openKey(newSystem, false);
               molfile = new PDBFilter(newSystem, this);
            } else {
               newSystem = new FSystem(fileName, true);
               newSystem.setFile(f);
               if (!openKey(newSystem, true)){
                  graphics.getStatusBar().setText("  No Key File for : " + f.getName());
                  return;
               }

               if (!br.ready()){
                  graphics.getStatusBar().setText("  Could not open: " + f.getName());
                  return;
               }
               String firstAtom = br.readLine();
               if (firstAtom == null){
                  graphics.getStatusBar().setText("  Could not open: " + f.getName());
                  return;
               }
               br.close();
               fr.close();

               String data[] = firstAtom.trim().split(" +");
               if (data == null || data.length < 2){
                  return;
               }
               try{
                  int i = Integer.parseInt(data[0]);
               } catch (Exception e){
                  graphics.getStatusBar().setText("  Could not open: " + f.getName());
                  return;
               }

               // TINKER Cartesian Coordianates
               if (data.length > 3){
                  //System.out.println("TINKER Cartesian Coordinates");
                  newSystem.setFileType(SystemFilter.XYZ);
                  molfile = new XYZFilter(newSystem, mmfilter, this);
               }
               // TINKER Internal Coordinates
               else{
                  //System.out.println("TINKER Internal Coordinates");
                  newSystem.setFileType(SystemFilter.INT);
                  molfile = new INTFilter(newSystem, mmfilter, this);
               }
            }
         } catch (IOException e){
            graphics.getStatusBar().setText("  Could not open: " + f.getName());
            return;
         }

         setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         openThread = new Thread(molfile);
         openThread.start();
         setPanel(GRAPHICS);
      }
   }

   public void open(String name){
      if (name == null){
         return;
      }
      File f = new File(name);
      open(f);
   }

   /**
    * Opens the supplied File
    * @param f File to open
    */
   public void openFromPDB(){
      if (openThread != null && openThread.isAlive()){
         return;
      }

      String code = JOptionPane.showInputDialog("Enter the PDB Identifier (4 characters)", "");
      if (code == null || code.length() != 4){
         return;
      }

      String pdbAddress = "http://www.rcsb.org/pdb/cgi/export.cgi/" + code + ".pdb?pdbId=" + code +
                          ";format=PDB";

      String vrmlAddress = "http://www.rcsb.org/pdb/cgi/render.cgi/" + code +
                           ".wrl?format=vrml_default&pdbId=" + code;

      try{
         FSystem newSystem = new FSystem(code, false);
         // Get the PDB File
         PDBFilter pdbFilter = new PDBFilter(newSystem, pdbAddress, vrmlAddress, this);
         setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         openThread = new Thread(pdbFilter);
         openThread.start();
         setPanel(GRAPHICS);
      } catch (Exception e){
         return;
      }

   }

   /**
    * Close the connection to a running simulation
    */
   public void release(){
      if (simulation != null){
         simulation.release();
         simulation = null;
         menu.setConnect(true);
      }
   }

   /**
    * Set the split panes to their default proportions
    */
   public void resizePanes(int move){
      if (move == 0){
         spDivider = 0;
         menu.setMenuShowing(false);
         menu.menuClick();
         menu.setSystemShowing(false);
         menu.systemClick();
      } else{
         splitPane.setDividerLocation(splitPane.getDividerLocation() + move);
      }
   }

   /**
    * Save the currently selected System to disk
    * @param file File to save the system to
    */
   public void save(File file){
      FSystem m = tree.getActive();
      if (m != null && !m.isClosing()){
         SystemFilter filter;
         if (m.getFileType() == SystemFilter.XYZ){
            filter = new XYZFilter();
         } else{
            return;
         }

         File savefile = null;
         if (file != null){
            savefile = file;
         } else{
            JFileChooser fileChooser = getFileChooser();
            fileChooser.setCurrentDirectory(cwd);
            fileChooser.setAcceptAllFileFilterUsed(true);
            int result = fileChooser.showSaveDialog(this);
            if (result == JFileChooser.APPROVE_OPTION){
               savefile = fileChooser.getSelectedFile();
               cwd = savefile.getParentFile();
            }
         }

         if (savefile != null){
            filter.setFSystem(m);
            m.setFile(savefile);
            m.setName(savefile.getName());
            filter.writefile();
         }
      }
   }

   /**
    * Save preferences to the user node
    */
   public void savePrefs(){
      Preferences prefs = Preferences.userNodeForPackage(MainPanel.class);
      JFrame frame = (JFrame) SwingUtilities.getRoot(this);
      prefs.putInt("MainPanel_x", frame.getLocation().x);
      prefs.putInt("MainPanel_y", frame.getLocation().y);
      prefs.putInt("MainPanel_width", frame.getWidth());
      prefs.putInt("MainPanel_height", frame.getHeight());
      prefs.putBoolean("MainPanel_system", menu.isSystemShowing());
      prefs.putInt("MainPanel_divider", splitPane.getDividerLocation());
      prefs.putBoolean("MainPanel_menu", menu.isMenuShowing());
      prefs.putBoolean("MainPanel_axis", menu.isAxisShowing());
      if (ip == null){
         ip = new String("");
      }
      if (address != null){
         String s = address.getHostAddress();
         if (s != null){
            prefs.put("MainPanel_ip", s);
         }
         prefs.putInt("MainPanel_port", socketAddress.getPort());
      }

      prefs.put("MainPanel_cwd", cwd.toString());

      jobPanel.savePrefs();
      keyPanel.savePrefs();
      if (graphics != null){
         graphics.savePrefs();
      }
      if (jconsole != null){
         jconsole.savePrefs();
      }
   }

   public void setPanel(int panel){
      tabbedPane.setSelectedIndex(panel);
   }

   public static final int GRAPHICS = 0;
   public static final int KEYWORDS = 1;
   public static final int MODELING = 2;
   public static final int LOGS = 3;

   /**
    * Driver to test functionality
    */
   private void test(){
      File f = new File(".");
      File testfiles[] = f.listFiles();
      FSystem system;

      GraphicsContext.detail = 3;

      JWindow jwindow = new JWindow();
      jwindow.getContentPane().add(graphics);
      graphicsDevice = getGraphicsConfiguration().getDevice();
      try{
         graphicsDevice.setFullScreenWindow(jwindow);
         for (int i = 0; i < testfiles.length; i++){
            if (xyzFileFilter.accept(testfiles[i])){
               System.out.println("" + testfiles[i].toString());
               open(testfiles[i]);
               synchronized (this){
                  while (openThread != null && openThread.isAlive()){
                     try{
                        wait(500);
                     } catch (Exception e){
                        System.out.println("" + e);
                     }
                  }
               }
               system = tree.getActive();
               if (system != null){
                  graphics.updateScene(system, false, true, GraphicsContext.SPACEFILL, false, 0);
                  synchronized (this){
                     while (graphics.sceneRendering()){
                        try{
                           wait(500);
                        } catch (Exception e){
                           System.out.println("" + e);
                        }
                     }
                  }

                  graphics.updateScene(system, false, true, GraphicsContext.TUBE, true,
                     GraphicsContext.byCHARGE);

                  synchronized (this){
                     while (graphics.sceneRendering()){
                        try{
                           wait(500);
                        } catch (Exception e){
                           System.out.println("" + e);
                        }
                     }
                  }

                  graphics.updateScene(system, false, true, GraphicsContext.TUBE, true,
                     GraphicsContext.byGROUP);
                  synchronized (this){
                     while (graphics.sceneRendering()){
                        try{
                           wait(500);
                        } catch (Exception e){
                           System.out.println("" + e);
                        }
                     }
                  }
                  if (system.getKeyFile() != null){
                     keyPanel.keySave(new File("" + system.getKeyFile().getName() + ".test"));
                  }
                  save(new File("" + system.getName() + ".test"));
                  closeActive();
               }
            }
         }
      } catch (Exception e){
         System.out.println("" + e);
      } finally{
         graphicsDevice.setFullScreenWindow(null);
      }
   }

   /**
    * Refresh various Panels based on the active System
    * @param updateRender If true, the default rendering is attempted
    */
   public void update(boolean updateRender){
      FSystem active = tree.getActive();
      if (active == null){
         return;
      }

      if (active != null){
         while (graphics == null){
            try{
               wait(100);
            } catch (Exception e){
               System.out.println("" + e);
            }
         }
         graphics.Render(active, updateRender);
      }
      tree.addTreeNode(active);
      tree.onlySelection(active);
      tree.updateStatus();
   }

   /**
    * JColorChooser for reuse
    */
   private static JColorChooser colorChooser = null;
   public static JColorChooser getColorChooser(){
      if (colorChooser == null){
         colorChooser = new JColorChooser();
      }
      return colorChooser;
   }

   /**
    * JFileChooser for reuse
    */
   private static JFileChooser chooser;
   private static File cwd;
   public static JFileChooser getFileChooser(){
      if (chooser == null){
         chooser = new JFileChooser();
      }
      chooser.resetChoosableFileFilters();
      chooser.setFileSelectionMode(chooser.FILES_ONLY);
      chooser.setFileHidingEnabled(false);
      chooser.setAcceptAllFileFilterUsed(true);
      chooser.setCurrentDirectory(getCWD());
      chooser.setMultiSelectionEnabled(false);
      chooser.setSelectedFile(null);
      return chooser;
   }

   /**
    * Currently used for testing full screen graphics
    */
   public static GraphicsDevice graphicsDevice;

   /**
    * KeyFileFilter for reuse
    */
   public static KeyFileFilter keyfilefilter = new KeyFileFilter();

   /**
    * Reference to the Keyword Editor
    */
   public KeywordPanel keyPanel;

   /**
    * Facilitates playback control of TINKER trajectory (*.arc) files
    */
   private boolean oscillate = false;

   /**
    * XYZFileFilter for reuse
    */
   public static XYZFileFilter xyzFileFilter = new XYZFileFilter();
   public static ARCFileFilter arcFileFilter = new ARCFileFilter();
   public static INTFileFilter intFileFilter = new INTFileFilter();
   public static DYNFileFilter dynFileFilter = new DYNFileFilter();
   public static InducedFileFilter indFileFilter = new InducedFileFilter();
   public static PRMFileFilter prmFileFilter = new PRMFileFilter();
   public static PDBFileFilter pdbFileFilter = new PDBFileFilter();

   /**
    * Force Field Explorer Application Frame
    */
   private JFrame frame;

   /**
    * Java3D Canvas
    */
   public Graphics3D graphics;

   /**
    * Panel containing the Graphics window
    */
   private GraphicsPanel graphPanel;
   public FullScreenWindow fullScreenWindow;
   private JLabel status;

   /**
    * Where to email for help/to submit bugs
    */
   private static JTextArea help = null;
   private static void initHelp(){
      help = new JTextArea(" Send FFE questions, bugs and suggestions to: \n" +
             " schnied@dasher.wustl.edu\n" + " Send TINKER questions, bugs and suggestions to: \n" +
             " ponder@dasher.wustl.edu\n" +
             " Copyright (c) Jay William Ponder 2004, All rights reserved");
      EtchedBorder bb = new EtchedBorder(EtchedBorder.RAISED);
      help.setBorder(new TitledBorder(bb, "Feedback", TitledBorder.LEFT, TitledBorder.ABOVE_TOP));
      help.setEditable(false);
   }

   /**
    * This JSplitPane holds the Graphics and Beanshell Interpreter
    */
   private JFrame shellFrame;
   private JConsole jconsole;
   private String ip = new String("");
   private int port = 2000;
   private InetAddress address = null;
   private InetSocketAddress socketAddress = new InetSocketAddress(port);

   /**
    * Modeling Window
    */
   private JobPanel jobPanel;

   /**
    * The menu
    */
   private FMenuBar menu;

   /**
    * Parses TINKER parameter files
    */
   private MMFilter mmfilter;

   /**
    * Systems are opened in their own thread so the GUI doesn't become unresponsive.
    * Only one system can be opened at a time, and is referenced by openThread
    */
   private Thread openThread = null;

   /**
    * This JSplitPane holds the innerPane and 3D Graphics
    */
   private JTabbedPane tabbedPane;
   private JSplitPane splitPane;
   private int spDivider;

   /**
    * Result window
    */
   private LogPanel logPanel;

   /**
    * Root where FSystems are attached
    */
   private Root root;

   /**
    * Encapsulates a connected simulation
    */
   private Simulation simulation;

   /**
    * The JTree
    */
   private Tree tree;
}
