/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.panels;

import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import ffe.lang.*;

/**
 * The Tree Class creates and manages a JTree view of the data. It also
 * handles the selection mechanism and keeps track of the active system and nodes.
 */
public final class Tree
   extends JTree
   implements TreeSelectionListener{

   private Root root;
   private MainPanel fpanel;
   private DefaultTreeModel treeModel;
   private DefaultTreeSelectionModel treeSelectionModel;

   private FSystem active = null; // Reference to the active FSystem
   private FNode activeNode = null; // Reference to the last Selected Node
   private ArrayList activeNodes = new ArrayList();

   private JLabel status = null;
   private JLabel step = null;
   private JLabel energy = null;

   /**
    * The Default Constructor initializes JTree properties, then updates is representation
    * based on the Struture of the Tree that extends from the Root arguement.
    */
   public Tree(MainPanel f){
      fpanel = f;
      root = fpanel.getRoot();
      initTree();
   }

   /**
    * Initialize the Tree representation based on the Root data node
    */
   public void initTree(){
      addTreeSelectionListener(this);
      //setExpandsSelectedPaths(true);
      //setScrollsOnExpand(true);
      setLargeModel(true);
      setEditable(false);
      putClientProperty("JTree.lineStyle", "Angled");
      setShowsRootHandles(true);
      DefaultTreeCellRenderer tcr = new DefaultTreeCellRenderer();
      tcr.setClosedIcon(new ImageIcon(getClass().getClassLoader().getResource(
         "ffe/icons/folder.png")));
      tcr.setOpenIcon(new ImageIcon(getClass().getClassLoader().getResource(
         "ffe/icons/folder_open.png")));
      setCellRenderer(tcr);

      treeModel = new DefaultTreeModel(root);
      treeSelectionModel = new DefaultTreeSelectionModel();
      setModel(treeModel);
      setSelectionModel(treeSelectionModel);
      setVisible(false);
   }

   public void addTreeNode(FNode nodeToAdd){
      if (nodeToAdd == null){
         return;
      }
      synchronized (this){
         DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) treeModel.getRoot();
         int index = rootNode.getChildCount();
         treeModel.insertNodeInto(nodeToAdd, rootNode, index - 1);
         addSelection(nodeToAdd);
         setVisible(true);
      }
   }

   public void removeTreeNode(FNode nodeToRemove){
      if (nodeToRemove == null){
         return;
      }
      synchronized (this){
         treeModel.removeNodeFromParent(nodeToRemove);
         if (root.getChildCount() <= 0){
            setVisible(false);
         }
      }
   }

   public FSystem[] getSystems(){
      synchronized (this){
         FSystem[] systems = new FSystem[root.getChildCount()];
         int index = 0;
         for (Enumeration e = root.children(); e.hasMoreElements(); ){
            systems[index++] = (FSystem) e.nextElement();
         }
         return systems;
      }
   }

   public DefaultTreeModel getTreeModel(){
      return treeModel;
   }

   public DefaultTreeSelectionModel getTreeSelectionModel(){
      return treeSelectionModel;
   }

   public void setHighlighting(boolean h){
      synchronized (this){
         if (GraphicsContext.highlightSelections != h){
            GraphicsContext.highlightSelections = h;
            for (ListIterator li = activeNodes.listIterator(); li.hasNext(); ){
               FNode node = (FNode) li.next();
               node.setSelected(h);
            }
            fpanel.getGraphics3D().updateScene(activeNodes, false, false, 0, true,
               GraphicsContext.bySELECT);
         }
      }
   }

   /**
    * Returns the active FSystem.
    */
   public FSystem getActive(){
      return active;
   }

   public FNode getActiveNode(){
      return activeNode;
   }

   public ArrayList getActiveNodes(){
      return activeNodes;
   }

   /**
    * Sets the FSystem s to be active.
    */
   public void setActive(FSystem s){
      synchronized (this){
         active = s;
         updateStatus();
         if (active == null){
            return;
         }
         if (fpanel.getKeywordPanel() != null){
            fpanel.getKeywordPanel().loadActive(active);
         }
         if (fpanel.getJobPanel() != null){
            fpanel.getJobPanel().loadActive(active);
         }
      }
   }

   public void setActive(int i){
      synchronized (this){
         if (i < root.getChildCount()){
            setActive((FSystem) root.getChildAt(i));
         } else if (root.getChildCount() == 0){
            setActive(null);
         }
      }
   }

   public void updateStatus(){
      if (active == null){
         status.setText("  ");
         step.setText("  ");
         energy.setText("  ");
         return;
      }

      if (active.getFile() != null){
         status.setText("  " + active.toFFString());
      } else{
         status.setText("  " + active.toString());
      }

      if (active.isSimulation()){
         step.setText(active.getTimeString() + " ");
      } else if (active.isOptimization()){
         step.setText(active.getStepString() + " ");
      } else if (active.getCycles() > 1){
         step.setText("" + active.getCurrentCycle() + "/" + active.getCycles());
      } else{
         step.setText("");
      }

      if (active.isSimulation()){
         energy.setText(active.getEnergyString() + " ");
      } else if (active.isOptimization()){
         energy.setText(active.getEnergyString() + " ");
      } else if (active.getCycles() > 1){
         energy.setText("");
      } else{
         energy.setText("");
      }
   }

   public void setStatus(JLabel s, JLabel t, JLabel e){
      status = s;
      step = t;
      energy = e;
   }

   /**
    * Removes a FSystem from the Root and sets the first child FSystem to be active.
    */
   public void removeFSystem(FSystem s){
      synchronized (this){
         if (s == null){
            return;
         }
         if (root.getIndex(s) < 0){
            return;
         }
         root.remove(s);
         if (root.getChildCount() > 0){
            active = (FSystem) root.getChildAt(0);
         } else{
            active = null;
         }
         System.out.println("" + active + " set active");
         setActive(active);
      }
   }

   ArrayList newPaths = new ArrayList();
   ArrayList removedPaths = new ArrayList();
   ArrayList prePaths = new ArrayList();
   ArrayList picks = new ArrayList();
   TreePath nullPath = null;

   public void valueChanged(TreeSelectionEvent e){
      synchronized (this){
         // Determine the Active System
         FNode lastNode = (FNode) getLastSelectedPathComponent();
         if (lastNode != null){
            activeNode = lastNode;
            FSystem s = activeNode.getFSystemNode();
            if (s != null){
               setActive(s);
            }
         }

         TreePath[] paths = e.getPaths();
         if (paths == null){
            return;
         }

         // Reuse the same the ArrayLists
         ArrayList temp = prePaths;
         prePaths = newPaths;
         newPaths = temp;

         // Determine new and removed paths
         newPaths.clear();
         removedPaths.clear();
         for (int i = 0; i < paths.length; i++){
            if (e.isAddedPath(i)){
               newPaths.add(paths[i]);
            } else{
               removedPaths.add(paths[i]);
            }
         }

         // Create a non-redundant set of new/removed paths
         TreePath pathi, pathj;
         for (int i = 0; i < newPaths.size(); i++){
            pathi = (TreePath) newPaths.get(i);
            if (pathi == nullPath){
               continue;
            }
            for (int j = i + 1; j < newPaths.size(); j++){
               pathj = (TreePath) newPaths.get(j);
               if (pathi == nullPath || pathj == nullPath){
                  continue;
               }
               if (pathi.isDescendant(pathj)){
                  newPaths.set(j, nullPath);
               } else if (pathj.isDescendant(pathi)){
                  newPaths.set(i, nullPath);
               }
            }
         }
         while (newPaths.remove(nullPath)){
         }

         for (int i = 0; i < removedPaths.size(); i++){
            pathi = (TreePath) removedPaths.get(i);
            if (pathi == nullPath){
               continue;
            }
            for (int j = i + 1; j < removedPaths.size(); j++){
               pathj = (TreePath) removedPaths.get(j);
               if (pathi == nullPath || pathj == nullPath){
                  continue;
               }
               if (pathi.isDescendant(pathj)){
                  removedPaths.set(j, nullPath);
               } else if (pathj.isDescendant(pathi)){
                  removedPaths.set(i, nullPath);
               }
            }
         }
         while (removedPaths.remove(nullPath)){
         }

         // Remove the RemovedPaths from the Existing List
         for (int i = 0; i < prePaths.size(); i++){
            pathi = (TreePath) prePaths.get(i);
            for (int j = 0; j < removedPaths.size(); j++){
               pathj = (TreePath) removedPaths.get(j);
               if (pathj.isDescendant(pathi)){
                  prePaths.set(i, nullPath);
                  break;
               }
            }
         }
         while (prePaths.remove(nullPath)){
         }

         // Combine new Paths and Existing Paths non-redundantly
         for (int i = 0; i < newPaths.size(); i++){
            pathi = (TreePath) newPaths.get(i);
            if (pathi == nullPath){
               continue;
            }
            for (int j = 0; j < prePaths.size(); j++){
               pathj = (TreePath) prePaths.get(j);
               if (pathj == nullPath){
                  continue;
               }
               if (pathi == nullPath || pathj == nullPath){
                  continue;
               }
               if (pathi.isDescendant(pathj)){
                  prePaths.set(j, nullPath);
               } else if (pathj.isDescendant(pathi)){
                  newPaths.set(i, nullPath);
               }
            }
         }
         while (newPaths.remove(nullPath)){
         }
         while (prePaths.remove(nullPath)){
         }
         newPaths.addAll(prePaths);

         activeNodes.clear();
         for (int i = 0; i < newPaths.size(); i++){
            pathi = (TreePath) newPaths.get(i);
            activeNodes.add(pathi.getLastPathComponent());
         }

         if (activeNode != null){
            TreePath activePath = new TreePath(activeNode);
            expandPath(activePath.getParentPath());
            //scrollPathToVisible(activePath);
            int r = treeSelectionModel.getLeadSelectionRow();
            scrollRowToVisible(r);
         }

         // We now have a non-redundant set of Active Paths; and a non-redundant
         // set of removed paths
         picks = new ArrayList();
         // Clear highlight of deselected nodes
         for (ListIterator li = removedPaths.listIterator(); li.hasNext(); ){
            TreePath r = (TreePath) li.next();
            boolean change = true;
            for (ListIterator lj = newPaths.listIterator(); lj.hasNext(); ){
               TreePath n = (TreePath) lj.next();
               if (n.isDescendant(r)){
                  change = false;
               }
            }
            if (change){
               FNode f = (FNode) r.getLastPathComponent();
               f.setSelected(false);
               picks.add(f);
            }
         }

         for (ListIterator li = newPaths.listIterator(); li.hasNext(); ){
            TreePath n = (TreePath) li.next();
            boolean change = true;
            for (ListIterator lj = prePaths.listIterator(); lj.hasNext(); ){
               TreePath p = (TreePath) lj.next();
               if (p.isDescendant(n)){
                  change = false;
               }
            }
            if (change){
               FNode f = (FNode) n.getLastPathComponent();
               f.setSelected(true);
               picks.add(f);
            }
         }
         if (GraphicsContext.highlightSelections){
            fpanel.getGraphics3D().updateScene(picks, false, false, 0, true,
               GraphicsContext.bySELECT);
         } else if (GraphicsContext.labelAtoms || GraphicsContext.labelResidues){
            fpanel.getGraphics3D().setLabelsUpdated();
         }
      }
   }

   public void collapseAll(){
      int row = getRowCount() - 1;
      while (row >= 0){
         collapseRow(row);
         row--;
      }
   }

   public void addSelection(FNode f){
      if (f == null){
         return;
      }
      synchronized (this){
         TreePath path = new TreePath(f.getPath());
         addSelectionPath(path);
         f.setSelected(true);
      }
   }

   public void addSelections(List a){
      synchronized (this){
         for (ListIterator li = a.listIterator(); li.hasNext(); ){
            FNode f = (FNode) li.next();
            addSelection(f);
         }
      }
   }

   public void removeSelection(FNode f){
      if (f == null){
         return;
      }
      synchronized (this){
         TreePath path = new TreePath(f.getPath());
         for (Enumeration e = getExpandedDescendants(path); e.hasMoreElements(); ){
            TreePath treePath = new TreePath((DefaultMutableTreeNode) e.nextElement());
            collapsePath(treePath);
         }
         removeSelectionPath(path);
         f.setSelected(false);
      }
   }

   public void removeSelections(List a){
      synchronized (this){
         for (ListIterator li = a.listIterator(); li.hasNext(); ){
            FNode f = (FNode) li.next();
            removeSelection(f);
         }
      }
   }

   public void selectAll(){
      if (active == null){
         return;
      }
      synchronized (this){
         List l = Arrays.asList(getSystems());
         addSelections(l);
      }
   }

   public void onlySelection(FNode f){
      synchronized (this){
         if (activeNodes != null){
            int num = activeNodes.size();
            TreePath paths[] = new TreePath[num];
            for (int i = 0; i < num; i++){
               paths[i] = new TreePath(((FNode) activeNodes.get(i)).getPath());
            }
            removeSelectionPaths(paths);
         }
         collapseAll();
         addSelection(f);
      }
   }

   public void groupSelection(FNode f1, FNode f2){
      if (f1 == null || f2 == null){
         return;
      }
      synchronized (this){
         TreePath paths[] = new TreePath[2];
         paths[0] = new TreePath(f1.getPath());
         paths[1] = new TreePath(f2.getPath());
         RowMapper rm = fpanel.getTree().getTreeSelectionModel().getRowMapper();
         int[] rows = rm.getRowsForPaths(paths);
         setSelectionInterval(rows[0], rows[1]);
      }
   }

   public void toggleSelection(FNode f){
      if (f == null){
         return;
      }
      synchronized (this){
         TreePath path = new TreePath(f.getPath());
         if (isPathSelected(path)){
            removeSelectionPath(path);
         } else{
            addSelectionPath(path);
         }
      }

   }

   public void toggleSelections(ArrayList a){
      synchronized (this){
         for (ListIterator li = a.listIterator(); li.hasNext(); ){
            FNode f = (FNode) li.next();
            toggleSelection(f);
         }
      }
   }

}
