/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.parsers;

import java.io.*;
import java.util.*;

import ffe.lang.*;

/**
 * The MMFilter Class is used to parse and store molecular mechanics data
 * from TINKER Keyword (*.KEY) and Parameter (*.PRM) files.
 */
public final class MMFilter{

   // Various Lennard Jones representations
   private int radiustype = RMIN;
   private int radiussize = RADIUS;
   private float rscale = 1.0f;
   private float rfactor = 1.0f;

   private static final int RMIN = 0;
   private static final int SIGMA = 1;
   private static final int RADIUS = 0;
   private static final int DIAMETER = 1;
   private File keyfile = null; // The key file should contain a reference

   // to the force field in use

   /**
    * The default Constructor intialized the HashTables, then attempts to parse
    * a Force Field Data File.
    */
   public MMFilter(){}

   /**
    * Find and returns a Force Field Atom type for the corresponding pdb symbol
    * specified by the Residue name (res) and PDB Symbol (pdb_symbol).
    */
   public MMAtom getMMAtom(String res, String pdb_symbol){
      MMAtom mmatom;
      String nres = (String) aalookup.get(res); //Convert from a 3-letter abrev. to full name
      if (nres == null){
         mmatom = (MMAtom) atom.get("1");
         return mmatom;
      }
      Hashtable res_types = (Hashtable) atomlookup.get(nres);
      if (res_types != null){ //Get the pdb to OPLS MM conversion key
         String key = (String) res_types.get(pdb_symbol);
         if (key == null){
            key = (String) res_types.get(pdb_symbol.substring(0, pdb_symbol.length() - 1));
         }
         if (key == null && pdb_symbol.length() > 2){
            key = (String) res_types.get(pdb_symbol.substring(0, pdb_symbol.length() - 2));
         }
         if (key != null){ //If the key is found, retrieve the MM data for this atom
            mmatom = (MMAtom) atom.get(key);
            if (mmatom == null){
               //System.out.println("Key: " + key + " not found MM Atom Hashtable");
               mmatom = (MMAtom) atom.get("1");
            }
            return mmatom;
         }
      }
      //System.out.println("Not Found: " + res + ", " + pdb_symbol + " - Default Hydrogen used");
      mmatom = (MMAtom) atom.get("1");
      return mmatom;
   }

   /**
    * Return a Force Field Atom type specified by prm_num
    */
   public MMAtom getMMAtom(int prm_num){
      String key = new String("" + prm_num);
      MMAtom mmatom = (MMAtom) atom.get(key);
      if (mmatom == null){
         System.out.println("Atom type " + prm_num + " not found in " + getForceField());
      }
      return mmatom;
   }

   /*
        public MMAtom getCloseMMAtom(char c){
       for (Enumeration e = atom.elements(); e.hasMoreElements();){
           MMAtom m = (MMAtom) e.nextElement();
           if (m.atoms.charAt(0) == c) {
    System.out.println("Type " + m.atoms.charAt(0) + " given to " + c);
               return m;
           }
       }
       return null;
        }*/

   //Parse a Force Field data file
   public boolean parse(){

      atom = new Hashtable();
      //bond = new Hashtable();
      //angle = new Hashtable();
      //improper = new Hashtable();
      //dihedral = new Hashtable();
      vdw = new Hashtable();
      charge = new Hashtable();
      atomlookup = new Hashtable();

      try{
         String recordType;
         InputStream fis = null;
         BufferedReader br;
         StreamTokenizer stoken;
         if (forceFieldParam != null && !forceFieldParam.equalsIgnoreCase("NONE")){
            // Append the file extension if necessary
            if (forceFieldParam.indexOf(".prm") < 0){
               forceFieldParam = new String(forceFieldParam + ".prm");
            }
            // Check for an absolute path
            File ff = new File(forceFieldParam);
            if (!ff.exists()){
               // Check for a path relative to the key file location
               if (keyfile != null){
                  ff = new File(keyfile.getParentFile(), forceFieldParam);
                  if (!ff.exists()){
                     return false;
                  }
               }
            }
            fis = new FileInputStream(ff);
            if (fis == null){
               System.out.println("Error attempting to open the parameter file");
               return false;
            }

            br = new BufferedReader(new InputStreamReader(fis));
            stoken = new StreamTokenizer(br);
            stoken.wordChars('*', '*');
            stoken.wordChars('+', '+');
            stoken.commentChar('#');
            stoken.eolIsSignificant(true);

            while (findRecord(stoken)){
               recordType = stoken.sval;
               if (recordType.equalsIgnoreCase("Atom")){
                  parseAtom(stoken);
                  //else if (recordType.equalsIgnoreCase("Bond")) parseBond(stoken);
                  //else if (recordType.equalsIgnoreCase("Angle")) parseAngle(stoken);
                  //else if (recordType.equalsIgnoreCase("Torsion")) parseDihedral(stoken);
                  //else if (recordType.equalsIgnoreCase("Imptors")) parseImproper(stoken);
               } else if (recordType.equalsIgnoreCase("VDW")){
                  parseVDW(stoken);
                  // Put Partial Charges and Monopoles in the same structure
               } else if (recordType.equalsIgnoreCase("Charge")){
                  parseCharge(stoken);
               } else if (recordType.equalsIgnoreCase("Multipole")){
                  parseMultipole(stoken);
                  //else if (recordType.equalsIgnoreCase("Biotype")) parseBiotype(stoken);
               } else if (recordType.equalsIgnoreCase("Radiussize")){
                  parseKey(stoken, recordType);
               } else if (recordType.equalsIgnoreCase("Radiustype")){
                  parseKey(stoken, recordType);
               }
            }
         }

         if (keyfile != null){
            FileInputStream is = new FileInputStream(keyfile);
            if (is != null){
               br = new BufferedReader(new InputStreamReader(is));
               stoken = new StreamTokenizer(br);
               stoken.wordChars('*', '*');
               stoken.wordChars('+', '+');
               stoken.commentChar('#');
               stoken.eolIsSignificant(true);
               while (findRecord(stoken)){
                  recordType = stoken.sval;
                  if (recordType.equalsIgnoreCase("Atom")){
                     parseAtom(stoken);
                     //else if (recordType.equalsIgnoreCase("Bond")) parseBond(stoken);
                     //else if (recordType.equalsIgnoreCase("Angle")) parseAngle(stoken);
                     //else if (recordType.equalsIgnoreCase("Torsion")) parseDihedral(stoken);
                     //else if (recordType.equalsIgnoreCase("Imptors")) parseImproper(stoken);
                  } else if (recordType.equalsIgnoreCase("VDW")){
                     parseVDW(stoken);
                  } else if (recordType.equalsIgnoreCase("Charge")){
                     parseCharge(stoken);
                     //else if (recordType.equalsIgnoreCase("Biotype")) parseBiotype(stoken);
                  } else if (recordType.equalsIgnoreCase("Radiussize")){
                     parseKey(stoken, recordType);
                  } else if (recordType.equalsIgnoreCase("Radiustype")){
                     parseKey(stoken, recordType);
                  }
               }
            }
         }
         if (atom.size() == 0){
            //System.out.println("No atom types read in");
            return false;
         }
         finalizeAtomTypes();
      } catch (IOException e){
         System.out.println(e);
      }
      return true;
   }

   private void finalizeAtomTypes(){
      MMAtom mm;
      float[] temp;

      for (Enumeration e = atom.elements(); e.hasMoreElements(); ){
         mm = (MMAtom) e.nextElement();
         if (mm.stused){
            temp = (float[]) charge.get(mm.specifictype);
         } else{
            temp = (float[]) charge.get(mm.type);
            //default
         }
         if (temp == null){
            mm.charge = 0;
         } else{
            mm.charge = temp[0];
            //Ad VDW parameters to Atom type records
         }
         temp = (float[]) vdw.get(mm.type);
         //default - shouldn't get used
         if (temp == null){
            mm.sigma = 2.0f;
            mm.epsilon = 0.5f;
         } else{
            mm.sigma = rfactor * temp[0] / rscale;
            mm.epsilon = temp[1];
         }

      }
   }

   //Find a new record entry, ie a record keyword at the beginning of a line
   private boolean findRecord(StreamTokenizer st){
      if (st == null){
         return false;
      }
      int status;
      String record;
      boolean lineStart = false;

      while (true){
         try{
            status = st.nextToken();
            //if the end of file is reached, there are no more records
            if (status == st.TT_EOF){
               return false;
            }
            //if an end of line is reached, the next Token is a possible record
            else if (status == st.TT_EOL){
               lineStart = true;
            } else if (lineStart == true && status == st.TT_WORD){
               record = st.sval;
               //System.out.println("" + record + st.lineno());
               if (records.contains(record.toLowerCase())){
                  return true;
               } else{
                  lineStart = false;
               }
            }
         } catch (IOException e){
            System.out.println("" + e);
         }
      }
   }

   private void parseBiotype(StreamTokenizer st){
      try{
         //Atom Symbol
         st.nextToken(); //Skip Biotype #
         st.nextToken();
         String atomsymbol = new String(st.sval);
         //Description
         st.nextToken();
         String resname = new String(st.sval);
         //Atom Type
         st.nextToken();
         String key = new String("" + (int) st.nval);

         if (!key.equals("999")){
            if (atomlookup.containsKey(resname)){
               Hashtable restypes = (Hashtable) atomlookup.get(resname);
               restypes.put(atomsymbol, key);
            } else{
               Hashtable ht = new Hashtable();
               ht.put(atomsymbol, key);
               atomlookup.put(resname, ht);
            }
         }
      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseAtom(StreamTokenizer st){
      try{
         //Atom Type
         st.nextToken();
         String type = new String("" + (int) st.nval);
         type = type.intern();

         //Atom Class
         st.nextToken();
         String specifictype = null;
         if (st.ttype == st.TT_NUMBER){
            specifictype = type;
            type = new String("" + (int) st.nval);
            type = type.intern();
            st.nextToken();
         }

         //Atoms Represented
         String basic_type = new String(st.sval);
         basic_type.intern();

         //Description
         st.nextToken();
         String descript = new String(st.sval);

         //Atomic Number
         st.nextToken();
         int a = (int) st.nval;

         //Atomic Mass
         st.nextToken();
         float m = (float) st.nval;

         //Hybridization
         st.nextToken();
         int h = (int) st.nval;

         MMAtom mmatom;
         if (specifictype != null){ //If specific atom types are used
            if (MMAtom.mmatompool.isEmpty()){
               mmatom = new MMAtom(type, specifictype, basic_type, descript, a, m, h);
            } else{
               mmatom = (MMAtom) MMAtom.mmatompool.get(0);
               mmatom.reInit(type, specifictype, basic_type, descript, a, m, h);
            }
            atom.put(specifictype, mmatom);
         } else{ //No atom class
            if (MMAtom.mmatompool.isEmpty()){
               mmatom = new MMAtom(type, basic_type, descript, a, m, h);
            } else{
               mmatom = (MMAtom) MMAtom.mmatompool.get(0);
               mmatom.reInit(type, basic_type, descript, a, m, h);
            }
            atom.put(type, mmatom);
         }
      } catch (IOException e){
         System.out.println("" + e + st.toString());
      }
   }

   private void parseVDW(StreamTokenizer st){
      try{
         float[] vdws = new float[2];

         //Atom Type
         st.nextToken();
         String key = new String("" + (int) st.nval);
         //Sigma
         st.nextToken();
         vdws[0] = (float) st.nval;
         //Epsilon
         st.nextToken();
         vdws[1] = (float) st.nval;

         vdw.put(key, vdws);

      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseKey(StreamTokenizer st, String key){
      try{
         st.nextToken();
         String value = new String(st.sval);
         if (key.equalsIgnoreCase("RADIUSTYPE")){
            if (value.equalsIgnoreCase("R-MIN")){
               radiustype = RMIN;
               rfactor = 1.0f;
            } else{
               radiustype = SIGMA;
               rfactor = (float) Math.pow(2.0, 1.0 / 6.0);
            }
         } else if (key.equalsIgnoreCase("RADIUSSIZE")){
            if (value.equalsIgnoreCase("DIAMETER")){
               radiussize = DIAMETER;
               rscale = 2.0f;
            } else{
               radiussize = RADIUS;
               rscale = 1.0f;
            }
         }
      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseCharge(StreamTokenizer st){
      try{
         float[] c = new float[1];
         //Atom Type
         st.nextToken();
         String key = new String("" + (int) st.nval);
         //Charge
         st.nextToken();
         c[0] = (float) st.nval;

         charge.put(key, c);

      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseMultipole(StreamTokenizer st){
      try{
         float[] c = new float[1];
         //Atom Type
         st.nextToken();
         String key = new String("" + (int) st.nval);
         // Skip two values
         st.nextToken();
         st.nextToken();
         // Partial Charge
         st.nextToken();
         c[0] = (float) st.nval;

         charge.put(key, c);

      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseBond(StreamTokenizer st){
      try{
         //Atom 1
         st.nextToken();
         String a1 = new String("" + (int) st.nval);
         //Atom 2
         st.nextToken();
         String a2 = new String("" + (int) st.nval);
         String key = a1 + a2;
         //Force Constant
         st.nextToken();
         float fc = (float) st.nval;
         //Equilibrium Bond Length
         st.nextToken();
         float eq = (float) st.nval;
         bond.put(key, new MMTerm(key, fc, eq));
      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseAngle(StreamTokenizer st){
      try{
         //Atom 1
         st.nextToken();
         String a1 = new String("" + (int) st.nval);
         //Atom 2
         st.nextToken();
         String a2 = new String("" + (int) st.nval);
         //Atom 3
         st.nextToken();
         String a3 = new String("" + (int) st.nval);
         //Force Constant
         st.nextToken();
         float fc = (float) st.nval;
         //Equilibrium;
         st.nextToken();
         float eq = (float) st.nval;

         String key = a1 + a2 + a3;
         angle.put(key, new MMTerm(key, fc, eq));
      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseDihedral(StreamTokenizer st){
      try{
         //Atom 1
         st.nextToken();
         String a1 = new String("" + (int) st.nval);
         //Atom 2
         st.nextToken();
         String a2 = new String("" + (int) st.nval);
         //Atom 3
         st.nextToken();
         String a3 = new String("" + (int) st.nval);
         //Atom 4
         st.nextToken();
         String a4 = new String("" + (int) st.nval);
         //Force Constant
         st.nextToken();
         float fc = (float) st.nval;
         //Phase
         st.nextToken();
         float ph = (float) st.nval;
         //Periodicity
         st.nextToken();
         int pe = (int) st.nval;

         String key = a1 + a2 + a3 + a4;
         dihedral.put(key, new MMTerm(key, fc, ph, pe));

      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   private void parseImproper(StreamTokenizer st){
      try{
         MMTerm mmterm = null;
         //Atom 1
         st.nextToken();
         String a1 = new String("" + (int) st.nval);
         //Atom 2
         st.nextToken();
         String a2 = new String("" + (int) st.nval);
         //Atom 3
         st.nextToken();
         String a3 = new String("" + (int) st.nval);
         //Atom 4
         st.nextToken();
         String a4 = new String("" + (int) st.nval);

         st.nextToken();
         if (st.ttype != st.TT_EOL){
            String key = a1 + a2 + a3 + a4;
            //Force Constant
            float fc = (float) st.nval;
            //Phase
            st.nextToken();
            float ph = (float) st.nval;
            //Periodicity
            st.nextToken();
            int pe = (int) (float) st.nval;

            st.nextToken();
            if (st.ttype != st.TT_EOL){
               //Force Constant 2
               float fc2 = (float) st.nval;
               //Phase 2
               st.nextToken();
               float ph2 = (float) st.nval;
               //Period 2
               st.nextToken();
               int pe2 = (int) st.nval;
               mmterm = new MMTerm(key, fc, ph, pe, fc2, ph2, pe2);
            } else{
               mmterm = new MMTerm(key, fc, ph, pe);
            }
            improper.put(key, mmterm);
         }
      } catch (IOException e){
         System.out.println("" + e);
      }
   }

   /**
    * A Hashtable of MMAtom objects
    */
   public Hashtable atom;

   /**
    * Hashtable of MMTerm objects stored by concatenation of Keys
    */
   public Hashtable bond, angle, improper, dihedral;
   public Hashtable atomlookup;
   private Hashtable vdw, charge;
   private String forceFieldParam = null;

   public void setForceField(File key, String param){
      if (param == null || param.equalsIgnoreCase("none")){
         forceFieldParam = null;
      } else{
         forceFieldParam = param;
      }
      keyfile = key;
   }

   public void setForceFieldLookup(String ff){
      forceFieldParam = (String) fflookup.get(ff.toUpperCase().trim());
      if (forceFieldParam == null){
         System.out.println(ff + " force field unknown");
         return;
      }
   }

   public void setForceFieldFile(File ff){
      forceFieldParam = ff.toString();
   }

   public String getForceField(){
      return forceFieldParam;
   }

   public void destroy(){
      MMAtom a;
      for (Enumeration e = atom.elements(); e.hasMoreElements(); ){
         a = (MMAtom) e.nextElement();
         if (a.used == false){
            a.destroy();
         }
      }
   }

   /**
    * A Hashtable of Amino Acid Name Strings, stored by three-letter abreviations
    */
   public static final Hashtable aalookup = new Hashtable();
   private static final Vector records = new Vector();
   private static final Hashtable fflookup = new Hashtable();
   static{
      records.add("atom");
      records.add("bond");
      records.add("angle");
      records.add("torsion");
      records.add("imptors");
      records.add("vdw");
      records.add("charge");
      records.add("multipole");
      records.add("biotype");
      records.add("radiustype");
      records.add("radiussize");

      aalookup.put("GLY", "Glycine");
      aalookup.put("ALA", "Alanine");
      aalookup.put("VAL", "Valine");
      aalookup.put("LEU", "Leucine");
      aalookup.put("ILE", "Isoleucine");
      aalookup.put("SER", "Serine");
      aalookup.put("THR", "Threonine");
      aalookup.put("CYS", "Cystine (-SS-)");
      aalookup.put("PRO", "Proline");
      aalookup.put("PHE", "Phenylalanine");
      aalookup.put("TYR", "Tyrosine");
      aalookup.put("TRP", "Tryptophan");
      aalookup.put("HIS", "Histidine (HD)");
      aalookup.put("ASP", "Aspartic Acid");
      aalookup.put("ASN", "Asparagine");
      aalookup.put("GLU", "Glutamic Acid");
      aalookup.put("GLN", "Glutamine");
      aalookup.put("MET", "Methionine");
      aalookup.put("LYS", "Lysine");
      aalookup.put("ARG", "Arginine");
      aalookup.put("A", "Adenine");
      aalookup.put("T", "Thymine");
      aalookup.put("G", "Guanine");
      aalookup.put("C", "Cytosine");
      aalookup.put("HOH", "HOH");
      aalookup.put("HEM", "HEM");

      fflookup.put("AMBER94", "amber94.prm");
      fflookup.put("CHARMM19", "charmm19.prm");
      fflookup.put("CHARMM27", "charmm27.prm");
      fflookup.put("DUDEK-PROTEIN", "dudek.prm");
      fflookup.put("ENCAD", "encad.prm");
      fflookup.put("HOCH-STERN", "hoch.prm");
      fflookup.put("MERCK-MMFF", "merck.prm");
      fflookup.put("MM2-1991", "mm2.prm");
      fflookup.put("MM3-2000", "mm3.prm");
      fflookup.put("MM3-PROTEIN", "mm3pro.prm");
      fflookup.put("MMFF94-PROTEIN", "mmffpro.prm");
      fflookup.put("OPLS-UA", "oplsua.prm");
      fflookup.put("OPLS-AA", "oplsaa.prm");
      fflookup.put("SMOOTH", "smooth.prm");
      fflookup.put("SMOOTH-ALL-ATOM", "smoothaa.prm");
      fflookup.put("AMOEBA", "tinker.prm");
      fflookup.put("AMOEBA-PROTEIN", "tinkerpro.prm");
      fflookup.put("AMOEBA-WATER", "water.prm");
   }
}
