/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.parsers;

import java.io.*;
import java.net.URL;
import java.util.*;

import ffe.lang.*;
import ffe.panels.MainPanel;

/**
 * The PDBFilter class parses data from a Protein DataBank (*.PDB) file. The following
 *  records are recognized: HEADER, ATOM, HETATM, LINK,
 *  SSBOND, HELIX, SHEET, TURN. The rest are ignored.
 */
public final class PDBFilter
   extends SystemFilter{

   // An open BufferedReader is passed to the PDBFilter when a connection is
   // made to www.rcsb.org (or a mirror) to download a structure.
   private String pdbURL = null;
   private String vrmlURL = null;

   /**
    * Default Constructor
    */
   public PDBFilter(){
      super();
   }

   /**
    * Parse the PDB File form a local disk
    */
   public PDBFilter(FSystem f, MainPanel tink){
      super(f, tink);
   }

   /**
    * Parse the PDB File from a URL
    * @return boolean
    */
   public PDBFilter(FSystem f, String pdb, String vrml, MainPanel mainPanel){
      super(f, mainPanel);
      pdbURL = pdb;
      vrmlURL = vrml;
   }

   /**
    * Parse the PDB File
    */
   public boolean readfile(){
      String rawdata, altloc, atomType;
      String residueType, polymerType;
      boolean link;
      int residueNum;
      MMAtom mmatom;

      ArrayList links = new ArrayList();
      Vector structs = new Vector();
      double[] d = new double[3];
      Atom a = null;
      String[] connect, struct;
      FileWriter fw = null;
      BufferedReader br = null;
      BufferedWriter bw = null;

      fsystem.setFileType(SystemFilter.PDB);
      try{
         setFileRead(false);
         if (pdbURL == null){
            // Open a data stream to the PDB file
            if (fsystem == null){
               return false;
            }
            File pdbFile = fsystem.getFile();
            if (pdbFile == null || !pdbFile.exists() || !pdbFile.canRead()){
               return false;
            }
            FileReader fr = new FileReader(pdbFile);
            br = new BufferedReader(fr);
         } else{
            try{
               URL url = new URL(pdbURL);
               br = new BufferedReader(new InputStreamReader(url.openStream()));
               int retry = 0;
               while (!br.ready() && retry < 10){
                  synchronized (this){
                     statusBar.setText("  Waiting on Network");
                     wait(50);
                     retry++;
                  }
               }
            } catch (Exception e){
               return false;
            }
            // The downloaded PBD file will be echoed to the local file system
            // (if the file does not already exist)
            String fileName = fsystem.getName() + ".pdb";
            String path = fpanel.getCWD().getAbsolutePath();
            File pdbFile = new File(path + File.separatorChar + fileName);
            if (!pdbFile.exists()){
               fw = new FileWriter(pdbFile);
               bw = new BufferedWriter(fw);
            }
            fsystem.setFile(pdbFile);
         }
         rawdata = br.readLine();

         //First atom is #1, to match xyz file format
         int atomnum = 1;
         StringBuffer status = new StringBuffer("  Opening " + fsystem.getFile().getName() + "  ");

         // While the END parameter is not read in, load atoms
         while ((rawdata != null) && (!rawdata.startsWith("END"))){
            if (rawdata.startsWith("ATOM")){
               altloc = new String(rawdata.substring(16, 17).toUpperCase());
               if (!altloc.equals(" ")){
                  fsystem.addAltLocation(altloc);
               }
               if (altloc.equals(" ") || altloc.equals("A")){
                  atomType = new String(rawdata.substring(13, 16).trim());
                  residueType = new String(rawdata.substring(17, 20).trim());
                  residueNum = Integer.decode(rawdata.substring(22, 26).trim()).intValue();
                  polymerType = new String(rawdata.substring(21, 22));
                  if (polymerType.equalsIgnoreCase(" ")){
                     polymerType = "Blank".intern();
                  }
                  d[0] = Double.valueOf(rawdata.substring(30, 38).trim()).doubleValue();
                  d[1] = Double.valueOf(rawdata.substring(38, 46).trim()).doubleValue();
                  d[2] = Double.valueOf(rawdata.substring(46, 54).trim()).doubleValue();
                  mmatom = getMMAtom(atomType);
                  a = new Atom(atomnum++, atomType, mmatom, d, residueType, residueNum, polymerType);
                  fsystem.addFNode(a);
               }
            } else if (rawdata.startsWith("HETATM")){
               altloc = new String(rawdata.substring(16, 17));
               if (!altloc.equals(" ")){
                  fsystem.addAltLocation(altloc);
               }
               if (altloc.equals(" ") || altloc.equals("A")){
                  residueType = new String(rawdata.substring(17, 20).trim());
                  atomType = new String(rawdata.substring(13, 16).trim());
                  residueNum = Integer.decode(rawdata.substring(22, 26).trim()).intValue();
                  polymerType = new String(rawdata.substring(21, 22));
                  if (polymerType.equalsIgnoreCase(" ")){
                     polymerType = "Blank".intern();
                  }
                  d[0] = Double.valueOf(rawdata.substring(30, 38).trim()).doubleValue();
                  d[1] = Double.valueOf(rawdata.substring(38, 46).trim()).doubleValue();
                  d[2] = Double.valueOf(rawdata.substring(46, 54).trim()).doubleValue();
                  mmatom = getMMAtom(atomType);
                  a = new Atom(atomnum++, atomType, mmatom, d, residueType, residueNum, polymerType);
                  fsystem.addFNode(a);
               }
            } else if (rawdata.startsWith("HEADER")){
               String systemName = new String(rawdata.substring(10, rawdata.length()).trim());
               //fsystem.setName(systemName);
            } else if (rawdata.startsWith("SSBOND")){
               link = true;
               connect = new String[6];
               connect[0] = new String(rawdata.substring(15, 16)); //Polymers
               connect[1] = new String(rawdata.substring(29, 30));
               connect[2] = new String(rawdata.substring(17, 21).trim()); //Residues
               connect[3] = new String(rawdata.substring(31, 35).trim());
               connect[4] = new String("SG"); //Atoms
               connect[5] = new String("SG");
               links.add(connect);
            } else if (rawdata.startsWith("LINK")){
               link = true;
               connect = new String[6];
               connect[0] = new String(rawdata.substring(21, 22)); //Polymers
               connect[1] = new String(rawdata.substring(51, 52));
               connect[2] = new String(rawdata.substring(22, 26).trim()); //Residues
               connect[3] = new String(rawdata.substring(52, 56).trim());
               connect[4] = new String(rawdata.substring(12, 16).trim()); //Atoms
               connect[5] = new String(rawdata.substring(42, 46).trim());
               links.add(connect);
            } else if (rawdata.startsWith("HELIX")){
               struct = new String[6];
               struct[0] = new String(rawdata.substring(0, 6).trim()); //HELIX
               struct[1] = new String(rawdata.substring(19, 20)); //Polymers
               struct[2] = new String(rawdata.substring(31, 32));
               struct[3] = new String(rawdata.substring(21, 25).trim()); //Residue
               struct[4] = new String(rawdata.substring(33, 37).trim());
               struct[5] = new String(rawdata.substring(38, 40).trim()); //Helix Type
               structs.add(struct);
            } else if (rawdata.startsWith("SHEET")){
               struct = new String[6];
               struct[0] = new String(rawdata.substring(0, 6).trim()); //SHEET
               struct[1] = new String(rawdata.substring(21, 22)); //Polymers
               struct[2] = new String(rawdata.substring(32, 33));
               struct[3] = new String(rawdata.substring(22, 26).trim()); //Residue
               struct[4] = new String(rawdata.substring(33, 37).trim());
               struct[5] = new String(rawdata.substring(38, 40).trim()); //Strand Sense
               structs.add(struct);
            } else if (rawdata.startsWith("TURN")){
               struct = new String[6];
               struct[0] = new String(rawdata.substring(0, 6).trim()); //TURN
               struct[1] = new String(rawdata.substring(19, 20)); //Polymers
               struct[2] = new String(rawdata.substring(30, 31));
               struct[3] = new String(rawdata.substring(20, 24).trim()); //Residue
               struct[4] = new String(rawdata.substring(31, 35).trim());
               structs.add(struct);
            }
            if (bw != null && rawdata != null){
               bw.write(rawdata);
               bw.newLine();
            }
            if (atomnum % 1000 == 0){
               status.append(".");
               statusBar.setText(status.toString());
            }
            rawdata = br.readLine();
         }
         if (bw != null){
            bw.flush();
            bw.close();
         }
         br.close();

         // Assign Secondary Structure Based on PDB Info
         for (Enumeration e = structs.elements(); e.hasMoreElements(); ){
            struct = (String[]) e.nextElement();
            if (struct[1].equalsIgnoreCase(" ")){
               struct[1] = "Blank".intern();
            }
            Polymer p = fsystem.getPolymer(struct[1], false);
            if (p != null){
               for (int i = Integer.parseInt(struct[3]); i <= Integer.parseInt(struct[4]); i++){
                  Residue r = p.getResidue(i);
                  if (r != null){
                     r.setSSType(Residue.parseSSType(struct[0]));
                  }
               }
            }
         }

         setFileRead(true);
      } catch (IOException e){
         statusBar.setText("  Opening " + fsystem.getFile().getName() + " Failed");
         return false;
      }

      // Now read in VRML data

      /*
             if (vrmlURL == null){
         String name = fsystem.getName() + ".wrl";
         File vrmlFile = new File(fpanel.getCWD() + File.separator + name);
         if (vrmlFile != null && vrmlFile.exists() && vrmlFile.canRead()){
            statusBar.setText("  Loading " + name);
            fsystem.setVRML(fpanel.getGraphics3D().loadVRML(vrmlFile));
         }
             } else{
         File vrmlFile = new File(fpanel.getCWD() + File.separator + fsystem.getName() + ".wrl");
         if (!vrmlFile.exists()){
            int retry = 0;
            boolean done = false;
            int lineNumber = 0;
            StringBuffer status = new StringBuffer("  Downloading VRML  ");
            statusBar.setText(status.toString());
            try{
               URL copyURL = new URL(vrmlURL);
               InputStreamReader ir = new InputStreamReader(copyURL.openStream());
               BufferedReader vbr = new BufferedReader(ir);
               BufferedWriter vbw = new BufferedWriter(new FileWriter(vrmlFile));
               while (!done && retry < 100){
                  while (!vbr.ready() && retry < 100){
                     synchronized (this){
                        statusBar.setText("  Waiting on Network");
                        wait(100);
                        retry++;
                     }
                  }
                  String data = vbr.readLine();
                  lineNumber++;
                  vbw.write(data);
                  vbw.newLine();
                  if (data.toUpperCase().indexOf("END OF OUTPUT") >= 0){
                     done = true;
                  }
                  if (lineNumber % 100 == 0){
                     status.append(".");
                     statusBar.setText(status.toString());
                  }
               }
               vbr.close();
               vbw.flush();
               vbw.close();
            } catch (Exception e){

            } finally{
               if (retry == 100 || !done){
                  statusBar.setText("  VRML Download Timed Out");
                  vrmlFile.delete();
               }
            }
         }
         try{
            statusBar.setText("  Loading VRML");
            if (vrmlFile.exists()){
               fsystem.setVRML(fpanel.getGraphics3D().loadVRML(vrmlFile));
            } else{
               URL url = new URL(vrmlURL);
               fsystem.setVRML(fpanel.getGraphics3D().loadVRML(url));
            }
         } catch (Exception e){
            statusBar.setText("  VRML Failed to Load");
            return false;
         }
             }
       */

      return true;

   }

   public boolean writefile(){
      File pdbfile = fsystem.getFile();
      if (pdbfile == null){
         return false;
      }
      try{
         FileWriter fw = new FileWriter(pdbfile);
         BufferedWriter bw = new BufferedWriter(fw);
         FSystem m = getFSystem();
         //m.writePDB(bw, 1);
         bw.write("END");
         bw.close();
         fw.close();
      } catch (IOException e){
         System.out.println("" + e);
         return false;
      }
      return true;
   }

   public static MMAtom getMMAtom(String s){
      String c = s.toLowerCase();
      if (c.startsWith("h")){
         return (MMAtom) mmatoms.get("H");
      }
      if (c.startsWith("c")){
         return (MMAtom) mmatoms.get("C");
      }
      if (c.startsWith("n")){
         return (MMAtom) mmatoms.get("N");
      }
      if (c.startsWith("o")){
         return (MMAtom) mmatoms.get("O");
      }
      if (c.startsWith("p")){
         return (MMAtom) mmatoms.get("P");
      }
      if (c.startsWith("s")){
         return (MMAtom) mmatoms.get("S");
      } else{
         return (MMAtom) mmatoms.get("X");
      }
   }

   static Hashtable mmatoms;
   static{
      mmatoms = new Hashtable();
      mmatoms.put("H", new MMAtom("H", "H", "Hydrogen", 1, 12.0f, 4, 1.2f, 0.0f, 0.0f));
      mmatoms.put("C", new MMAtom("C", "C", "Carbon", 6, 12.0f, 4, 1.85f, 0.0f, 0.0f));
      mmatoms.put("N", new MMAtom("N", "N", "Nitrogen", 7, 14.0f, 3, 1.54f, 0.0f, 0.0f));
      mmatoms.put("O", new MMAtom("O", "O", "Oxygen", 8, 16.0f, 2, 1.4f, 0.0f, 0.0f));
      mmatoms.put("P", new MMAtom("P", "P", "Phosphorus", 15, 31.0f, 4, 1.9f, 0.0f, 0.0f));
      mmatoms.put("S", new MMAtom("S", "S", "Sulfur", 16, 32.0f, 2, 1.85f, 0.0f, 0.0f));
      mmatoms.put("X", new MMAtom("X", "X", "Other", 3, 1.0f, 4, 1.85f, 0.0f, 0.0f));
   }
}
