/**
 * <p>Title: Force Field Explorer</p>
 * <p>Description: Force Field Explorer provides a user interface to TINKER</p>
 * <p>Copyright: Copyright (c) 2004 Jay Ponder</p>
 * <p>Institution: Washington University</p>
 * @author Michael J. Schnieders
 * @version 4.2
 */

package ffe.parsers;

import java.io.*;
import java.text.*;
import java.util.*;
import javax.vecmath.Vector3d;

import ffe.lang.*;
import ffe.panels.MainPanel;

/**
 * The XYZFilter class parses TINKER Cartesian Coordinate (*.XYZ) files.
 */
public class XYZFilter
   extends SystemFilter{

   /**
    * Default Constructor
    */
   public XYZFilter(){
      super();
   }

   public XYZFilter(FSystem system, MMFilter mmfilter, MainPanel fpanel){
      super(system, mmfilter, fpanel);
      setType(SystemFilter.XYZ);
   }

   /**
    * Parse the XYZ File
    */
   public boolean readfile(){

      // FFE only reads XYZ files if it is associated with a parameter file.
      File xyzfile = fsystem.getFile();
      Keyword parameters = fsystem.getKeyword("PARAMETERS");
      if (parameters == null){
         return false;
      }

      String parmname = parameters.getEntry(0);
      if (parmname == null){
         return false;
      }
      parmname = parmname.replaceAll("\"", "");

      if (mmfilter == null){
         mmfilter = new MMFilter();
      }
      mmfilter.setForceField(getFSystem().getKeyFile(), parmname);
      fsystem.setForceField((new File(parmname)).getName().toUpperCase());

      if (!mmfilter.parse()){
         file_read = false;
         return false;
      }

      try {
         setFileRead(false);

         FileReader fr = new FileReader(xyzfile);
         BufferedReader br = new BufferedReader(fr);
         String data = br.readLine();

         if (data == null){
            return false;
         }

         String tokens[] = data.trim().split(" +", 2);
         int num_atoms = Integer.parseInt(tokens[0]);
         if (num_atoms < 1){
            return false;
         }

         if (tokens.length == 2){
            getFSystem().setName(tokens[1]);
         }

         int bonds[][] = new int[num_atoms][4];
         double d[] = new double[3];

         atomList = new ArrayList();
         StringBuffer status = new StringBuffer("  Opening " + xyzfile.getName() + " with " + num_atoms + " atoms");
         for (int i = 0; i < num_atoms; i++){
            if (!br.ready()){
               return false;
            }
            data = br.readLine();
            if (data == null){
               statusBar.setText("  Check atom " + (i + 1) + " in " + fsystem.getFile().getName());
               return false;
            }
            tokens = data.trim().split(" +");
            if (tokens == null || tokens.length < 6){
               statusBar.setText("  Check atom " + (i + 1) + " in " + fsystem.getFile().getName());
               return false;
            }

            String atom_name = tokens[1];
            d[0] = Double.parseDouble(tokens[2]);
            d[1] = Double.parseDouble(tokens[3]);
            d[2] = Double.parseDouble(tokens[4]);
            int atom_type = Integer.parseInt(tokens[5]);
            MMAtom mmatom = (MMAtom) mmfilter.getMMAtom(atom_type);
            if (mmatom == null){
               statusBar.setText("  Check Atom Type for Atom " + (i + 1) + " in " +
                  fsystem.getFile().getName());
               return false;
            }
            Atom a = null;
            if (Atom.atompool.isEmpty()){
               a = new Atom(i + 1, atom_name, mmatom, d);
            } else{
               a = (Atom) Atom.atompool.remove(0);
               a.reInit(i + 1, atom_name, mmatom, d);
            }
            atomList.add(a);

            // Bond Data
            int numbonds = tokens.length - 6;
            for (int b = 0; b < 4; b++){
               if (b < numbonds){
                  int bond = Integer.parseInt(tokens[6 + b]);
                  if (bond > num_atoms || bond < 1){
                     statusBar.setText("  Check the Bond for Atom " + (i + 1) + "to atom" + bond +
                        " in " + fsystem.getFile().getName());
                     return false;
                  }
                  bonds[i][b] = bond;
               } else{
                  bonds[i][b] = 0;
               }
            }
            if (i % 1000 == 0){
               status.append(".");
               statusBar.setText(status.toString());
            }
         }
         if (br.ready()){
            data = br.readLine();
            if (data != null){
               tokens = data.trim().split(" +");
               int arcnum = Integer.parseInt(tokens[0]);
               if (arcnum == num_atoms){
                  setType(XYZFilter.ARC);
               }
            }
         }

         br.close();
         fr.close();
         bondList = new ArrayList();
         for (int i = 1; i <= num_atoms; i++){
            int a1 = i;
            int j = -1;
            while (j < 3 && bonds[i - 1][++j] > 0){
               int a2 = bonds[i - 1][j];
               if (a1 < a2){
                  if (a1 > num_atoms || a1 < 1 || a2 > num_atoms || a2 < 1){
                     statusBar.setText("  Check the Bond Bewteen " + a1 + " and " + a2 + " in " +
                        fsystem.getFile().getName());
                     return false;
                  }
                  Atom atom1 = (Atom) atomList.get(a1 - 1);
                  Atom atom2 = (Atom) atomList.get(a2 - 1);
                  if (atom1 == null || atom2 == null){
                     statusBar.setText("  Check the Bond Bewteen " + a1 + " and " + a2 + " in " +
                        fsystem.getFile().getName());
                     return false;
                  }
                  if (Bond.bondpool.isEmpty()){
                     bondList.add(new Bond(atom1, atom2, 1));
                  } else{
                     Bond tempbond = (Bond) Bond.bondpool.remove(0);
                     tempbond.reInit(atom1, atom2, 1);
                     bondList.add(tempbond);
                  }
               }
            }
         }
         if (getType() == XYZFilter.ARC){
            return readtrajectory();
         }
         return true;
      } catch (IOException e){
         System.out.println("" + e);
      }
      return false;
   }

   public boolean writefile(){
      File xyzfile = fsystem.getFile();
      if (xyzfile == null){
         return false;
      }
      try{
         FileWriter fw = new FileWriter(xyzfile);
         BufferedWriter bw = new BufferedWriter(fw);

         // XYZ File First Line
         int numatoms = fsystem.getAtomList().size();
         String blanks = new String("               ");

         int len = (new String("" + numatoms)).length();
         bw.write(blanks.substring(0, 6 - len) + numatoms + " " + fsystem.toString() + "\n");

         Atom a, a2;
         Bond b;
         ArrayList bonds;
         StringBuffer line;
         StringBuffer lines[] = new StringBuffer[numatoms];
         String indexS, id, type, xS, yS, zS;

         // XYZ File Atom Lines
         List atoms = fsystem.getAtomList();
         Vector3d offset = fsystem.getOffset();
         NumberFormat form = NumberFormat.getInstance();
         DecimalFormat formatCoord = null;
         if (form instanceof DecimalFormat){
            formatCoord = (DecimalFormat) form;
            formatCoord.applyPattern("#####0.000000");
         }
         DecimalFormat formatBond = (DecimalFormat) NumberFormat.getInstance();
         formatBond.applyPattern("#####0");
         for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
            a = (Atom) li.next();
            indexS = new String("" + a.getXYZIndex());
            line = new StringBuffer(blanks.substring(0, 6 - indexS.length()) + indexS + "  ");
            id = a.getID();
            line.append(id + blanks.substring(0, 3 - id.length()));
            xS = formatCoord.format(a.getX() - offset.x);
            yS = formatCoord.format(a.getY() - offset.y);
            zS = formatCoord.format(a.getZ() - offset.z);
            line.append(blanks.substring(0, 12 - xS.length()) + xS);
            line.append(blanks.substring(0, 12 - yS.length()) + yS);
            line.append(blanks.substring(0, 12 - zS.length()) + zS);
            type = new String("" + a.getAtomType());
            line.append(blanks.substring(0, 6 - type.length()) + type);
            bonds = a.getBonds();
            if (bonds != null){
               for (ListIterator lj = bonds.listIterator(); lj.hasNext(); ){
                  b = (Bond) lj.next();
                  a2 = b.getOtherAtom(a);
                  xS = formatBond.format(a2.xyzindex);
                  line.append(blanks.substring(0, 6 - xS.length()) + xS);
               }
               lines[a.getXYZIndex() - 1] = line.append("\n");
            }
         }
         for (int i = 0; i < numatoms; i++){
            try{
               bw.write(lines[i].toString());
            } catch (Exception e){
               System.out.println("" + i);
            }
         }
         bw.close();
         fw.close();
      } catch (IOException e){
         System.out.println("" + e);
         return false;
      }
      return true;
   }

   public boolean readtrajectory(){
      // If the first entry was read succesfully, reopen the
      // archive file and read the rest of the coordinates
      try{
         setFileRead(false);
         FileReader fr = new FileReader(fsystem.getFile());
         BufferedReader bin = new BufferedReader(fr);
         String data = null;
         int numatoms = atomList.size();

         // Read past the first entry
         for (int i = 0; i < numatoms + 1; i++){
            data = bin.readLine();
         }

         int xyznum, cycle = 1;
         double x, y, z;
         double coords[][] = new double[numatoms][3];
         while (bin.ready()){
            data = bin.readLine();
            try{
               int num = Integer.parseInt(data.trim().split(" +")[0]);
               if (num != numatoms){
                  return false;
               }
            } catch (Exception e){
               return false;
            }
            for (int i = 0; i < numatoms; i++){
               //Atom Data
               data = bin.readLine();
               String[] tokens = data.trim().split(" +");
               if (tokens == null || tokens.length < 6){
                  statusBar.setText("  Check atom " + (i + 1) + ", Frame " + (cycle + 1)  + " in " + fsystem.getFile().getName());
                  return false;
               }
               coords[i][0] = Double.parseDouble(tokens[2]);
               coords[i][1] = Double.parseDouble(tokens[3]);
               coords[i][2] = Double.parseDouble(tokens[4]);
            }
            for (ListIterator li = atomList.listIterator(); li.hasNext(); ){
               Atom a = (Atom) li.next();
               xyznum = a.xyzindex;
               x = coords[xyznum - 1][0];
               y = coords[xyznum - 1][1];
               z = coords[xyznum - 1][2];
               Vector3d v3d = new Vector3d(x, y, z);
               a.addTrajectoryCoords(v3d, cycle);
            }
            cycle++;
            if (cycle % 10 == 0){
               statusBar.setText("  Read Frame: ".intern() + cycle);
            }
         }
         fsystem.setCycles(cycle);
         fsystem.setFileType(SystemFilter.ARC);
         setFileRead(true);
         return true;
      } catch (FileNotFoundException e){
         System.out.println("" + e);
      } catch (IOException e){
         System.out.println("" + e);
      }
      return false;
   }

   public static boolean readOnto(File newFile, FSystem oldSystem){
      if (newFile == null || !newFile.exists() || oldSystem == null){
         return false;
      }
      try{
         FileReader fr = new FileReader(newFile);
         BufferedReader br = new BufferedReader(fr);
         String data = br.readLine();
         if (data == null){
            return false;
         }

         String tokens[] = data.trim().split(" +");
         int num_atoms = Integer.parseInt(tokens[0]);
         if (num_atoms != oldSystem.getAtomList().size()){
            return false;
         }

         double d[][] = new double[3][num_atoms];

         for (int i = 0; i < num_atoms; i++){
            if (!br.ready()){
               return false;
            }
            data = br.readLine();
            if (data == null){
               System.out.println("Check atom " + (i + 1));
               return false;
            }

            tokens = data.trim().split(" +");
            if (tokens == null || tokens.length < 6){
               System.out.println("Check atom " + (i + 1));
               return false;
            }

            d[0][i] = Double.parseDouble(tokens[2]);
            d[1][i] = Double.parseDouble(tokens[3]);
            d[2][i] = Double.parseDouble(tokens[4]);
         }

         double[] x = new double[3];
         List atoms = oldSystem.getAtomList();

         for (ListIterator li = atoms.listIterator(); li.hasNext(); ){
            Atom a = (Atom) li.next();
            int index = a.getXYZIndex() - 1;
            x[0] = d[0][index];
            x[1] = d[1][index];
            x[2] = d[2][index];
            a.moveto(x);
         }
         oldSystem.centerFSystem();
         oldSystem.setFile(newFile);
         br.close();
         fr.close();
         return true;
      } catch (Exception e){
         return false;
      }
   }
}
